#light "off"
module Microsoft.Fine.Absyn

open Util

type ident = Sugar.ident
type lident = Sugar.LongIdent

type withinfo<'a,'t> = {
  v: 'a; 
  sort: 't;
  p: Range.range; 
} 

type var<'t>  = withinfo<lident,'t>
type bvar<'t> = withinfo<(ident * ident),'t> 
  (* Bound vars have a name for pretty printing, 
     and a unique name generated during desugaring. 
     Only the latter is used during type checking.  *)
type fieldname = lident
type bvdef = ident * ident  (* a pp-name and a real name *)

type typ' =  
  | Typ_btvar  of bvar<kind>
  | Typ_const  of var<kind>
  | Typ_record of list<fieldname * typ> 
  | Typ_fun    of option<bvdef> * typ * typ (* stash the formal param for pp *)
  | Typ_univ   of bvdef * kind * typ (* stash the formal param for pp *)
  | Typ_dtuple of list<option<bvdef> * typ>
  | Typ_refine of bvdef * typ * formula (* add another arg to stash a proof term? *)
  | Typ_app    of typ * typ
  | Typ_dep    of typ * exp
  | Typ_affine of typ
  | Typ_lam    of bvdef * typ * typ   (* \x:t. T *)
  | Typ_unknown
  | Typ_uvar   of Unionfind.uvar<uvar_basis> * kind
and typ = withinfo<typ', kind>

and uvar_basis = 
  | Uvar of (typ -> kind -> bool) (* A well-formedness check to ensure that all names are in scope *)
  | Fixed of typ
      
and exp' =
  | Exp_bvar       of bvar<typ>
  | Exp_fvar       of var<typ>
  | Exp_constant   of Sugar.sconst
  | Exp_constr_app of var<typ> * list<typ> * list<exp> * list<exp> 
  | Exp_abs        of bvdef * typ * exp 
  | Exp_tabs       of bvdef * kind * exp            (* produced during type checking *)
  | Exp_app        of exp * exp
  | Exp_tapp       of exp * typ             (* produced during type checking *)
  | Exp_match      of exp * list<pat * exp> * exp  (* always includes a default case *)
  | Exp_cond       of exp * exp * exp       
  | Exp_recd       of option<lident> * list<typ> * list<exp> * list<fieldname * exp>  
  | Exp_proj       of exp * fieldname 
  | Exp_ascribed   of exp * typ * evidence
  | Exp_let        of bool * list<bvdef * typ * exp> * exp (* let (rec?) x1 = e1 AND ... AND xn = en in e *)
  | Exp_gvar       of int  (* only for internal use by proof extraction. unbound variable deBruijn index *) 
  | Exp_extern_call of Sugar.externref * ident * typ * list<typ> * list<exp>
  | Exp_primop     of ident * list<exp>
  | Exp_bot                                 (* error case *)

and exp = withinfo<exp',typ>      

and pat = 
  | Pat_variant of lident * list<typ> * list<exp> * list<bvar<typ>> 

and evidence=list<exp*exp>
      
(* int  *)
and kind =
  | Kind_boxed of kind (* B_* or B_A *)
  | Kind_star
  | Kind_affine
  | Kind_tcon of kind * kind (* * -> k, A -> k, or (t -> k) -> k *)
  | Kind_dcon of typ * kind
  | Kind_qkind of list<(bvdef * kind)> * kind (* 'a::* => list 'a => ... *)
  | Kind_unknown 

and formula = typ

type uvar = Unionfind.uvar<uvar_basis>

type tparam =
  | Tparam_typ  of bvdef * kind (* idents for pretty printing *)
  | Tparam_term of bvdef * typ

type aqual = Private | Public

type sigelt =
  | Sig_tycon_kind   of lident * list<tparam> * kind * bool (* bool identifies a prop *)
  | Sig_typ_abbrev   of lident * list<tparam> * kind * typ
  | Sig_record_typ   of lident * list<tparam> * kind * typ
  | Sig_datacon_typ  of lident * list<tparam> * typ * bool * aqual * option<lident> (* bool identifies an assumption;lident identifies type name *)
  | Sig_value_decl   of lident * typ 
  | Sig_extern_value of Sugar.externref * lident * typ
  | Sig_extern_typ   of Sugar.externref * sigelt (* lident * typ * kind *)
      
type signature = list<sigelt>

type letbinding = list<bvdef * typ * exp> (* let recs may have more than one element *)

type modul = {
  name: lident;
  extends: option<lident>;
  pos: Range.range;
  signature: signature;
  letbindings: list<letbinding * bool>; (* the boolean signals a let rec *)
  main:option<exp>;
  exports: signature;
}

(* ************************Utilities*************************************** *)
let unbox = function
  | Kind_boxed k -> k
  | k -> k

let dummyRange = 0L
let withsort v s = {v=v; sort=s; p=dummyRange}
let setsort w t = {v=w.v; sort=t; p=w.p}
let withpos e r = {v=e; sort=withsort Typ_unknown Kind_unknown; p=r}
let t_withpos t r = {v=t; sort=Kind_unknown; p=r}
let withinfo e s r = {v=e; sort=s; p=r}

let new_bvar ppn rn r = withpos (ppn, rn) r
let pp_name bvd = fst bvd
let real_name bvd = snd bvd
let bvar_ppname (bv:bvar<'a>) : ident = let (pp, _) = bv.v in pp
let bvar_real_name (bv:bvar<'a>) : ident = let (_,rn) = bv.v in rn

let bvar_eq (bv1:bvar<typ>) (bv2:bvar<typ>) = 
  (bvar_real_name bv1).idText = (bvar_real_name bv2).idText
let fvar_eq fv1 fv2 = Sugar.lid_equals fv1.v fv2.v

let bvd_to_bvar (bvd:bvdef) =
  let pp = pp_name bvd in
  let rn = real_name bvd in
  let pos = rn.idRange in
    new_bvar pp rn pos

let bvd_to_bvar_s (bvd:bvdef) sort =
  let pp = pp_name bvd in
  let rn = real_name bvd in
  let pos = rn.idRange in
    {v=(pp, rn); sort=sort; p=pos}

let bvar_to_bvd bv = (bvar_ppname bv, bvar_real_name bv) 

(* let bvar_index bv = let (ix, _, _) = bv.v in ix *)

let gensym r = match r with 
    None -> Sugar.nng.Apply "gensym" dummyRange
  | Some r -> Sugar.nng.Apply "gensym" r

let genproofsym r = match r with 
    None -> Sugar.nng.Apply "__proofsym" dummyRange
  | Some r -> Sugar.nng.Apply "__proofsym" r
      
let rec gensyms = function 
    0 -> []
  | n -> (gensym None)::gensyms (n-1)

let genident r = 
  let sym = gensym r in
    match r with 
        None -> Sugar.ident(sym, dummyRange)
      | Some r -> Sugar.ident(sym, r)

let new_bvd () = let id = genident None in (id,id)
let new_proof_bvd () = 
  let id = Sugar.ident(genproofsym None, dummyRange) in 
    (id,id)
let is_proof_bvd (_,(rn:ident)) = rn.idText.StartsWith("__proofsym")
  
let bvdef_of_str s = let id = Sugar.ident(s, dummyRange) in (id, id)
(* type bvar_maps = ((int->int->bvar<kind>->typ) * (int->int->bvar<typ>->exp'))     *)

      
let rec compress' typ' = match typ' with
  | Typ_uvar (uv,k) -> 
      begin
        match Unionfind.find uv with 
            Uvar _ -> typ'
          | Fixed typ -> compress' typ.v
      end
  | _ -> typ' 
and compress t = 
  let t' = compress' t.v in
    if LanguagePrimitives.PhysicalEquality t' t.v then t
    else withinfo t' t.sort t.p
      
let (freevarsTyp:typ -> (list<bvar<kind>> * list<bvar<typ>>), 
     freevarsExp:exp -> (list<bvar<kind>> * list<bvar<typ>>)) =
  let is_bound (rn:ident) (idl:list<ident>) = 
    List.exists (fun (id':ident) -> id'.idText = rn.idText) idl in
  let rec fvs bvs out t = match (compress t).v with
    | Typ_btvar bv -> 
        let rn = bvar_real_name bv in
          if (is_bound rn bvs) then out else 
            let (out_tvs, out_xvs) = out in
              (bv::out_tvs, out_xvs)

    | Typ_const _ -> out

    | Typ_record fnt_l -> 
        List.fold_left (fun out (_, t) -> fvs bvs out t) out fnt_l 

    | Typ_fun (None, t, t') -> 
        let out' = fvs bvs out t in
          fvs bvs out' t

    | Typ_lam (bvd, t, t')
    | Typ_fun (Some bvd, t, t') -> 
        let out' = fvs bvs out t in
        let bvs' = (real_name bvd)::bvs in
          fvs bvs' out' t'

    | Typ_univ(bvd, k, t) -> 
        let bvs' = (real_name bvd)::bvs in
          fvs bvs' out t
            
    | Typ_dtuple (bvdopt_t_l) -> 
        let out, _ = List.fold_left 
          (fun (out, bvs) (bvdopt, t) -> 
             let out' = fvs bvs out t in
             let bvs' = match bvdopt with
                 None -> bvs
               | Some bvd -> (real_name bvd)::bvs in
               (out', bvs')) (out, bvs) bvdopt_t_l in
          out

    | Typ_refine (bvd, t, f) -> 
        let out' = fvs bvs out t in
        let bvs' = (real_name bvd)::bvs in
          fvs bvs' out' f

    | Typ_app (t1,  t2) -> 
        let out' = fvs bvs out t1 in
          fvs bvs out' t2
            
    | Typ_dep (t, e) -> 
        let out' = fvs bvs out t in
          fvs_exp bvs out' e
       
    | Typ_affine t -> fvs bvs out t
        
    | Typ_unknown 
    | Typ_uvar _ -> out 
  and fvs_exp bvs out e = 
    let out = fvs bvs out e.sort in
      match e.v with
        | Exp_bvar bv ->
            let rn = bvar_real_name bv in
              if (is_bound rn bvs) then out else 
                let (out_tvs, out_xvs) = out in
                  (out_tvs, bv::out_xvs)
        | Exp_fvar _      
        | Exp_constant _ -> out
        | Exp_constr_app (_, tl, el1, el) -> 
            let out' = List.fold_left (fvs bvs) out tl in 
              List.fold_left (fvs_exp bvs) out' (el1@el)
        | Exp_primop(op, el) -> 
            List.fold_left (fvs_exp bvs) out el
        | Exp_abs (bvd, t, e) -> 
            let out' = fvs bvs out t in
            let bvs' = (real_name bvd)::bvs in
              fvs_exp bvs' out' e
                
        | Exp_tabs(bvd, k, e) -> 
            let bvs' = (real_name bvd)::bvs in
              fvs_exp bvs' out e
                
        | Exp_app (e1, e2) -> 
            let out' = fvs_exp bvs out e1 in
              fvs_exp bvs out' e2
                
        | Exp_tapp (e, t) -> 
            let out' = fvs_exp bvs out e in
              fvs bvs out' t 
                
        | Exp_match (e, eqns, def) -> 
            let out1 = fvs_exp bvs out e in
            let out2 = List.fold_left 
              (fun out (p,e) -> 
                 let out', bvs' = fvs_pat bvs out p in
                   fvs_exp bvs' out' e) out1 eqns in
              fvs_exp bvs out2 def
                
        | Exp_cond (e, e1, e2) -> 
            let out1 = fvs_exp bvs out e in
            let out2 = fvs_exp bvs out1 e1 in
              fvs_exp bvs out2 e2
                
        | Exp_recd (_, targs, phantoms, fn_e_l) -> 
            let out1 = List.fold_left (fvs bvs) out targs in
            let out2 = List.fold_left (fvs_exp bvs) out1 phantoms in
              List.fold_left (fun out (fn, e) -> 
                                fvs_exp bvs out e) out2 fn_e_l
                
        | Exp_proj (e, fn) -> 
            fvs_exp bvs out e

        | Exp_ascribed (e, t, ev) -> 
            let out' = fvs_exp bvs out e in
            let out' = fvs bvs out' t in
              List.fold_left (fun out (bv, e) -> 
                                let out' = fvs_exp bvs out bv in
                                  fvs_exp bvs out' e) out' ev

        | Exp_let (false, [(bvd, t, e)], body) -> 
            let out1 = fvs_exp bvs out e in
            let out2 = fvs bvs out1 t in
            let bvs' = (real_name bvd)::bvs in
              fvs_exp bvs' out2 body

        | Exp_let (true, bvd_t_e_l, body) -> 
            let bvs' = List.fold_left 
              (fun bvs (bvd, _, _) -> 
                 real_name bvd::bvs) bvs bvd_t_e_l in
            let out1 = List.fold_left 
              (fun out (_, _, e) -> fvs_exp bvs' out e) out bvd_t_e_l in
              fvs_exp bvs' out1 body
                
        | Exp_extern_call(_, _, _, _, el) -> 
            List.fold_left (fvs_exp bvs) out el 

(*         | Exp_with_proofs (e1, el) ->  *)
(*             List.fold_left (fvs_exp bvs) out (e1::el) *)
              
        | Exp_bot -> out 
  and fvs_pat bvs out = function
    | Pat_variant (_, tl, el, bvar_l) -> 
        let out' = List.fold_left (fvs bvs) out tl in
        let out'' = List.fold_left (fvs_exp bvs) out' el in
        let bvs' = List.fold_left (fun bvs bvar -> (bvar_real_name bvar)::bvs) bvs bvar_l in
          out'', bvs' in
    ((fun t -> fvs [] ([],[]) t), 
     (fun e -> fvs_exp [] ([], []) e))

let rec kind_fold_map 
    (f: 'env -> 'benv -> typ' -> ('env * typ' * option<kind>))
    (g: 'env -> 'benv -> exp' -> ('env * exp'))
    (ext: 'benv -> bvdef -> ('benv * bvdef))
    (env:'env) (benv:'benv) (k:kind) : ('env * kind) =
  match k with 
    | Kind_star 
    | Kind_affine -> env, k
    | Kind_tcon(k, k') -> 
        let env, k = kind_fold_map f g ext env benv k in
        let env, k' = kind_fold_map f g ext env benv k' in
          env, Kind_tcon(k, k')
    | Kind_dcon(t, k') -> 
        let env, t = typ_fold_map f g ext env benv t in
        let env, k' = kind_fold_map f g ext env benv k' in
          env, Kind_dcon(t, k')
    | Kind_qkind(qs, k) ->
        let env, k = kind_fold_map f g ext env benv k in
          env, Kind_qkind(qs, k)
    | Kind_boxed k -> 
        let env, k = kind_fold_map f g ext env benv k in
          env, Kind_boxed k
    | Kind_unknown -> env, k
            
and typ_fold_map 
    (f: 'env -> 'benv -> typ' -> ('env * typ' * option<kind>))
    (g: 'env -> 'benv -> exp' -> ('env * exp'))
    (ext: 'benv -> bvdef -> ('benv * bvdef))
    (env:'env) (benv:'benv) (t:typ) : ('env * typ) =
  let t_top = compress t in
  let t = t_top.v in
  let env', t', k = match t with 
    | Typ_uvar (_,k) -> 
        let env, t', kopt = f env benv t in
          (match t' with 
             | Typ_uvar(uv, k) -> 
                 let env, k = kind_fold_map f g ext env benv k in
                   env, Typ_uvar(uv, k), k                
             | _ -> 
                 let k = match kopt with 
                    | None -> t_top.sort 
                    | Some k -> k in
                 let env, k = kind_fold_map f g ext env benv k in
                   env, t', k)
    | Typ_btvar _ 
    | Typ_const _ 
    | Typ_unknown -> 
        let env', t', kopt = f env benv t in
          (match kopt with 
             | None -> (*                  let env', k = kind_fold_map f g ext env benv t_top.sort in *)
                 env', t', t_top.sort
             | Some k -> env', t', k)
    | _ -> 
        let env', t' =  match t with 
          | Typ_record fnt_l -> 
              let env', fnt_l' = List.fold_left 
                (fun (env',out) (fn, t) -> 
                   let env',t' = typ_fold_map f g ext env benv t in
                     env', (fn,t')::out) (env, []) fnt_l in
                env', Typ_record (List.rev fnt_l')
                  
          | Typ_fun (nopt, t1, t2) -> 
              let env', t1' = typ_fold_map f g ext env benv t1 in
              let benv', nopt'  = match nopt with 
                | None -> benv, nopt
                | Some bvd -> 
                    let benv, bvd' = ext benv bvd in
                      benv, Some bvd' in
              let env', t2' = typ_fold_map f g ext env' benv' t2 in
                env', Typ_fun(nopt', t1', t2')
                  
          | Typ_univ (bvd, k, t) -> 
              let benv', bvd' = ext benv bvd in
              let env', t' = typ_fold_map f g ext env benv' t in
              let env', k' = kind_fold_map f g ext env benv' k in
                env', Typ_univ(bvd', k', t')
                  
          | Typ_dtuple nt_l -> 
              let env',_, nt_l' = List.fold_left 
                (fun (env, benv, out) (n, t) -> 
                   let env', t' = typ_fold_map f g ext env benv t in
                   let benv', n' = match n with 
                       None -> benv, n
                     | Some bvd -> let benv', n' = ext benv bvd in benv', Some n' in
                     env', benv', (n',t')::out) (env, benv, []) nt_l in
                env', Typ_dtuple (List.rev nt_l')
                  
          | Typ_refine (bvd, t, form) -> 
              let env', t' = typ_fold_map f g ext env benv t in
              let benv', bvd' = ext benv bvd in
              let env'', form' = typ_fold_map f g ext env' benv' form in
                env'', Typ_refine(bvd', t', form')

          | Typ_app(t1, t2) -> 
              let env', t1' = typ_fold_map f g ext env benv t1 in
              let env'', t2' = typ_fold_map f g ext env' benv t2 in
                env'', Typ_app(t1', t2')
                  
          | Typ_dep (t, e) -> 
              let env', t' = typ_fold_map f g ext env benv t in
              let env'', e' = exp_fold_map f g ext env' benv e in
                env'', Typ_dep(t', e')

          | Typ_lam(x, t, t') ->
              let env', t = typ_fold_map f g ext env benv t in
              let benv', bvd' = ext benv x in
              let env'', t' = typ_fold_map f g ext env' benv' t' in
                env'', Typ_lam(x, t, t')
                  
          | Typ_affine t -> 
              let env', t' = typ_fold_map f g ext env benv t in
                env', Typ_affine t' in
        let env', k = kind_fold_map f g ext env benv t_top.sort in
          env', t', k
  in
    env', withinfo t' k t_top.p

and exp_fold_map 
    (f: 'env -> 'benv -> typ' -> ('env * typ' * option<kind>))
    (g: 'env -> 'benv -> exp' -> ('env * exp'))
    (ext: 'benv -> bvdef -> ('benv * bvdef)) 
    (env:'env) (benv:'benv) (e:exp) : ('env * exp) = 
  let env', exp' = match e.v with
    | Exp_bot 
    | Exp_bvar _ 
    | Exp_fvar _
    | Exp_constant _ -> g env benv e.v
    | Exp_constr_app (v, tl, el1, el) -> (* change v to an expression instead of a var? *)
        let env1, _ = g env benv (Exp_fvar v) in
        let env', tlrev = List.fold_left 
          (fun (env', out) t -> 
             let env', t' = typ_fold_map f g ext env' benv t in
               env', t'::out) (env1, []) tl in
        let env', el1rev = List.fold_left 
          (fun (env', out) e -> 
             let env', e' = exp_fold_map f g ext env' benv e in
               env', e'::out) (env', []) el1 in
        let env', elrev = List.fold_left 
          (fun (env', out) e -> 
             let env', e' = exp_fold_map f g ext env' benv e in
               env', e'::out) (env', []) el in
          env', Exp_constr_app(v, List.rev tlrev, List.rev el1rev, List.rev elrev)

    | Exp_primop(op, el) -> 
        let env', elrev = List.fold_left 
          (fun (env', out) e -> 
             let env', e' = exp_fold_map f g ext env' benv e in
               env', e'::out) (env, []) el in
          env', Exp_primop(op, List.rev elrev)
            
    | Exp_abs (bvd, t, e) ->
        let env', t' = typ_fold_map f g ext env benv t in
        let benv', bvd' = ext benv bvd in
        let env'', e' = exp_fold_map f g ext env' benv' e in
          env'', Exp_abs(bvd', t', e')
             
    | Exp_tabs(bvd, k, e) -> 
        let benv', bvd' = ext benv bvd in
        let env', e' = exp_fold_map f g ext env benv' e in
          env', Exp_tabs(bvd', k, e')

    | Exp_app (e1, e2) -> 
        let env', e1' = exp_fold_map f g ext env benv e1 in
        let env'', e2' = exp_fold_map f g ext env' benv e2 in
          env'', Exp_app(e1', e2')
            
    | Exp_tapp (e, t) -> 
        let env', e' = exp_fold_map f g ext env benv e in
        let env'', t' = typ_fold_map f g ext env' benv t in
          env'', Exp_tapp(e',t')
            
    | Exp_match(e, eqns, def) -> 
        let env', e' = exp_fold_map f g ext env benv e in
        let env'', eqnsrev = List.fold_left 
          (fun (env', eqns) (p, e) -> 
             let env', benv', p' = match p with 
(*                | Pat_constant _ -> env', benv, p *)
               | Pat_variant (lid, tl, phantoms, bvl) -> 
                   let env', tl' = List.fold_left (fun (env', out) t -> 
                                                     let env', t' = typ_fold_map f g ext env' benv t in
                                                       env', t'::out) (env', []) tl in
                   let env', phantoms' = List.fold_left (fun (env', out) e -> 
                                                           let env', e' = exp_fold_map f g ext env' benv e in
                                                             env', e'::out) (env', []) phantoms in
                   let env', benv', bvl' = 
                     List.fold_left (fun (env', benv, bvl) bv -> 
                                       let env', bvt' = typ_fold_map f g ext env' benv bv.sort in
                                       let benv', bvd' = ext benv (bvar_to_bvd bv) in
                                         env', benv', (bvd_to_bvar_s bvd' bvt')::bvl) (env', benv, []) bvl in
                     env', benv', Pat_variant(lid, List.rev tl', List.rev phantoms', List.rev bvl') in
             let env', e' = exp_fold_map f g ext env' benv' e in
               env', (p,e')::eqns) (env', []) eqns in
        let env'', def' = exp_fold_map f g ext env'' benv def in
          env'', Exp_match(e', List.rev eqnsrev, def')
            
    | Exp_cond (e, e1, e2) -> 
        let env1, e' = exp_fold_map f g ext env benv e in
        let env2, e1' = exp_fold_map f g ext env1 benv e1 in
        let env3, e2' = exp_fold_map f g ext env2 benv e2 in
          env3, Exp_cond(e',e1',e2')
            
    | Exp_recd (lidopt, tl, phantoms, fn_e_l) -> 
        let env', tlrev = List.fold_left 
          (fun (env', out) t -> 
             let env', t' = typ_fold_map f g ext env' benv t in
               env', t'::out) (env, []) tl in
        let env', phantoms_rev = List.fold_left 
          (fun (env', out) e -> 
             let env', e' = exp_fold_map f g ext env' benv e in
               env', e'::out) (env', []) phantoms in
        let env', fn_e_l_rev = List.fold_left 
          (fun (env, out) (fn, e) -> 
             let env', e' = exp_fold_map f g ext env benv e in
               env', (fn,e')::out) (env', []) fn_e_l in
          env', Exp_recd(lidopt, List.rev tlrev, List.rev phantoms_rev, List.rev fn_e_l_rev)
            
    | Exp_proj (e, fn) -> 
        let env', e' = exp_fold_map f g ext env benv e in
          env', Exp_proj(e', fn)
            
    | Exp_ascribed (e,t,ev) -> 
        let env', e' = exp_fold_map f g ext env benv e in
        let env'', t' = typ_fold_map f g ext env' benv t in
        let env'', ev' = List.fold_left
          (fun (env, out) (bv, e) -> 
             let env', bv' = exp_fold_map f g ext env benv bv in
             let env'', e' = exp_fold_map f g ext env' benv e in
               (env'', (bv',e')::out)) (env'', []) ev in
          env'', Exp_ascribed(e',t', List.rev ev')
            
    | Exp_let (false, bvd_t_e_l, e) -> 
        let env', benv', bvd_t_e_l_rev = List.fold_left 
          (fun (env, benv, out) (bv,t,e) -> 
             let env', e' = exp_fold_map f g ext env benv e in
             let env'', t' = typ_fold_map f g ext env' benv t in
             let benv', bv' = ext benv bv in 
               (env', benv', (bv', t', e')::out)) (env,benv,[]) bvd_t_e_l in
        let env'', e' = exp_fold_map f g ext env' benv' e in
          env'', Exp_let(false, List.rev bvd_t_e_l_rev, e')
            
    | Exp_let (true, bvd_t_e_l, e) -> 
        let benv', bvd_t_e_l' = List.fold_left 
          (fun (benv, out) (bv,t,e) -> 
             let benv', bv' = ext benv bv in
               benv', (bv', t, e)::out) (benv,[]) bvd_t_e_l in
        let env', bvd_t_e_l' = List.fold_left 
          (fun (env, out) (bv, t, e) -> 
             let env', t' = typ_fold_map f g ext env benv' t in
             let env'', e' = exp_fold_map f g ext env' benv' e in
               env'', (bv, t', e')::out) (env, []) bvd_t_e_l' in
        let env'', e' = exp_fold_map f g ext env' benv' e in
          env'', Exp_let(true, bvd_t_e_l', e')
            
    | Exp_gvar _ -> raise Impos in
    
  let env'', sort' = typ_fold_map f g ext env' benv e.sort in
    env'', withinfo exp' sort' e.p 


(*********************************************************************************)
      
let rec descend_typ_map 
    (f: 'env -> typ' -> ('env * typ'))
    (g: 'env -> exp' -> ('env * exp'))
    (env:'env) (t:typ) : ('env * typ) =
  let t_top = compress t in
  let env, t = f env t_top.v in
  let env', t' = match t with 
    | Typ_uvar _ 
    | Typ_unknown -> env, t

    | Typ_btvar bv -> 
        let env', bvsort = descend_kind_map f g env bv.sort in
        let bv' = setsort bv bvsort in
          env', Typ_btvar bv'

    | Typ_const fv -> 
        let env', fvsort = descend_kind_map f g env fv.sort in
        let fv' = setsort fv fvsort in
          env', Typ_const fv'

    | Typ_record fnt_l -> 
        let env', fnt_l' = List.fold_left 
          (fun (env',out) (fn, t) -> 
             let env',t' = descend_typ_map f g env t in
               env', (fn,t')::out) (env, []) fnt_l in
          env', Typ_record (List.rev fnt_l')
            
    | Typ_fun (nopt, t1, t2) -> 
        let env', t1' = descend_typ_map f g env t1 in
        let env', t2' = descend_typ_map f g env' t2 in
          env', Typ_fun(nopt, t1', t2')
            
    | Typ_univ (bvd, k, t) -> 
        let env', t' = descend_typ_map f g env t in
          env', Typ_univ(bvd, k, t')
            
    | Typ_dtuple nt_l -> 
        let env', nt_l' = List.fold_left 
          (fun (env, out) (n, t) -> 
             let env', t' = descend_typ_map f g env t in
               env', (n,t')::out) (env, []) nt_l in
          env', Typ_dtuple (List.rev nt_l')
            
    | Typ_refine (bvd, t, form) -> 
        let env', t' = descend_typ_map f g env t in
        let env'', form' = descend_typ_map f g env' form in
          env'', Typ_refine(bvd, t', form')

    | Typ_lam(x, t, t') ->
        let env', t = descend_typ_map f g env t in
        let env'', t' = descend_typ_map f g env' t' in
          env'', Typ_lam(x, t, t')

    | Typ_app(t1, t2) -> 
        let env', t1' = descend_typ_map f g env t1 in
        let env'', t2' = descend_typ_map f g env' t2 in
          env'', Typ_app(t1', t2')
            
    | Typ_dep (t, e) -> 
        let env', t' = descend_typ_map f g env t in
        let env'', e' = descend_exp_map f g env' e in
          env'', Typ_dep(t', e')
            
    | Typ_affine t -> 
        let env', t' = descend_typ_map f g env t in
          env', Typ_affine t' 
  in
  let env'', tsort = descend_kind_map f g env' t_top.sort in
    env'', withsort t' tsort 
      
and descend_exp_map 
    (f: 'env -> typ' -> ('env * typ'))
    (g: 'env -> exp' -> ('env * exp'))
    (env:'env) (e_top:exp) : ('env * exp) = 
  let env, e = g env e_top.v in
  let env', exp' = match e with
    | Exp_bot 
    | Exp_constant _ -> env, e
    | Exp_bvar bv -> 
        let env', sort' = descend_typ_map f g env bv.sort in
          env', Exp_bvar(setsort bv sort')
    | Exp_fvar fv -> 
        let env', sort' = descend_typ_map f g env fv.sort in
          env', Exp_fvar(setsort fv sort')
    | Exp_constr_app (v, tl, el1, el) -> 
        let env, vsort = descend_typ_map f g env v.sort in
        let v = setsort v vsort in
        let env', tlrev = List.fold_left 
          (fun (env, out) t -> 
             let env', t' = descend_typ_map f g env t in
               env', t'::out) (env, []) tl in
        let env', el1rev = List.fold_left 
          (fun (env', out) e -> 
             let env'', e' = descend_exp_map f g env' e in
               env'', e'::out) (env', []) el1 in
        let env'', elrev = List.fold_left 
          (fun (env', out) e -> 
             let env'', e' = descend_exp_map f g env' e in
               env'', e'::out) (env', []) el in
          env'', Exp_constr_app(v, List.rev tlrev, List.rev el1rev, List.rev elrev)
    | Exp_primop(id, el) -> 
        let env', elrev = List.fold_left 
          (fun (env', out) e -> 
             let env'', e' = descend_exp_map f g env' e in
               env'', e'::out) (env, []) el in
          env', Exp_primop(id, List.rev elrev)
    | Exp_abs (bvd, t, e) ->
        let env', t' = descend_typ_map f g env t in
        let env'', e' = descend_exp_map f g env' e in
          env'', Exp_abs(bvd, t', e')
            
    | Exp_tabs(bvd, k, e) -> 
        let env', e' = descend_exp_map f g env e in
          env', Exp_tabs(bvd, k, e')
            
    | Exp_app (e1, e2) -> 
        let env', e1' = descend_exp_map f g env e1 in
        let env'', e2' = descend_exp_map f g env' e2 in
          env'', Exp_app(e1', e2')
            
    | Exp_tapp (e, t) -> 
        let env', e' = descend_exp_map f g env e in
        let env'', t' = descend_typ_map f g env' t in
          env'', Exp_tapp(e',t')
            
    | Exp_match(e, eqns, def) -> 
        let env', e' = descend_exp_map f g env e in
        let env'', eqnsrev = List.fold_left 
          (fun (env', eqns) (p, e) -> 
             let env', p' = match p with 
                 (*                | Pat_constant _ -> env', p *)
               | Pat_variant (lid, tl, phantoms, bvl) -> 
                   let env', tl' = List.fold_left (fun (env', out) t -> 
                                                     let env', t' = descend_typ_map f g env' t in
                                                       env', t'::out) (env', []) tl in
                   let env', phantoms' = List.fold_left (fun (env', out) e -> 
                                                           let env', e' = descend_exp_map f g env' e in
                                                             env', e'::out) (env', []) phantoms in
                   let env', bvl' = 
                     List.fold_left (fun (env', bvl) bv -> 
                                       let env', bvt' = descend_typ_map f g env' bv.sort in
                                         env', (setsort bv bvt')::bvl) (env', []) bvl in
                     env', Pat_variant(lid, List.rev tl', List.rev phantoms', List.rev bvl') in
             let env', e' = descend_exp_map f g env' e in
               env', (p',e')::eqns) (env', []) eqns in
        let env'', def' = descend_exp_map f g env'' def in
          env'', Exp_match(e', List.rev eqnsrev, def')
            
    | Exp_cond (e, e1, e2) -> 
        let env1, e' = descend_exp_map f g env e in
        let env2, e1' = descend_exp_map f g env1 e1 in
        let env3, e2' = descend_exp_map f g env2 e2 in
          env3, Exp_cond(e',e1',e2')
            
    | Exp_recd(lidopt, tl, el1, fn_e_l) -> 
        let env', tlrev = List.fold_left 
          (fun (env, out) t -> 
             let env', t' = descend_typ_map f g env t in
               env', t'::out) (env, []) tl in
        let env', el1rev = List.fold_left 
          (fun (env', out) e -> 
             let env'', e' = descend_exp_map f g env' e in
               env'', e'::out) (env', []) el1 in
        let env', fn_e_l_rev = List.fold_left 
          (fun (env, out) (fn, e) -> 
             let env', e' = descend_exp_map f g env e in
               env', (fn,e')::out) (env', []) fn_e_l in
          env', Exp_recd(lidopt, List.rev tlrev, List.rev el1rev, List.rev fn_e_l_rev)
            
    | Exp_proj (e, fn) -> 
        let env', e' = descend_exp_map f g env e in
          env', Exp_proj(e', fn)
            
    | Exp_ascribed (e,t,ev) -> 
        let env', e' = descend_exp_map f g env e in
        let env'', t' = descend_typ_map f g env' t in
        let env'', ev' =  List.fold_left 
          (fun (env, out) (bv, e) -> 
             let env', bv' = descend_exp_map f g env bv in
             let env'', e' = descend_exp_map f g env' e in
               (env'', (bv',e')::out)) (env'',[]) ev in
          env'', Exp_ascribed(e',t', List.rev ev')
            
    | Exp_let (b, bvd_t_e_l, e) -> 
        let env', bvd_t_e_l_rev = List.fold_left 
          (fun (env, out) (bv,t,e) -> 
             let env', e' = descend_exp_map f g env e in
             let env'', t' = descend_typ_map f g env' t in
               (env', (bv, t', e')::out)) (env,[]) bvd_t_e_l in
        let env'', e' = descend_exp_map f g env' e in
          env'', Exp_let(b, List.rev bvd_t_e_l_rev, e')
            
    | Exp_gvar _ -> raise Impos in
  let env'', sort' = descend_typ_map f g env' e_top.sort in
    env'', withinfo exp' sort' e_top.p 

and descend_kind_map 
    (f: 'env -> typ' -> ('env * typ'))
    (g: 'env -> exp' -> ('env * exp'))
    (env:'env) (k_top:kind) : ('env * kind) = match k_top with
      | Kind_boxed _
      | Kind_star 
      | Kind_affine -> env, k_top
      | Kind_tcon(k, k') -> 
          let env', k' = descend_kind_map f g env k' in
            env', Kind_tcon(k, k')
      | Kind_dcon(t, k) -> 
          let env', t' = descend_typ_map f g env t in
          let env'', k' = descend_kind_map f g env k in
            env'', Kind_dcon(t', k')
      | Kind_qkind(qs, k) -> 
          let env', k' = descend_kind_map f g env k in
            env', Kind_qkind(qs, k')
      | _ -> raise Impos
          
let range_of_sigelt = function              
  | Sig_tycon_kind (lid, _, _, _)    
  | Sig_typ_abbrev  (lid, _, _, _)
  | Sig_record_typ  (lid, _, _, _)
  | Sig_datacon_typ (lid, _, _, _, _, _) 
  | Sig_value_decl (lid, _) -> Sugar.range_of_lid lid

let rec_compress t = 
  typ_fold_map 
    (fun () () t -> (), compress' t, None)
    (fun () () x -> (), x) 
    (fun () x -> ((), x)) () () t
    
(* collect unification variables in a type *)
let uvars_in_typ t = 
  let collect_uvars uvs () t = match t with 
    | Typ_uvar (uv, k) -> 
        (match List.tryFind (fun uv' -> Unionfind.equivalent uv' uv) uvs with 
           | Some _ -> uvs, t, None
           | None ->  uv::uvs, t, None)
    | _ -> uvs, t, None in
  let exp_folder_noop env () e = (env,e) in
  let ext_benv_noop () x = (),x in
  let uvs, tsub = typ_fold_map collect_uvars exp_folder_noop ext_benv_noop [] () t in
    uvs

let rec uvars_in_uvar uv = match Unionfind.find uv with 
    Uvar _ -> [uv]
  | Fixed t -> 
      let uvt = uvars_in_typ t in
      let uvs = List.concat (List.map uvars_in_uvar uvt) in
        uv::uvs

(* all bvar names in a kind are expected to be unique *)
let substitute_kind kind (bvd:bvdef) with_t : kind =
  let bvd_rn = (snd bvd).idText in
  let fold_t () () t = match t with
      Typ_btvar bv when (bvar_real_name bv).idText = bvd_rn -> (), with_t.v, Some (with_t.sort)
    | _ -> (), t, None in
  let fold_e () () e = (), e in
    snd (kind_fold_map fold_t fold_e (fun () x -> ((),x)) () () kind)

let substitute_kind_l kind (smap:list<bvdef * typ>)  : kind =
  let fold_t () () t = match t with
      Typ_btvar bv ->
        (try 
           let (_,with_t) = 
             List.find (fun (bvd:bvdef, t) -> (bvar_real_name bv).idText = (real_name bvd).idText ) smap in
             (), with_t.v, Some (with_t.sort)
         with
             _ -> (), t, None)
    | _ -> (), t, None in
  let fold_e () () e = (), e in
    snd (kind_fold_map fold_t fold_e (fun () x -> ((),x)) () () kind) 

(* all bvar names in a type are expected to be unique *)
let substitute typ (bvd:bvdef) with_t : typ =
  let bvd_rn = (snd bvd).idText in
  let fold_t () () t = match t with
      Typ_btvar bv when (bvar_real_name bv).idText = bvd_rn -> (), with_t.v, Some (with_t.sort)
    | _ -> (), t, None in
  let fold_e () () e = (), e in
    snd (typ_fold_map fold_t fold_e (fun () x -> ((),x)) () () typ)

let substitute_l typ (smap:list<bvdef*typ>) : typ = 
  let fold_t () () t = match t with
      Typ_btvar bv -> 
        (try 
           let (_,with_t) = 
             List.find (fun (bvd:bvdef, t) -> (bvar_real_name bv).idText = (real_name bvd).idText ) smap in
             (), with_t.v, Some (with_t.sort)
         with
             _ -> (), t, None)
    | _ -> (), t, None in
  let fold_e () () e = (), e in
    snd (typ_fold_map fold_t fold_e (fun () x -> ((),x)) () () typ)
     
(* all bvar names in a type are expected to be unique *)
let substitute_exp typ (bvd:bvdef) with_e =
  let bvd_rn = (snd bvd).idText in
  let fold_e () () e = match e with
      Exp_bvar bv when (bvar_real_name bv).idText = bvd_rn -> (), (with_e.v)
    | _ -> (), e in
  let fold_t () () t = (), t, None in
    snd (typ_fold_map fold_t fold_e (fun () x -> ((),x)) () () typ)

let substitute_exp_l typ (subst:list<(bvdef*exp)>) = 
  let fold_e () () e = match e with
      Exp_bvar bv -> 
        let bvn = (bvar_real_name bv).idText in
          (match List.tryFind (fun (bvd:bvdef, e) -> (snd bvd).idText=bvn) subst with
               None -> (), e
             | Some (_, e') -> (), e'.v)
    | _ -> (), e in
  let fold_t () () t = (), t, None in
    snd (typ_fold_map fold_t fold_e (fun () x -> ((),x)) () () typ)

let e_substitute_exp_l exp (subst:list<(bvdef*exp)>) = 
  let fold_e () () e = match e with
      Exp_bvar bv -> 
        let bvn = (bvar_real_name bv).idText in
          (match List.tryFind (fun (bvd:bvdef, e) -> (snd bvd).idText=bvn) subst with
               None -> (), e
             | Some (_, e') -> (), e'.v)
    | _ -> (), e in
  let fold_t () () t = (), t, None in
    snd (exp_fold_map fold_t fold_e (fun () x -> ((),x)) () () exp)

let instantiate_tparams t tps (targs:list<typ>) (eargs:list<exp>) = 
  let args = (List.map Inl targs)@(List.map Inr eargs) in
    List.fold_left2 (fun out tp arg -> match tp, arg with
                         Tparam_typ(bvd, k), Inl t -> 
                           substitute out bvd t
                       | Tparam_term(bvd, _), Inr e -> 
                           substitute_exp out bvd e
                       | _ -> raise (Bad "Unexpected instantiation of type parameter")) t tps args
      
let rec open_typ_with_exp typ exp = 
  match typ.v with 
      Typ_fun(None, targ, tret) -> tret
    | Typ_fun(Some bvd, targ, tret) -> substitute_exp tret bvd exp
    | Typ_affine t -> open_typ_with_exp t exp
    | _ -> (trace -1197 "open_typ_with_exp got unexpected typ" typ; raise Impos)

let rec open_typ_with_typ tuniv targ = 
  match tuniv.v with 
    | Typ_univ(bvd, k, t) -> substitute t bvd targ 
    | Typ_affine t -> open_typ_with_typ t targ
    | _ -> (trace -1197 "open_typ_with_typ got unexpected typ" tuniv; raise Impos)
        
let freshen_typ t benv : typ = 
  let map_bv benv (bv:bvar<'a>) = 
    let rn = bvar_real_name bv in
      try 
        let (_,bv') = List.find (fun (bv:bvdef, bv') -> (snd bv).idText = rn.idText) benv in
          bvd_to_bvar_s bv' bv.sort
      with 
        | :? System.Collections.Generic.KeyNotFoundException -> raise Impos in
  let fold_t () benv t = match t with
    | Typ_btvar bv -> let bv' = map_bv benv bv in (), Typ_btvar bv', None
    | _ -> (), t, None in
  let fold_e () benv e = match e with 
      Exp_bvar bv -> let bv' = map_bv benv bv in (), Exp_bvar bv'
    | _ -> (), e in
  let ext benv bvd = 
    let newname = genident None in
    let bvd' = (newname, newname) in
      (bvd,bvd')::benv, bvd' in
    snd (typ_fold_map fold_t fold_e ext () benv t)
        

let rec ascribe e t ev = match e.v with 
    Exp_ascribed (e, _, ev') -> ascribe e t (ev'@ev)
  | _ -> withsort (Exp_ascribed(e, t, ev)) t
      
let rec unascribe e = match e.v with 
  | Exp_ascribed (e, _, _) -> unascribe e 
  | _ -> e

let equalsExp e1 e2 = (* only implemented for non-function values *)
  let rec same_exp e1 e2 = 
    match (unascribe e1).v, (unascribe e2).v with
      | Exp_bvar bv1, Exp_bvar bv2 -> (bvar_real_name bv1).idText = (bvar_real_name bv2).idText
      | Exp_fvar fv1, Exp_fvar fv2 -> (Sugar.text_of_lid fv1.v) = (Sugar.text_of_lid fv2.v) 
      | Exp_constant c1, Exp_constant c2 -> c1 = c2 
      | Exp_constr_app (v1, tl, elp1, el), Exp_constr_app (v2, tl2, elp2, el2) ->  
          let rec same_exp_list = function
            | [] -> true
            | (e1, e2)::tl -> (same_exp e1 e2) && same_exp_list tl in
            ((List.length tl) = (List.length tl2) &&
                (List.length elp1) = (List.length elp2) &&
                (List.length el) = (List.length el2) &&
                (Sugar.lid_equals v1.v v2.v) &&
                (same_exp_list (List.zip (elp1@el) (elp2@el2))))
              
      | Exp_recd (_, _, _, fne_l_1), Exp_recd (_, _, _, fne_l_2) -> 
          let rec same_fn_lists = function
              [] -> true
            | ((f1, e1), (f2,e2))::tl -> 
                (Sugar.lid_equals f1 f2) && (same_exp e1 e2) && (same_fn_lists tl) in
            (List.length fne_l_1 = List.length fne_l_2) &&
              (same_fn_lists (List.zip fne_l_1 fne_l_2))
              
      | _ -> false in    
    same_exp e1 e2


let rec is_dependent t = 
  let eq_bvd_bvar (bvd:bvdef) (xv:bvar<typ>) =
    (real_name bvd).idText = (bvar_real_name xv).idText in
  match t.v with 
  | Typ_dtuple (bvd_t_l) ->
      let rec is_dtuple bvs = function
        | [] -> No
        | (bvd_opt, t)::rest -> 
            let tvs, xvs = freevarsTyp t in
            let isect = Util.intersect bvs xvs eq_bvd_bvar in
              match isect with 
                  [] -> (match bvd_opt with 
                             None -> is_dtuple bvs rest
                           | Some bv -> is_dtuple (bv::bvs) rest)
                | _ -> Yes in
        is_dtuple [] bvd_t_l

  | Typ_fun (None, t, t') -> 
      is_dependent t' 
        
  | Typ_fun (Some bv, t, t') -> 
      (match is_dependent t' with
         | No -> 
             let tvs, xvs = freevarsTyp t' in 
               (match Util.intersect [bv] xvs eq_bvd_bvar with
                    [] -> No
                  | _ -> Yes)
         | x -> x)

  | _ -> Maybe

let depends_on (bvd:bvdef) t = 
  let bvrn = (real_name bvd).idText in
  let _, xvs = freevarsTyp t in
    if List.exists (fun (xv:bvar<typ>) -> (bvar_real_name xv).idText = bvrn) xvs then
      Maybe
    else No
  

let rec whnf t = 
  let t' = 
    match t.v with 
      | Typ_dep(t1, e) -> 
          let t1 = whnf t1 in
            (match t1.v with 
               | Typ_lam(x, t1_a, t1_r) -> 
                   let t1_r' = substitute_exp t1_r x e in 
                     whnf t1_r'
               | _ -> withinfo (Typ_dep(t1, e)) t.sort t.p)
      | _ -> t in
    t'

