#light "off"
// (c) Microsoft Corporation. All rights reserved
module Microsoft.Fine.Desugar 

open Sugar
open Absyn
open AbsynUtils
open Const
open Util
open DesugarEnv
open Profiling 

type range = Range.range

exception Error of string * range 

let dstyp_ctr = new_counter "desugar_typ"
let dsform_ctr = new_counter "desugar_formula"
let dsexp_ctr = new_counter "desugar_exp"
let dsmatch_ctr = new_counter "desugar_match"
let dstdef_ctr = new_counter "desugar_typ_defn"
let dstdecl_ctr = new_counter "desugar_typ_decl"
let dstermdecl_ctr = new_counter "desugar_term_decl"
let peq_by_constr_ctr = new_counter "partitionEqnsByConstructor" 
let peq_mixture_ctr = new_counter "partitionMixture" 
let cra_ctr = new_counter "constructor_rule_applicable" 
let vra_ctr = new_counter "variable_rule_applicable" 
let mra_ctr = new_counter "mixture_rule_applicable" 


exception ErrorEnv of string * range * env

let DEBUG_LEVEL = 10
(* let W t = withsort t Kind_unknown *)

let collect_typ_decls decls = 
  let accum = List.fold_left (fun accum d -> match d with 
                                | Def_extern_val _
                                | Def_extern_typ _
                                | Def_val _ 
                                | Def_open _ 
                                | Def_tycon _ 
                                | Def_assume _ -> d::accum
                                | _ -> accum) [] decls in
    List.rev accum

let var_from_fvar fv = match fv.v with
    Exp_fvar qname -> qname
  | _ -> raise Impos

let lid_from_fvar fv = match fv.v with
    Exp_fvar qname -> 
      qname.v
  | _ -> raise Impos

let ppname_from_bvar bv = match bv.v with
    Exp_bvar bv -> bvar_ppname bv
  | _ -> raise Impos

let get_pat_var = function
    Pat_as(Pat_wild _, v, _, _) -> Some v
  | _ -> None

let rec is_tuple_pat = function
    Pat_tuple _ -> true
  | Pat_paren(p,r) -> is_tuple_pat p
  | _ -> false
      
(* Type definitions used in pattern matching compiler *)
type eqnlhs = list<exp>
type eqnbranch = {patterns:list<Sugar.synpat>;
                  pendingsubs:list<binding>; (* pending substs *)
                  whenclause:option<Sugar.synexpr>;
                  branch:Sugar.synexpr; (* branch *)
                  range:Range.range}
type constreqn = {constrname:Disj<lident,sconst>; nargs:int; argppnames:list<option<ident>>; crange:Range.range; eqnbranches:list<eqnbranch>}
type eqns_t = (int*env*eqnlhs*list<eqnbranch>)
exception ErrorPat of string * eqnlhs * list<eqnbranch>

let is_datacon env = function
  | Expr_id_get x ->
      (match try_lookup_datacon env [x] with 
           Some _ -> true
         | _ -> false)
  | Expr_lid_get (b, lid, r) ->
      (match try_lookup_datacon env lid with
           Some _ -> true
         | _ -> false)
  | _ -> false

let extend_env_with_quants env lid = 
  let env = match try_lookup_val_decl_typ env lid with
    | None -> env
    | Some t -> 
        let rec collect_quants out t = match t.v with 
          | Typ_univ(bvd, k, t) -> collect_quants ((bvd, k)::out) t
          | _ -> out in
        let quants = collect_quants [] t in
          List.fold_left (fun env ((pp, rn), k) -> 
                            push_binding_with_real_name env (DSBinding_typ_var(pp, k)) rn) env quants  in
    env

let is_constr_app env e = match e.v with
    Exp_constr_app(v, tl, _, el) -> Some(v, tl, el)
  | Exp_fvar v -> 
      (match try_lookup_datacon env v.v with 
           Some _ -> Some(v, [], [])
         | _ -> None)
  | _ -> None 
let is_primop e = match e.v with 
  | Exp_primop(v, el) ->  Some(v,el) 
  | Exp_fvar ({v=[v]; sort=_; p=_}) when is_operator v.idText -> Some(v, [])
  | _ -> None 

(* extend type environment with type variables appearing in t *)
let rec collect_tvars_typ env t : (list<bvdef * kind> * env) =
  let rec skind_eq (k1:Absyn.kind) (k2:Absyn.kind) = match k1, k2 with 
    | Kind_unknown, Kind_unknown
    | Kind_star, Kind_star
    | Kind_affine, Kind_affine  -> true
    | Kind_tcon(k11, k12), Kind_tcon(k21, k22) -> skind_eq k11 k21 && skind_eq k12 k22
    | Kind_dcon _, Kind_dcon _ -> false
    | _ -> false in
  let rec ctv bindings env t =
    match t with
      | Type_var (Typar(id,kindOpt), r) ->
          (match try_lookup_typ_var env id with
             | Some tv -> 
                 (match kindOpt, tv.sort with 
                    | Some k1, k2 when not (skind_eq (desugar_kind env k1) k2) -> 
                        let msg = spr "Multiple kind annotations on type variable %s" id.idText in
                          raise (Error (msg, r))
                    | _ -> bindings, env)
             | None -> (* TODO: some of these will be Kind_affine *)
                 let kind = match kindOpt with 
                     None -> Kind_star
                   | Some k -> desugar_kind env k in
                 let b = DSBinding_typ_var (id, kind) in
                 let env', realname = push_local_binding env b in
                   ((id, realname), kind)::bindings, env')
      | Type_anon _
      | Type_term_index _
      | Type_lid _ -> bindings, env
      | Type_app (t, tl, r) ->
          let bindings', env' = ctv bindings env t in
            List.fold_left (fun (bindings', env') t -> ctv bindings' env' t) (bindings',env') tl
      | Type_tuple (xtl, r) ->
          List.fold_left (fun (b,env) (x,t) -> ctv b env t) (bindings, env) xtl
      | Type_fun ((x,t), t', r) ->
          List.fold_left (fun (b,env) t -> ctv b env t) (bindings, env) [t;t']
      | Type_refinement (x, t, f, r) ->
          let b',env' = ctv bindings env t in
          let b'', env'' = ctv_form b' env' f in
            b'', env''
      | Type_affine(t, r) -> ctv bindings env t
      | Type_lam (id,t,t',_) -> 
          let bindings, env = ctv bindings env t in
            ctv bindings env t'
      | Type_formula(f,r) -> ctv_form bindings env f
      | _ -> raise (NYI "Unexpected type in CTV")
  and ctv_form bindings env f : (list<bvdef*kind> * env) =
    match f with
      | Form_atom (Atom_true, _) -> bindings, env
      | Form_atom (Atom_false, _) -> bindings, env
      | Form_atom (Atom_pred t, _) -> ctv bindings env t
      | Form_not (f,_) -> ctv_form bindings env f
      | Form_conj (f1, f2, _)
      | Form_disj (f1, f2, _) 
      | Form_impl (f1, f2, _) -> 
          let b, e = ctv_form bindings env f1 in
            ctv_form b e f2
      | Form_all (x, t, f, _)
      | Form_exists (x, t, f, _) ->
          let b',env' = ctv bindings env t in
            ctv_form b' env' f
  in
    ctv [] env t

and collect_tvars_kind env : (Sugar.kind -> (list<bvdef * kind> * env)) = function
  | Sugar.Kind_star 
  | Sugar.Kind_affine -> [], env
  | Sugar.Kind_tcon (k, k') -> 
      let out, env = collect_tvars_kind env k in
      let out', env = collect_tvars_kind env k' in
        out@out', env
  | Sugar.Kind_dcon(t, k) -> 
      let out, env = collect_tvars_typ env t in
      let out', env = collect_tvars_kind env k in
        out@out', env
  
and desugar_kind env k = 
  let quants, env = collect_tvars_kind env k in 
  let rec aux = function
    | Sugar.Kind_star -> Kind_star
    | Sugar.Kind_affine -> Kind_affine
    | Sugar.Kind_tcon (k, k') -> Kind_tcon (aux k, aux k')
    | Sugar.Kind_dcon (t, k) -> Kind_dcon (desugar_typ env t, aux k) in
  let k' = aux k in 
    match quants with 
      | [] -> k'
      | _ -> Kind_qkind(quants, k')

and desugar_typ env t : Absyn.typ =
  let rec desugar_to_typ_or_exp env t : Disj<typ, (exp*range)> = match t with
    | Type_lid (lid, r) -> (* GOTCHA: Fix parser ambiguity here? *)
        (match try_lookup_typ_name env lid with
             Some t -> Inl t
           | None -> 
               (* The parser cannot distinguish typ params from term
                  params; So, if we can't resolve lid to a type name,
                  try to resolve it as a term param.  GOTCHA: This may
                  be a problem if term variable names can shadow type
                  names. *)
               match try_lookup_lid env lid with
                   Some e -> Inr (e, r)
                 | None -> 
                     raise (ErrorEnv ("Unbound type name or index: "^(text_of_lid lid), r, env)))
    | Type_app (t, tl, r) -> 
        let t = desugar_typ env t in
        let tapp =
          List.fold_left (fun tapp t ->
                            match (desugar_to_typ_or_exp env t) with
                              | Inl t -> Wt (Typ_app (tapp, t)) 
                              | Inr (e,p) ->Wt (Typ_dep (tapp, e))) t tl in
          Inl tapp
            
    | Type_tuple (xtl, r) ->
        let rec desugar_tuple_typ env = function
          | [] -> raise (Error ("Unexpected empty tuple type", r))
          | [(x,t)] -> desugar_typ env t
          | (x1,t1)::tl -> 
              let t1 = desugar_typ env t1 in
              let tup = match x1 with 
                  None -> 
                    let t2 = desugar_tuple_typ env tl in
                      Typ_dtuple([(None, t1); (None, t2)])
                | Some x1 -> 
                    let env', real_x1 = push_local_binding env (DSBinding_var(x1,t1)) in
                    let t2 = desugar_tuple_typ env' tl in
                      Typ_dtuple [(Some(x1, real_x1), t1); (None, t2)] in
                (Wt tup) in
          Inl (desugar_tuple_typ env xtl)
                 
    | Type_fun ((x,t), t', r) ->
        let t = desugar_typ env t in
        let bvd_opt, env' = match x with
          | None -> None, env
          | Some x -> 
              let env', realname = 
                push_local_binding env (DSBinding_var(x, t)) in
                Some((x,realname)), env' in
        let t' = desugar_typ env' t' in
          Inl (Wt (Typ_fun (bvd_opt, t, t')))
            
    | Type_var (Typar(id,_), r) ->
        (match try_lookup_typ_var env id with
             Some t -> Inl t
           | None -> 
               raise (ErrorEnv ("Unbound type variable: "^id.idText, r, env)))
                
    | Type_term_index (e, r) ->
        let e = desugar_expression env e in
          Inr (e, r)
            
    | Type_refinement (x, t, f, r) ->
        let t = desugar_typ env t in
        let env', realname = push_local_binding env (DSBinding_var(x, t)) in
        let f = strip_pf_typ (desugar_formula env' f) in
          Inl (Wt (Typ_refine ((x,realname), t,f)))

    | Type_affine(t,r) -> 
        let t = desugar_typ env t in
          Inl (Wt (Typ_affine t))
            
    | Type_anon(r) ->
        Inl (withinfo Typ_unknown Kind_unknown r)

    | Type_lam(x, t, t',_) -> 
        let t = desugar_typ env t in
        let env, realname = push_local_binding env (DSBinding_var(x, t)) in
        let t' = desugar_typ env t' in
          Inl (Wt (Typ_lam((x, realname), t, t')))

    | Type_formula(f, r) -> 
        let f = strip_pf_typ (desugar_formula env f) in 
          Inl f

    | _ -> raise (NYI "Unknown type constructor")

  in
  let thunk () = 
    let t, (quants, env_with_quants) =  match t with 
      | Type_qtyp(tpars, t, r) ->
          let out = List.fold_left 
            (fun (bindings, qenv) (Typar(id, kopt)) -> 
               (match try_lookup_typ_var qenv id with 
                  | Some _ -> raise (Error ("Declared type variables must be unique; '"^id.idText^ " has already been used", r))
                  | None -> 
                      let kind = match kopt with
                        | None -> Kind_star 
                        | Some k -> desugar_kind qenv k in
                      let b = DSBinding_typ_var (id, kind) in
                      let qenv', realname = push_local_binding qenv b in
                        ((id, realname), kind)::bindings, qenv')) ([], env) tpars in
            t, out 
      | _ -> if generalize env then t, collect_tvars_typ env t else t, ([], env) in
      match desugar_to_typ_or_exp env_with_quants t with
          Inl t ->
            List.fold_left (fun t (bvd, k) -> Wt (Typ_univ (bvd, k, t))) t quants
        | Inr (e,r) -> raise (Error ("Expected type; got expression", r)) in
    profile thunk dstyp_ctr
      
and desugar_formula env ff = 
  let thunk () = match ff with 
    | Form_atom (a,r) ->
        let t = desugar_atom env a in
          wrap_pf_typ t

    | Form_not (f, r) ->
        let t = desugar_formula env f in
          Wt (wrap_pf_typ' (Typ_app(not_typ, t)))

    | Form_conj (f1, f2, r) ->
        let t1 = desugar_formula env f1 in
        let t2 = desugar_formula env f2 in
        let t1' = Wt (Typ_app (and_typ, t1)) in
          Wt (wrap_pf_typ' (Typ_app (t1', t2)))

    | Form_disj (f1, f2, r) ->
        let t1 = desugar_formula env f1 in
        let t2 = desugar_formula env f2 in
        let t1' = Wt (Typ_app (or_typ, t1)) in
          Wt (wrap_pf_typ' (Typ_app (t1', t2)))
            

    | Form_all (x, t, f, r) ->
        let t = desugar_typ env t in
        let env', realname = push_local_binding env (DSBinding_var(x, t)) in
        let f' = desugar_formula env' f in
          Wt (wrap_pf_typ' (Typ_fun (Some(x, realname), t, f')))

    | Form_exists (x, t, f, r) ->
        let t = desugar_typ env t in
        let env', realname = push_local_binding env (DSBinding_var(x, t)) in
        let f = desugar_formula env' f in
          Wt (wrap_pf_typ' (Typ_dtuple ([(Some((x,realname)),t);(None, f)])))

    | Form_impl (f1, f2, r) -> 
        let t1 = desugar_formula env f1 in
        let t2 = desugar_formula env f2 in
        let t1' = Wt (Typ_app (implies_typ, t1)) in
          Wt (wrap_pf_typ' (Typ_app (t1', t2))) in
    profile thunk dsform_ctr

and desugar_atom env = function
  | Atom_true -> true_typ
  | Atom_false -> false_typ
  | Atom_pred t -> desugar_typ env t
    
and desugar_expression env (ee:Sugar.synexpr):  exp = 
  let thunk () = match ee with
    | Expr_id_get x -> 
        (match try_lookup_lid env [x] with
             Some e -> e
           | None -> 
               raise (ErrorEnv ("Unbound variable: "^x.idText, x.idRange, env)))
          
    | Expr_lid_get (b, lid, r) -> 
        (match try_lookup_lid env lid with
             Some e -> e 
           | None -> 
               (try
                  let lid', fieldname = Util.pfx lid in
                  let e = Expr_lid_get(b, lid', r) in
                  let proj = Expr_lvalue_get(e, [fieldname], r) in
                    desugar_expression env proj
                with _ -> 
                  raise (ErrorEnv ("Unbound variable: "^(text_of_lid lid), range_of_lid lid, env))))
                
    | Expr_typed (e,t,r) -> (* TODO: This can also be a pack expression *)
        let e = desugar_expression env e in
        let t = desugar_typ env t in
          withpos (Exp_ascribed (e, t, [])) r
            
    | Expr_const (c, r) ->
        withpos (Exp_constant c) r
          
    | Expr_lvalue_get (e, lid, r) ->
        let e = desugar_expression env e in
        let ltxt = Sugar.text_of_lid lid in
          if ltxt  = "One" then 
            withpos (Exp_proj(e, Const.tuple_proj_one)) r
          else if ltxt = "Two" then
            withpos (Exp_proj(e, Const.tuple_proj_two)) r
          else
            List.fold_left 
              (fun e fn -> 
                 let lid = [fn] in 
                 let fn = match try_lookup_field_name env lid with
                   | Some l -> l
                   | None -> 
                       raise (ErrorEnv ("Unbound field name: "^(text_of_lid lid), range_of_lid lid, env)) in
                   withpos (Exp_proj (e, fn)) r) e lid 
            
    | Expr_paren(e,r) -> (* parenthesized expressions kept in AST to distinguish A.M((x,y)) from A.M(x,y) *) (* TODO: Not sure what diff this makes *)
        let e = desugar_expression env e in
        let r = Range.union_ranges e.p r in
          withpos e.v r

    | Expr_tuple (el, r) ->
        let rec desugar_tuple_exp = function
            [] -> raise Impos
          | [e] -> desugar_expression env e 
          | e::tl -> 
              let e1 = desugar_expression env e in
              let e2 = desugar_tuple_exp tl in
              let c = match try_lookup_lid env Const.tuple_UU_lid with
                | Some c -> c
                | None -> raise Impos in (* tuple_UU_lid should always be in there *)
                withpos (Exp_constr_app (var_from_fvar c, [], [], [e1;e2])) r in
          desugar_tuple_exp el 
            
    | Expr_list (el, r) ->
        let consExp, nilExp = match try_lookup_lid env cons_lid, try_lookup_lid env nil_lid with
            Some c, Some n -> c, n
          | _ -> raise Impos (* Should always be in there *) in
        let consVar, nilVar = match consExp.v, nilExp.v with 
            Exp_fvar cvar, Exp_fvar nvar -> cvar, nvar 
          | _ -> raise Impos in
        let nilExp = withpos (Exp_constr_app (nilVar, [Wt Typ_unknown], [], [])) r in
          List.fold_right
            (fun e listExp ->
               let e = desugar_expression env e in
               let r' = Range.union_ranges e.p r in
                 withpos (Exp_constr_app (consVar, [Wt Typ_unknown], [], [e; listExp])) r')
            el nilExp
            
    | Expr_recd (eOpt, rel, r) ->
        let _ = (match eOpt with
                   | Some e -> raise (NYI "Record 'with' syntax")
                   | _ -> ()) in
        let fields = List.fold_left (fun accum ((path, fn), e) ->
                                       let qname = 
                                         match try_lookup_field_name env (path@[fn]) with
                                             Some q -> q
                                           | _ -> 
                                               raise (ErrorEnv ("Unbound field name: "^(text_of_lid (path@[fn])), fn.idRange, env)) in
                                       let e = desugar_expression env e in
                                         (qname, e)::accum) [] rel in
          withpos (Exp_recd(None, [], [], (List.rev fields))) r
            
    | Expr_lambda (spats, body, r) ->
        let rec spat_as_var_typ = function
          | SPat_as (id,r) ->
              (id, Wt Typ_unknown)
          | SPat_typed (p, t, r) ->
              match (spat_as_var_typ p) with
                | (id, {v=Typ_unknown; sort=_; p=_}) ->
                    let t = desugar_typ env t in
                      (id, t)
                | _ -> raise (Error ("Too many type ascriptions", r)) in
        let spat_as_binding p = 
          let (v,t) = spat_as_var_typ p in 
            DSBinding_var(v, t) in
        let rec spats_as_bindings_topt = function
          | SPats (pl, r) ->
              let bindings = List.map spat_as_binding pl in
                bindings, None
          | SPats_typed (spats, t, r) ->
              let bindings, rtopt = spats_as_bindings_topt spats  in
              let rtopt = match rtopt with 
                | None -> Some (desugar_typ env t) 
                | Some _ -> raise (Error ("Multiple return type ascriptions on a function", r)) in
                bindings, rtopt in
        let bindings, retTypOpt = spats_as_bindings_topt spats in
        let env', bvds_rev = List.fold_left 
          (fun (env, out) ((DSBinding_var(ppname,t)) as b) -> 
             let env', realname = push_local_binding env b  in
               env', ((ppname, realname), t)::out) (env, []) bindings in
        let body = desugar_expression env' body in
        let body = match retTypOpt with
            None -> body
          | Some t -> withpos (Exp_ascribed (body, t, [])) body.p in
          List.fold_left 
            (fun body (bvd, t) -> 
               withpos (Exp_abs (bvd, t, body)) r) body bvds_rev
            
            
    | Expr_match (e, eqns, _, r) -> (* TODO: discard third arg from sugar *)
        profile (fun () -> desugar_match env e eqns r) dsmatch_ctr
          
    | Expr_app (e1, e2, r) -> (* TODO: desugar this differently if e1 is a constructor app *)
        let exp1 = desugar_expression env e1 in
        let exp2 = desugar_expression env e2 in
          (match is_constr_app env exp1 with 
             | Some (v, tl, el) -> (* e1 is a constr app; so produce an Exp_constr_app *)
                 withpos (Exp_constr_app (v, tl, [], el@[exp2])) r
             | _ -> 
                 (match is_primop exp1 with 
                    | None ->  withpos (Exp_app (exp1, exp2)) r
                    | Some(op, args) -> withpos (Exp_primop(op, args@[exp2])) r))
            
    | Expr_tyapp (e1, tl, r) -> (* doubt that the parser will ever produce this *)
        let e1 = desugar_expression env e1 in
        let tl = List.map (desugar_typ env) tl in
          (match is_constr_app env e1 with 
             | Some (v, [], []) -> 
                 withpos (Exp_constr_app(v, tl, [], [])) r
             | _ -> 
                 List.fold_left (fun e t -> withpos (Exp_tapp(e, t)) r) e1 tl)

    | Expr_let (false, _, [Binding(_, pat, BindingExpr(_,e), r)], body, r') when is_tuple_pat pat -> 
        (* special case: unpack expressions; desugared as match expressions *) 
        desugar_expression env (Expr_match(e, [Clause(pat, None, body, r)], false, r'))
          
    | Expr_let (isRec, _, bindings, body, r) -> 
        (* TODO: isUse is unnecessary; remove from Sugar *)
        let bodyenv, bindings = desugar_let_bindings env isRec bindings in 
        let body = desugar_expression bodyenv body in
          withpos (Exp_let(isRec, bindings, body)) r                

    | Expr_seq (maybe_do, e1, e2, r) -> (* TODO: ignore? false indicates "do a then b then return a" *)
        let e1 = desugar_expression env e1 in
        let dummyname = genident (Some r) in
        let env', realname = push_local_binding env (DSBinding_var(dummyname, Wt Typ_unknown)) in
        let e2 = desugar_expression env' e2 in
          withpos (Exp_let (false, [((dummyname, realname), Wt Typ_unknown, e1)], e2)) r
            
    | Expr_cond (e, ethen, Some(eelse), rIftoThen, riftoend) ->
        let e' = Expr_match(e, [Clause(Pat_const (Const_bool true, dummyRange), None, ethen, rIftoThen);
                                Clause(Pat_const (Const_bool false, dummyRange), None, eelse, rIftoThen)],
                            false, riftoend) in
          desugar_expression env e'
            (*       let e = desugar_expression env e in *)
            (*       let ethen = desugar_expression env ethen in *)
            (*       let eelse = desugar_expression env eelse in *)
            (*         withpos (Exp_cond(e, ethen, eelse)) riftoend *)
            
    | Expr_cond (e, ethen, None, rIftoThen, riftoend) ->
        raise (Error ("Every conditional statement requires both a then- and else-branch", riftoend))
  in profile thunk dsexp_ctr
       
and desugar_let_bindings env isRec bindings : (env * list<bvdef*typ*exp>) = 
  match isRec, bindings with 
    | false, [Binding(bk, p, BindingExpr(tyOpt, body), r)] -> 
        let ppname, exp' =
          (match p with
             | Pat_as(Pat_wild _, name, _, _) ->
                 let env = extend_env_with_quants env [name] in
                 let body = desugar_expression env body in
                   (match tyOpt with 
                      | None -> name, body 
                      | Some (t,r) -> 
                          let t = desugar_typ env t in
                          let body = withpos (Exp_ascribed (body, t, [])) (Range.union_ranges r body.p) in
                            name, body)

             | Pat_wild r -> 
                 let name = genident (Some r) in
                 let body = desugar_expression env body in
                   (match tyOpt with 
                      | None -> name, body 
                      | Some (t,r) -> 
                          let t = desugar_typ env t in
                          let body = withpos (Exp_ascribed (body, t, [])) (Range.union_ranges r body.p) in
                            name, body)

             | Pat_lid _ ->
                 desugar_pat_exp_as_lambda env p body tyOpt

             | _ -> raise (NYI "Non variable and application let bindings")
                 (* TODO: this will be the place to introduce unpacks
                    when the pattern is a tuple pat *)) in
                 (*         let bvname = bvar_name bv in  *)
        let binding = DSBinding_var (ppname, Wt Typ_unknown) in
        let env', realname = push_local_binding env binding in 
          env', [((ppname, realname), Wt Typ_unknown, exp')]
            
    | false, (Binding(bk, p, BindingExpr(tyOpt, body), r))::_ -> 
        raise (Error("Unexpected multiple non-recursive let-bindings", r))
          
    | true, bindings -> 
        let env, bvd_bindings_rev = List.fold_left
          (fun (env, out) ((Binding(bk, p, bexp, r)) as b) ->
             match p with
                 Pat_lid([fname], args, r) -> (* TODO: where does the type annot for a let rec come from? Wthat if fname is really a lid?*)
                   let binding = DSBinding_var (fname, Wt Typ_unknown) in
                   let env', realname = push_local_binding env binding in
                     env', ((fname, realname), Wt Typ_unknown, b)::out
               | _ -> raise (NYI "Non application let rec bindings")) (env,[]) bindings in
        let letbindings = List.fold_left
          (fun bindings (bvd, t, (Binding(bk, p, BindingExpr(tyOpt, body), r))) ->
             let _, lam = desugar_pat_exp_as_lambda env p body tyOpt in
               (bvd, t, lam)::bindings) [] bvd_bindings_rev in
          env, letbindings

and desugar_pat_exp_as_lambda env pat exp rettyOpt : (ident * exp) = (* ppname * exp *)
  match pat with
    | Pat_lid (fname, args, r) ->
        let (argbindings, env) =
          List.fold_left
            (fun (bindings, env) pat -> match pat with
               | Pat_typed (Pat_as(Pat_wild _, ppname, _, _), typ, r) ->
                   let typ = desugar_typ env typ in
                   let binding = DSBinding_var(ppname, typ) in
                   let env', realname = push_local_binding env binding in
                     ((ppname, realname), typ, r)::bindings, env'
                   
               | Pat_as(Pat_wild _, ppname, _, r) ->
                   let binding = DSBinding_var(ppname, Wt Typ_unknown) in
                   let env', realname = push_local_binding env binding in
                     ((ppname, realname), Wt Typ_unknown, r)::bindings, env'
                   
               | _ -> raise (Error ("Do not support patterns in formal parameters", r))) ([], env) args in
          (* If there's a val declaration for fname, then gather the type variables 
             in the val decl into the current environment *)
        let env = extend_env_with_quants env fname in 
        let body = desugar_expression env exp in
        let body = match rettyOpt with 
            None -> body 
          | Some (t,r) -> 
              let t = desugar_typ env t in
                withpos (Exp_ascribed (body, t, [])) (Range.union_ranges r (body.p)) in
        let lambda = List.fold_left (fun body (bvd, typ, r) -> 
                                       let body' = Exp_abs (bvd, typ, body) in
                                       let r' = Range.union_ranges r body.p in
                                         withpos body' r') body argbindings in
          (match fname with 
               [fname] -> (fname, lambda)
             | _ -> raise (Error ("Unexpected qualified name in let binding", range_of_lid fname)))
    | _ ->   raise (Error ("Unexpected pattern in a function declaration", range_of_synpat pat))
  
and desugar_match env exp eqns range =
  (* See slpj-book Ch 5: Efficient compilation of pattern matching *)
    
  let rec dmatch _ix env discrims eqns defaultExp : exp =
    let flatten_constr_arg_pats ap = match ap with 
        [Pat_paren(Pat_tuple(plist, _), r)] -> plist
      | _ -> ap in
    let rec partitionEqnsByConstructor (xx: list<eqnbranch>) : list<constreqn> = 
      let thunk () = match xx with 
        | [] -> []
        | cbranch::eqbranches ->
            match cbranch.patterns with 
              | [] -> raise Impos (* applicability check should have occured already *)
              | pat::otherPats -> 
                  let cName, argPats = match pat with
                    | Pat_lid (lid, argPats, r') ->
                        (match try_lookup_lid env lid with
                             Some fvar -> 
                               Inl (lid_from_fvar fvar), flatten_constr_arg_pats argPats
                           | None -> 
                               raise (ErrorEnv ("Unbound variable: "^(text_of_lid lid), range_of_lid lid, env)))
                    | Pat_const (c, r) -> 
                        Inr c, []
                    | _ -> raise Impos (* app check done *) in
                  let cgroup, rest =
                    List.partition (fun eqb -> match List.hd (eqb.patterns), cName with 
                                      | Pat_lid (lid, argPats, r'), Inl cname ->
                                          let qname = 
                                            match try_lookup_lid env lid with 
                                                Some q -> lid_from_fvar q
                                              | None -> raise (ErrorEnv ("Unbound variable: "^(text_of_lid lid), range_of_lid lid, env)) in
                                            qname = cname
                                      | Pat_const (c',r), Inr c when c=c' -> true
                                      | _ -> false) eqbranches in
                  let otherGroups = partitionEqnsByConstructor rest in
                  let nargs = List.length argPats in
                  let argppnames = List.map (function 
                                                 Pat_as (Pat_wild _, id, _, _) -> Some id
                                               | _ -> None) argPats in
                  let cgroup =
                    List.map (fun eqb -> match eqb.patterns, cName with 
                                | Pat_const (c', r)::otherPats, Inr const -> {eqb with patterns=otherPats}
                                | Pat_lid(_, argPats, r')::otherPats, Inl cname -> 
                                    let argPats = flatten_constr_arg_pats argPats in
                                      if List.length argPats <> nargs then
                                        raise (Error ("Inconsistent number of arguments to constructor " 
                                                      ^((Sugar.text_of_path -<- Sugar.path_of_lid) cname)^" in pattern", r'))
                                      else
                                        {eqb with patterns=argPats@otherPats}
                                | _ -> raise Impos) (cbranch::cgroup) in
                  let ceqn = {constrname=cName; nargs=nargs; argppnames=argppnames; crange=cbranch.range; eqnbranches=cgroup} in
                    ceqn::otherGroups
      in
        profile thunk peq_by_constr_ctr
    in


    let partitionMixture (eqbranches:list<eqnbranch>) : list<list<eqnbranch>> =
      let thunk () = 
        let rec phelper groups is_var_chunk eqbranches =
          let gather_group test eqbranches =(*  : list<eqnbranch> = *)
            let rec matching_prefix group eqbranches = (*  : (list<eqnbranch> * list<eqnbranch>) =  *)
              match eqbranches with
                | [] -> group, eqbranches
                | hd::tl -> if test hd then matching_prefix (hd::group) tl else group, eqbranches
            in
            let group, rest = matching_prefix [] eqbranches in
              List.rev group, rest
          in
          let is_eqb_pat_a_var eqb = 
            match eqb.patterns with 
              | (Pat_wild _)::_ 
              | (Pat_as(Pat_wild _, _, _, _))::_ -> true
              | _ -> false
          in
            match eqbranches with
              | [] -> List.rev groups
              | _  ->
                  let test = if is_var_chunk then is_eqb_pat_a_var else (not -<- is_eqb_pat_a_var) in
                  let hdGroup, rest = gather_group test eqbranches in
                    let groups = match hdGroup with
                      | [] -> groups
                      | _ -> hdGroup::groups in
                      phelper groups (not is_var_chunk) rest
        in
          phelper [] true eqbranches
      in
        profile thunk peq_mixture_ctr 
    in
      
    let constructor_rule_applicable eqbs = 
      let thunk () = List.for_all (fun eqb -> match eqb.patterns with 
                                     | [] -> false
                                     | pat::_ -> 
                                         match pat with
                                           | Pat_lid _ 
                                           | Pat_const _ -> true
                                           | _ -> false) eqbs in
        profile thunk cra_ctr in


    let variable_rule_applicable eqbs =       
      let thunk () = List.for_all
        (fun eqb -> match eqb.patterns with 
             [] -> false 
           | pat::tl ->  match pat with 
               | Pat_wild _ 
               | Pat_as(Pat_wild _, _, _, _) -> true
               | _ -> false) eqbs in
        profile thunk vra_ctr in


    let mixture_rule_applicable eqbs = 
      let thunk () = 
        ((not -<- variable_rule_applicable) eqbs) &&
          ((not -<- constructor_rule_applicable) eqbs) &&
          List.for_all (fun eqb -> match eqb.patterns with 
                          | [] -> false
                          | pat::tl -> match pat with
                              | Pat_lid _
                              | Pat_wild _ 
                              | Pat_const _
                              | Pat_as(Pat_wild _, _, _, _) -> true
                              | _ -> false) eqbs in 
        profile thunk mra_ctr in
      
    let mk_inr s eqns = 
      let (_, _, discrims, eqbs) = eqns in
      let _ = match discrims with 
        | [] -> 
            List.for_all (fun eqb -> match eqb.patterns with
                              [] -> true
                            | _ -> raise (NYI s)) eqbs
        | _ -> true in
        Inr eqns in
      
    (* If each eqn E_i begins with a variable x_i,
       we can reduce the match by substituting v_0 the
       discriminant variable for x_i in each branch b_i. *)
    let try_elim_vars (_ix, env, discrims, eqns) : Disj<exp, eqns_t> = 
      match discrims with
          [] -> Inr(_ix, env, discrims, eqns)
        | discrim_hd::discrim_tl ->
            if (variable_rule_applicable eqns) then 
              let eqbs_rev = List.fold_left
                (fun eqbs eqb -> (* (pat::pats, branch, r) eqns -> *)
                   match eqb.patterns with 
                       pat::rest -> 
                         (match pat with 
                            | Pat_wild _ -> 
                                let eqb' = {eqb with patterns=rest} in
                                  eqb'::eqbs
                            | Pat_as(Pat_wild _, x, _,  _) -> 
                                let ppname = ppname_from_bvar discrim_hd in
                                let bps = DSBinding_pending_subst (x, ppname) in
                                let eqb' = 
                                  {eqb with pendingsubs=bps::eqb.pendingsubs;
                                     patterns=rest} in   
                                  eqb'::eqbs
                            | _ -> raise Impos)
                     | _ -> raise Impos) [] eqns in
              let eqbs = List.rev eqbs_rev in
                mk_inr "vars prog" (_ix, env, discrim_tl, eqbs)
            else
              mk_inr "vars no prog" (_ix, env, discrims, eqns) (* no progress *)
    in
      
    let try_elim_constructors (_ix, env, discrims, eqbranches) : Disj<exp, eqns_t> =
      match discrims with
        | [] -> mk_inr "constr np" (_ix, env, discrims, eqbranches) (* no progress *)
        | discrim_hd::discrim_tl ->
            let rec eqnGroups2exp env groups : (Disj<list<pat*exp>, list<sconst*exp>> * exp * Range.range) = match groups with
              | [] -> Inl [], defaultExp, defaultExp.p
              | ceqns::rest ->
                  let newnames = (* List.map (fun n -> ident(n, discrim_hd.p)) *) 
                    (List.map (function Some id -> id | None -> ident (gensym None, discrim_hd.p)) ceqns.argppnames) in
                  let env', vars_bvds_rev =
                    List.fold_left (fun (env, vars) name ->
                                      let env', realname = push_local_binding env (DSBinding_var (name, Wt Typ_unknown)) in
                                      let bvar = withpos (name, realname) (ceqns.crange) in
                                      let v = withpos (Exp_bvar bvar) (ceqns.crange) in
                                        (env', (v, (name, realname))::vars)) (env, []) newnames in
                  let vars, bvds = List.split (List.rev vars_bvds_rev) in
                  let vars = vars@discrim_tl in
                  let thenBranch = dmatch (_ix + 1) env' vars ceqns.eqnbranches defaultExp in
                  let elseBranches, defaultExp, range = eqnGroups2exp env rest in
                  let range = Range.union_ranges ceqns.crange range in
                  let branches = match elseBranches, ceqns.constrname with
                    | Inl lpe, Inl cname -> 
                        let thisPat = Pat_variant(cname, [], [], List.map (fun (pp,rn) -> new_bvar pp rn pp.idRange) bvds) in
                          Inl ((thisPat, thenBranch)::lpe)
                    | Inr lce, Inr const -> Inr ((const, thenBranch)::lce)
                    | Inl [], Inr const -> Inr [const, thenBranch]
                    | _ -> raise (Error ("Mixture of constant and constructor patterns is never well-typed", range)) in
                    branches, defaultExp, range
            in
              if (constructor_rule_applicable eqbranches) then 
                let ceqnGroups = partitionEqnsByConstructor eqbranches in
                let branches, defaultExp, range = eqnGroups2exp env ceqnGroups in
                  match branches with 
                    | Inl pat_branches -> 
                        let matchExp = Exp_match(discrim_hd, pat_branches, defaultExp) in
                          Inl (withpos matchExp range)
                            
                    | Inr [(Const_bool true,  _); (Const_bool false, _)] 
                    | Inr [(Const_bool true,  _)] -> 
                        let thenBranch, elseBranch = match branches with 
                            Inr [(_, thenBranch); (_, elseBranch)] -> thenBranch, elseBranch
                          | Inr [(_, thenBranch)] -> thenBranch, defaultExp in
                        let ifstmt = withpos (Exp_cond(discrim_hd, thenBranch, elseBranch)) range in
                          Inl ifstmt
                        
                    | Inr const_branches -> 
                        let nested_cond = List.fold_right 
                          (fun (const, then_branch) else_branch -> 
                             let discr = match const with
                               | Const_bool true -> discrim_hd
                               | _ ->  
                                   let const = W (Exp_constant const) in
                                   let eq_test = mk_eqtest discrim_hd const in
                                     eq_test in
                             let mk_else x = W (Exp_cond(x, then_branch, else_branch)) in
                               withpos (hoist discr mk_else) range) const_branches defaultExp in
                          Inl (nested_cond)
              else
                mk_inr "constr np 2" (_ix, env, discrims, eqbranches) (* no progress *)
    in
      
    let try_mixture (_ix, env, discrims, eqbranches) : Disj<exp,eqns_t> =
      if (mixture_rule_applicable eqbranches) then 
        (let groups = partitionMixture eqns in
         let matchExp =
           List.fold_right (dmatch (_ix + 1) env discrims) groups defaultExp in
           Inl matchExp)
      else
        mk_inr "mix np" (_ix, env, discrims, eqbranches) (* no progress *)
    in
      
    let try_empty_discrims (_ix, env, discrims, eqbranches) : Disj<exp,eqns_t> = match discrims with
      | hd::tl -> mk_inr "empty np" (_ix, env, discrims, eqbranches) (* no progress *)
      | [] ->
          let allEmptyPats = List.for_all
            (fun eqb -> match eqb.patterns with
               | [] -> true
               | pat::_ -> 
                   raise (Error ("Unexpected extra pattern variables", (range_of_synpat pat)))) eqbranches in
            match eqbranches with
              | [] -> Inl defaultExp
              | eqb::tl -> (* ordering of cases requires taking the first branch *)
                  let env = List.fold_right 
                    (fun psub env ->  (* TODO: Shadow variables between pending sub and target will blow up? *) 
                       let env, _ = push_local_binding env psub in 
                         env) eqb.pendingsubs env in
                  let branch =
                    match eqb.whenclause with 
                        Some wc -> 
                          let wc = desugar_expression env wc in
                          let then_branch = desugar_expression env (eqb.branch) in
                          let else_branch = dmatch (_ix + 1) env [] tl defaultExp in
                            (*                           let eq_test = mk_eqtest wc (W (Exp_constant (Const_bool true))) in *)
                          let range = Range.union_ranges (Range.union_ranges wc.p then_branch.p) else_branch.p in
                            withpos (Exp_cond(wc, then_branch, else_branch)) range 
                      | None -> 
                          desugar_expression env (eqb.branch) in
                    Inl branch
    in

    let try_elim_paren_pats (_ix, env, discrims, eqbranches) : Disj<exp, eqns_t> =
      let eqb' = List.map (fun eqb -> match eqb.patterns with 
                             | pat::tl -> 
                                 (match pat with 
                                    | Pat_paren(p, r) -> 
                                        {eqb with patterns=p::tl}
                                    | _ -> eqb)
                             | _ -> eqb) eqbranches in
        mk_inr "paren" (_ix, env, discrims, eqb')
    in

    let try_elim_tuple_pats (_ix, env, discrims, eqbranches) : Disj<exp, eqns_t> = 
      let eqb' = List.map (fun eqb -> match eqb.patterns with 
                             | pat::tl -> 
                                 (match pat with 
                                    | Pat_tuple(pats, r) -> 
                                        let rec desugar_tuple_pat = function
                                            [] -> raise Impos
                                          | [pat] -> pat
                                          | hd::tl -> 
                                              let pat_tl = desugar_tuple_pat tl in
                                                Pat_lid (Const.tuple_UU_lid, [hd;pat_tl], r) in
                                        let tupleConsPat = desugar_tuple_pat pats in
                                          {eqb with patterns=tupleConsPat::tl}
                                    | _ -> eqb)
                             | _ -> eqb) eqbranches in
        mk_inr "tuple" (_ix, env, discrims, eqb')
    in

    let (-||-) f g x = match g x with
      | Inl result -> Inl result
      | Inr y -> 
(*           if (x <> y) then  *)
(*             print_state y; *)
          f y in
      
    let pipeline =
      (try_mixture -||- try_elim_tuple_pats -||- try_elim_paren_pats -||- 
         try_elim_constructors -||- try_elim_vars -||- try_empty_discrims) in
      
(*       debug "\n Equations on recursive call \n"; *)
(*       print_state (_ix, env, discrims, eqns); *)
      (* Sometimes CBN would be so nice! *)
      (* (pipeline -||- try_mixture -||- try_elim_constructors -||- try_elim_vars -||- try_empty_discrims) *)
      match pipeline (_ix, env, discrims, eqns) with
        | Inl res -> res
        | Inr (_ix, env', discrims', eqns') -> 
            if LanguagePrimitives.PhysicalEquality (env', discrims', eqns') (env, discrims, eqns) then 
              (* no progress *)
              raise (ErrorPat ("Pattern matching compiler entered an infinite loop", discrims, eqns))
            else
              ((* debug "."; *)
               dmatch (_ix + 1) env' discrims' eqns' defaultExp)
  in
    
    (* TODO:
       Desugar list patterns into constructor apps
       Handle tuple patterns
       Deal with pending substitutions
    *)
  let mk_eqnbranch = function
    | Clause (pat, whenclause, branch, r) -> 
        {patterns=[pat]; pendingsubs=[]; whenclause=whenclause; branch=branch; range=r}
  in
  let eqbranches = List.map mk_eqnbranch eqns in 

  let dummyname = genident (Some range) in
  let env', realname = push_local_binding env (DSBinding_var (dummyname, Wt Typ_unknown)) in (* add a binding for the new discriminant var *)
  let newbvar = new_bvar dummyname realname range in
  let discriminant = withpos (Exp_bvar newbvar) range in

  let defaultExp = withpos Exp_bot range in (* type checker can try to prove that this case is unreachable *)
  let matchexp = dmatch 0 env' [discriminant] eqbranches defaultExp in (* do the real work of desugaring the match *)
    
  let exp' = desugar_expression env exp in (* desugar the real discriminant and let-bind it to the new var *)
    
  let letexp = Exp_let (false, [((dummyname, realname), Wt Typ_unknown, exp')], matchexp) in
    withpos letexp range
        

let desugar_union_constr env tycon uctd : typ = match uctd with
  | ConstrFields(fieldDecls) -> raise (NYI "ML-style declarations for variant types")(*  ML-style declaration *)
(*       let desugar_fdecl = function *)
(*         | Field (None, t, _, _) -> *)
(*             desugar_typ env t *)
(*         | Field (_, _, _, r) -> *)
(*             raise (Error ("Unexpected record field identifier in variant constructor argument", r)) *)
(*       in *)
(*       let argTyps = List.map desugar_fdecl fieldDecls in (\* no dependences *\) *)
(*       let final_typ = match tycon with *)
(*         | Sig_tycon_kind (lid, tps, k, _) -> *)
(*             let final_typ = Typ_const (withinfo_s lid k (range_of_lid lid)) in *)
(*               (\* convert tparams into de Bruijn indexed bvars *\) *)
(*             let bvars = *)
(*               List.fold_right (fun tp ((tix, xix), accum) -> match tp with *)
(*                                  | Tparam_typ (id, k) -> *)
(*                                      let btvar = withinfo_s (tix, id) k id.idRange in *)
(*                                        ((tix+1, xix), (Inl (Typ_btvar btvar))::accum) *)
(*                                  | Tparam_term (id, t) -> *)
(*                                      let bvar = withinfo_s (xix, id) t id.idRange in *)
(*                                      let bvarexp = withinfo_s (Exp_bvar bvar) t id.idRange in *)
(*                                        ((tix, xix+1), (Inr bvarexp)::accum)) tps ((0,0),[]) in *)
(*               List.fold_left (fun ft -> function *)
(*                                 | Inl btvar -> *)
(*                                     Typ_app (ft, btvar) *)
(*                                 | Inr bvar -> *)
(*                                     Typ_dep (ft, bvar)) final_typ (snd bvars) *)
(*         | _ -> raise Impos in *)
(*       let ctyp = *)
(*         match argTyps with *)
(*           | [] -> final_typ *)
(*           | [t] -> Typ_fun (None, t, final_typ) *)
(*           | _ -> Typ_fun (None, Typ_dtuple (List.map (fun x -> (None, x)) argTyps), final_typ) (\* TODO: really? Should this be curried? *\) *)
(*       in *)
(*         ctyp *)
  | ConstrType styp -> (* GADT-style declaration *)
      let typ = desugar_typ env styp in
      let rec instance_of plid t = match t.v with
        | Typ_const idt when (Sugar.lid_equals idt.v plid) -> true
        | Typ_app (t, _) 
        | Typ_dep (t, _) -> instance_of plid t
        | _ -> false
      in
      let rec final_typ t = match t.v with
        | Typ_refine(_, t', _) 
        | Typ_fun (_,_,t') 
        | Typ_univ (_,_,t') -> final_typ t'
        | _ -> t in
      let lid, inst = match tycon with
        | Sig_tycon_kind (lid, _, _, _) ->
            lid, instance_of lid (final_typ typ)
        | _ -> raise Impos
      in
        if ((not inst) && (text_of_lid lid <> "Prims.NTuple")) then
          let msg = spr "Constructor does not construct expected type: expected %s, got %s " (text_of_lid lid) (Pretty.strTyp (final_typ typ)) in
            raise (Error (msg, range_of_lid lid))
        else typ
          
let desugar_typ_defn env tycon_lid tycon_kind tparams isprop td priv : signature = 
  let thunk () = match td with
    | TyconCore_repr_hidden (r) ->
        [Sig_tycon_kind (tycon_lid, tparams, tycon_kind, isprop)]
          
    | TyconCore_union (funionConstrDecls, r) ->
        let tyconSig = Sig_tycon_kind (tycon_lid, tparams, tycon_kind, isprop) in
        let env' = push_sigelt env tyconSig in
        let desugar_variant_constr = function
          | UnionConstr(unionCaseId, unionConstrTypeDecl, range) ->
              let qname = match extern_ref env with
                | Some eref -> 
                    (Sugar.eref_to_lid eref)@[unionCaseId]
                      (*                     (match Sugar.eref_to_lid tycon_lid with  *)
                      (*                          Some(dll, classname) ->  *)
                      (*                            let _, tcname = Util.pfx classname in *)
                      (*                            let constrName = ident(tcname.idText^"/_"^unionCaseId, unionCaseId.idRange) in *)
                      (*                              (Sugar.eref_to_lid eref)@[constrName] *)
                      (*                        | _ -> raise Impos) *)
                | None -> (current_module env) @ [unionCaseId] in
                if not (unique_name env qname) then
                  let msg = spr "Data constructor names must be unique; the name %s has already been used" (text_of_lid qname) in
                    raise (Error (msg, r))
                else
                  Sig_datacon_typ (qname, tparams, (desugar_union_constr env' tyconSig unionConstrTypeDecl), 
                                   false, (if priv then Private else Public), Some tycon_lid) 
        in
          tyconSig::(List.map desugar_variant_constr funionConstrDecls)
            
    | TyconCore_recd (fieldDecls, r) -> 
        let fn_t_l = List.map (fun (Field(fieldname, t, r)) -> 
                                 let fn = 
                                   if extern_typ env then 
                                     match extern_ref env with
                                         Some eref -> 
                                           (Sugar.eref_to_lid eref)@[fieldname]
                                       | _ -> raise Impos 
                                   else (current_module env)@[fieldname] in
                                   fn, desugar_typ env t) fieldDecls in
        let record = t_withpos (Typ_record fn_t_l) r in
          [Sig_record_typ (tycon_lid, tparams, tycon_kind, record)]
            
    | TyconCore_abbrev (typ, r) -> 
        let env = set_generalize env false in
        let t = desugar_typ env typ in
          [Sig_typ_abbrev(tycon_lid, tparams, tycon_kind, t)] in
    profile thunk dstdef_ctr
          
      
let desugar_typ_decls env decl : env = 
  let rec dtd_helper env decl : env * signature = match decl with
    | Def_open (id,_) ->
        (* TODO: Fix this to qualify the namespace.
           open Foo
           open Bar   <-- is this open Bar, or open Foo.Bar?
        *)
        push_namespace env id, []
    | Def_tycon (tds, isprop, priv, _) -> (* type t1 = ... and tn = ... *)
        let td_to_sig (env, sigelts) = function
            TyconDefn (Tycon(name, kind, tpdecls, _, _),  tdRepr, r) ->
              let qname = match extern_ref env with
                | Some eref -> (Sugar.eref_to_lid eref)@name 
                | None -> (current_module env) @ name in
                if not (unique_typ_name env qname) then
                  let msg = spr "Type names must be unique; the name %s has already been used" (text_of_lid qname) in
                    raise (Error (msg, r))
                else
                  let env', tparams_rev = List.fold_left 
                    (fun (env, tparams) -> function
                       | TyparDecl (Typar(a,k_opt)) -> (* TODO:record kind in Typar *)
                           let k = match k_opt with
                               None -> Kind_star
                             | Some k -> desugar_kind env k  in
                           let env', realname = push_local_binding env (DSBinding_typ_var(a,k)) in
                             env', (Tparam_typ((a, realname), k)::tparams)
                       | TermparDecl (x, t) ->
                           let t = desugar_typ env t in
                           let env', realname = push_local_binding env (DSBinding_var(x,t)) in
                             env', (Tparam_term((x, realname), t)::tparams)) (env, []) tpdecls in
                  let tparams = List.rev tparams_rev in
                  let kind = desugar_kind env' kind in
                  let ty_data_cons = desugar_typ_defn env' qname kind tparams isprop tdRepr priv in
                  let env' = List.fold_left push_sigelt env ty_data_cons in 
                    (env', sigelts@ty_data_cons) in
        let rec_bindings =
          List.fold_left (fun names -> function
                              TyconDefn (Tycon([name], _, _, _, _),  _, _) -> (* TODO: is name always an ident and not lid? *)
                                (DSBinding_tycon (name, Kind_unknown)::names)) [] tds in
        let recenv = List.fold_left push_rec_binding env rec_bindings in
        let (_, signature) = List.fold_left td_to_sig (recenv, []) tds in
        let tycons, datacons = List.partition (function
                                                 | Sig_tycon_kind _ -> true
                                                 | _ -> false) signature in
(*         let rest, datacons = List.partition (function *)
(*                                                | Sig_datacon_typ _ -> true *)
(*                                                | _ -> false) signature in *)
(*         let recs, rest = List.fold_left (fun (recs, rest) s -> match s with  *)
(*                                            | Sig_tycon_kind _ -> s::recs, rest *)
(*                                            | Sig_record_typ (lid, tps, k, _)  *)
(*                                            | Sig_typ_abbrev (lid, tps, k, _) -> (Sig_tycon_kind (lid, tps, k, false))::recs, s::(rest)) ([],[]) rest in *)
(*         let rest = List.rev rest in *)
(*         let recs = [] in (\*  List.rev recs in *\) *)
(*         let env = List.fold_left push_sigelt env recs in *)
        let env = List.fold_left push_sigelt env tycons in
          List.fold_left push_sigelt env datacons, (tycons@datacons)
            
    | Def_assume (idopt, formula, r) -> 
        let ftyp = desugar_formula env formula in
        let lid = match idopt with 
          | Some id -> (current_module env) @ [id]
          | None -> (current_module env) @ [genident (Some r)] in
          if not (unique_name env lid) then
            let msg = spr "Assumptions must bind unique names; the name %s has already been used" (text_of_lid lid) in
              raise (Error (msg, r))
          else
            let b = Sig_datacon_typ(lid, [], ftyp, true, Public, None) in
              push_sigelt env b, [b]
                
    | Def_val (id, t, r) -> 
        let t = desugar_typ env t in
        let lid = (current_module env) @ [id] in
          if not (unique_name env lid) then
            let msg = spr "Value declarations must bind unique names; the name %s has already been used" id.idText in
              raise (Error (msg, r))
          else
            let b = Sig_value_decl (lid, t) in 
              push_sigelt env b, [b]

    | Def_extern_val(eref, id, t) -> 
        let t = desugar_typ env t in
        let lid = (current_module env)@[id] in
        let b = Sig_extern_value(eref, lid, t) in
          push_sigelt env b, [b]
            
    | Def_extern_typ(eref, td) ->
        let decl' = Def_tycon([td], false, false, dummyRange) in
        let env' = set_extern_typ env eref in
        let _, sigelts = dtd_helper env' decl' in
        let sigelts' = List.map (fun se -> Sig_extern_typ(eref, se)) sigelts in
          List.fold_left push_sigelt env sigelts', sigelts' 
            
    | _ -> raise Impos in
    profile (fun () -> let env, _  = dtd_helper env decl in env) dstdecl_ctr

let desugar_term_decls modul env decls : modul = 
  let thunk () = 
    let modul', _ =  List.fold_left 
      (fun (modul, env) decl -> 
         match decl with 
           | Def_let (isRec, bindings, r) -> 
               let env', bindings = desugar_let_bindings env isRec bindings in
               let modul' = {modul with letbindings=modul.letbindings@[(bindings,isRec)]} in
                 (modul', env')
           | Def_expr (e, r) -> 
               let exp = desugar_expression env e in 
               let modul = 
                 match modul.main with 
                     None -> {modul with main=Some exp}
                   | Some _ -> raise (Error ("Multiple main expressions in module", r)) in
                 (modul, env)
           | _ -> (modul, env)) (modul, env) decls in
      modul' in
    profile thunk dstermdecl_ctr
        
let desugar_mimpl openModules : Sugar.moduleImpl -> modul = function
  | ModuleOrNamespaceImpl (lid, extendsOpt, decls, range) ->
      let env = initial_env openModules lid in
      let env = List.fold_left desugar_typ_decls env (collect_typ_decls decls) in
      let modul = env_as_modul env extendsOpt in 
        desugar_term_decls modul env decls 

type dsenv = list<lident*modul>
let desugar (openModuleList:dsenv) : Sugar.ParsedImplFile -> (dsenv * list<modul>) = function
  | ParsedImplFile modules -> 
      try 
        let oml, out = 
          List.fold_left (fun (oml, out) frag -> 
                            let mimpl =
                              match frag with
                                | AnonTopModuleImpl (decls, r) -> 
                                    ModuleOrNamespaceImpl([], None, decls, r)
                                | NamedTopModuleImpl m -> m in
                            let modul = desugar_mimpl oml mimpl in
                              (modul.name, modul)::oml, modul::out) (openModuleList, []) modules in
          oml, List.rev out
      with
        | Error (m, r) -> 
            Printf.printf "ERROR at %s: %s\n" (Range.string_of_range r) m;
            raise (Failure "Desugaring failed") 
              
        | ErrorEnv (m,r,env) -> 
            if !Util.verbosity > 4 then 
              begin
                debug_dump 4 "Environment:" ();
                dump_env env;
                Printf.printf "\n================================================================================\n";
              end;
            Printf.printf "ERROR at %s: %s\n" (Range.string_of_range r) m;
            raise (Failure "Desugaring failed") 
              
        | ErrorPat (m, lhs, branches) -> 
            if !Util.verbosity > 4 then 
              begin
                Printf.printf "\n================================================================================\n";
                Printf.printf "\nEqn lhs=\n";
                printfn "%A" lhs;
                Printf.printf "\nEqn branches=\n";
                printfn "%A" branches;
                Printf.printf "\n================================================================================\n";
              end;
            Printf.printf "ERROR %s\n" m;
            raise (Failure "Desugaring failed") 
              
        | NYI s -> 
            Printf.printf "\nFeature not yet implemented: %s\n" s;
            raise (Failure "Desugaring failed") 
              
                
                
