#light "off"
// (c) Microsoft Corporation. All rights reserved

module Microsoft.Fine.DesugarEnv

open Sugar
open Absyn 
open AbsynUtils
open Const
open Util
open Profiling

let id_ctr_cum, id_ctr = new_counter "Env.lookup_id", ref ""
let lid_ctr_cum, lid_ctr = new_counter "Env.lookup_lid", ref ""
let dc_ctr_cum, dc_ctr = new_counter "Env.lookup_datacon", ref ""
let fn_ctr_cum, fn_ctr = new_counter "Env.lookup_field_name", ref ""
let vd_ctr_cum, vd_ctr = new_counter "Env.lookup_val_decl_typ", ref ""

let freshen_counters n = 
  id_ctr := (new_counter ("Env.lookup_id "^n));
  lid_ctr := (new_counter ("Env.lookup_lid "^n));
  dc_ctr := (new_counter ("Env.lookup_datacon "^n));
  fn_ctr := (new_counter ("Env.lookup_field_name "^n));
  vd_ctr := (new_counter ("Env.lookup_val_decl_typ "^n))
    
type binding = 
  | DSBinding_typ_var of ident * kind
  | DSBinding_var of ident * typ
  | DSBinding_tycon of ident * kind
  | DSBinding_pending_subst of ident * ident

type env = {
  curmodule: lident;
  modules:list<lident * modul>;  (* already fully type checked modules *)
  open_namespaces: list<lident>; (* fully qualified names, in order of precedence *)

  sigaccum:signature;            (* type declarations being accumulated for the current module *)
  externs:signature;
  localbindings:list<ident * binding>;  (* local name bindings for name resolution, paired with an env-generated unique name *)
  recbindings:list<binding>;    (* names bound by type t1 .. and t2 = .. *)
  generalize_tvars:bool;
  extern_typ:bool;
  extern_ref:option<externref>
}
(* exception Found_var_binding of lident * exp *)
(* exception Found_typ_binding of typ *)
exception Not_found_binding of lident 

let open_modules e = e.modules
let current_module env = env.curmodule
let generalize env = env.generalize_tvars 
let set_generalize env b = {env with generalize_tvars = b}

let newenv n =  {curmodule=n; modules=[]; open_namespaces=[];
                 sigaccum=[]; externs=[];localbindings=[]; recbindings=[]; 
                 generalize_tvars=true; extern_typ=false; extern_ref=None}

let env_as_modul env ext = {name = env.curmodule; extends=ext; pos=(range_of_lid env.curmodule); 
                            signature = List.rev env.sigaccum;
                            letbindings=[]; main=None;exports=[]} 

let set_extern_typ env eref = {env with extern_typ=true; extern_ref=Some eref}
let extern_typ env = env.extern_typ
let extern_ref env = env.extern_ref
let clear_extern_typ env = {env with extern_typ=false; extern_ref=None}

let initial_env open_modules module_lid =
  let open_ns =
    if lid_equals module_lid Const.prims_lid then [] else [Const.prims_lid] in
  let externs = List.concat (List.map (fun (_,modul) -> (List.filter (function
                                                                        | Sig_extern_value _ 
                                                                        | Sig_extern_typ _ -> true
                                                                        | _ -> false) modul.signature)) open_modules) in
  let env = {curmodule=module_lid; modules=open_modules; open_namespaces=open_ns;
             sigaccum=[]; externs=externs; localbindings=[]; recbindings=[]; 
             generalize_tvars=true; extern_typ=false; extern_ref=None} in
    debug_dump 1 "Initial desugaring environment" env;
    freshen_counters (text_of_lid module_lid);
    env

let range_of_binding = function
  | DSBinding_typ_var(id, _)
  | DSBinding_var(id, _) 
  | DSBinding_tycon (id, _) -> id.idRange
  | DSBinding_pending_subst (id, id') -> Range.union_ranges id.idRange id'.idRange

let push_binding_with_real_name env b rn = 
  {env with localbindings=(rn, b)::env.localbindings}
    
let push_local_binding env b = 
  let fresh_name = genident (Some (range_of_binding b)) in
    {env with localbindings=(fresh_name, b)::env.localbindings}, fresh_name
      
let push_rec_binding env b = 
  {env with recbindings=b::env.recbindings}
    
let push_sigelt env s = 
  {env with sigaccum=s::env.sigaccum}
    
let push_namespace env lid = 
  {env with open_namespaces = lid::env.open_namespaces}

let try_lookup_typ_var env (id:ident) =
  let fopt = List.tryFind (fun (fn, b) -> match b with 
                             | DSBinding_typ_var (id',k) -> id.idText=id'.idText 
                             | _ -> false) env.localbindings in
    match fopt with 
      | Some (fn, DSBinding_typ_var (id',k)) -> 
          let t = withsort (Typ_btvar (withinfo (id,fn) k id.idRange)) k in
            Some t
      | _ -> None 


let try_lookup_unqualified_typ_name env (lid:lident) (finder:lident -> sigelt -> option<typ>) : option<typ> = 
  (* GOTCHA: Sometimes the parser will parse a term index as a type index *)
  match lid with 
    | [n] -> 
        (* Unqualified name: try to resolve using, in order, 
           1. rec bindings
           2. sig bindings in current module
           3. each open namespace, in order *)
        let found = find_map env.recbindings 
          (function 
             | DSBinding_tycon (id,k) when id.idText=n.idText -> 
                 let name = match env.extern_ref with 
                     None -> env.curmodule @ [id] 
                   | Some eref -> (Sugar.eref_to_lid eref) @[id] in
                   
                   Some (withsort (Typ_const (withinfo name k n.idRange)) k)
             | _ -> None) in
        let found' = found =^^=> 
          (fun _ -> 
             find_map (env.sigaccum@env.externs) (finder (env.curmodule @ [n]))) in
          found' =^^=>
              (fun _ -> 
                 find_map env.open_namespaces 
                   (fun ns -> 
                      let modul_opt = List.tryFind (fun (ns', _) -> lid_equals ns ns') env.modules in
                        match modul_opt with 
                            None -> None
                          | Some (_,m) -> find_map m.signature (finder (ns@[n]))))
    | _ -> 
        (* Resolve qnames first in the current module, then as a
           relatively qualified name using open namespaces.  Then
           try to resolve as a fully qualified name *)
        let module_name, id = pfx lid in
        let found = if Sugar.lid_equals module_name env.curmodule then
          find_map (env.sigaccum@env.externs) (finder lid) else None in
          found =^^=>
            (fun _ -> 
               find_map (env.open_namespaces @ [[]])
                 (fun ns -> 
                    let qname = ns@lid in
                    let prefix, id = pfx qname in 
                    let modul_opt = List.tryFind (fun (ns', _) -> lid_equals ns' prefix) env.modules in
                      match modul_opt with 
                          None -> None
                        | Some (_,m) -> find_map m.signature (finder qname)))
        
let try_lookup_typ_name env lid = 
  let rec find_in_signature lid = function
    | Sig_tycon_kind (lid',_,k,_) when Sugar.lid_equals lid lid' -> 
        let t = withsort (Typ_const (withinfo lid' k (Sugar.range_of_lid lid'))) k in
          Some t
    | Sig_record_typ(lid',_,_,_)
    | Sig_typ_abbrev (lid',_,_,_) when Sugar.lid_equals lid lid' -> 
        let t = withsort (Typ_const (t_withpos lid' (Sugar.range_of_lid lid'))) Kind_unknown in
          Some t 
    | Sig_extern_typ (eref, s) as ss -> 
        let mname, id = Util.pfx lid in
          if not (Sugar.lid_equals mname env.curmodule) then None
          else
            let lid' = (Sugar.eref_to_lid eref)@[id] in
              find_in_signature lid' s
    | _ -> None
  in
    (* GOTCHA: Sometimes the parser will parse a term index as a type index *)
(*     (if (Sugar.text_of_lid env.curmodule <> "Prims") then *)
(*       print_any ("lookup typ name in env: ", env.externs@env.sigaccum)); *)
    match lid with 
      | [n] -> 
          (match try_lookup_typ_var env n with
             | None -> try_lookup_unqualified_typ_name env lid find_in_signature
             | Some t -> Some t)
      | _ -> 
          try_lookup_unqualified_typ_name env lid find_in_signature 
            
let try_lookup_id env (id:ident) =
  if is_operator id.idText then
    let e = withpos (Exp_fvar(W [id])) id.idRange in
      Some e
  else if (id.idText = "not") then 
    let notlid = lid_of_path ["Prims"; "_dummy_op_Negation"] id.idRange in
    let e = withpos (Exp_fvar(W notlid)) id.idRange in
      Some e
  else if (id.idText = "op_Equality") then 
    let eqlid = lid_of_path ["Prims"; "op_Equality"] id.idRange in
    let e = withpos (Exp_fvar(W eqlid)) id.idRange in
      Some e
  else
    let rec do_find (searchKey:ident) = function
        [] -> None
      | (fn,b)::tl -> 
          match b with 
            | DSBinding_var (id', t) when id'.idText=searchKey.idText -> 
                let v = withinfo (searchKey, fn) t (searchKey.idRange) in
                let exp = withinfo (Exp_bvar v) t searchKey.idRange  in
                  Some exp
            | DSBinding_pending_subst (key, key') when key.idText = searchKey.idText ->
                do_find key' tl 
            | _ -> do_find searchKey tl in
      do_find id env.localbindings
        
let try_lookup_module env path = 
  try
    let (mlid, modul) = 
      List.find (fun (mlid, modul) -> (path_of_lid mlid) = path) env.modules in
      Some modul
  with
      _ -> None

let rec try_lookup_unqualified_name env lid finder = 
  match lid with 
    | [n] -> 
        (* Unqualified name: try to resolve using, in order, 
           1. sig bindings in current module
           2. each open namespace, in order *)
        let found = find_map (env.sigaccum@env.externs) (finder (env.curmodule@lid)) in
          found =^^=>
            (fun _ -> 
               find_map env.open_namespaces
                 (fun ns -> 
                    let ns_path = Sugar.path_of_lid ns in
                    let m = try_lookup_module env ns_path in
                      match m with
                          None -> None
                        | Some m -> 
                            find_map m.signature (finder (ns@lid))))

    | hd::tl when hd=Sugar.extern_id ->
        let _, n = Util.pfx tl in
          try_lookup_unqualified_name env [n] finder
        
    | _ -> 
        (* Resolve qnames first as a relatively qualified name using open namespaces. 
           Then try to resolve as a fully qualified name *)
        find_map (env.open_namespaces@[[]])
          (fun ns -> 
             let qname = ns@lid in
             let prefix, id = pfx qname in 
             let signature = 
               match prefix with 
                 | [] -> Some (env.sigaccum@env.externs)
                 | _ -> 
                     if Sugar.lid_equals env.curmodule prefix then
                       Some (env.sigaccum@env.externs)
                     else
                       let m = try_lookup_module env (path_of_lid prefix) in
                         match m with 
                             None -> None
                           | Some m -> Some m.signature in
               bind_opt signature 
                 (fun s -> find_map s (finder qname)))
          
let try_lookup_lid env (lid:lident) = 
  let ll () = 
    let rec finder r key_lid = function
      | Sig_datacon_typ (lid', tps, t, _, aq, _) when Sugar.lid_equals key_lid lid' -> 
          let e = Exp_fvar (withpos lid' r) in (*  need to figure out instantiations of type parms *)
            Some(key_lid, (withpos e r))
      | Sig_value_decl (lid', t) when Sugar.lid_equals key_lid lid' ->
          let e = Exp_fvar (withpos lid' r) in (*  need to figure out instantiations of type parms *)
            Some (key_lid, (withpos e r))
      | Sig_extern_value (lang, lid', t) when Sugar.lid_equals key_lid lid' -> 
          let e = Exp_fvar (withpos lid' r) in (*  need to figure out instantiations of type parms *)
            Some (key_lid, (withpos e r))
      | Sig_extern_typ (eref, s) as ss-> 
          let mname, id = Util.pfx key_lid in
            if not (Sugar.lid_equals mname env.curmodule) then None
            else
              let lid' = (Sugar.eref_to_lid eref)@[id] in
                finder r lid' s
      | _ -> None  in
      match lid with 
        | [n] -> 
            (match try_lookup_id env n with
               | Some e -> Some e
               | None -> 
                   match try_lookup_unqualified_name env lid (finder n.idRange) with 
                       Some (_, e) -> Some e
                     | _ -> None)
        | _ -> 
            match try_lookup_unqualified_name env lid (finder (range_of_lid lid)) with
                Some (_, e) -> Some e
              | _ -> None in
    profilel ll [!lid_ctr; lid_ctr_cum]
      
let try_lookup_val_decl_typ env (lid:lident) = 
  let f () = 
    let lid = env.curmodule@lid in
    let finder = function
      | Sig_value_decl(lid', t) when Sugar.lid_equals lid' lid -> Some t
      | Sig_extern_value (lang, lid', t) when Sugar.lid_equals lid' lid -> Some t
      | _ -> None in
      find_map (env.sigaccum@env.externs) finder in
    profilel f [!vd_ctr; vd_ctr_cum]
                                                                                      
let try_lookup_datacon env (lid:lident) = 
  let f () = 
    let rec finder r key_lid = function
      | Sig_datacon_typ (lid', tps, t, false, aq, _) when Sugar.lid_equals key_lid lid' -> 
          let e = Exp_fvar (withpos lid' r) in (*  need to figure out instantiations of type parms *)
            Some (key_lid, (withpos e r))
      | Sig_extern_typ (eref, s) as ss -> 
          let mname, id = Util.pfx key_lid in
            if not (Sugar.lid_equals mname env.curmodule) then None
            else
              let lid' = (Sugar.eref_to_lid eref)@[id] in
                finder r lid' s 
      | _ -> None in
      match try_lookup_unqualified_name env lid (finder (range_of_lid lid)) with
        Some (_, e) -> Some e
      | _ -> None in
    profilel f [!dc_ctr; dc_ctr_cum]

let try_lookup_field_name env lid = 
  let f () = 
    let rec finder full_lid = function
      | Sig_record_typ (_, _, _, {v=Typ_record fn_t_l; sort=_; p=_}) -> 
          find_map fn_t_l 
            (fun (fn,t) -> 
               if Sugar.lid_equals full_lid fn then 
                 Some (full_lid, withpos Exp_bot dummyRange)
               else None)

      | Sig_extern_typ(eref, s) -> 
          let mname, id = Util.pfx full_lid in
            if not (Sugar.lid_equals mname env.curmodule) then None
            else
              let lid' = (Sugar.eref_to_lid eref)@[id] in
                finder lid' s
          
      | _ -> None in
      match try_lookup_unqualified_name env lid finder with
          Some(l, _) -> Some l 
        | _ -> None in
    profilel f [!fn_ctr; fn_ctr_cum]
      
let unique_name env lid = 
  match try_lookup_lid env lid with
      None -> true
    | Some _ -> false
        
let unique_typ_name env lid = 
  match try_lookup_typ_name env lid with
      None -> true
    | Some _ -> false

let dump_env (env:env) = printfn "%A" (env.sigaccum@env.externs)
