#light "off"
module Microsoft.Fine.Fine

open Microsoft.FSharp.Compiler
open Microsoft.Fine
open Microsoft.Fine.Lexhelp
open Util
open Getopt
open Profiling 

#if RELEASE
let prims = ref "../lib/prims.f9"
let prooflibvals = ref "../lib/prooflibvals.f9"
#else
let prims = ref "prims.f9" 
let prooflibvals = ref "prooflib/prooflibvals.f9"
#endif

let quiet = ref true
let writePrims = ref false 
let to_dcil = ref false
let cache = ref false
let usecache = ref false
let spl = ref false
let eraseproofs = ref false

let _ = 
  Profiling.new_counter "Typechecking";
  Profiling.new_counter "Translation";
  Profiling.new_counter "ToDcil";
  Profiling.new_counter "DcilChecking"
    
let typechecking_ctr = "Typechecking"
let translation_ctr = "Translation"
let toDcil_ctr = "ToDcil"
let dcilChecking_ctr = "DcilChecking"

let specs = 
[
  ( 'v', "verbosity", None, Some (fun x -> Util.verbosity := int_of_string x; quiet:=false));
  ( noshort, "trace", None, Some (fun x -> Util.verbosity := -(int_of_string x)));
  ( noshort, "disable_solver", Some (fun x -> Solver.DISABLE_SOLVER := true), None);
  ( noshort, "skip_translation", Some (fun x -> Util.skip_trans := true), None);
  ( noshort, "extract_proofs", Some (fun x -> Util.extract_proofs := true), None);
  ( noshort, "typecheck_proofs", Some (fun x -> (Util.extract_proofs := true; Util.typecheck_proofs:=true)), None);
  ( noshort, "to_dcil", Some (fun x -> to_dcil := true), None);
  ( noshort, "loud", Some (fun x -> quiet := false), None);
  ( noshort, "proof_stats", Some (fun x -> AbsynUtils.proof_stats_flag := true), None);
  ( noshort, "genIL", Some (fun x -> genIL := true; to_dcil:=true; ), None);
  ( noshort, "writePrims", Some (fun x -> writePrims := true; genIL:=true;to_dcil:=true), None);
  ( noshort, "skipUnit", Some (fun x -> Util.skipUnit := true), None);
  ( noshort, "profile", Some (fun x -> Profiling.flag := true), None);
  ( noshort, "odir", None, Some (fun x -> Ilgen.outputDir := Some x));
  ( noshort, "silent", Some (fun x -> Util.silent := true), None);
  ( noshort, "prims", None, Some (fun x -> prims := x));
  ( noshort, "prooflibvals", None, Some (fun x -> prooflibvals := x));
  ( noshort, "cache", Some (fun x -> cache := true), None);
  ( noshort, "usecache", Some (fun x -> usecache := true), None);
  ( noshort, "spl", Some (fun x -> spl := true), None);
  ( noshort, "ssdebug", Some (fun x -> SimpleSolver.debug := true), None);
  ( noshort, "ss", Some (fun x -> SimpleSolver.disable := false), None);
  ( noshort, "noss", Some (fun x -> SimpleSolver.disable := true), None);
  ( noshort, "prn", Some (fun x -> Util.print_real_names := true),  None);
  ( noshort, "dump_src", Some (fun x -> Util.dump_src := true),  None);
  ( noshort, "dump_tgt", Some (fun x -> Util.dump_tgt := true),  None);

]
  
let process_args () = 
  let file_list = ref [] in
  let _ = parse_cmdline specs (fun i -> file_list := !file_list @ [i]) in
    !file_list

let parse (filename:string) = 
  let parse_to_sugar lexer (lexbuf:Microsoft.FSharp.Text.Lexing.LexBuffer<char> (*UnicodeLexing.Lexbuf*)) defaultNamespace filename = 
    let lower = String.lowercase filename in
    let impl = Pars.implementationFile lexer lexbuf in 
      impl in
  let lexbuf = Microsoft.FSharp.Text.Lexing.LexBuffer<char>.FromTextReader(new System.IO.StreamReader(filename)) in
  let lightSyntaxStatus = LightSyntaxStatus (false) in
  let warnOnFirstToken = false in
  let lexargs = mkLexargs ((fun () -> "."), filename, [], lightSyntaxStatus,(new Lexhelp.LexResourceManager()), ref []) in
  let sugar = 
    Lexhelp.usingLexbufForParsing (lexbuf,filename,None) 
      (fun lexbuf ->
         let tokenizer:Lexfilter.LexFilter = 
           Lexfilter.create
             warnOnFirstToken
             lightSyntaxStatus
             (Lex.token lexargs true) lexbuf
         in
           parse_to_sugar tokenizer.lexer lexbuf None filename) in
    sugar

let parse_and_tc (tcenv, dsenv) filename = 
  let sugar = parse filename in
  let dsenv', mods = Desugar.desugar dsenv sugar in
  let mods = List.map Anf.anf_module mods in
(*     List.iter (fun (m:Absyn.modul) -> if (Sugar.text_of_lid m.name) = "SLib" then Pretty.printModule m) mods; *)
    let typedMods, env = List.fold_left 
    (fun (typedModules, env) (modul:Absyn.modul) -> 
       let env = Tcenv.set_current_module env modul.name in
       let m = profile (fun () -> Tc.typing_modul env modul) typechecking_ctr in
         (match m with 
              Some m -> 
                let env' = Tcenv.push_module env m in                        
                  m::typedModules, env'
            | None -> raise (Failure ("Type checking failed: " ^filename)))) ([], tcenv) mods in
    List.rev typedMods, (env, dsenv')

let expand_types env mods = 
  let env = Tcenv.clear_current_state env in
    List.map (Expandtyp.expand_typ_abbrevs env) mods 
      
let derefine env mods = 
  if !extract_proofs then 
    let dmods, env = List.fold_left 
      (fun (dmods, env) (modul:Absyn.modul) -> 
         let env = Tcenv.set_current_module env modul.name in
         let dmod = Derefine.derefine_modul env modul in
         let env' = Tcenv.push_module env dmod in                        
           dmod::dmods, env')
      ([], env) mods in
      (env, List.rev dmods)
  else env, mods
      
let translate rest = 
  let _, t_rest = List.fold_left 
    (fun (pfx, out) (m:Absyn.modul) -> 
       if !spl && Pretty.str_of_lident m.name = "ProofLib" then m::pfx, out
       else if (Pretty.str_of_lident m.name <> "ProofLib" &&
                Pretty.str_of_lident m.name <> "Prims") 
            then let m' = Translation.transModule None pfx m in 
                 (m::pfx, m'::out)
            else (pfx, out)) ([],[]) rest in
    List.rev t_rest

let nclasses mods = 
  List.fold_left (fun out (m:Target.tModule)  -> out + Hashtbl.fold (fun name cd n -> n+1) m.decls 0) 0 mods
  
let tc_target (pfx :Target.tModule list) mods = 
    let _ = List.fold_left TargetChecker.checkModule 
            (Target.predefined::(List.map (fun (m:Target.tModule) -> m.decls) pfx))
            mods in
      ()

let read_cache_file () = 
  let is = System.IO.File.OpenRead (".fine.cache") in
  let formatter = new System.Runtime.Serialization.Formatters.Binary.BinaryFormatter() in
  let res = formatter.Deserialize(is) in
  let result = match ((unbox res) : 'a list) with 
    | [(tprelude, tproof, ilenv)] -> tprelude, tproof, ilenv
    | _ -> failwith "deserialization failed: expected one item" in
    is.Close(); result
      
let write_cache_file (tprelude:Target.tModule) (tproof:Target.tModule) ilenv =
  let os = System.IO.File.OpenWrite (".fine.cache") in
  let formatter = new System.Runtime.Serialization.Formatters.Binary.BinaryFormatter() in
    formatter.Serialize(os,box ([(tprelude, tproof, ilenv)]));
    os.Flush(); os.Close()
      
let translate_and_check tcenv dprelude dproof =
  if !usecache then read_cache_file ()
  else
    let tprelude = Translation.transModule (Some tcenv) [] dprelude in
    let tproof = Translation.transModule None [dprelude] dproof in
      if !typecheck_proofs then 
        (let _ = TargetChecker.checkModule [Target.predefined] tprelude in
         let _ = TargetChecker.checkModule [Target.predefined;tprelude.decls] tproof in ());
      let ilenv = Ilgen.initial_env tprelude in
      let _ = if !writePrims then
        (let primsMod = Ilgen.tmod_to_absil ilenv tprelude in
         let _ = Ilgen.write_il tprelude primsMod in
           (*        let _ = PrettyTarget.printModule tprelude in *)
           ()) in
      let ilenv = Ilgen.extend_env ilenv tproof in
      let _ = (if !writePrims then
                 let ilproofmod = Ilgen.tmod_to_absil ilenv tproof in
                 let _ = Ilgen.write_il tproof ilproofmod in ()) in
        if !cache then write_cache_file tprelude tproof ilenv;
        tprelude, tproof, ilenv
          
(* let check_preludes () =  *)
(*   if !usecache then read_cache_file () *)
(*   else *)
(*     let prelude = !prims in *)
(*     let [preludeMod], env =        *)
(*       parse_and_tc (Tcenv.initial_env Const.prims_lid, []) prelude in *)
(*     let _ = Proof.init preludeMod in (\* setup Z3 state *\) *)
(*     let [proofMod], env = parse_and_tc env (!prooflibvals) in *)
(*     let tcenv, _ = env in *)
(*     let [preludeMod;proofMod] = expand_types tcenv [preludeMod; proofMod] in *)
(*     let (denv, [dprelude; dproof]) = derefine (Tcenv.initial_env Const.prims_lid) [preludeMod; proofMod] in *)
(*     let tprelude = Translation.transModule [] dprelude in *)
(*     let tproof = Translation.transModule [dprelude] dproof in *)
(*     let _ = TargetChecker.checkModule [Target.predefined] tprelude in *)
(*     let _ = TargetChecker.checkModule [Target.predefined;tprelude.decls] tproof in  *)
(*     let ilenv = Ilgen.initial_env tprelude in *)
(*     let _ = if !writePrims then  *)
(*       (let primsMod = Ilgen.tmod_to_absil ilenv tprelude in  *)
(*        let _ = Ilgen.write_il tprelude primsMod in *)
(* (\*        let _ = PrettyTarget.printModule tprelude in *\) *)
(*          ()) in *)
(*     let ilenv = Ilgen.extend_env ilenv tproof in *)
(*     let _ = (if !writePrims then  *)
(*                let ilproofmod = Ilgen.tmod_to_absil ilenv tproof in *)
(*                let _ = Ilgen.write_il tproof ilproofmod in ()) in *)
(*       if !cache then write_cache_file env denv tprelude tproof ilenv; *)
(*       env, denv, tprelude, tproof, ilenv *)

let main _ =    
  try 
    (* first compile prelude and prooflibvals *)
    let start_time = System.DateTime.Now in
    let messageWithTime s = 
      let elapsed = System.DateTime.Now - start_time in
        Printf.printf "\n %f: %s" (elapsed.TotalSeconds) s in
    let filenames = process_args () in
    let prelude = !prims in
    let _ = messageWithTime "starting prims" in
    let [preludeMod], env =
      parse_and_tc (Tcenv.initial_env Const.prims_lid, []) prelude in
    let _ = messageWithTime "completed prims, starting prooflib" in
    let _ = Proof.init preludeMod in (* setup Z3 state *)
    let [proofMod], env = parse_and_tc env (!prooflibvals) in
    let _ = messageWithTime "finished prooflib" in
    let prelude_tcenv, _ = env in
    let _ = messageWithTime "starting parser and src checker" in
      (*     Util.mainmodule := true; *)
    let env, fmods = List.fold_left (fun (env, fmods) filename -> 
                                       let fmods', env' = parse_and_tc env filename in
                                         env', fmods@fmods') (env,[]) filenames in
      if !Util.dump_src && !Util.skip_trans then List.iter (fun m -> Pretty.printModule m) fmods; 
      if not !Util.skip_trans then 
        begin
          let [preludeMod;proofMod] = expand_types prelude_tcenv [preludeMod; proofMod] in
          let (denv, [dprelude; dproof]) = derefine (Tcenv.initial_env Const.prims_lid) [preludeMod; proofMod] in
          let main_tcenv, _ = env in
          let fmods = expand_types main_tcenv fmods in
          let dmods = 
            let _ = messageWithTime "starting derefinement" in
            let _, dmods = derefine denv fmods in dmods in
            if !Util.dump_src then List.iter (fun m -> Pretty.printModule m) dmods; 
            if !to_dcil then 
              let tprelude, tproof, ilenv = translate_and_check prelude_tcenv dprelude dproof in
              let _ = messageWithTime "starting translation to DCIL" in
              let tmods = translate dmods in
                if !Util.dump_tgt then List.iter (fun m -> PrettyTarget.printModule m) tmods; 
              let _ = Printf.printf "\nTranslated to %d classes\n" (nclasses tmods) in
              let _ = messageWithTime "starting DCIL checking" in
              let _ = if !typecheck_proofs then tc_target [tprelude;tproof] tmods in
                if !genIL then 
                  begin
                    let _ = messageWithTime "starting writing to CIL" in
                    let ilenv, ilmods = List.fold_left 
                      (fun (env, ilmods) tm -> 
                         let env' = Ilgen.extend_env env tm in
                         let ilmod = Ilgen.tmod_to_absil env' tm in
                           env', (tm, ilmod)::ilmods) (ilenv, []) tmods in
                      List.iter (fun (tmod, ilmod) -> Ilgen.write_il tmod ilmod) (List.rev ilmods)
                  end
        end;
      messageWithTime "finished compilation\n";
      Profiling.print_profile ();
      List.iter (fun (m:Absyn.modul) -> pr "Checked module: %s\n" (Sugar.text_of_lid m.name)) fmods;
  with 
    | Failure msg -> pr "%s\n" msg
    | Sugar.WrappedError(Failure m, r) -> (pr "\n%s\n" m)
    | Sugar.WrappedError(e, r) -> (printfn "%A" e; pr "%s\n" (Range.string_of_range r))
        
      
        
