#light "off"

module Microsoft.Fine.Pretty
open Microsoft.Fine
open Absyn

exception PrettyBad

let pr = Printf.printf
let spr = Printf.sprintf

let is_marked e = ((e.p < (int64 0)) && (e.p > (int64 -1500)))

let strLanguage = function
    FSharp -> "F#"
  | CSharp -> "C#"
  | Fine -> "Fine"
let rec unascribe e = match e.v with 
  | Exp_ascribed (e, _, _) -> unascribe e 
  | _ -> e
let str_of_lident (x:lident) =
  String.concat "." (List.map (fun (y:ident) -> y.idText) x)

let strBvd (bvd:bvdef) = (snd bvd).idText

let rec strTyp tt =
  let tt =  whnf (compress tt) in
    match tt.v with
  | Typ_record(l) ->
      let sl = List.map (fun (li,t) -> spr "%s:%s" (str_of_lident li) (strTyp t)) l in
      spr "{ %s }" (String.concat ", " sl)
  | Typ_dtuple [(nopt, t1); (_,t2)] ->
      let s1 = match nopt with
          None -> strTyp t1
        | Some bvd -> spr "%s:%s" (strBvd bvd) (strTyp t1) in
        spr "(%s * %s)" s1 (strTyp t2)
  | Typ_refine(bvd,t,f) -> 
    let r = (fst bvd).idRange in
    let ident = Sugar.ident("x", r) in
    let x = (ident, ident) in
    let newf = freshen_type f [(bvd, x)] in
    spr "{x:%s | %s}" (strTyp t) (strTyp newf)
  | Typ_affine t -> spr "!(%s)" (strTyp t)
  | Typ_univ((pp,rn),k,t) -> spr "(%s::%s). %s" (if !Util.print_real_names then (rn.idText) else (pp.idText)) (strKind k) (strTyp t)
  | Typ_btvar(bv) -> strBtvar bv
  | Typ_const(x) -> str_of_lident x.v
  | Typ_app(t1,t2) -> spr "%s[%s]" (strTyp t1) (strTyp t2)
  | Typ_fun(None, t1,t2) -> spr "%s -> %s" (strTyp t1) (strTyp t2)
  | Typ_fun(Some ((pp,rn)), t1,t2) -> spr "(%s|%s:%s) -> %s" pp.idText rn.idText (strTyp t1) (strTyp t2)
  | Typ_dep(t,e) -> spr "%s %s" (strTyp t) (strExp e)
  | Typ_unknown -> "Typ_unknown"
  | Typ_uvar(u,_) -> "Typ_uvar"
  | Typ_lam((x,_),t,t') -> spr "fun (%s:%s) -> %s" x.idText (strTyp t) (strTyp t')
  | _ -> "UNKNOWN TYPE"

and freshen_type t benv : typ = 
  let map_bv benv (bv:bvar<'a>) = 
    let rn = bvar_real_name bv in
      try 
        let (_,bv') = List.find (fun (bv:bvdef, bv') -> (snd bv).idText = rn.idText) benv in
          bvd_to_bvar_s bv' bv.sort
      with 
        | :? System.Collections.Generic.KeyNotFoundException -> bv in
  let fold_t () benv t = match t with
    | Typ_btvar bv -> let bv' = map_bv benv bv in (), Typ_btvar bv', None
    | _ -> (), t, None in
  let fold_e () benv e = match e with 
      Exp_bvar bv -> let bv' = map_bv benv bv in (), Exp_bvar bv'
    | _ -> (), e in
  let ext benv bvd = 
    let newname = genident None in
    let bvd' = (newname, newname) in
      (bvd,bvd')::benv, bvd' in
    snd (Absyn.typ_fold_map fold_t fold_e ext () benv t)
and strEv = function
  | [] -> ""
  | ev -> spr "{%s}" (String.concat ";" (List.map (fun (e, e') -> spr "%s=%s" (strExp (unascribe e)) (strExp (unascribe e'))) ev))
and strExp' e = match e.v with
  | Exp_bvar(x) -> strBvar x
  | Exp_fvar(x) -> str_of_lident x.v
  | Exp_app(e1,e2) -> spr "(%s %s)" (strExp e1) (strExp e2)
  | Exp_tabs(bvd, k, e) -> spr "/\ %s.%s"  (strBvd bvd) (strExp e)
  | Exp_tapp (e, t) -> spr "%s [%s]" (strExp e) (strTyp t)
  | Exp_match (e, branches, def) -> 
      let strBranch (Pat_variant(lid, tl, _, xl), exp) = 
        spr "\n\t| (%s %s) -> %s" (Sugar.text_of_lid lid) (String.concat " " (List.map strBvar xl)) (strExp exp) in
      let brstr = String.concat " " (List.map strBranch branches) in
        spr "(match %s with %s \n | _ -> %s)" (strExp e) brstr (strExp def)
  | Exp_cond (e1, e2, e3) -> 
      spr "if %s then %s \n else %s" (strExp e1) (strExp e2) (strExp e3)
  | Exp_recd(_, _, _, l) ->
      let l2 = List.map (fun (fn,e) -> spr "%s=%s" (str_of_lident fn) (strExp e)) l in
      spr "{ %s }" (String.concat ", " l2)
  | Exp_proj(e,fn) -> spr "(%s).%s" (strExp e) (str_of_lident (List.tl fn))
  | Exp_ascribed(e,t,ev) -> spr "((%s) : %s %s)" (strExp e) (strTyp t) (strEv ev)
(*   | Exp_constr_app(dc,ts,_,[]) -> str_of_lident dc.v  *)
  | Exp_constr_app(dc,ts,_,es) -> (*  when List.length es <> 0->  *)
      let s1 = str_of_lident dc.v in
      let s2 = String.concat ", " (List.map strTyp ts) in
      let s3 = String.concat " " (List.map strExp es) in
        spr "(%s [%s] %s)" s1 s2 s3
          (*         spr "(%s %s)" s1 s3 *)
  | Exp_primop(op, el) -> 
      spr "%s(%s)" op.idText (String.concat ", " (List.map strExp  el))
  | Exp_let(b,l,e) ->
      let s1 = if b then "rec " else "" in
      let l2,l3 = List.unzip (List.map (fun (bv:bvdef,_,e) -> strBvd bv, strExp e) l) in
      let s2,s3 = String.concat "," l2, String.concat "," l3 in
      let s4 = strExp e in
      spr "\nlet %s%s = %s in\n %s" s1 s2 s3 s4
  | Exp_abs((_,rn),t,e) -> spr "\%s:%s.  %s" rn.idText (strTyp t) (strExp e)
  | Exp_constant(Sugar.Const_bool b) -> spr "%b" b
  | Exp_constant(c) -> c.ToString()
  | Exp_gvar(i) -> spr "{#%d}" i
  | Exp_bot -> "Exp_bot"
  | Exp_extern_call (_, id, t, ts, es) ->
    spr "ext call %s[%s]%s" (id.idText)
      (String.concat "," (List.map strTyp ts)) 
      (String.concat "," (List.map strExp es))
  | _ -> "UNKNOWN EXPRESSION"

and strExp e = 
  if not (is_marked e) then strExp' e
  else spr "<MARKED-%d> %s </MARKED-%d>" e.p (strExp' e) e.p
    
and strKind = function
  | Kind_boxed k  -> (spr "boxed(%s)" (strKind k))
  | Kind_star -> "*"
  | Kind_affine -> "A"
  | Kind_tcon(k1, k2) -> spr "%s => %s" (strKind k1) (strKind k2)
  | Kind_dcon(t,k) -> spr "%s => %s" (strTyp t) (strKind k)
  | Kind_qkind(qs,k) -> spr "<%s> %s" (String.concat ", " (List.map (fun ((pp, _):bvdef, _) -> pp.idText) qs)) (strKind k)
  | Kind_unknown -> "Kind_unknown"
      
      
and strBvar bv = if !Util.print_real_names then (bvar_real_name bv).idText else (bvar_ppname bv).idText
and strBtvar bv = if !Util.print_real_names then spr "'%s" (bvar_real_name bv).idText  else spr "'%s" (bvar_ppname bv).idText
  
let strTparam = function
  | Tparam_typ((_,x),k) -> spr "%s::%s" x.idText (strKind k)
  | Tparam_term((_,x),t) -> spr "%s:%s" x.idText (strTyp t)

let strTparamList = function
  | [] -> ""
  | l -> spr "<%s>" (String.concat "," (List.map strTparam l))

let rec strSigelt = function
  | Sig_tycon_kind(li,l,k, false) ->
      spr "type %s%s :: %s" (str_of_lident li) (strTparamList l) (strKind k)
  | Sig_tycon_kind(li,l,k, true) ->
      spr "prop %s%s :: %s" (str_of_lident li) (strTparamList l) (strKind k)
  | Sig_record_typ(li,l,_,t) ->
      spr "record typ %s %s %s" (str_of_lident li) (strTparamList l) (strTyp t)
  | Sig_typ_abbrev(li,l,_,t) ->
      spr "typ abbrev %s %s %s" (str_of_lident li) (strTparamList l) (strTyp t)
  | Sig_datacon_typ(li,l,t,_,_,_) ->
      spr "datacon %s%s : %s" (str_of_lident li) (strTparamList l) (strTyp t)
  | Sig_value_decl(li,t) ->
      spr "val %s = %s" (str_of_lident li) (strTyp t)
  | Sig_extern_value (eref, li, t) -> 
      spr "extern val %s.%s : %s" eref.dll (str_of_lident li) (strTyp t)
  | Sig_extern_typ (eref, s) -> 
      spr "extern %s.%s %s" eref.dll (str_of_lident eref.classname) (strSigelt s)
        
let printTyp t = pr "%s" (strTyp t)

let strLet lb = 
  String.concat " " (List.flatten (List.map
                                     (fun (bindings, isRec) -> 
                                        (List.map
                                           (fun ((bvd:bvdef),t,e) -> spr "\n%s = %s\n" (fst bvd).idText (strExp e)) bindings)) lb))
    
let printModule m =
  pr "\n(*************************************************)";
  pr "\n(* module name *)\n";
  pr "%s\n" (str_of_lident m.name);
  (match (m.extends) with
    | None -> ()
    | Some lid -> pr ": %s\n" (str_of_lident lid));
  pr "\n(* signature *)\n";
  List.iter (fun s -> pr "%s\n" (strSigelt s)) m.signature;
  pr "\n(* letbindings *)\n";
  pr "%s" (strLet m.letbindings);
  pr "\n(* exports *)\n";
  List.iter (fun s -> pr "%s\n" (strSigelt s)) m.exports;
  pr "\n(*************************************************)\n"
  

