#light "off"
// (c) Microsoft Corporation. All rights reserved
module Microsoft.Fine.PrettyTarget

(* pretty-printer of the target *)
open Sugar
open Microsoft.Fine.Target

let pr = Printf.printf
let spr = Printf.sprintf

(* ====================== utilities ================= *)
let strIdent id = id
let strLident lid = String.concat "." lid
let strBvdef bvdef = (strIdent(fst bvdef)) ^ "#" ^ (strIdent (snd bvdef))
let strClassName classname = if !Util.verbose_names then spr "<n>%s</n>" (strLident classname) else (strLident classname)
let strFieldName fieldname = fieldname
let strMethodName methodname = methodname
  
let strVisibility (vis: tVisibility) =
  match vis with
    | TVis_public -> "public"
    | TVis_internal -> "internal"

let rec strTvar (tvar : tVar<tKind>) =
  let (id, k) = tvar in spr "%s " (strIdent id)
  (*spr "%s : %s" (strIdent id) (strKind k)*)

and strVvar (vvar: tVar<tType>) =
  let (id, t) = vvar in
  spr "%s" (strIdent id) (*(strType t)*)

and strTname (tname: tName<tKind>) =
  let (lid, k) = tname in
  spr "%s: %s" (strLident lid) (strKind k)

and strVname (vname: tName<tType>) =
  let (lid, ty) = vname in
  spr "%s: %s" (strLident lid) (strType ty)

and strKind (knd: tKind) =
  match knd with
    | TKind_star -> "*"
    | TKind_affine -> "A"              
    | TKind_arrow (t, knd') -> (strType t ) ^ " -> " + (strKind knd')
    | TKind_karrow (k1, k2) -> (strKind k1) ^ " -> " + (strKind k2)

and strType (t: tType): string =
  match t with
    | TType_index u -> spr "!%d" u
    | TType_var tvar -> spr "Tvar( %s )"  (strTvar tvar)
    | TType_name name -> spr "Tname( %s )" (strTname name)
    | TType_fun (bvdopt, t1, t2) ->
      (match bvdopt with
        | None -> spr "%s -> %s" (strType t1) (strType t2)
        | Some bvd -> spr "\%s: %s -> %s" (strBvdef bvd) 
                      (strType t1) (strType t2) )
    | TType_tfun (bvd, kind, tbody) ->
        spr "\%s :: %s. %s" (strBvdef bvd) (strKind kind) (strType tbody)
    | TType_concrete tconcrete -> spr "data %s " (strTypeConcrete tconcrete)
    | TType_dep (tfun, v) -> spr "%s <%s>" (strType tfun) (strVal v)
        (*| TType_singleton v -> spr "Singleton(%s)" (strVal v)*)
    | TType_tapp (tfun, targ) -> spr "%s [%s]" (strType tfun) (strType targ)
    | TType_affine ty -> spr "! %s" (strType ty)

and strTypeConcrete (tconcrete: tTypeConcrete) =
  let (classname, targs, vargs) = tconcrete in
  let strTargs = String.concat "," (List.map strType targs) in
  let strVargs = String.concat "," (List.map strVal vargs) in
    if !Util.verbose_names then 
      spr "<ttc>%s<t>%s</t><v>%s</v></ttc>" (strClassName classname) strTargs strVargs
    else spr "%s[ %s; %s]" (strClassName classname) strTargs strVargs

and strVal v =
  match v with
    | TVal_var vvar -> strVvar vvar
    | TVal_obj (tconcrete, _, vs) ->
      let strVs = String.concat "," (List.map strVal vs) in
      spr "%s(%s)" (strTypeConcrete tconcrete) strVs
    | TVal_constant cons -> (cons.ToString())
    | TVal_ldfld (v, fname) -> spr "%s.%s" (strVal v) fname

let strMethodRef(mname, targs) =
  spr "%s<%s>" mname (String.concat "," (List.map strType targs))

let strPrimop bop =
  match bop with
    | AND -> "&&"
    | OR -> "||"
    | ADD -> "+"
    | SUB -> "-"
    | MUL -> "*"
    | DIV -> "/"

let rec strExp (e: tExp) =
  match e with
    | TExp_val v -> strVal v
    | TExp_name n -> strVname n
    | TExp_ldfld (vobj, fname) -> spr "%s.%s" (strVal vobj) fname
    | TExp_extern_call ((eref:externref), methodref, _, _, vargs) -> 
        let strVargs = String.concat "," (List.map strVal vargs) in
          spr "extern call %s::%s.%s (%s)"
            eref.dll
            (Sugar.text_of_lid eref.classname) 
            (strMethodRef methodref) strVargs 
    | TExp_call (vobj, eargs, mref) ->
      let strEargs = String.concat "," (List.map strVal eargs) in
      spr "call %s.%s(%s)" (strVal vobj) (strMethodRef mref) strEargs 
    | TExp_let (vvar, e1, e2) ->
      spr "\nlet %s = %s \n  in %s" (strVvar vvar) (strExp e1) (strExp e2)
    | TExp_isinst (v, tconcrete_e_list, edefault, t) ->
      let rec prBranches tes =
        match tes with
          | [] -> spr "_ -> %s " (strExp edefault)
          | (cname, ts, vars, e) :: rest ->
            spr "\n    %s<%s, %s> -> %s \n %s"
             (strClassName cname)
             (String.concat "," (List.map strType ts))
             (String.concat "," (List.map strVvar vars))
             (strExp e)
            (prBranches rest) in
      spr "\n %s isinst %s : %s \n" (strVal v) (prBranches tconcrete_e_list) (strType t)
    | TExp_cond(eb, et, ef) ->
      spr "if(%s) then %s else %s" (strExp eb) (strExp et) (strExp ef)
    | TExp_primop(bop, vl) -> 
      spr "%s (%s)" (strPrimop bop) (String.concat ", " (List.map strVal vl))
    | TExp_bot -> "bot"

let strFieldDecl fdecl = spr "%s: %s" (fst fdecl) (strType (snd fdecl))

let strStaticFieldDecl sfdecl =
  let (fname, t, expopt) = sfdecl in
  match expopt with
    | None -> spr "%s: %s" fname (strType t)
    | Some e -> spr "%s: %s = %s" fname (strType t) (strExp e)

let strMethodDecl (mdecl: tMethodDecl) =
  let (retTy, mname, tvars, params, body) = mdecl in
  let strBody =
    match body with
      | None -> ""
      | Some e -> strExp e in
  spr "%s %s<%s>(%s){%s}" 
  (strType retTy)
  mname
  (String.concat "," (List.map strTvar tvars))
  (String.concat "," (List.map strVvar params))
  strBody

let strEvidence evs =
  String.concat "," (List.map (fun (var, v) -> spr "%s -> %s" (strVvar var) (strVal v)) evs)
  
let strClassDecl (cdecl:tClassDecl) =
  let strExtends =
    match cdecl.extends with
      | None -> ""
      | Some tsuper -> ": " ^ (strType tsuper) in
  let strckind = 
    match cdecl.kind with
      | None -> ""
      | Some kind -> " :: " ^ (strKind kind) in
  spr "\n  === %s class name: %s:%s \n    ptvars: %s \n    tvars: %s \n    pvvars: %s \n    vvars: %s \n    evidences: %s \n    extends: %s \n    fields: %s \n    staticFields: %s \n methods: %s :: %s"  
  (strVisibility cdecl.visibility)
  (strClassName cdecl.name) (cdecl.namestring)
  (String.concat "," (List.map strTvar cdecl.ptvars))
  (String.concat "," (List.map strTvar cdecl.tvars))
  (String.concat "," (List.map strVvar cdecl.pvvars))
  (String.concat "," (List.map strVvar cdecl.vvars))
  (strEvidence cdecl.evidences)
  strExtends
  (String.concat "," (List.map strFieldDecl cdecl.fields))
  (String.concat ",\n" (List.map strStaticFieldDecl cdecl.staticFields))
  (String.concat "," (List.map strMethodDecl cdecl.methods))
  strckind

let strLidentOpt lidOpt =
  match lidOpt with
    | None -> ""
    | Some lid -> strLident lid

let printModule (m:tModule) = 
  pr "\n======== module %s: %s ==========" 
     (strLident m.name) (strLidentOpt m.extends);
  Hashtbl.iter (fun cn cdecl -> (pr "\n ====== class decl %s" (strClassDecl cdecl))) (m.decls);
  Hashtbl.iter (fun cn cdecl -> (pr "\n >>>>>> extern class decl %s" (strClassDecl cdecl))) (m.externs);
  pr "\n === module class = %s" (strClassDecl m.modCdecl);
  let _ = match m.entry with 
      None -> ()
    | Some e -> 
        pr "\n======== entry point ==========";
        printfn "%A" e;
        pr "\n%s" (strExp e) in
    pr "\n=========== end of %s ==========" (strLident m.name);
