#light "off"
// (c) Microsoft Corporation. All rights reserved

module Microsoft.Fine.Profiling

module On = struct

exception ProfilingException of string

type counter = {name:string; total_time:System.TimeSpan ref; nstarts:int ref; started_at:System.DateTime ref;}
let max_ctr_len = ref 0
let counters: (string, counter) Hashtbl.t = Hashtbl.create 20
let ts0 = System.TimeSpan(0L)
let t0 = System.DateTime.Now

let find_counter n = 
  try  
    Some (Hashtbl.find counters n)
  with
    | :? System.Collections.Generic.KeyNotFoundException -> None 

let now () = System.DateTime.Now

let new_counter name = 
  match find_counter name with
    | Some _ -> 
(*         let _ = Util.pr "Already found counter %s\n" name in *)
          name 
    | None -> 
(*         let _ = Util.pr "Creating counter %s\n" name in *)
        let ctr = {name=name; total_time=ref ts0; nstarts=ref 0; started_at=ref t0} in
        let _ = Hashtbl.add counters name ctr in
          if String.length name > !max_ctr_len then 
            max_ctr_len := String.length name;
          name
            
let start_counter name = 
  match find_counter name with
    | Some ctr -> 
        incr ctr.nstarts;
        if !(ctr.nstarts) = 1 then ctr.started_at := now()
    | None -> raise (ProfilingException ("Counter " ^name^ " does not exist"))

let stop_counter name =
  match find_counter name with
    | Some ctr -> 
        decr ctr.nstarts;
        if !(ctr.nstarts) < 0 then raise (ProfilingException ("Starts and ends on counter "^name^" don't match up"));
        if !(ctr.nstarts) = 0 then 
          let elapsed = now() - !ctr.started_at in
            ctr.total_time := elapsed + !ctr.total_time
    | None -> raise (ProfilingException ("Counter " ^name^ " does not exist"))

let profile (f: unit -> 'b) (ctrname:string) : 'b = 
  try 
    let _ = start_counter ctrname in
    let result = f () in
    let _ = stop_counter ctrname in
      result
  with 
      e when (stop_counter ctrname; false) -> raise e (*nasty hack to ensure a counter is 
                                                        always stopped without messing up the stack trace *)
let profilel (f: unit -> 'b) (ctrs:list<string>) : 'b = 
  try 
    let _ = List.iter start_counter ctrs in
    let result = f () in
    let _ = List.iter stop_counter ctrs in
      result
  with 
      e when (List.iter stop_counter ctrs; false) -> raise e (*nasty hack to ensure a counter is 
                                                        always stopped without messing up the stack trace *)


let print_profile () = 
  let ctr2string ctr = 
    let n = !max_ctr_len - (String.length ctr.name) in
    let rec mk_pad out n = if n = 0 then out else mk_pad (" " ^ out) (n - 1) in
    let pad = mk_pad "" n in
      Printf.sprintf "%s:%s\t %fsec\n" ctr.name pad (!ctr.total_time).TotalSeconds in
  let all_ctrs = Hashtbl.fold (fun name ctr accum -> ctr::accum) counters [] in
  let all_ctrs = List.sortWith (fun c1 c2 -> if (!c2.total_time).TotalSeconds >= (!c1.total_time).TotalSeconds then 1 else -1) all_ctrs in
  let output = List.fold_left 
    (fun accum c -> 
       if !c.nstarts > 0 then Printf.printf "Warning: counter %s is unreliable; %d starts were not stopped\n" c.name !c.nstarts;
       accum ^ (ctr2string c)) "" all_ctrs in
    Printf.printf "%s" output
end

module Off = struct

let new_counter n = n
let start_counter n = ()
let stop_counter n = ()
let profile (f: unit -> 'b) (ctrname:string) : 'b = f ()
let profilel (f: unit -> 'b) (ctrs:list<string>) : 'b = f ()
let print_profile () = ()

end

let flag = ref false

let new_counter n = On.new_counter n
let start_counter n = if !flag then On.start_counter n else Off.start_counter n
let stop_counter n = if !flag then On.stop_counter n else Off.stop_counter n
let profile  (f: unit -> 'b) (ctrname:string) : 'b = 
  if !flag then On.profile f ctrname else Off.profile f ctrname
let profilel (f: unit -> 'b) (ctrs:list<string>) : 'b = 
  if !flag then On.profilel f ctrs else Off.profilel f ctrs
let print_profile () = 
  if !flag then On.print_profile () else Off.print_profile ()

