(*
  Copyright (c) 2009, Regents of the University of California

  Authors: Ravi Chugh
  
  All rights reserved.
  
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following 
  conditions are met:
  
  1. Redistributions of source code must retain the above copyright 
  notice, this list of conditions and the following disclaimer.

  2. Redistributions in binary form must reproduce the above 
  copyright notice, this list of conditions and the following disclaimer 
  in the documentation and/or other materials provided with the distribution.

  3. Neither the name of the University of California, San Diego, nor 
  the names of its contributors may be used to endorse or promote 
  products derived from this software without specific prior 
  written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*)


#light "off"

module Microsoft.Fine.Proof

open System.Text.RegularExpressions
open Microsoft.Z3
open Microsoft.Fine
open Absyn
open ProofUtils
open ProofCommon
open ProofZZZ
open ProofCombinators

module Id = IdTable



(****************************************************************************)
(* FINE EXPS/TYPS --> Z3 TERMS                                              *)
(****************************************************************************)

(* TODO 10/4 so unfortunate. instead, scrape exp for strings first *)
let (getOnDemandOps : unit -> taggedop list),
    (storeOnDemandOp : taggedop -> unit) =
  let l = ref [] in
  (fun () -> let x = !l in l := []; x),
  (fun o -> l := o :: !l)

let noneExp, noneTyp, noneExpMsg =
  let b = false in (* set to true for debugging *) 
  (fun e -> if b then pr "noneExp: %s\n" (Pretty.strExp e); None),
  (fun t -> if b then pr "noneTyp: %s\n" (Pretty.strTyp t); None),
  (fun e s -> if b then pr "noneExp @ %s: %s\n" s (Pretty.strExp e); None)



type varmap = list<string*int>
let varctr = ref 1000
let get_typinstid (vmap:varmap) (v:ident) : (int * varmap) = 
  match List.try_assoc v.idText vmap with 
      Some id  -> id, vmap
    | None -> 
        let _ = incr varctr in
        let id = !varctr in
          id, ((v.idText,id)::vmap)

let try_lookup_id vmap (ident:ident) = 
  List.try_assoc ident.idText vmap

let rec try_lookup_id_typ vmap t = match t.v with
    Typ_btvar bv -> try_lookup_id vmap (bvar_real_name bv)
  | Typ_refine(_, t, _) -> try_lookup_id_typ vmap t
  | _ -> None

let rec expToTerm vmap (info:info) (e:exp) : term option = match e.v with
  | Exp_ascribed(e',_,_) -> expToTerm vmap info e'
  | Exp_constant(Sugar.Const_bool true) -> Some termTrue
  | Exp_constant(Sugar.Const_bool false) -> Some termFalse
  | Exp_constant(Sugar.Const_string (b,_)) ->
      let s = System.Text.Encoding.Unicode.GetString(b,0,b.Length) in
      let i =
        if IdTable.mem Strings.idTable s then IdTable.getId Strings.idTable s
        else begin
          let i = IdTable.processVal Strings.idTable s in
          storeOnDemandOp (BG, DefFun (symOfString s, [||], stringSort(), None));
          i
        end
      in
      Some (App (symOfString s, [||]))
  | Exp_bvar x ->
      let tiio = match tiioOfTyp (Tcenv.expand_typ_rec info.envn e.sort) with 
        | None -> try_lookup_id_typ vmap e.sort 
        | Some i -> Some i in
      let s = (bvar_real_name x).idText in
      (match findIdx s info.binders with
         | Some i -> Some (BoundV (i, tiio))
         | None   -> Some (FreeV (s, tiio)))
  | Exp_constr_app(dc,[t1;t2],_,[e1;e2]) when Sugar.lid_equals dc.v Const.tuple_UU_lid ->
      let t1 = Tcenv.expand_typ_rec info.envn t1 in
      let t2 = Tcenv.expand_typ_rec info.envn t2 in
      (match tiioOfTi (TITuple (tiOfTyp t1, tiOfTyp t2)) with
         | None -> noneExp e
         | Some tii ->
             (match expToTerm vmap info e1, expToTerm vmap info e2 with
                | Some f1, Some f2 -> Some (App (symOfTuple tii, [|f1;f2|]))
                | _                -> noneExp e))
  | Exp_constr_app(dc,ts,_,es) ->
      let ts = List.map (Tcenv.expand_typ_rec info.envn) ts in
      let tiios = List.map tiioOfTyp ts in
      if existsNone tiios then noneExp e
      else let fos = List.map (expToTerm vmap info) es in
           if existsNone fos then noneExp e
           else let fs = projSomes fos in 
                let tiis = projSomes tiios in
                let di = IdTable.getId Dcons.idTable (slOfLi dc.v) in
                Some (App (symOfDconinst (di,tiis), Array.of_list fs))
(* TODO recd *)
(*
  | Exp_recd(l) ->
      let fields, es = List.split l in
      let typs =
        List.map (fun e -> Tcenv.expand_typ_rec info.envn e.sort) es in
      let til = List.map tiOfTyp typs in
      (match tiioOfTi (TIRecord (List.zip fields til)) with
         | None -> None
         | Some tii ->
             let fos = List.map (expToTerm info) es in
             if existsNone fos then None
             else Some (App (symOfRecd tii, Array.of_list (projSomes fos))))
*)
(* TODO recd
  | Exp_proj(e',fn) ->
      let t = Tcenv.expand_typ_rec info.envn e'.sort in
      (match t.v with
         | Typ_record(l) ->
           let fields, typs = List.split l in
           let til = List.map tiOfTyp typs in
           (match tiioOfTi (TIRecord (List.zip fields til)) with
              | None -> None
              | Some tii ->
                  (match expToTerm info e' with
                     | None -> None
                     | Some f ->
                         let j = IdTable.getId Fields.idTable (slOfLi fn) in
                         Some (App (symOfRecdProj tii j, [|f|]))))
         | _ -> None)
*)
  | Exp_recd(Some(li),[],[],l) -> (* monomorphic record *)
      let fields, es = List.split l in
      (match tiioOfTi (TIRecordMono (Id.getId Records.idTable (slOfLi li))) with
         | None -> noneExp e
         | Some tii ->
             let fos = List.map (expToTerm vmap info) es in
             if existsNone fos then noneExp e
             else Some (App (symOfRecd tii, Array.of_list (projSomes fos))))
(*
  | Exp_proj(e',fn) ->
      let t = Tcenv.expand_typ_rec info.envn e'.sort in
      (* TODO t.v will no longer look like a Typ_record. look for a Typ_const instead! *)
      (match t.v with Typ_record(l) ->
         let firstfield = slOfLi (List.hd (fst (List.split l))) in
         (* only checking monomorphic records *)
         (match tiioOfTi (TIRecordMono (Id.getId Records.idTable firstfield)) with
            | None -> noneExpMsg e "proj1"
            | Some tii ->
                (match expToTerm info e' with
                   | None -> noneExpMsg e "proj2"
                   | Some f ->
                       let j = IdTable.getId Fields.idTable (slOfLi fn) in
                       Some (App (symOfRecdProj tii j, [|f|]))))
      | _ -> noneExpMsg e "proj3")
*)
  | Exp_proj(e',fn) ->
      let t = Tcenv.expand_typ_rec info.envn e'.sort in
        (match (AbsynUtils.unrefine t).v with 
             Typ_const(x) ->
               (match tiioOfTi (TIRecordMono (Id.getId Records.idTable (slOfLi x.v))) with
                  | None -> noneExpMsg e "proj1"
                  | Some tii ->
                      (match expToTerm vmap info e' with
                         | None -> noneExpMsg e "proj2"
                         | Some f ->
                             let j = IdTable.getId Fields.idTable (slOfLi fn) in
                               Some (App (symOfRecdProj tii j, [|f|]))))
           | _ -> noneExpMsg e "proj3")
  | Exp_fvar x -> failwith (spr "expToTerm: Exp_fvar %s" (sli x.v))
  | _ -> noneExp e 
      

let isFullyAppliedProp info t = match t.sort with 
  | Kind_star ->  Tcenv.is_prop info.envn t 
  | _ -> false

let rec fullyAppliedPropToTerm vmap (info:info) (acc:term list) t = match t.v with
  | Typ_const(x) ->
      let s, sl = sli x.v, slOfLi x.v in
      if s = sli Const.true_lid then
        (match acc with
           | [] -> Some True
           | _  -> raise (Never "fullyAppliedPropToTerm 1"))
      else if s = sli Const.false_lid then
        (match acc with
           | [] -> Some False
           | _  -> raise (Never "fullyAppliedPropToTerm 2"))
      else if IdTable.mem Props.idTable sl then
        let propid = IdTable.getId Props.idTable sl in
        Some (App (symOfPropid propid, Array.of_list acc))
      else if IdTable.mem Dcons.idTable sl then
        let dconid = IdTable.getId Dcons.idTable sl in
          (* TODO need to allow polymorphic datacons *)
        let funsym = symOfDconinst (dconid,[]) in
          Some (App (funsym, Array.of_list acc))
            (*
              else if IdTable.mem Records.idTable sl then
              let recdconid = IdTable.getId Records.idTable sl in
              let funsym = symOfRecd (tiiOfTi (TIRecordMono recdconid)) in
              Some (App (funsym, Array.of_list acc))
            *)
      else failwith (spr "fullyAppliedPropToTerm %s" s)

  | Typ_btvar bv -> 
      Some (App((bvar_real_name bv).idText, Array.of_list acc))

  | Typ_dep(t1, e) ->
      (match expToTerm vmap info e with
         | Some f -> fullyAppliedPropToTerm vmap info (f::acc) t1
         | None   -> noneTyp t)

  | Typ_app({v=Typ_const(x);sort=_;p=_}, _) when Sugar.lid_equals x.v Const.eq_lid -> 
      (match acc with 
         | [a;b] -> Some(Eq(a,b))
         | _ -> raise (Never "fullyAppliedPropToTerm 3"))
        
  | _ -> noneTyp t


(*
(* TODO 11/15 temporary *)
let stripPfT t = try stripPfT t with _ -> t
*)

let rec otherTypToTerm vmap (info:info) (t:typ) : term option = match t.v with

  | Typ_const(x) when Sugar.lid_equals x.v Const.true_lid -> noneTyp t
  | Typ_const(x) when Sugar.lid_equals x.v Const.false_lid -> noneTyp t

  (* logical connectives *)
  | Typ_app (t1,t2) ->
      (match t1.v with
         | Typ_const(x) when Sugar.lid_equals x.v Const.not_lid ->
             (match typToTerm vmap info (stripPfT t2) with
                | Some s -> Some (Not s)
                | _      -> noneTyp t)
         | Typ_app (u1,u2) ->
             (match u1.v with
                | Typ_const(x) ->
                    let a, b = stripPfT u2, stripPfT t2 in
                    (match typToTerm vmap info a, typToTerm vmap info b with
                       | Some s1, Some s2 ->
                           if Sugar.lid_equals x.v Const.and_lid then Some (And (s1,s2))
                           else if Sugar.lid_equals x.v Const.or_lid then Some (Or (s1,s2))
                           else if Sugar.lid_equals x.v Const.implies_lid then Some (Imp (s1,s2))
                           else if Sugar.lid_equals x.v Const.iff_lid then Some (Iff (s1,s2))
                           else noneTyp t
                       | _ -> noneTyp t)
                | _ -> noneTyp t)
         | _ -> noneTyp t)

  (* propositions *)
  | Typ_fun (None, t1, t2) ->
      (match typToTerm vmap info (stripPfT t1), typToTerm vmap info (stripPfT t2) with
         | Some s1, Some s2 -> Some (Imp (s1, s2))
         | _                -> noneTyp t)

  (* universals *)
  | Typ_fun (Some bvd, t1, t2) ->
      (match t1.sort with 
        | Kind_star -> 
            let id = snd bvd in
            let env' = Tcenv.push_local_binding info.envn (Tcenv.Binding_var (id,t1)) in
            let binders' = id.idText :: info.binders in
              (match typToTerm vmap { envn=env'; binders=binders' } (stripPfT t2) with
                 | Some s -> Some (Forall ([|tiioOfTyp t1|], [|symOfFormDom (bvd,t1)|], s))
                 | _ -> noneTyp t)
        | _ -> noneTyp t )

  (* existentials *)
  | Typ_dtuple [(Some(bvd),t1);(_,t2)] ->
      let id = real_name bvd in
      let env' = Tcenv.push_local_binding info.envn (Tcenv.Binding_var (id,t1)) in
      let binders' = id.idText :: info.binders in
      (match typToTerm vmap { envn=env'; binders=binders' } (stripPfT t2) with
         | Some s -> Some (Exists (tiioOfTyp t1, symOfFormDom (bvd,t1), s))
         | _ -> noneTyp t)

  | _ -> noneTyp t
      
and typToTerm vmap (info:info) (t:typ) : term option =
  if isFullyAppliedProp info t
  then fullyAppliedPropToTerm vmap info [] t
  else otherTypToTerm vmap info t
    

(****************************************************************************)
(* INTERFACE WITH TYPE CHECKER                                              *)
(****************************************************************************)

let queryCount = queryCount

let init (_:Absyn.modul) : unit =
  let l = [(BG, DefSort None)] in (* sortOther *)
  processOps emptyInfo l "Proof.init"

let finish () : unit =
  streamQueries.Flush (); streamQueries.Close ();
  printAutoGenSigsToFile ();
  ProofExtract.printStatsToFile ();
  z3.Dispose ()



let skipSigeltMessage = "not_handling_this_sigelt"

let curTconid               = ref (-1)
(*
let curTconidIsEquatable    = ref false
let curTconidIsMonomorphic  = ref false
*)
let curDconsArgTyps         = ref []

let doProcessSigelt (info:info) = profile "Proof.doProcessSigelt" (fun () -> function
    | Sig_tycon_kind(li,ts,k0,false) -> begin
      setCurMod (Tcenv.current_module info.envn);
      let title = spr "tycon_%s" (sli li) in
      let tconid = Id.processVal Tcons.idTable (slOfLi li) in
      let k = Tcenv.tycon_kind_from_tparams ts k0 in
      if tconIsEquatable k then begin
        Ht.add Tcons.isEquatable tconid true;
        Ht.add Tcons.dconIds tconid [];
        match k with 
          | Kind_star when (sli li <> "Prims.NTuple") -> 
              Ht.add Tcons.isMonomorphic tconid true;
              Ht.add Tcons.insts tconid [[]];
              let sigelts, ops = processEquatableTypinst info (TIMono tconid) in
                title, sigelts, ops
          | _ -> 
              Ht.add Tcons.isMonomorphic tconid false;
              Ht.add Tcons.insts tconid [];
              title, [], []
      end else begin
        Ht.add Tcons.isEquatable tconid false;
        title, [], []
      end
    end

  (* to force a tuple or record type to be processed as equatable,
     must provide an explicit type abbreviation that has no remaining
     type parameters. *)
  | Sig_typ_abbrev(li,[],_,t) -> begin
      setCurMod (Tcenv.current_module info.envn);
      let title = spr "typ_abbrev_%s" (sli li) in
      let t = Tcenv.expand_typ_rec info.envn t in (* expand abbreviated typ *)
      let sigelts, ops = processForEquatableTypinsts info (tiOfTyp t) in
        title, sigelts, ops
    end

  | Sig_record_typ(li,ts,_,{v=Typ_record(l);sort=_;p=_}) -> begin
      setCurMod (Tcenv.current_module info.envn);
      let title = spr "typ_record_typ_%s" (sli li) in
      let recdconid = Id.processVal Records.idTable (slOfLi li) in
      let fields, typs = List.split l in
      let typs = List.map (Tcenv.expand_typ_rec info.envn) typs in
      let l = List.zip fields typs in
      Ht.add Records.fieldsTyps recdconid l;
      (* assign ids to field names *)
      let fieldids = List.map (fun fld -> Id.processVal Fields.idTable (slOfLi fld)) (List.map fst l) in
      (* associate first field id with record id *)
      Ht.add Fields.firstFieldIdToRecordId (List.hd fieldids) recdconid;

        if List.length ts = 0 then begin
(*
pr "mono record %s\n" (sli li);
*)
        let modname = (List.hd li).idText in
        let firstfield = slOfLi (fst (List.hd l)) in
        if modname = "__Extern" then Ht.add firstFieldToExternRecordName firstfield li;
        let sigelts, ops = processForEquatableTypinsts info (TIRecordMono recdconid) in
        title, sigelts, ops
      end else begin (* TODO handle polymorphic records *)
(*
pr "poly record %s\n" (sli li);
*)
        title, [], []
      end
    end

  (* proposition *)
  | Sig_tycon_kind(_,[],Kind_qkind _, _) -> skipSigeltMessage, [], []

  | Sig_tycon_kind(li,[],k,true) -> begin
      setCurMod (Tcenv.current_module info.envn);
      let title = spr "prop_%s" (sli li) in
      let ts = List.map (Tcenv.expand_typ_rec info.envn) (typArgsOfKind k) in
      let tis = List.map tiOfTyp ts in
(*
let _ = pr "typinsts of %s args: %s\n" (sli li) (String.concat " " (List.map strTypinst tis)) in
*)
      let sigelts, ops = processListForEquatableTypinsts info tis in
      let tiios = List.map tiioOfTi tis in
      if existsNone tiios then 
        let comm = (BG, Comment (spr "cannot axiomatize prop %s" (sli li))) in
        (title, sigelts, comm::ops)
      else
        let tiis = projSomes tiios in
        let moreSigelts, moreOps = processProp info li tiis in
        (title, sigelts@moreSigelts, ops@moreOps)
    end

  (* user-supplied assumption *)
  | Sig_datacon_typ(li,[],t,true,_,_) -> begin
      setCurMod (Tcenv.current_module info.envn);
      let title = spr "assumption_%s" (sli li) in
      let t = Tcenv.expand_typ_rec info.envn t in
        (*       let ops = *)
        (*         if (new Regex("^Skolemize")).IsMatch (sli (List.tl li)) then *)
        (*           let i = incr skAxiomCount; !skAxiomCount in *)
        (* (\* *)
        (*           (match mkAid AKSkolemize with *)
        (*              | ASkolemize i -> (Ht.add skolemizeAxiom i (li,t); []) *)
        (*              | _ -> raise (Never "aid sk")) *)
        (* *\) *)
        (*           let _ = Ht.add skolemizeAxiom i (li,t) in *)
        (*           let t = stripPfT t in *)
        (*           (match typToTerm [] info t with *)
        (*             | Some x -> [(BG, JustMake (x, i))] *)
        (*             | None   -> []) *)
        (*         else *)
        let aid =
          (match mkAid AKUser with
             | AUser i -> (Ht.add userAxiom i (li,t); AUser i)
             | _ -> raise (Never "aid user")) in
        let t = stripPfT t in
        let ops = (match typToTerm [] info t with
                     | Some x -> [(BG, Assume (x, Some (sli li), aid))]
                     | None   -> bgWarn (spr "cannot axiomatize assumption %s\n;   %s" 
                                           (sli li) (strTyp t))) in
          title, [], ops
    end

  (* ordinary data constructor *)
  | Sig_datacon_typ(li,tps,t,false,_,_) -> begin
      setCurMod (Tcenv.current_module info.envn);
      let title = spr "datacon_%s" (sli li) in
      if checkIfProofConstructor (sli li) (mkVarTTyp li t) then title, [], []
      else if Const.is_tuple_data_lid li then title, [], []
      else begin
(*
pr "%s : %s\n" (sli li) (Pretty.strTyp t);
*)
        let t = pushTparamsIntoTyp tps t in
        let t = Tcenv.expand_typ_rec info.envn t in
        let argTyps, retTyp, typBvdefs = flattenArrowTyp t in
        let tconid = Id.getId Tcons.idTable (slOfLi (outerTypOfTyp retTyp)) in
(*
pr "tconid %d dcon %s : %s\n" tconid (sli li) (Pretty.strTyp t); flush stdout;
*)
        if tconid <> !curTconid then begin (* starting to process dcons of a new tcon *)
          curTconid := tconid;
          curDconsArgTyps := []; 
        end;
        if Ht.find Tcons.isEquatable tconid then begin
          let dconid = Id.processVal Dcons.idTable (slOfLi li) in
          Ht.add Dcons.typ dconid t;
          Ht.add Dcons.tconId dconid tconid;
          Ht.replace Tcons.dconIds tconid (dconid::(Ht.find Tcons.dconIds tconid));
          Ht.add Dcons.argTyps dconid argTyps;
          Ht.add Dcons.retTyp dconid retTyp;
          (* assume dcon has same num of type args as the type it constructs *)
          Ht.add Dcons.typBvdefs dconid typBvdefs;
          if Ht.find Tcons.isMonomorphic tconid then begin
            let sigelts, ops, allArgTiilOpt = processDconinst info dconid [] in
            let neqSigelts, neqOps =
              (match allArgTiilOpt with 
                 | None -> [], []
                 | Some(allArgTiil) -> begin
                     let tii = tiiOfTi (TIMono !curTconid) in
                     let l1, l2 = genDiseqs info tii dconid [] allArgTiil !curDconsArgTyps in
                     curDconsArgTyps := (dconid,allArgTiil)::!curDconsArgTyps;
                     l1, l2
                   end) in
            title, sigelts @ neqSigelts, ops @ neqOps
          end else
            title, [], []
        end else
          title, [], []
      end
    end

(*
  | Sig_value_decl(li,t) -> begin
*)
  | Sig_value_decl(li,t) when List.hd (slOfLi li) <> "ProofLib" -> begin
      setCurMod (Tcenv.current_module info.envn);
      let title = spr "value_decl_%s" (sli li) in
      let t = Tcenv.expand_typ_rec info.envn t in
      let sigelts, ops = processForEquatableTypinsts info (tiOfTyp t) in
      title, sigelts, ops
    end

  | _ -> skipSigeltMessage, [], []
)


let process_sigelt (env:Tcenv.env) (se:sigelt) : sigelt list =
  let info = inf env in
  let title, newSigelts, ops = doProcessSigelt info se in
(* TODO not quite, need to still define eq types and eq pred
  let newSigelts = if !Util.extract_proofs then newSigelts else [] in
*)
  assert (List.for_all (function (BG,_) -> true | _ -> false) ops);
  if title = skipSigeltMessage
    then () (* avoid writing all ProofLib sigelts to queries/all.smt *)
    else begin
      (* TODO ugh *)
      let l0 = getOnDemandOps () in
      processOps info (l0@ops) (spr "process_sigelt_%s" title)
    end;
  newSigelts

let generated_symbols () : signature =
  failwith "generated_symbols is deprecated!"

let pb_ctr = Profiling.new_counter "Proof.process_binding"
let pbv_ctr = Profiling.new_counter "Proof.process_binding var"
let pbm_ctr = Profiling.new_counter "Proof.process_binding match"
let expt_ctr = Profiling.new_counter "Proof.process_binding expand_typ_rec"


let pred_kind_to_sort_array_opt vmap k = 
  let tii_of_typ_opt t = 
    try Some (tiiOfTi (tiOfTyp t))
    with 
        _ -> 
          (match t.v with 
             | Typ_btvar v -> try_lookup_id vmap (bvar_real_name v)
             | _ -> None) in
  let rec aux out = function
    | Kind_star -> Some (List.to_array (List.rev out))
    | Kind_dcon(t, k) -> 
        match tii_of_typ_opt t with 
            None -> None
          | Some tii -> aux ((Some tii)::out) k in
    aux [] k
            
let processBinding env (l, vmap) binding : (list<taggedop> * varmap) =
  let thunk () = match binding with 

    | Tcenv.Binding_typ (v, Kind_affine) -> l, vmap

    | Tcenv.Binding_typ (v, Kind_star) -> 
        let id, vmap' = get_typinstid vmap v in
(*           Printf.printf "Adding a sort %d for type variable %s\n" id v.idText; *)
          (FG, DefSort (Some id))::(FG, Comment "Added type variable!")::l, vmap'
            
    | Tcenv.Binding_typ (v, k) -> 
        let id, vmap = get_typinstid vmap v in
          (match pred_kind_to_sort_array_opt vmap k with 
             | None -> l, vmap
             | Some s -> 
                 (FG, DefPred (v.idText, s, Some "Quantified pred"))::(FG, Comment "Added quantified prop!")::l, vmap)
            
    | Tcenv.Binding_var(id,tt) -> 
        let thunk () =  begin
          let defSymbolWithTyp uu = 
            let usort = 
              match tiioOfTyp uu with 
                | None -> try_lookup_id_typ vmap uu 
                | Some i -> Some i in
              [(FG, DefFun (symOfIdent id uu, [||], usort, None))] in
          let tt = compress tt in
          let tt = Profiling.profile (fun () -> Tcenv.expand_typ_rec env tt) expt_ctr in
          let more =
            (match tt.v with
                 (* define new function symbol and assume refinement. this should never happen when doing type-preserving translation. *)
               | Typ_refine(bvd,uu,vv) -> begin
                   let def = defSymbolWithTyp uu in
                   let newe = bvdToExp (id,id) tt in
                   let vv' = substitute_exp vv bvd newe in
                     def @ 
                       (match typToTerm vmap (inf env) vv' with
                          | None   -> fgWarn "can't convert refinement"
                          | Some o -> [(FG, Assume (o, None, mkAid AKBindingVar))]) 
                 end
                   (* assert proven formula and track proof term *)
               | Typ_app ({v=Typ_const(x);sort=_;p=_},t') when Sugar.lid_equals x.v Const.pf_lid -> 
                   (*let _ = pr "bindingvar proof\n"; flush stdout in*)
                   (match typToTerm vmap (inf env) t' with
                      | None   -> fgWarn "can't convert proven formula"
                      | Some o ->
                          (match mkAid AKBindingVar with
                             | ABindingVar(i) ->
                                 let _ = Ht.add bindingVarIdToPf i (BarePf id.idText) in
                                   (*
                                     let _ = pr "bindingvaridtopf %d %s\n" i id.idText; flush stdout in
                                   *)
                                   [(FG, Assume (o, None, ABindingVar i))]
                             | _ ->
                                 raise (Never "processBinding: Binding_var")))
                     (*
                     (* assert proven formula and track proof term *)
                       | Typ_dtuple [(Some(bvd),t3);(_,t4)] ->
                       (match t4.v with
                       | Typ_app ({v=Typ_const(x);sort=_;p=_},phi) when Sugar.lid_equals x.v Const.pf_lid -> 
                       let id' = real_name bvd in
                     (*let _ = pr "bindingvar dtuple proof %s\n" id.idText in*)
                       (match typToTerm (inf env) phi with
                       | None   -> fgWarn "can't convert proven formula"
                       | Some o ->
                       (match mkAid AKBindingVar with
                       | ABindingVar(i) ->
                       let _ = Ht.add bindingVarIdToPf i (SndPf id) in
                       let _ = Ht.add binderOfDepTupleProof id id' in
                       fgComment "PROOF IN TUPLE"
                       @
                       [(FG, DefFun (symOfIdent id' t3, [||], tiioOfTyp t3, None));
                       (FG, Assume (o, None, ABindingVar i))]
                       | _ ->
                       raise (Never "processVar: Binding_var dtuple")))
                       | _ -> defSymbolWithTyp tt)
                     *)
                     (* just define new function symbol *)
               | _ -> defSymbolWithTyp tt)
          in
            (fgComment (spr "BindingVar %s %s" id.idText (strTyp tt))
             @ more
             @ l), vmap
        end in
          Profiling.profile thunk pbv_ctr


    | Tcenv.Binding_match(e1,e2) -> 
        let thunk () = begin
          let t1 = Profiling.profile (fun () -> Tcenv.expand_typ_rec env e1.sort) expt_ctr in
          let t2 = Profiling.profile (fun () -> Tcenv.expand_typ_rec env e2.sort) expt_ctr in
          let more =

            (*
            (* careful: this Ht.mem depends on equality of idents *)
              (match e1.v with Exp_bvar(x) when Ht.mem binderOfDepTupleProof (real_name x.v) ->
              (match e2.v with Exp_constr_app(dc,_,_,[e3]) when Sugar.lid_equals dc.v Const.tuple_UU_lid ->
              (match e3.v with Exp_bvar(x2) ->
              let e4 = mkExpTyp (...
              (match expToTerm info e1, expToTerm info e3 with
              | Some a, Some b -> [(FG, (Assume (Eq (a,b), None, mkAid AKBindingMatch)))]

            *)

            let tiioOfTyp t = match tiioOfTyp t with 
                None -> try_lookup_id_typ vmap t
              | x -> x in
              (match tiioOfTyp t1, tiioOfTyp t2 with
                 | Some tii1, Some tii2 -> 
                     let info = inf env in
                       (match expToTerm vmap info e1, expToTerm vmap info e2 with
                          | Some a, Some b -> [(FG, (Assume (Eq (a,b), None, mkAid AKBindingMatch)))]
                          | None, None -> fgWarn "couldn't convert either exp"
                          | None, _    -> fgWarn "couldn't convert first exp"
                          | _,    None -> fgWarn "couldn't convert second exp")
                 | None, None -> fgWarn (spr "couldn't convert either expression's typ %s %s" (strTyp t1) (strTyp t2))
                 | None, _    -> fgWarn "couldn't convert e1's typ"
                 | _,    None -> fgWarn "couldn't convert e2's typ")
          in
            fgComment (spr "BindingMatch %s %s" (strExp e1) (strExp e2))
            @ more
            @ l, vmap
        end in
          Profiling.profile thunk pbm_ctr in
    Profiling.profile thunk pb_ctr

let processAllBindings env = 
  let l, vmap = Tcenv.fold_env env (fun (out, vmap) b -> 
                                      let newelts, vmap' = processBinding env ([],vmap) b in
                                      let newelts = List.rev newelts in
                                        newelts@out, vmap') ([],[]) in  
    List.rev l, vmap

let prove_and_extract env t : bool * exp' option =
  incr queryCount;
  let info = inf env in
  let t = Tcenv.expand_typ_rec env t in
  let l, vmap = processAllBindings env in
(*
pr "prove_and_extract %s\n" (Pretty.strTyp t); flush stdout;
*)
  let t =
    (match t.v with
       (* TODO if outer typ is a pf<>, does that mean this is a proof term from before? i.e. don't need to query again? *)
       | Typ_app ({v=Typ_const(x);sort=_;p=_},t') when Sugar.lid_equals x.v Const.pf_lid -> t'
       | _ -> t) in

  match typToTerm vmap info t with
    | Some tm -> begin
        (* TODO ugh *)
        let l0 = getOnDemandOps () in
        let l' = [(FG, Query tm)] in
        processOps info (l0@l@l') (spr "prove_and_extract_%04d" !queryCount);
        !z3QueryResult
      end
    | None -> begin
        let l' = fgWarn (spr "can't convert typ %s" (strTyp t)) in
        processOps info (l@l') (spr "prove_and_extract_%04d" !queryCount);
        (false, None)
      end

let prove_eq_and_extract env e1 e2 : bool * exp' option =
  incr queryCount;
  let info = inf env in
  let l, vmap = processAllBindings env in
(*
pr "prove_eq_and_extract %s %s\n" (Pretty.strExp e1) (Pretty.strExp e2); flush stdout;
*)
  match expToTerm vmap info e1, expToTerm vmap info e2 with
    | Some a, Some b -> begin
        (* TODO ugh *)
        let l0 = getOnDemandOps () in
        let l' = [(FG, Query (Eq (a, b)))] in
        processOps info (l0@l@l') (spr "prove_eq_and_extract_%04d" !queryCount);
        !z3QueryResult
      end
    | None, _ -> begin
        let l' = fgWarn (spr "can't convert exp %s" (strExp e1)) in
        processOps info (l@l') (spr "prove_eq_and_extract_%04d" !queryCount);
        (false, None)
      end
    | _, None -> begin
        let l' = fgWarn (spr "can't convert exp %s" (strExp e2)) in
        processOps info (l@l') (spr "prove_eq_and_extract_%04d" !queryCount);
        (false, None)
      end


