(*
  Copyright (c) 2009, Regents of the University of California

  Authors: Ravi Chugh
  
  All rights reserved.
  
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following 
  conditions are met:
  
  1. Redistributions of source code must retain the above copyright 
  notice, this list of conditions and the following disclaimer.

  2. Redistributions in binary form must reproduce the above 
  copyright notice, this list of conditions and the following disclaimer 
  in the documentation and/or other materials provided with the distribution.

  3. Neither the name of the University of California, San Diego, nor 
  the names of its contributors may be used to endorse or promote 
  products derived from this software without specific prior 
  written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*)


#light "off"

module Microsoft.Fine.ProofCombinators

open Microsoft.Fine
open Absyn
open AbsynUtils
open ProofUtils
open ProofCommon
open Util
module Id = IdTable


(****************************************************************************)
(* OTHER UTILS                                                              *)
(****************************************************************************)

let liOfSl = List.map (function s -> new ident(s,dummyRange))

let slOfLi = List.map (function (id:ident) -> id.idText)

let ssl (sl:string list) : string = String.concat "." sl

let rec strTypinst : typinst -> string =
  let strOfList l = String.concat "#" (List.map strTypinst l) in
  function
    | TIOther       -> "TIOther"
    | TIMono(i)     -> spr "TIMono_%d" i
    | TIPoly(i,l)   -> spr "TIPoly_%d<%s>" i (strOfList l)
    | TITuple(x,y)  -> spr "TITuple<%s#%s>" (strTypinst x) (strTypinst y)
(*
    | TIRecord(l) ->
        let s =
          String.concat "#"
            (List.map (function (fn,ti) ->
               spr "fld_%s_%s" (sli fn) (strTypinst ti)
            ) l)
        in spr "TIRecord_%s" s
*)
    | TIRecordMono(i) -> spr "TIRecordMono_%d" i

let strOfDconinst ((i,js):dconinst) =
  spr "%d_%s" i (String.concat "#" (List.map soi js))


(****************************************************************************)
(* FINE ABSYN UTILS                                                         *)
(****************************************************************************)

let mkId (s:string) : ident               = new ident(s,dummyRange)

let mkTypKind (t:typ') (k:kind) : typ     = { v=t; sort=k; p=dummyRange }
let mkTypStar (t:typ') : typ              = mkTypKind t Kind_star
let mkTyp (t:typ') : typ                  = mkTypKind t Kind_unknown
let dummyTyp                              = mkTyp Typ_unknown

let mkExpTyp (e:exp') (t:typ) : exp       = { v=e; sort=t; p=dummyRange }
let mkExp (e:exp') : exp                  = mkExpTyp e dummyTyp
let mkTagExp p (e:exp') : exp             = { v=e; sort=dummyTyp; p= int64 p }
let dummyExp                              = mkExp Exp_bot

let mkVarTTyp li (t:typ) : var<typ>       = { v=li; sort=t; p=dummyRange }
let mkVarT li : var<typ>                  = mkVarTTyp li dummyTyp
let dummyVarT : var<typ>                  = mkVarT [mkId "Nada"]

let mkVarKKind li (k:kind) : var<kind>    = { v=li; sort=k; p=dummyRange }
let mkVarK li : var<kind>                 = mkVarKKind li Kind_unknown

let mkTypConst li : typ                   = mkTyp (Typ_const (mkVarK li))

let bvdToExp (x,y) (t:typ) : exp          = mkExpTyp (Exp_bvar { v=(x,y); sort=t; p=dummyRange }) t

let idToExp x (t:typ) : exp               = bvdToExp (x,x) t

let mkDconApp dc typs exps       = mkExp (Exp_constr_app (dc, typs, [], exps))
let mkTagDconApp p dc typs exps  = mkTagExp p (Exp_constr_app (dc, typs, [], exps))
let mkExpApp e1 e2               = mkExp (Exp_app (e1, e2))
let mkExpAbs e1 t e2             = mkExp (Exp_abs (e1, t, e2))
let mkTagExpAbs p e1 t e2        = mkTagExp p (Exp_abs (e1, t, e2))

let liToExp li                   = mkDconApp (mkVarT li) [] []

let mkFvarTyp li (t:typ)         = mkExpTyp (Exp_fvar (mkVarTTyp li t)) t
let mkFvar li                    = mkFvarTyp li dummyTyp

let expTrue  = mkExpTyp (Exp_constant (Sugar.Const_bool true)) Const.bool_typ
let expFalse = mkExpTyp (Exp_constant (Sugar.Const_bool false)) Const.bool_typ

(* TODO recd *)
(*
let mkRecd fields es  = mkExp (Exp_recd (List.zip fields es))
*)
let mkRecd fields es  = mkExp (Exp_recd (None, [], [], List.zip fields es))
let mkProj e f        = mkExp (Exp_proj (e, f))

let mkTup t1 t2 e1 e2 = mkDconApp (mkVarT Const.tuple_UU_lid) [t1;t2] [e1;e2]
let mkProj1 e         = mkProj e Const.tuple_proj_one
let mkProj2 e         = mkProj e Const.tuple_proj_two

let mkInt i           = mkExp (Exp_constant (Sugar.Const_int32 (int32 i)))

(*
let expOfSl sl        = liToExp (liOfSl sl)
let proofTrue         = expOfSl ["Prims";"T"]
*)

let proofTrue         = liToExp (liOfSl ["Prims";"T"])


let mkPfT         = AbsynUtils.wrap_pf_typ
let stripPfT      = AbsynUtils.strip_pf_typ

(* Apply a list of typs and exps to type constructor. *)
let mkTypApp tc typs exps = 
  let rec f ts es = match ts, es with
    | _, e::es'  -> mkTyp (Typ_dep (f ts es', e))
    | t::ts', [] -> mkTyp (Typ_app (f ts' [], t))
    | [], []     -> tc 
  in f (List.rev typs) (List.rev exps)

let mkNot t = mkTypApp Const.not_typ     [t]   []
let mkPfNot = mkPfT -<- mkNot

let mkOr t u = mkTypApp Const.or_typ      [t;u] []
let mkPfOr  t u   = mkPfT (mkOr t u)

let mkAnd t u = mkTypApp Const.and_typ      [t;u] []
let mkPfAnd  t u   = mkPfT (mkAnd t u)

let mkImp t u = mkTypApp Const.implies_typ      [t;u] []
let mkPfImp  t u   = mkPfT (mkImp t u)

let mkIff t u   = mkTypApp Const.iff_typ     [t;u] []
let mkPfIff t u = mkPfT (mkIff t u)

let mkEq tc e f = (mkTypApp tc [] [e;f])
let mkPfEq tc e f = mkPfT (mkEq tc e f)


(* Takes n types (already wrapped with pf) and creates a nested sequence of
   binary ors. *)
let rec mkPfOr_n n ts = match n, ts with
  | 2, [t;u] -> mkPfOr t u
  | n, h::tl when n = List.length ts -> mkPfOr h (mkPfOr_n (n-1) tl)
  | _ -> raise (Never "mkOr")

let rec mkPfAnd_n n ts = match n, ts with
  | 2, [t;u] -> mkPfAnd t u
  | n, h::tl when n = List.length ts -> mkPfAnd h (mkPfAnd_n (n-1) tl)
  | _ -> raise (Never "mkAnd")

(* Takes a proposition T(x1,...,xn) and returns the lident of the type
   constructor and the expression arguments. *)
let destructFaProp t =
  let rec f acc t = match t.v with
    | Typ_const(x)  -> (x.v, acc)
    | Typ_dep(t1,e) -> f (e::acc) t1
    | _             -> failwith (spr "destructFaProp: %s" (Pretty.strTyp t))
  in f [] t

(* Takes a right-deep arrow and returns the bvdefs/domains of the innermost
   body. Automatically unwraps pf<> at each level. *)
let getFormalsDomainsRangeOfArrow t =
  let rec f acc t = match t.v with
    | Typ_fun(Some((_,formal)),dom,t') -> f ((formal.idText,dom)::acc) t'
    | Typ_app(t1,t2) ->
        (match t1.v with (* discard pf wrappers *)
           | Typ_const(x) when sli x.v = sli Const.pf_lid -> f acc t2
           | _ -> acc, t)
    | _ -> acc, t
  in f [] t

let destructDepArrow t = match t.v with
  | Typ_fun(Some bv,t1,t2) -> bv,t1,t2
  | _                      -> failwith "destructDepArrow"

let rec mkDepArrow binders body =
  match binders with
    | (bvd,t)::tl -> mkTyp (Typ_fun (Some bvd, t, mkDepArrow tl body))
    | []          -> body

let mkArrow t1 t2 = mkTyp (Typ_fun (None, t1, t2))

let rec mkArrowL = function
  | []      -> raise (Never "mkArrowL")
  | t::[]   -> t
  | t::rest -> mkTyp (Typ_fun (None, t, mkArrowL rest))

(* More specific than mkDepArrow, since it wraps the outer type and all
   inner range types of arrows with pf<>. Note that body should already
   be pf-wrapped though. *)
let rec mkPfDepArrow binders body =
  match binders with
    | []          -> body
    | (bvd,t)::tl -> mkPfT (mkTyp (Typ_fun (Some bvd, t, mkPfDepArrow tl body)))

(* TODO added 11/17 *)
let rec mkPfDepTuple binders body =
  match binders with
    | []          -> body
    | (bvd,t)::tl -> mkPfT (mkTyp (Typ_dtuple ([(Some(bvd),t);(None, mkPfDepTuple tl body)])))

(* Apply a list of typs to an exp. *)
let rec applyTyps e = function
  | []    -> e
  | t::ts -> applyTyps (mkExp (Exp_tapp (e, t))) ts

(* Apply a list of exps to an exp. *)
let rec applyExps e = function
  | []    -> e
  | f::fs -> applyExps (mkExp (Exp_app (e, f))) fs

(* Applies type and value parameters to a function expression.
   Typechecker doesn't currently allow explicitly creating Exp_tapps. *)
let mkCall1 funexp typs exps =
  applyExps (applyTyps funexp typs) exps
  (* TODO
  applyExps (applyTyps funexp []) exps
  *)

(* Applies only value parameters to function expression and ascribes
   it a return type. Deprecated. *)
let mkCall2 funexp exps ret =
  mkExp (Exp_ascribed (applyExps funexp exps, ret, []))

(* Deprecated. *)
let mkTagCall2 p funexp exps ret =
  mkTagExp p (Exp_ascribed (applyExps funexp exps, ret, []))


(* If the input type is a Typ_app, walks the Typ_app tree and
   flattens the type parameters. Does not recursively drill into
   other kinds of types. *)
let flattenTypApps : typ -> typ list =
  let rec foo acc t = match t.v with
    | Typ_app(t1,t2) -> foo (foo acc t2) t1
    | _              -> t::acc in
  (fun t -> foo [] t)

let rec tconIsEquatable = function
  | Kind_tcon(k1,k2) -> tconIsEquatable k1 && tconIsEquatable k2
  | Kind_star        -> true
  | _                -> false

(* Substitute formals with type parameters in t. *)
let rec substBvdefs (l:(bvdef * typ) list) (t:typ) : typ =
  match l with
    | [] -> t
    | (bvd,act)::rest -> substBvdefs rest (substitute t bvd act)

(* Canonicalizes type with type tparams by creating additional universal
   types. *)
let rec pushTparamsIntoTyp (tps:tparam list) (t:typ) : typ =
  match List.rev tps with
    | (Tparam_typ (bvd,k))::rest ->
         pushTparamsIntoTyp rest (withsort (Typ_univ (bvd, k, t)) Kind_star)
    (* TODO check this, added on 10/23 *)
    | (Tparam_term(bvd,t1))::rest -> t (* since all tparams came first *)
    | [] -> t
(*
    | _  -> failwith "pushTparamsIntoTyp"
*)

(* TODO get term bvds also *)
(* Return the argument types, the final return type, and the bvdefs for all
   type abstractions. *)
let flattenArrowTyp : typ -> typ list * typ * bvdef list =
  let rec f (acc1,acc2) t = match t.v with
    | Typ_fun(_,t1,t2)    -> f ((t1::acc1),acc2) t2
    | Typ_univ(bvd,_,t1)  -> f (acc1,(bvd::acc2)) t1
    | _                   -> (t::acc1),acc2 in
  fun t ->
    let l, bvds = f ([],[]) t in
    List.rev (List.tl l), List.hd l, List.rev bvds

(* Returns type parameters from a dependent type constructor. *)
let typArgsOfKind : kind -> typ list =
  let rec f acc = function
    | Kind_star       -> acc
    | Kind_dcon(t,k') -> f (t::acc) k'
    | k               -> failwith (spr "typArgsOfKind: %s" (Pretty.strKind k))
  in fun k -> List.rev (f [] k)

(* Creates a new bvdef and the corresponding bvar expression. *)
let mkFormDom (s:string) (t:typ) =
  let bvd = mkId s, genident None in
  let e   = bvdToExp bvd t in
  (bvd,t), e

(* Uses the given string instead of generating a new symbol. *)
let mkFormDomFixed (s:string) (t:typ) =
  let bvd = mkId s, mkId s in
  let e   = bvdToExp bvd t in
  (bvd,t), e

let rec outerTypOfTyp (t:typ) : lident =
  match t.v with
    | Typ_app(t1,_)      -> outerTypOfTyp t1
    | Typ_dep(t1,_)      -> outerTypOfTyp t1
    | Typ_refine(_,t1,_) -> outerTypOfTyp t1
    | Typ_const(x)       -> x.v
    | _                  -> failwith (spr "outerTypOfTyp: %s" (Pretty.strTyp t))
  


(****************************************************************************)
(* TABLES FOR EVERYTHING                                                    *)
(****************************************************************************)

module Strings = struct
  let idTable : string Id.t = Id.create "Strings.idTable" id
end

(* TODO this assumes that fields of extern records are not polymorphic *)
let firstFieldToExternRecordName : (string list, lident) Ht.t = Ht.create "firstFieldToExternRecordName" ssl

module Records = struct
  let idTable : string list Id.t = Id.create "Records.idTable" ssl
  let fieldsTyps : (recdconid, (fieldname * typ) list) Ht.t = Ht.create "Records.fieldsTyps" soi
end

module Fields = struct
  let (idTable    :                          string list Id.t) =
    Id.create "Fields.idTable" ssl
  let firstFieldIdToRecordId: (fieldid, recdconid) Ht.t = Ht.create "Fields.recdId" soi
end

module Props = struct
  let (idTable    :                          string list Id.t) =
    Id.create "Props.idTable" ssl
end
  (*     Ht.create "Props.dconMono" soi *)

module Dcons = struct
  let (idTable    :                          string list Id.t),
      (typ        :                        (dconid, typ) Ht.t),
      (tconId     :                        (dconid, int) Ht.t),
      (argTyps    :                   (dconid, typ list) Ht.t),
      (retTyp     :                        (dconid, typ) Ht.t),
      (typBvdefs  :                 (dconid, bvdef list) Ht.t) =
    Id.create "Dcons.idTable" ssl,
    Ht.create "Dcons.typ" soi,
    Ht.create "Dcons.tconId" soi,
    Ht.create "Dcons.argTyps" soi,
    Ht.create "Dcons.retTyp" soi,
    Ht.create "Dcons.typBvdefs" soi
end

module Tcons = struct
  let (idTable    :                          string list Id.t),
      (dconIds    :                (tconid, dconid list) Ht.t),
      (insts      :              (tconid, int list list) Ht.t) =
    Id.create "Tcons.idTable" ssl,
    Ht.create "Tcons.dconIds" soi,
    Ht.create "Tcons.insts" soi
  let isEquatable : (tconid, bool) Ht.t = Ht.create "Tcons.isEquatable" soi
  let isMonomorphic : (tconid, bool) Ht.t = Ht.create "Tcons.isMonomorphic" soi
end

(* these are the typinsts that are equatable *)
module TypInsts = struct
  let (idTable    :                              typinst Id.t),
      (modname    :                  (typinstid, string) Ht.t),
      (eqTypKind  :                    (typinstid, kind) Ht.t) = 
    Id.create "TypInsts.idTable" strTypinst,
    Ht.create "TypInsts.modname" soi,
    Ht.create "TypInsts.eqTypKind" soi
end

module DconInsts = struct
  let (dconMono   :                 (dconinst, var<typ>) Ht.t),
      (retTiio    :         (dconinst, typinstid option) Ht.t) =
    Ht.create "DconInsts.dconMono" strOfDconinst,
    Ht.create "DconInsts.retTiio" strOfDconinst
end

module TupleInsts = struct
  let (* (dconProj1  :                (typinstid, var<typ>) Ht.t), *)
(*       (dconProj2  :                (typinstid, var<typ>) Ht.t), *)
(*       (dconEq1    :                (typinstid, var<typ>) Ht.t), *)
(*       (dconEq2    :                (typinstid, var<typ>) Ht.t), *)
(*       (dconMonoEq :                (typinstid, var<typ>) Ht.t), *)
      (argTiis    :   (typinstid, typinstid * typinstid) Ht.t) =
(*     Ht.create "TupleInsts.dconProj1" soi, *)
(*     Ht.create "TupleInsts.dconProj2" soi, *)
(*     Ht.create "TupleInsts.dconEq1" soi, *)
(*     Ht.create "TupleInsts.dconEq2" soi, *)
(*     Ht.create "TupleInsts.dconMonoEq" soi, *)
    Ht.create "TupleInsts.argTiis" soi
end

module RecordInsts = struct
  let (fieldTiil  :          (typinstid, typinstid list) Ht.t),
      (fieldnames :          (typinstid, fieldname list) Ht.t),
      (retTiio    :        (typinstid, typinstid option) Ht.t),
      (dconProj   :      (typinstid * fieldid, var<typ>) Ht.t),
      (dconEq     :      (typinstid * fieldid, var<typ>) Ht.t),
      (dconMonoEq :                (typinstid, var<typ>) Ht.t) =
    Ht.create "RecordInsts.fieldTiil" soi,
    Ht.create "RecordInsts.fieldnames" soi,
    Ht.create "RecordInsts.retTiio" soi,
    Ht.create "RecordInsts.dconProj" (fun (i,j) -> spr "%d,%d" i j),
    Ht.create "RecordInsts.dconEq" (fun (i,j) -> spr "%d,%d" i j),
    Ht.create "RecordInsts.dconMonoEq" soi
end

module Disequalities = struct
  let neqBool : var<typ> ref = ref (mkVarT [])
  let count : int ref = ref 0
  let dcon : (int, var<typ>) Ht.t = Ht.create "Disequalities.dcon" soi
end


(****************************************************************************)
(* UTILS THAT LOOKUP IN TABLES                                              *)
(****************************************************************************)

let mkStr i =
  let s = Id.getVal Strings.idTable i in
  let b = System.Text.Encoding.ASCII.GetBytes(s) in
  mkExp (Exp_constant (Sugar.Const_string (b, dummyRange)))

let rec longStrTypinst : typinst -> string =
  let strList l = String.concat "," (List.map longStrTypinst l) in
  function
    | TIOther      -> failwith "strTypinst TIOther"
    | TIMono(i)    -> ssl (Id.getVal Tcons.idTable i)
    | TIPoly(i,l)  -> let x = ssl (Id.getVal Tcons.idTable i) in
                      spr "%s<%s>" x (strList l)
    | TITuple(x,y) -> spr "tuple<%s>" (strList [x;y])
(*
    | TIRecord(l) ->
        let s =
          String.concat "#"
            (List.map (function (fn,ti) ->
               spr "fld_%s_%s" (sli fn) (longStrTypinst ti)
            ) l)
        in spr "TIRecord_%s" s
*)
    | TIRecordMono(i) -> spr "record_mono_%s" (ssl (Id.getVal Records.idTable i))

let longStrOfDconinst ((di,tiil):dconinst) : string =
  let til = List.map (Id.getVal TypInsts.idTable) tiil in
  let s = String.concat "," (List.map longStrTypinst til) in
  spr "%s<%s>" (ssl (Id.getVal Dcons.idTable di)) s

(* TODO *)
(* this version uses abbrev name if extern record *)
let rec synthesizeTyp : typinst -> typ =
  function
    | TIMono(i) -> mkTypConst (liOfSl (Id.getVal Tcons.idTable i))
    | TIPoly(i,l) ->
        let tc = mkTypConst (liOfSl (Id.getVal Tcons.idTable i)) in
        let targs = List.map synthesizeTyp l in
        mkTypApp tc targs []
    | TITuple(x,y) ->
        let t1, t2 = synthesizeTyp x, synthesizeTyp y in
        (* using None to indicate normal tuple type *)
        mkTyp (Typ_dtuple [(None,t1);(None,t2)])
(*
    | TIRecord(l) ->
        let fields, til = List.split l in

        (* TODO think about this *)
        let firstfield = slOfLi (List.hd fields) in
        if Ht.mem firstFieldToExternRecordName firstfield then
          mkTypConst (Ht.find firstFieldToExternRecordName firstfield)
        else 

        let typs = List.map synthesizeTyp til in
        mkTyp (Typ_record (List.zip fields typs))
*)

    | TIRecordMono(i) ->
        (* TODO think about this *)
        let l = Ht.find Records.fieldsTyps i in
        let fields, typs = List.split l in
        let firstfield = slOfLi (List.hd fields) in
        if Ht.mem firstFieldToExternRecordName firstfield
          then mkTypConst (Ht.find firstFieldToExternRecordName firstfield)
          else mkTypConst (liOfSl (Id.getVal Records.idTable i)) (* mkTyp (Typ_record l) *)

    | TIOther -> failwith "synthesizeTyp: TIOther"




let rec tiOfTyp : typ -> typinst =
  let convert_one t = match t.v with
(* TODO recd *)
(*
    | Typ_const(x) -> TIMono (Id.getId Tcons.idTable (slOfLi x.v))
    | Typ_record(l) ->
        let fields, typs = List.split l in
        TIRecord (List.zip fields (List.map tiOfTyp typs))
*)
    | Typ_const(x) ->
        let sl = slOfLi x.v in
        if Id.mem Tcons.idTable sl then TIMono (Id.getId Tcons.idTable sl)
        else if Id.mem Records.idTable sl then TIRecordMono (Id.getId Records.idTable sl)
(* TODO added 11/1, when I added Rewrite_NotTrue to ProofLib *)
        else if sl = slOfLi (Const.true_lid) then TIOther
        else if sl = slOfLi (Const.false_lid) then TIOther
        else failwith (spr "tiOfTyp: %s" (Pretty.strTyp t)) 
    | Typ_record _ -> failwith "tiOfTyp: shouldn't be getting a Typ_record"
    (* want to use None, but seems like bvdefs are being generated
       even for normal tuples? *)
    | Typ_dtuple([(_,t1);(_,t2)]) -> TITuple (tiOfTyp t1, tiOfTyp t2)
    | Typ_refine(_,t',_) -> tiOfTyp t'
    | Typ_uvar _ -> (pr "tiOfTyp %s\n  TIOther uvar\n" (Pretty.strTyp t); TIOther)
    | _ -> TIOther
  in
  (fun t ->
     match flattenTypApps t with
       | []     -> raise (Never "tiOfTyp")
       | t'::[] -> convert_one t'
       | t'::ts -> (match convert_one t' with
                      | TIMono x -> TIPoly (x, List.map tiOfTyp ts)
                      | _        -> TIOther))

let tiiOfTi (ti:typinst) : typinstid = Id.getId TypInsts.idTable ti 

let tiOfTii (tii:typinstid) : typinst = Id.getVal TypInsts.idTable tii

let tiioOfTi (ti:typinst) : typinstid option =
  if Id.mem TypInsts.idTable ti then Some (tiiOfTi ti) else None

let tiioOfTyp (t:typ) : typinstid option = tiioOfTi (tiOfTyp t)

let typOfTi ti = synthesizeTyp ti

let typOfTii = typOfTi -<- tiOfTii

let lookupAxiomTii info lid tii = 
  let axT = Tcenv.lookup_lid info.envn lid in
  let axVar = withsort lid axT in
  let t = typOfTii tii in
    axVar, t

let boolSort () =
  let tconid = Id.getId Tcons.idTable (slOfLi Const.bool_lid) in
  Some (tiiOfTi (TIMono tconid))

let stringSort () =
  let tconid = Id.getId Tcons.idTable (slOfLi Const.string_lid) in
  Some (tiiOfTi (TIMono tconid))

let boolTii () =
  match tiioOfTyp Const.bool_typ with
    | Some btii -> btii
    | None -> raise (Never "boolTii")


(****************************************************************************)
(* TO AND FROM Z3 FUNCTION SYMBOLS                                          *)
(****************************************************************************)

(* calling one of the symOf functions registers the mapping from a Z3
   function symbol to the expression that it names. *)

let (symOfPropid : propid -> string),
    (propidOfSym : string -> propid option) =
  let ht = Ht.create "symToPropid" id in
  (fun i ->
     let sym = ssl (Id.getVal Props.idTable i) in
     Ht.add ht sym i;
     sym),
  (fun s ->
     if Ht.mem ht s
       then Some (Ht.find ht s)
       else None)

let (symOfDconinst : dconinst -> string),
    (dconinstOfSym : string -> dconinst option) =
  let ht = Ht.create "symToDconinst" id in
  (fun (i,js) ->
     let s = String.concat "#" (List.map soi js) in
(*
     let sym = spr "Dconinst_%d_%s" i s in
*)
     let sym = spr "%s_%s" (ssl (IdTable.getVal Dcons.idTable i)) s in
     Ht.add ht sym (i,js);
     sym),
  (fun s -> if Ht.mem ht s then Some (Ht.find ht s) else None)

let (symOfTuple : typinstid -> string),
    (symOfTupleProj1 : typinstid -> string),
    (symOfTupleProj2 : typinstid -> string),
    (tupleCtorOfSym : string -> typinstid option),
    (tupleProj1OfSym : string -> typinstid option),
    (tupleProj2OfSym : string -> typinstid option) =
  let ht1 = Ht.create "symOfTuple" id in
  let ht2 = Ht.create "symOfTupleProj1" id in
  let ht3 = Ht.create "symOfTupleProj2" id in
  (fun tii -> let s = spr "Tup_tii%d" tii in Ht.add ht1 s tii; s),
  (fun tii -> let s = spr "TupProj_tii%d_1" tii in Ht.add ht2 s tii; s),
  (fun tii -> let s = spr "TupProj_tii%d_2" tii in Ht.add ht3 s tii; s),
  (fun s -> if Ht.mem ht1 s then Some (Ht.find ht1 s) else None),
  (fun s -> if Ht.mem ht2 s then Some (Ht.find ht2 s) else None),
  (fun s -> if Ht.mem ht3 s then Some (Ht.find ht3 s) else None)
    

let (symOfRecd : typinstid -> string),
    (symOfRecdProj : typinstid -> fieldid -> string),
    (recdCtorOfSym : string -> typinstid option),
    (recdProjOfSym : string -> (typinstid * fieldid) option)  =
  let ht1 = Ht.create "symOfRecd" id in
  let ht2 = Ht.create "symOfRecdProj" id in
  (fun tii -> let s = spr "Recd_%d" tii in Ht.add ht1 s tii; s),
  (fun tii j ->
     let s = spr "RecdProj_%d_%s" tii (ssl (Id.getVal Fields.idTable j)) in
     Ht.add ht2 s (tii,j);
     s),
  (fun s -> if Ht.mem ht1 s then Some (Ht.find ht1 s) else None),
  (fun s -> if Ht.mem ht2 s then Some (Ht.find ht2 s) else None)

let (symOfIdent : ident -> typ -> string),
    (identTiioOfSym : string -> typinstid option option),
    (unequatableTypOfSym : string -> typ option) =
  let ht1 = Ht.create "symToTiio" id in
  let ht2 = Ht.create "symToSavedTyp" id in
  (fun id t ->
     let tiio = tiioOfTyp t in
     let s = id.idText in
     Ht.add ht1 s tiio;
     Ht.add ht2 s t;
     s),
  (fun s -> if Ht.mem ht1 s then Some (Ht.find ht1 s) else None),
  (fun s -> if Ht.mem ht2 s then Some (Ht.find ht2 s) else None)

let symOfFormDom (fd:(bvdef*typ)) : string =
  let s = symOfIdent (real_name (fst fd)) (snd fd) in
  Ht.add symToFormDom s fd;
  s

let (symOfString : string -> string),
    (stringidOfSym : string -> int option) =
  let ht1 = Ht.create "symToString" id in
  (fun s ->
     let i = Id.getId Strings.idTable s in
(*
     let sym = spr "string%03d" i in
*)
     let sym = spr "str_%s" (s.Replace(' ','_')) in
     Ht.add ht1 sym i;
     sym),
  (fun s -> if Ht.mem ht1 s then Some (Ht.find ht1 s) else None)


(****************************************************************************)
(* NAMES OF SPECIALIZED AXIOMS AND THEOREMS                                 *)
(****************************************************************************)


(* these generated lidents always have length 2, where the first component
   is the module name and the second is the specialized principle name *)

let mkSpecTc (prefix:string) (tii:typinstid) : lident =
  [mkId (Ht.find TypInsts.modname tii); mkId (spr "%s_tii%03d" prefix tii)]

(* let liEqTyp           = mkSpecTc "Eq" *)
      
(* let liDconRefl        = mkSpecTc "Refl" *)
(* let liDconSymm        = mkSpecTc "Symm" *)
(* let liDconTrans       = mkSpecTc "Trans" *)
(* let liFuncIff         = mkSpecTc "_Iff" *)
(* let liFuncMono        = mkSpecTc "_Monotonicity" *)
(* let liOpEq            = mkSpecTc "Op_Eq" *)

(* let liDconTupProj i   = mkSpecTc (spr "AxTupProj%d" i) *)
(* let liDconTupEq i     = mkSpecTc (spr "AxTupEq%d" i) *)
(* let liDconTupMonoEq   = mkSpecTc "TupMonoEq" *)

let liDconRecdProj i  = mkSpecTc (spr "AxRecdProj_%d" i)
let liDconRecdEq i    = mkSpecTc (spr "AxRecdEq_%d" i)
let liDconRecdMonoEq  = mkSpecTc "RecdMonoEq"

let liDconNeqBool     = liOfSl ["Prims";"NeqBool"]

let liDconNeq i       = mkSpecTc (spr "Neq_%04d" i)

(* let liDconMonoPropid (i:propid) : lident = *)
(*   let li = liOfSl (Id.getVal Props.idTable i) in *)
(*   let modname = List.hd li in *)
(*   modname :: [mkId (spr "Mono_prop_%s" *)
(*                       (ssl (Id.getVal Props.idTable i)))] *)

let liDconMonoDconinst ((i,js):dconinst) : lident =
  let li = liOfSl (Id.getVal Dcons.idTable i) in
  let modname = List.hd li in
  modname :: [mkId (spr "Mono_dconinst_%s" (strOfDconinst (i,js)))]


let getEqTyp info t = 
  let eqK = Tcenv.lookup_typ_const info.envn (Wt Const.eq_lid) in
    match eqK with 
      | Kind_qkind ([(a, Kind_star)], Kind_dcon(_a1, Kind_dcon(_a2, Kind_star))) ->
          (* NIK: should we be really paranoid and check _a1 and _a2? *)
          let eqT = withsort (Typ_const (withsort Const.eq_lid eqK)) eqK in
            eqT
      | _ ->  raise (Never "unexpected kind for Prims.Eq")
          
let eqTypOfTyp info t = 
  let eqT = getEqTyp info t in 
  let k = Kind_dcon(t, Kind_dcon(t, Kind_star)) in
    withsort (Typ_app(eqT, t)) k
      
let eqTypOfTii info = (eqTypOfTyp info) -<- typOfTii

(****************************************************************************)
(* AXIOMS                                                                   *)
(****************************************************************************)

let checkIfProofConstructor = 
  let mk_logic_name s = Const.p2l ["Prims"; s] in
  let names =
    List.map (fun s -> sli (mk_logic_name s))
      [
        "LiftPf"; "BindPf"; "AndTrue"; "OrFalse";
        "ExcMiddle"; "Contra"; "DestructFalse";
        "OrIntro1"; "OrIntro2"; "OrElim"; "AndIntro";
        "AndElim1"; "AndElim2";
        "DeMorganAnd"; "DeMorganAndRev"; "DeMorganOr"; "DeMorganOrRev";
        "DblNotElim"; "DblNotElimRev"; "Distribute";
        "ImpliesIntro"; "ModusPonens"; "IffIntro"; "IffElim1"; "IffElim2";
        "IdImplies"; "IdImpliesRev";
        "MonoNot"; "MonoAnd"; "MonoOr"; "MonoImplies";
      ] in
  let ht : (string, var<typ>) Ht.t = Ht.create "pfComb" id in
  (fun f t -> if List.mem f names then (Ht.add ht f t; true) else false)

let pfComb s = mkVarT (liOfSl ["Prims"; s])

module Axioms = struct
  let _T                  = pfComb "T"
  let _LiftPf             = pfComb "LiftPf"
  let _BindPf             = pfComb "BindPf"
  let _AndTrue            = pfComb "AndTrue"
  let _OrFalse            = pfComb "OrFalse"
  let _ExcMiddle          = pfComb "ExcMiddle"
  let _Contra             = pfComb "Contra"
  let _DestructFalse      = pfComb "DestructFalse"
  let _OrIntro1           = pfComb "OrIntro1"
  let _OrIntro2           = pfComb "OrIntro2"
  let _OrElim             = pfComb "OrElim"
  let _AndIntro           = pfComb "AndIntro"
  let _AndElim1           = pfComb "AndElim1"
  let _AndElim2           = pfComb "AndElim2"
  let _DeMorganNotAnd     = pfComb "DeMorganNotAnd"
  let _DeMorganOr         = pfComb "DeMorganOr"
  let _DeMorganNotOr      = pfComb "DeMorganNotOr"
  let _DeMorganAnd        = pfComb "DeMorganAnd"
  let _DblNotElim         = pfComb "DblNotElim"
  let _DblNotElimRev      = pfComb "DblNotElimRev"
  let _Distribute         = pfComb "Distribute"
  let _ImpliesIntro       = pfComb "ImpliesIntro"
  let _ModusPonens        = pfComb "ModusPonens"
  let _IffIntro           = pfComb "IffIntro"
  let _IffElim1           = pfComb "IffElim1"
  let _IffElim2           = pfComb "IffElim2"
  let _IdImplies          = pfComb "IdImplies"
  let _IdImpliesRev       = pfComb "IdImpliesRev"
  let _MonoNot            = pfComb "MonoNot"
  let _MonoAnd            = pfComb "MonoAnd"
  let _MonoOr             = pfComb "MonoOr"
  let _MonoImplies        = pfComb "MonoImplies"
  let _NotAllExistsNot    = pfComb "NotAll_existsNot"
  let _PullQuantAnd       = pfComb "PullQuantAnd"
  let _PullQuantOr        = pfComb "PullQuantOr"
  let _PullQuantAndE      = pfComb "PullQuantAndE"
  let _PullQuantOrE       = pfComb "PullQuantOrE"
  let _ElimUnused         = pfComb "ElimUnused"
  let _MonoOrIff          = pfComb "MonoOrIff"
end


(****************************************************************************)
(* LIBRARY THEOREMS                                                         *)
(****************************************************************************)

(*
let pfFunc s = expOfSl ["ProofLib"; s]
*)
let pfFunc s = mkFvar (liOfSl ["ProofLib"; s])

module Theorems = struct

  let _Refl                 = pfFunc "_Refl"
  let _Symm                 = pfFunc "_Symm"
  let _Trans                = pfFunc "_Trans"

  let _ModusPonensIff       = pfFunc "_ModusPonensIff"

  let _ProofByContradiction = pfFunc "_ProofByContradiction"

  let _Contra flag          = pfFunc (spr "_Contra_%s" (strFlags [flag]))

  let _Monotonicity_Not     = pfFunc "_Monotonicity_Not"
  let _Monotonicity_Imp     = pfFunc "_Monotonicity_Imp"
  let _Monotonicity_Iff     = pfFunc "_Monotonicity_Iff"
  let _Monotonicity_Or n    = pfFunc (spr "_Monotonicity_Or_%d" n)
  let _Monotonicity_And n   = pfFunc (spr "_Monotonicity_And_%d" n)

  let _AndElim n k =
    pfFunc (spr "_AndElim_%d_%d" n k)

  let _Commute_And bij =
    let n, strBij = List.length bij, String.concat "" (List.map soi bij) in
    pfFunc (spr "_Commute_And_%d_%s" n strBij)

  let _Commute_Or bij =
    let n, strBij = List.length bij, String.concat "" (List.map soi bij) in
    pfFunc (spr "_Commute_Or_%d_%s" n strBij)

  let _DeMorgan_NotAnd n flags =
    pfFunc (spr "_DeMorgan_NotAnd_%d_%s" n (strFlags flags))
  
  let _DeMorgan_NotOr n flags =
    pfFunc (spr "_DeMorgan_NotOr_%d_%s" n (strFlags flags))
  
  let _DeMorgan_And n flags =
    pfFunc (spr "_DeMorgan_And_%d_%s" n (strFlags flags))
  
  let _DeMorgan_Or n flags =
    pfFunc (spr "_DeMorgan_Or_%d_%s" n (strFlags flags))
  
  let _UnitRes n k flags =
    pfFunc (spr "_UnitRes_%d_%d_%s" n k (strFlags flags))

  let _Rewrite_TrueImplies  = pfFunc "_Rewrite_TrueImplies"

  let _Rewrite_IdImplies_1 flag = pfFunc (spr "_Rewrite_IdImplies_1_%s" (strFlags [flag]))

  let _Rewrite_IdImplies_2 flag = pfFunc (spr "_Rewrite_IdImplies_2_%s" (strFlags [flag]))

  let _Rewrite_DblNotElim   = pfFunc "_Rewrite_DblNotElim"

  let _Rewrite_DeMorgan_NotAnd n flags =
    pfFunc (spr "_Rewrite_DeMorgan_NotAnd_%d_%s" n (strFlags flags))
  
  let _Rewrite_DeMorgan_NotOr n flags =
    pfFunc (spr "_Rewrite_DeMorgan_NotOr_%d_%s" n (strFlags flags))
  
  let _Rewrite_DeMorgan_And n flags =
    pfFunc (spr "_Rewrite_DeMorgan_And_%d_%s" n (strFlags flags))
  
  let _Rewrite_DeMorgan_Or n flags =
    pfFunc (spr "_Rewrite_DeMorgan_Or_%d_%s" n (strFlags flags))
  
  let _Flatten_And branchSizes =
    let s = String.concat "" (List.map soi branchSizes) in
    pfFunc (spr "_Flatten_And_%s" s)

  let _Flatten_Or branchSizes =
    let s = String.concat "" (List.map soi branchSizes) in
    pfFunc (spr "_Flatten_Or_%s" s)

  let _Unflatten_And branchSizes =
    let s = String.concat "" (List.map soi branchSizes) in
    pfFunc (spr "_Unflatten_And_%s" s)

  let _Unflatten_Or branchSizes =
    let s = String.concat "" (List.map soi branchSizes) in
    pfFunc (spr "_Unflatten_Or_%s" s)

  let _DefAxiomOne n i flag =
    pfFunc (spr "_DefAxiomOne_%d_%d_%s" n i (strFlags [flag]))

  let _DefAxiomTwo n i =
    pfFunc (spr "_DefAxiomTwo_%d_%d" n i)

  let _DefAxiomThree n flags =
    pfFunc (spr "_DefAxiomThree_%d_%s" n (strFlags flags))

  let _DefAxiomFour n =
    pfFunc (spr "_DefAxiomFour_%d" n)

  let _DefAxiomFive flag =
    pfFunc (spr "_DefAxiomFive_%s" (strFlags [flag]))

  let _DefAxiomSix flag =
    pfFunc (spr "_DefAxiomSix_%s" (strFlags [flag]))

  let _DefAxiomSeven flags =
    pfFunc (spr "_DefAxiomSeven_%s" (strFlags flags))
  
  let _DefAxiomEight = pfFunc "_DefAxiomEight"

  let _DefAxiomNine  = pfFunc "_DefAxiomNine"

  let _TEMP_Rewrite_OrFalseFalseP = pfFunc "_TEMP_Rewrite_OrFalseFalseP"
  let _TEMP_Rewrite i = pfFunc (spr "_TEMP_Rewrite_%d" i)

  (* temporary *)
  let _Lemma_1_0 = pfFunc "_Lemma_1_0"
  let _Lemma_1_1 = pfFunc "_Lemma_1_1"
  let _Lemma_2_01 = pfFunc "_Lemma_2_01"
  let _Lemma_2_11 = pfFunc "_Lemma_2_11"
  let _Lemma_3_101 = pfFunc "_Lemma_3_101"

end


(****************************************************************************)
(* SPECIALIZED AXIOMS/THEOREMS                                              *)
(****************************************************************************)

module AutoAxioms = struct

  let monoDconinst : dconinst -> var<typ> = Ht.find DconInsts.dconMono

(*   let monoProp : propid -> var<typ> = Ht.find Props.dconMono *)

(*   let tupProj (i:int) (tii:typinstid) : var<typ> = *)
(*     Ht.find (if i = 1 then TupleInsts.dconProj1 else TupleInsts.dconProj2) tii *)

(*   let tupEq (i:int) (tii:typinstid) : var<typ> = *)
(*     Ht.find (if i = 1 then TupleInsts.dconEq1 else TupleInsts.dconEq2) tii *)

(*   let tupMonoEq : typinstid -> var<typ> = Ht.find TupleInsts.dconMonoEq *)

  let recdProj : typinstid*fieldid -> var<typ> = Ht.find RecordInsts.dconProj

  let recdEq : typinstid*fieldid -> var<typ> = Ht.find RecordInsts.dconEq

  let recdMonoEq : typinstid -> var<typ> = Ht.find RecordInsts.dconMonoEq

end

(* module AutoTheorems = struct *)

(* (\* *)
(*   let iff  (tii: typinstid) : exp = liToExp (liFuncIff tii) *)
(*   let mono (tii: typinstid) : exp = liToExp (liFuncMono tii) *)
(* *\) *)
(* (\*   let iff  (tii: typinstid) : exp = mkFvar (liFuncIff tii) *\) *)
(* (\*   let mono (tii: typinstid) : exp = mkFvar (liFuncMono tii) *\) *)

(* end *)


(****************************************************************************)
(* CREATING SPECIALIZED AXIOMS AND THEOREMS                                 *)
(****************************************************************************)

(* NIK: I think this only works with non-dependent, non-affine types at the moment. 
   Should be generalized. *)
let createSpecializedProofCombinators (info:info) (tii:typinstid) : sigelt list =
  let t = typOfTii tii in
  let kEqT = Kind_dcon (t, Kind_dcon (t, Kind_star)) in (* the kind of the equality tcon: t => t => * *)
  let _ = Ht.add TypInsts.eqTypKind  tii kEqT in
    []

let createMonoRule info (tiil:typinstid list)
                   (retTypFun: exp list -> exp list -> typ) : typ =
  let til = List.map (Id.getVal TypInsts.idTable) tiil in
  let argTyps = List.map synthesizeTyp til in
  let n = List.length argTyps in
  let l1,l2 =
    List.mapi (fun i argTyp -> mkFormDom (spr "x%d" i) argTyp) argTyps,
    List.mapi (fun i argTyp -> mkFormDom (spr "y%d" i) argTyp) argTyps in
  let (xFds,xExps),(yFds,yExps) = List.split l1, List.split l2 in
  let eqTyps =
    map1N n (fun i ->
      let tii_i = List.nth tiil (i-1) in
      let tEqTi = eqTypOfTii info tii_i in (* mkTypConst (liEqTyp tii_i) in *)
      let ex,ey = List.nth xExps (i-1), List.nth yExps (i-1) in
      mkPfT (mkTypApp tEqTi [] [ex;ey])) in
  let ret = mkPfT (retTypFun xExps yExps) in
  let rec foo = function
    | (bvd,t)::rest, eqs      -> mkDepArrow [bvd,t] (foo (rest,eqs))
    | []           , eq::eqs' -> mkArrow eq (foo ([],eqs'))
    | []           , []       -> ret
  in
  foo (interleave xFds yFds, eqTyps)

let processProp info (li:lident)
                (tiil:typinstid list) : sigelt list * taggedop list =
  let propid = Id.processVal Props.idTable (slOfLi li) in
  let ops = [(BG, DefPred (symOfPropid propid,
                           Array.of_list (wrapSomes tiil), None))] in
    [], ops
      

let processTupleInst info (tii:typinstid) (ti1:typinst) (ti2:typinst)
    : taggedop list =

  let tii1, tii2  = tiiOfTi ti1, tiiOfTi ti2 in
  let sort1, sort2, sortTup = Some tii1, Some tii2, Some tii in
    Ht.add TupleInsts.argTiis tii (tii1,tii2);
    
    let symCtor     = symOfTuple tii in
    let symProj1    = symOfTupleProj1 tii in
    let symProj2    = symOfTupleProj2 tii in

    let opDefs = 
      [(BG, DefFun (symCtor,  [|sort1;sort2|], sortTup, None));
       (BG, DefFun (symProj1, [|sortTup|],     sort1,   None));
       (BG, DefFun (symProj2, [|sortTup|],     sort2,   None))] in
      
    let bvX         = BoundV (1, sort1) in
    let bvY         = BoundV (0, sort2) in
    let bvTup1      = BoundV (1, sortTup) in
    let bvTup2      = BoundV (0, sortTup) in
      
    let typ1, typ2  = synthesizeTyp ti1, synthesizeTyp ti2 in
    let typTup      = synthesizeTyp (TITuple (ti1, ti2)) in
      
    let tEqTup      = eqTypOfTyp info typTup in
    let tEqTyp1     = eqTypOfTyp info typ1 in
    let tEqTyp2     = eqTypOfTyp info typ2 in
      
    let defProjAxiom first =
      let (fdx,_),(fdy,_) = mkFormDom "x" typ1, mkFormDom "y" typ2 in
      let symX, symY = symOfFormDom fdx, symOfFormDom fdy in 
      let symProj, bv = if first then symProj1, bvX else symProj2, bvY in
        (BG, Assume (Forall ([|sort1;sort2|], [|symX;symY|],
                             Eq (App (symProj, [|App (symCtor, [|bvX;bvY|])|]),
                                 bv)),
                     None,
                     ATupProj (tii,if first then 1 else 2))) in
    
    let defEqAxiom first =
      let (fdtup1,_) = mkFormDom "tup1" typTup in
      let (fdtup2,_) = mkFormDom "tup2" typTup in
      let symTup1, symTup2 = symOfFormDom fdtup1, symOfFormDom fdtup2 in
      let symProj = if first then symProj1 else symProj2 in
        (BG, Assume (Forall ([|sortTup;sortTup|], [|symTup1;symTup2|],
                             Imp (Eq (bvTup1, bvTup2),
                                  Eq (App (symProj, [|bvTup1|]),
                                      App (symProj, [|bvTup2|])))),
                     None,
                     ATupEq (tii,if first then 1 else 2))) in
      
    let op1 = defProjAxiom true in
    let op2 = defProjAxiom false in
    let op3 = defEqAxiom true in
    let op4 = defEqAxiom false in
      opDefs @ [op1;op2;op3;op4]


let processRecordInst info (tii: typinstid)
                      (i: recdconid) (* assuming monomorphic for now *)
                      (fields: fieldname list)
                      (til: typinst list) : sigelt list * taggedop list =

  let n = List.length fields in
  let tiil = List.map tiiOfTi til in
  Ht.add RecordInsts.fieldTiil tii tiil;
  Ht.add RecordInsts.fieldnames tii fields;

  let sorts, sortRecd = Array.of_list (wrapSomes tiil), Some tii in

  let symCtor = symOfRecd tii in
  let ctorDef = (BG, DefFun (symCtor , sorts, Some tii, None)) in

  let bvXs    = map1N n (fun z -> BoundV (n-z, Some (List.nth tiil (z-1)))) in
  let bvXs    = Array.of_list bvXs in
  let rec1    = BoundV (1, Some tii) in
  let rec2    = BoundV (0, Some tii) in

  let typFields  = List.map synthesizeTyp til in
(*
  let typRecd    = synthesizeTyp (TIRecord (List.zip fields til)) in
*)
  let typRecd    = synthesizeTyp (TIRecordMono i) in

  let tEqFields  = List.map (eqTypOfTyp info) typFields in
  let tEqRecd    = eqTypOfTii info tii in (* mkTypConst (liEqTyp tii) in *)

  let defsForField fieldId fieldIdx =
    let theField = liOfSl (Id.getVal Fields.idTable fieldId) in

    (* stuff for AxRecdProj sigelt *)
    let l =
      map1N n (fun i -> mkFormDom (spr "x%d" i) (List.nth typFields (i-1))) in
    let fdXis, eXis = List.split l in
    let expRecdProj = mkProj (mkRecd fields eXis) theField in
    let expProjected = List.nth eXis fieldIdx in
    let tProj =
      mkPfDepArrow fdXis
        (mkPfEq (List.nth tEqFields fieldIdx) expRecdProj expProjected) in
    let projSigelt =
      if !Util.extract_proofs then
        let _ = 
          Ht.add RecordInsts.dconProj
            (tii,fieldId)
            (mkVarTTyp (liDconRecdProj fieldId tii) tProj) in
        [Sig_datacon_typ
           (liDconRecdProj fieldId tii, [], tProj, false, Public, None)]
      else
        let _ = 
          Ht.add RecordInsts.dconProj
            (tii,fieldId) (mkVarTTyp (liDconRecdProj fieldId tii) dummyTyp) in
        [] in

    (* stuff for AxRecdEq sigelt *)
    let (fdRecd1,eRecd1) = mkFormDom "recd1" typRecd in
    let (fdRecd2,eRecd2) = mkFormDom "recd2" typRecd in
    let eProj1, eProj2 = mkProj eRecd1 theField, mkProj eRecd2 theField in
    let tEq =
      mkPfDepArrow [fdRecd1; fdRecd2]
        (mkPfImp (mkPfEq tEqRecd eRecd1 eRecd2)
                 (mkPfEq (List.nth tEqFields fieldIdx) eProj1 eProj2)) in
    let eqSigelt =
      if !Util.extract_proofs then
        let _ = 
          Ht.add RecordInsts.dconEq
            (tii,fieldId)
            (mkVarTTyp (liDconRecdEq fieldId tii) tEq) in
        [Sig_datacon_typ
           (liDconRecdEq fieldId tii, [], tEq, false, Public, None)]
      else
        let _ = 
          Ht.add RecordInsts.dconEq
            (tii,fieldId) (mkVarTTyp (liDconRecdEq fieldId tii) dummyTyp) in
        [] in

    (* stuff for Z3 ops *)
    let symProj = symOfRecdProj tii fieldId in
    let symXis = Array.of_list (List.map symOfFormDom fdXis) in
    let symRecd1, symRecd2 = symOfFormDom fdRecd1, symOfFormDom fdRecd2 in
    let ops =
      [(BG, DefFun (symProj, [|sortRecd|], sorts.[fieldIdx], None));
       (BG, Assume (Forall (sorts, symXis,
                            Eq (App (symProj, [|App (symCtor, bvXs)|]),
                                bvXs.[fieldIdx])),
                    None,
                    ARecdProj (tii, fieldId)));
       (BG, Assume (Forall ([|sortRecd;sortRecd|], [|symRecd1;symRecd2|],
                            Imp (Eq (rec1, rec2),
                                 Eq (App (symProj, [|rec1|]),
                                     App (symProj, [|rec2|])))),
                    None,
                    ARecdEq (tii, fieldId)))] in
    let extra =  Eq (App (symProj, [|rec1|]),
                     App (symProj, [|rec2|])) in
      
      projSigelt@eqSigelt, ops, extra
  in

  let fieldIds =
    List.map (Id.getId Fields.idTable) (List.map slOfLi fields) in
  let fieldIdxs = list0N (n-1) in
  let sigelts, ops, extras =
    List.fold_left
      (fun (acc1,acc2,acc3) (i,j) ->
         let l1,l2,extra = defsForField i j in (acc1@l1, acc2@l2, extra::acc3))
      ([],[],[])
      (List.zip fieldIds fieldIdxs) in
  let rec_eq_iff = 
    let fdRecd1, _ = mkFormDom "recd1" typRecd in
    let symRecd1 = symOfFormDom fdRecd1 in
    let fdRecd2, _ = mkFormDom "recd2" typRecd in
    let symRecd2 = symOfFormDom fdRecd2 in
      (BG, Assume (Forall ([|sortRecd;sortRecd|], [|symRecd1;symRecd2|],
                           Iff (Eq (rec1, rec2),
                                List.fold_left (fun out ex -> And (out, ex)) True extras)),
                   None, AOther)) in
  let ops = ops@[rec_eq_iff] in
  let seMono =
    if not !Util.extract_proofs then begin
      Ht.add RecordInsts.dconMonoEq tii (mkVarTTyp (liDconRecdMonoEq tii) dummyTyp);
      []
    end else begin
      let retTypFun xs ys =
        let recd1, recd2 = mkRecd fields xs, mkRecd fields ys in
        mkTypApp tEqRecd [] [recd1;recd2] in
      let tMono = createMonoRule info tiil retTypFun in
      Ht.add RecordInsts.dconMonoEq
        tii (mkVarTTyp (liDconRecdMonoEq tii) tMono);
      [Sig_datacon_typ
         (liDconRecdMonoEq tii, [], tMono, false, Public, None)]
    end
  in

  (sigelts@seMono, ctorDef::ops)


let genDiseqs info (tii:typinstid)
              (j:dconid)
              (typArgTiil:typinstid list)
              (allArgTiil:typinstid list)
              (l:(dconid * (typinstid list)) list)
    : sigelt list * taggedop list =

  if not !doDisequalities then [],[] else

  let typsOfTiil = List.map (fun tii -> synthesizeTyp (tiOfTii tii)) in

  let aTyps = typsOfTiil allArgTiil in

  let doOne (j',allArgTiil') =
(*
    pr "doOne\n  %d %s\n  %d %s\n" j (strIntList allArgTiil)
                                   j' (strIntList allArgTiil');
*)
    let bTyps = typsOfTiil allArgTiil' in
    let al = List.mapi (fun i t -> mkFormDom (spr "a%d" (i+1)) t) aTyps in
    let bl = List.mapi (fun i t -> mkFormDom (spr "b%d" (i+1)) t) bTyps in
    let (afds,aes),(bfds,bes) = List.unzip al, List.unzip bl in
    let dc1, dc2 = Id.getVal Dcons.idTable j, Id.getVal Dcons.idTable j' in
    let tEqT = eqTypOfTii info tii in (* mkTypConst (liEqTyp tii) in *)
      (* omitting type parameters to data constructors *)
    let constructed1 = mkDconApp (mkVarT (liOfSl dc1)) [] aes in
    let constructed2 = mkDconApp (mkVarT (liOfSl dc2)) [] bes in
    let tNeq =
      mkPfDepArrow (afds @ bfds)
        (mkPfNot (mkPfEq tEqT constructed1 constructed2)) in
    let neqId = (incr Disequalities.count; !Disequalities.count) in
    let sigelt =
      if !Util.extract_proofs then
        let _ = Ht.add Disequalities.dcon neqId
                  (mkVarTTyp (liDconNeq neqId tii) tNeq) in
        [Sig_datacon_typ
           (liDconNeq neqId tii, [], tNeq, false, Public, None)]
      else
        [] in

    let lenA, lenB = List.length aTyps, List.length bTyps in
    let symDc1     = symOfDconinst (j, typArgTiil) in
    let symDc2     = symOfDconinst (j', typArgTiil) in

    let op =
      if lenA + lenB = 0 then
        (BG, Assume (Not (Eq (App (symDc1, [||]), App (symDc2, [||]))),
                     None,
                     ANeq (neqId, tii)))
      else
        let syms  = Array.of_list (List.map symOfFormDom (afds @ bfds)) in
        let sorts = Array.of_list (wrapSomes (allArgTiil @ allArgTiil')) in
        let bBvs =
          List.mapi
            (fun i tii -> BoundV (lenB - 1 - i, Some tii))
            allArgTiil' in
        let aBvs =
          List.mapi
            (fun i tii -> BoundV (lenB + lenA - 1 - i, Some tii))
            allArgTiil in
        (BG, Assume (Forall (sorts, syms,
                             Not (Eq (App (symDc1, Array.of_list aBvs),
                                      App (symDc2, Array.of_list bBvs)))),
                     None,
                     ANeq (neqId, tii)))
    in

    sigelt,[op]
  in
  List.fold_left
    (fun (acc1,acc2) x -> let l1,l2 = doOne x in acc1@l1, acc2@l2)
    ([],[]) l


(* this is ugly *)
let curMod                  = ref "!!!"
let setCurMod (li:lident)   = curMod := (List.hd li).idText


(* Creates Fine principles and Z3 ops for an equatable typinst.
   Does not drill recursively and process equatable sub-typinsts.
   Call this on a typinst that has already been determined to be equatable.
 *)
let rec processEquatableTypinst (info:info) (ti:typinst) : sigelt list * taggedop list =
  let tii = Id.processVal TypInsts.idTable ti in
    Ht.add TypInsts.modname tii !curMod;
    let sigelts = createSpecializedProofCombinators info tii in
    let ops = [(BG, Comment (spr "typinst: %s" (longStrTypinst ti)));
               (BG, DefSort (Some tii))] in
      
    let moreSigelts, moreOps = match ti with
    | TIMono(i) when "Prims.bool" = ssl (Id.getVal Tcons.idTable i) ->
        (* adding Prims.True, Prims.False, and their disequality manually
           since these constructors do not appear directly in prims.f9 *)
        let tEqBool = eqTypOfTii info tii in (* mkTypConst (liEqTyp tii) in *)
        let tNeqBool = mkPfNot (mkPfEq tEqBool expTrue expFalse) in
        let sigelt =
          Sig_datacon_typ
            (liDconNeqBool, [], tNeqBool, false, Public, None) in
        Disequalities.neqBool := mkVarTTyp liDconNeqBool tNeqBool;
        let ops =
          [(BG, DefFun (stringPrimsTrue, [||], Some tii, None));
           (BG, DefFun (stringPrimsFalse, [||], Some tii, None));
           (BG, Assume (Not (Eq (termTrue, termFalse)),
                        None,
                        ANeqBool))] in
        [sigelt], ops

    | TIPoly(i,til) -> begin (* push instantiation to its datacons *)
        let dconids = Ht.find Tcons.dconIds i in
        let tiil = List.map tiiOfTi til in
        Ht.replace Tcons.insts i (tiil::(Ht.find Tcons.insts i)); 
        let monoSigelts, dconOps, neqSigelts, neqOps, _ =
          List.fold_left
            (fun (monoSigelts,
                  dconOps,
                  neqSigelts,
                  neqOps,
                  dconsArgTyps) j ->
               let (l1,l2,allArgTiilOpt) = processDconinst info j tiil in
               let ((l3,l4),l5) =
                 (match allArgTiilOpt with
                    | None -> ([], []), []
                  (* explain tiil are the type parameter actuals,
                     argTiil are the term parameter types after substituting
                     the type actuals! *)
                    | Some(allArgTiil) -> 
                        (genDiseqs info tii j tiil allArgTiil dconsArgTyps,
                        [(j,allArgTiil)])
                 ) in
               (monoSigelts @ l1,
                dconOps @ l2,
                neqSigelts @ l3,
                neqOps @ l4,
                dconsArgTyps @ l5)
            )
            ([],[],[],[],[])
            dconids
        in
        (monoSigelts @ neqSigelts, dconOps @ neqOps)
      end

    | TITuple(ti1,ti2) -> [], processTupleInst info tii ti1 ti2

(*
    | TIRecord(l) ->
        let fields, til = List.split l in processRecordInst tii fields til
*)

    | TIRecordMono(i) ->
        let fields, tl = List.split (Ht.find Records.fieldsTyps i) in
        let til = List.map tiOfTyp tl in
        processRecordInst info tii i fields til

    | _ -> [], []
  in
  (sigelts@moreSigelts, ops@moreOps)
    

(* Processes typinst recursively, returning new sigelts and ops for all
   equatable types that have not been processed before. *)
and processForEquatableTypinsts (info:info) (ti:typinst) : sigelt list * taggedop list =
  if Id.mem TypInsts.idTable ti then [], [] (* already processed *)
  else match ti with
    | TIMono(_) -> (* all monomorphic tcons are equatable *)
        processEquatableTypinst info ti
    | TIPoly(_,l) ->
        let sigelts, ops = processListForEquatableTypinsts info l in
        let tiiol = List.map tiioOfTi l in
        (* if any of the parameter types is not equatable, then this
           instantiation of the polymorphic tcon is not equatable *)
        if existsNone tiiol then (sigelts, ops)
        else let moreSigelts, moreOps = processEquatableTypinst info ti in
             (sigelts@moreSigelts, ops@moreOps)
    | TITuple(ti1,ti2) ->
        let sigelts, ops = processListForEquatableTypinsts info [ti1;ti2] in
        (match tiioOfTi ti1, tiioOfTi ti2 with
           | Some _, Some _ ->
               (* both type params must be equatable for this tuple
                  instantiation to be equatable *)
               let moreSigelts, moreOps = processEquatableTypinst info ti in
               (sigelts@moreSigelts, ops@moreOps)
           | _ -> (sigelts, ops))
(*
    | TIRecord(l) ->
        let fields, til = List.split l in
        (* TODO this check is nasty, since it's also in processsigelt *)
        let firstfield = slOfLi (List.hd fields) in
        if Ht.mem firstFieldToExternRecordName firstfield then [], []
        else
          let sigelts, ops = processListForEquatableTypinsts til in
          let tiiol = List.map tiioOfTi til in
          if existsNone tiiol then (sigelts, ops)
          else let moreSigelts, moreOps = processEquatableTypinst ti in
               (sigelts@moreSigelts, ops@moreOps)
*)
    | TIRecordMono(i) ->
        let fields, ts = List.split (Ht.find Records.fieldsTyps i) in
(* 11/9 why was I not processing extern records
        (* TODO this check is nasty, since it's also in processsigelt *)
        let firstfield = slOfLi (List.hd fields) in
        if Ht.mem firstFieldToExternRecordName firstfield then [], []
*)
        if false then [], []
        else
          let til = List.map tiOfTyp ts in
          let tiiol = List.map tiioOfTi til in
          let sigelts, ops = processListForEquatableTypinsts info til in
          if existsNone tiiol then
            (*
            let _ = pr "not process equatable mono record %d\n" i; flush stdout in
            *)
             (sigelts, ops)
          else let moreSigelts, moreOps = processEquatableTypinst info ti in
            (*
            let _ = pr "process equatable mono record %d\n" i; flush stdout in
            *)
            (sigelts@moreSigelts, ops@moreOps)
    | TIOther -> [], [] (* not equatable, as determined by tiOfTyp *)


and processListForEquatableTypinsts info (l:typinst list)
    : sigelt list * taggedop list =
  List.fold_left (fun (acc1,acc2) ti ->
    let l1,l2 = processForEquatableTypinsts info ti in acc1@l1, acc2@l2
  ) ([],[]) l


and processDconinst info (di:dconid) (tiil:typinstid list) 
    : sigelt list * taggedop list * (typinstid list) option =

  (* argTyps, retTyp of dcon as it was defined *)
  let argTyps       = Ht.find Dcons.argTyps di in
  let retTyp        = Ht.find Dcons.retTyp di in
  let typBvdefs     = Ht.find Dcons.typBvdefs di in

  let til           = List.map (Id.getVal TypInsts.idTable) tiil in
  let argTypActuals = List.map synthesizeTyp til in
  let bvsArgs       = List.zip typBvdefs argTypActuals in

  (* argTyps, retTyp of dcon instantiated with type params *)
  let argTyps       = List.map (substBvdefs bvsArgs) argTyps in
  let retTyp        = substBvdefs bvsArgs retTyp in

  let retArgsTil    = List.map tiOfTyp (retTyp::argTyps) in

  let sigelts, ops  = processListForEquatableTypinsts info retArgsTil in

  let retArgsTiiol  = List.map tiioOfTi retArgsTil in

  let allArgsTiiol    = List.tl retArgsTiiol in
  let allArgsTiilOpt  = if existsNone allArgsTiiol then None
                        else Some (projSomes allArgsTiiol) in

  (* if any of the argument or return types, after substituting the actual
     type parameters, is not equatable, then the instantiated dcon is
     not equatable *)
  if existsNone retArgsTiiol then begin
    let retTiio, argsTiio = List.hd retArgsTiiol, List.tl retArgsTiiol in
    let moreOps =
      [(BG, Comment (spr "cannot generate monotonicity for dconinst %s, \
                       but will generate Z3 function"
                         (longStrOfDconinst (di,tiil))));
       (BG, DefFun (symOfDconinst (di,tiil),
                    Array.of_list argsTiio, retTiio, None))]
    in
    Ht.add DconInsts.retTiio (di,tiil) retTiio;
    (sigelts, ops@moreOps, allArgsTiilOpt)

  end else begin
    let tiis = projSomes retArgsTiiol in
    let retTii, argsTii = List.hd tiis, List.tl tiis in
    let dconLi = liOfSl (Id.getVal Dcons.idTable di) in
    let monoSigelt =
      if (List.length argTyps = 0) || (not !Util.extract_proofs) then []
      else begin
        let retTypFun xs ys =
          let dc = mkVarTTyp dconLi (Ht.find Dcons.typ di) in
          (* TODO omitting type parameters ok? *)
          let data1 = mkDconApp dc [] xs in
          let data2 = mkDconApp dc [] ys in
            mkTypApp (eqTypOfTii info retTii) [] [data1;data2] in
        let mono = createMonoRule info argsTii retTypFun in
        let monoLi = liDconMonoDconinst (di,tiil) in
        Ht.add DconInsts.dconMono (di,tiil) (mkVarTTyp monoLi mono);
        [Sig_datacon_typ (monoLi, [], mono, false, Public, None)]
      end
    in
    let argSorts, retSort = wrapSomes argsTii, Some retTii in
    let moreOps =
      [(BG, Comment (spr "dconinst: %s" (longStrOfDconinst (di,tiil))));
       (BG, DefFun (symOfDconinst (di,tiil),
                    Array.of_list argSorts, retSort, None))]
    in
    Ht.add DconInsts.retTiio (di,tiil) retSort;

(*
    (* TODO need Fine dcon and ops for injectivity *)

    let injOps = doInjectivity (di,tiil) retTii argsTii in

    (sigelts@monoSigelt, ops@moreOps@injOps, allArgsTiilOpt)
*)

    (sigelts@monoSigelt, ops@moreOps, allArgsTiilOpt)
(*
*)

  end


(****************************************************************************)
(* PRINT AUTO-GENERATED PRINCIPLES                                          *)
(****************************************************************************)

let dumpAutoGenSigs_ oc =
  let printCaption s = log oc (spr "\n(********** %s **********)\n" s) in
  let printDcon li t =
    log oc (spr "\n  | %s : %s\n" (sli li) (Pretty.strTyp t)) in
  let printVal li t = 
    log oc (spr "\nval %s : %s\n" (sli li) (Pretty.strTyp t)) in
  log oc "\n(* Typ_unknown means that element was not generated *)\n";
    Id.iter TypInsts.idTable (fun tii _ ->
                                printCaption (Pretty.strTyp (synthesizeTyp (tiOfTii tii))));
    Id.iter Props.idTable (fun propid sl ->
                             printCaption (spr "Proposition %d %s" propid (ssl sl)));
    Ht.iter DconInsts.dconMono (fun (di,tiil) dconMono ->
                                  printCaption (spr "DconInst %s" (longStrOfDconinst (di,tiil)));
                                  printDcon (liDconMonoDconinst (di,tiil)) dconMono.sort);
    Ht.iter TupleInsts.argTiis 
      (fun tii _ ->
         let tii1, tii2  = Ht.find TupleInsts.argTiis tii in
           printCaption (spr "TupleInst: %d = (%d, %d)" tii tii1 tii2));
    Id.iter Fields.idTable (fun fieldid sl ->
                              printCaption (spr "Field %d = %s" fieldid (ssl sl)));
    Ht.iter RecordInsts.fieldTiil (fun tii tiil ->
                                     let fields = Ht.find RecordInsts.fieldnames tii in
                                       printCaption (spr "RecordInst %d = %s"
                                                       tii (String.concat ", " (List.map soi tiil)));
                                       List.iter (fun field ->
                                                    let j = Id.getId Fields.idTable (slOfLi field) in
                                                    let proj = AutoAxioms.recdProj (tii,j) in
                                                    let eq   = AutoAxioms.recdEq (tii,j) in
                                                      printDcon proj.v proj.sort;
                                                      printDcon eq.v eq.sort
                                                 ) fields;
                                       let mono = AutoAxioms.recdMonoEq tii in
                                         printDcon mono.v mono.sort
                                  );
    printCaption "Disequalities";
    let neqBool = !Disequalities.neqBool in
      printDcon neqBool.v neqBool.sort;
      Ht.iter Disequalities.dcon (fun i dc -> printDcon dc.v dc.sort);
      ()
        
let dumpAutoGenSigs oc =
  try dumpAutoGenSigs_ oc
  with _ -> log oc "\n\n(* STOPPING DUMP *)\n\n"

let printTables () = dumpAutoGenSigs stdout

let printAutoGenSigsToFile () =
  let streamAutoGen = new System.IO.StreamWriter("queries/autogen.f9") in
  dumpAutoGenSigs streamAutoGen;
  streamAutoGen.Flush (); streamAutoGen.Close ();
  ()

