(*
  Copyright (c) 2009, Regents of the University of California

  Authors: Ravi Chugh
  
  All rights reserved.
  
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following 
  conditions are met:
  
  1. Redistributions of source code must retain the above copyright 
  notice, this list of conditions and the following disclaimer.

  2. Redistributions in binary form must reproduce the above 
  copyright notice, this list of conditions and the following disclaimer 
  in the documentation and/or other materials provided with the distribution.

  3. Neither the name of the University of California, San Diego, nor 
  the names of its contributors may be used to endorse or promote 
  products derived from this software without specific prior 
  written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*)


#light "off"

module Microsoft.Fine.ProofCommon

open System.Diagnostics
open System.Text.RegularExpressions
open Microsoft.Z3
open Microsoft.Fine
open ProofUtils
open Absyn


let _ = Profiling.new_counter "Z3"
let _ = Profiling.new_counter "ProofZZZ.printCompactProof"
let _ = Profiling.new_counter "Proof.doProcessSigelt"
let _ = Profiling.new_counter "ProofExtract.termToTyp"
let _ = Profiling.new_counter "ProofExtract.termToExp"
let _ = Profiling.new_counter "ProofExtract.destructPats"
let _ = Profiling.new_counter "ProofExtract.synthesizeProof"

let profile x y = Profiling.profile y x


type ProofDebugLog() = class
(*   [<Conditional("DEBUG")>] *)
  static member Log (oc,s) = fpr oc "%s" s
end

let log oc s = ProofDebugLog.Log(oc,s)



let strTyp t = (Pretty.strTyp t).Replace('\n',' ')
let strExp e = (Pretty.strExp e).Replace('\n',' ')

let sli       = Pretty.str_of_lident
let strTerm   = fun _ -> "_Z3.Term_"


let doDisequalities     = ref true


let queryCount          = ref 0

let queryFile ()        = spr "queries/query-%04d.smt" !queryCount
let proofFile ()        = spr "queries/proof-%04d.txt" !queryCount
let longproofFile ()    = spr "queries/longproof-%04d.txt" !queryCount
let fineproofFile ()    = spr "queries/fineproof-%04d.f9" !queryCount
let streamQueries       = new System.IO.StreamWriter("queries/all.smt")

let stringPrimsTrue     = "Prims.True"
let stringPrimsFalse    = "Prims.False"
let sortOther           = "other"

type tconid     = int
type dconid     = int
type fieldid    = int
type propid     = int
type recdconid  = int

type typinst =
  | TIMono    of tconid
  | TIPoly    of tconid * typinst list
  | TITuple   of typinst * typinst
(*
  | TIRecord  of (fieldname * typinst) list (* TODO change fieldname to string list *)
*)
  | TIRecordMono of recdconid
(*
  | TIRecordPoly of recdconid * typinst list
*)
  | TIOther

type typinstid  = int
type dconinst   = dconid * typinstid list
type sort       = typinstid option

let strSort = function Some tii -> spr "tii_%03d" tii | None -> sortOther

let strSortToSort s =
  let re = new Regex("^tii_([0-9][0-9][0-9])$") in
  let m = re.Match s in
  if m.Success
    then Some (ios (m.Groups.get_Item(1).Value))
    else None


type assumptionId =
  | AUser          of int
  | ABindingVar    of int
  | ABindingMatch  of int
  | ATupProj       of typinstid * int  (* 1 or 2 *)
  | ATupEq         of typinstid * int  (* 1 or 2 *)
  | ARecdProj      of typinstid * fieldid
  | ARecdEq        of typinstid * fieldid
  | ANeqBool
  | ANeq           of int * typinstid
  | AOther (* remove this placeholder *)

let strOfAssumptionId = function
  | AUser i           -> spr "AUser %d" i
  | ABindingVar i     -> spr "ABindingVar %d" i
  | ABindingMatch i   -> spr "ABindingMatch %d" i
  | ATupProj(tii,i)   -> spr "ATupProj_%d %d" tii i
  | ATupEq(tii,i)     -> spr "ATupEq_%d %d" tii i
  | ARecdProj(tii,j)  -> spr "ARecdProj_%d_%d" tii j
  | ARecdEq(tii,j)    -> spr "ARecdEq_%d_%d" tii j
  | ANeqBool          -> "ANeqBool"
  | ANeq(i,tii)       -> spr "ANeq_%d_%d" i tii
  | AOther            -> "AOther"

type assumptionKind =
  | AKUser
  | AKBindingVar
  | AKBindingMatch
  | AKNeqBool

let mkAid =
  let next x = incr x; !x in
  let k1, k2, k3 = ref 0, ref 0, ref 0 in
  (function
     | AKUser           -> AUser (next k1)
     | AKBindingVar     -> ABindingVar (next k2)
     | AKBindingMatch   -> ABindingMatch (next k3)
     | _                -> raise (Never "mkAid"))

type caption = string

type term =
  | True
  | False
  | BoundV   of int * sort
  | FreeV    of string * sort
  | App      of string * term array
  | Not      of term
  | And      of term * term
  | Or       of term * term
  | Imp      of term * term
  | Iff      of term * term
  | Eq       of term * term
  | Forall   of sort array * string array * term (* can make these scalars instead of arrays *)
  | Exists   of sort * string * term

type op =
  | DefSort  of sort
  | DefPred  of string * sort array * caption option
  | DefFun   of string * sort array * sort * caption option
  | Assume   of term * caption option * assumptionId
  | Query    of term
  | Comment  of caption
  | Warning  of caption
(*   | JustMake of term * int *)

type fgbg      = FG | BG
type taggedop  = fgbg * op

type info      = { envn: Tcenv.env;        binders: string list }
let inf e      = { envn = e;               binders = [] }
let emptyInfo  = { envn = Tcenv.empty_env; binders = [] }

let termTrue            = App (stringPrimsTrue, [||])
let termFalse           = App (stringPrimsFalse, [||])

let fgComment s         = [(FG, Comment s)]
let fgWarn s            = [(FG, Warning s)]
let bgWarn s            = [(BG, Warning s)]


(* TODO might not need this, if not doing anything with tupled proofs *)
type pfbinding =
  | BarePf of string
  | SndPf of string


(****************************************************************************)
(* Z3 CONTEXT AND TERM CACHE                                                *)
(****************************************************************************)

let z3 =
  let par  = new Config() in
  let _    = par.SetParamValue("PROOF_MODE", "2") in
  let _    = par.SetParamValue("DISPLAY_PROOF", "true") in
  let _    = par.SetParamValue("MODEL", "true") in
  let ctx  = new Context(par) in
  let _    = par.Dispose () in
  ctx

let idTerm : Term IdTable.t = IdTable.create "idTerm" strTerm

(* Assigns unique ids to Z3 Term and all its subterms. *)
let rec assignIdsToTerm t =
  if IdTable.mem idTerm t then ()
  else begin
    (match z3.GetTermKind t with
       | TermKind.App         -> Array.iter assignIdsToTerm (z3.GetAppArgs t)
       | TermKind.Quantifier  -> assignIdsToTerm (z3.GetQuantifier t).Body
       | TermKind.Var         -> ()
       | _                    -> ());
    ignore (IdTable.processVal idTerm t)
  end
    
    
(****************************************************************************)
(* A FEW GLOBAL TABLES                                                      *)
(****************************************************************************)

let userAxiom : (int, lident * typ) Ht.t = Ht.create "userAxiom" soi

(* let skAxiomCount = ref 0 *)

(* let skTermToInt : (Term, int) Ht.t = Ht.create "skTermToInt" strTerm *)

(* let skolemizeAxiom : (int, lident * typ) Ht.t = Ht.create "skolemizeAxiom" soi *)

(*
let skolemizeAxioms : ((lident * typ) Zset.t) ref = ref (Zset.empty compare)
*)

(*
let assumeTerm : (Term, assumptionId) Ht.t = Ht.create "assumeTerm" strTerm
*)

let bgAssumeTerm : (Term, assumptionId) Ht.t = Ht.create "bgAssumeTerm" strTerm

let fgAssumeTerm : (Term, assumptionId) Ht.t = Ht.create "fgAssumeTerm" strTerm

let symToFormDom : (string, bvdef * typ) Ht.t = Ht.create "symToFormDom" id
  
(* TODO this is unreliable! *)
let varTermToSort : (Term, sort) Ht.t = Ht.create "varTermToSort" strTerm

(* since binding vars are not being reset in between queries, this table isn't being reset either *)
let bindingVarIdToPf : (int, pfbinding) Ht.t = Ht.create "bindingVarIdToPf" soi

let binderOfDepTupleProof : (ident, ident) Ht.t = Ht.create "binderOfDepTupleProof" (fun id -> id.idText)



(****************************************************************************)
(* Z3 TERM TO STRING                                                        *)
(****************************************************************************)

let rec indent k = if k = 0 then "" else "  " ^ indent (k-1)

let rec stringTerm_ k t = match z3.GetTermKind t with
  | TermKind.Numeral -> "numeral..."
  | TermKind.App ->
      let f, xs = z3.GetAppDecl t, z3.GetAppArgs t in
      let r = z3.GetRange f in
      let s1 = spr "%s(%s\n" (indent k) (f.GetDeclName()) in
      let s2 = String.concat "" (Array.to_list (Array.map (stringTerm (k+1)) xs)) in
      spr "%s%s" s1 s2
  | TermKind.Var -> spr "%s#%d\n" (indent k) (z3.GetVarIndex t)
  | TermKind.Quantifier ->
      let q = z3.GetQuantifier t in
      let pats = Array.map z3.GetPatternTerms q.Patterns in
      let binders = List.zip (Array.to_list q.Names) (Array.to_list q.Sorts) in
      let x = if q.IsForall then "forall" else "exists" in
      let y = String.concat " " (List.map
        (fun (a:Microsoft.Z3.Symbol,b:Microsoft.Z3.Sort) ->
          spr "(%s %s)" (z3.GetSymbolString(a)) (b.GetName())
        ) binders) in
      spr "%s%s %s\n%s" (indent k) x y (stringTerm (k+1) q.Body)
  | TermKind.Unknown -> "unknown termkind..."
  | _ -> raise (Never "stringTerm_")

and stringTerm k t = spr "%s" (stringTerm_ k t)
    
let printTerm oc k t = log oc (spr "%s" (stringTerm k t))


