(*
  Copyright (c) 2009, Regents of the University of California

  Authors: Ravi Chugh
  
  All rights reserved.
  
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following 
  conditions are met:
  
  1. Redistributions of source code must retain the above copyright 
  notice, this list of conditions and the following disclaimer.

  2. Redistributions in binary form must reproduce the above 
  copyright notice, this list of conditions and the following disclaimer 
  in the documentation and/or other materials provided with the distribution.

  3. Neither the name of the University of California, San Diego, nor 
  the names of its contributors may be used to endorse or promote 
  products derived from this software without specific prior 
  written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*)


#light "off"

module Microsoft.Fine.ProofUtils


exception Never of string


let pr  = Printf.printf
let spr = Printf.sprintf
let fpr = Printf.fprintf 

let ios = int_of_string
let soi = string_of_int
let iof = int_of_float
let foi = float_of_int



let strHead (s:string) : string = string (s.[0])

let strTail (s:string) =
  let rec f k acc =
    if k = 0 then f (k+1) acc 
    else if k = s.Length then acc
    else f (k+1) (acc ^ (string s.[k]))
  in
  f 0 ""

let strFlags l =
  String.concat "" (List.map (function true -> "1" | false -> "0") l)

let strIntList l = String.concat "" (List.map soi l)

(* assumes each number is one digit *)
let stringToIntList s =
  let rec f acc s =
    if s = "" then List.rev acc
    else f (strHead s :: acc) (strTail s) in
  f [] s




let array0N n = Array.init (succ n) (fun i -> i)
let array1N n = Array.init n (fun i -> i + 1)

let list0N n = Array.to_list (array0N n)
let list1N n = Array.to_list (array1N n)

let map1N n f = List.map f (list1N n)

(* lookup in associative list *)
let rec find p = function
  | []       -> raise Not_found
  | (x,y)::t -> if p x then y else find p t

let findIdx x l =
  let rec f k = function
    | [] -> None
    | y::tl -> if x = y then Some k else f (succ k) tl
  in f 0 l

let interleave l1 l2 =
  if List.length l1 <> List.length l2 then failwith "interleave" else
  let rec f acc = function
    | [], []       -> acc
    | x::xs, y::ys -> f (y::x::acc) (xs,ys)
    | _            -> raise (Never "interleave")
  in
  List.rev (f [] (l1,l2))

let invertBijection l =
  List.map
    (fun i -> match findIdx i l with
                | Some j -> j+1
                | None   -> failwith "invertBijection: arg wasn't a bij")
    (list1N (List.length l))
    


let existsNone l = List.exists (function None -> true | _ -> false) l

let wrapSomes l = List.map (function x -> Some x) l

let projSome = (function Some x -> x | _ -> raise (Never "projSome"))

let projSomes l = List.map projSome l


let splitAntesAndGoal l =
  let l_rev = List.rev l in
  List.rev (List.tl l_rev), List.hd l_rev

module Ht = struct

  type ('a,'b) t = ('a,'b) Hashtbl.t * string * ('a -> string)

  let create s f : ('a,'b) t = (Hashtbl.create 17, s, f)

  let add ((ht,_,_):('a,'b) t) (k:'a) (v:'b) = Hashtbl.add ht k v

  let replace ((ht,_,_):('a,'b) t) (k:'a) (v:'b) = Hashtbl.replace ht k v

  let mem ((ht,_,_):('a,'b) t) (k:'a) = Hashtbl.mem ht k

  let iter ((ht,_,_):('a,'b) t) f = Hashtbl.iter f ht

  let fold f ((ht,_,_):('a,'b) t) x = Hashtbl.fold f ht x

  let find ((ht,s,f):('a,'b) t) (k:'a) =
    if Hashtbl.mem ht k then Hashtbl.find ht k
    else failwith (spr "Ht.find: %s not in table %s" (f k) s)

  let clear ((ht,_,_):('a,'b) t) = Hashtbl.clear ht

  let remove ((ht,_,_):('a,'b) t) (x:'a) = Hashtbl.remove ht x

end


module IdTable = struct

  type 'a t = int ref
                * (int, 'a) Hashtbl.t 
                * ('a, int) Hashtbl.t 
                * string 
                * ('a -> string)

  let create s f = ref 0, Hashtbl.create 17, Hashtbl.create 17, s, f

  let processVal ((k,ht1,ht2,_,_):'a t) (x:'a) =
    if Hashtbl.mem ht2 x then Hashtbl.find ht2 x
    else (incr k; let i = !k in Hashtbl.add ht1 i x; Hashtbl.add ht2 x i; i)

  let getVal ((_,ht1,_,s,_):'a t) i =
    if Hashtbl.mem ht1 i then Hashtbl.find ht1 i
    else failwith (spr "IdTable.getVal: %d not in table %s" i s)

  let getId ((_,_,ht2,s,f):'a t) (x:'a) =
    if Hashtbl.mem ht2 x then Hashtbl.find ht2 x
    else failwith (spr "IdTable.getId: %s not in table %s" (f x) s)

  let iter ((_,ht1,_,_,_):'a t) f = Hashtbl.iter f ht1

  let clear ((k,ht1,ht2,_,_):'a t) =
    Hashtbl.clear ht1;
    Hashtbl.clear ht2;
    k := 0

  let mem ((_,_,ht2,_,_):'a t) = Hashtbl.mem ht2

end

