(*
  Copyright (c) 2009, Regents of the University of California

  Authors: Ravi Chugh
  
  All rights reserved.
  
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following 
  conditions are met:
  
  1. Redistributions of source code must retain the above copyright 
  notice, this list of conditions and the following disclaimer.

  2. Redistributions in binary form must reproduce the above 
  copyright notice, this list of conditions and the following disclaimer 
  in the documentation and/or other materials provided with the distribution.

  3. Neither the name of the University of California, San Diego, nor 
  the names of its contributors may be used to endorse or promote 
  products derived from this software without specific prior 
  written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*)


#light "off"

module Microsoft.Fine.ProofZZZ

open Microsoft.Z3
open Microsoft.Fine
open ProofUtils
open ProofCommon


(****************************************************************************)
(* Z3 PROOF PRINTER                                                         *)
(****************************************************************************)

let printCompactProof (oc:System.IO.TextWriter) (t:Term) =
  if !Util.silent then () else
  let alreadyStringified      = ref [] in
  let proofIndices            = ref [] in
  let shortString             = Hashtbl.create 17 in
  let longString              = Hashtbl.create 17 in
  let formulaOfProof          = Hashtbl.create 17 in
  let expandShort             = Hashtbl.find shortString in
  let expandShortList         = List.map expandShort in
  let expandLong              = Hashtbl.find longString in
  let expandLongList          = List.map expandLong in

  let stringifyTermOnce t =
    match z3.GetTermKind t with
      | TermKind.App -> "proof" = ((z3.GetRange (z3.GetAppDecl t)).GetName())
      | _ -> false
  in

  let rec stringify formals t =
    let k = IdTable.getId idTerm t in
(* TODO 11/6
    if List.mem k !alreadyStringified then k
*)
(*
    if List.mem k !alreadyStringified && (match z3.GetTermKind t with TermKind.Var -> false | _ -> true) then k
*)
(*
    if false then k (* rethink caching. i don't think it's ever useful actually. *)
*)
    if stringifyTermOnce t && List.mem k !alreadyStringified then k
    else begin
      let short, long = match z3.GetTermKind t with
        | TermKind.App ->
            let f, xs = z3.GetAppDecl t, z3.GetAppArgs t in
            let funsym = f.GetDeclName () in 
            (* the cache ids of all arguments *)
            let ys = Array.to_list (Array.map (stringify formals) xs) in
            begin match (z3.GetRange f).GetName() with (* check sort *)
               | "proof" -> begin
                   proofIndices := k :: !proofIndices;
                   let argIds, goalId = splitAntesAndGoal ys in
                   Hashtbl.add formulaOfProof k goalId;
                   let args = String.concat " " (expandShortList argIds) in
                   spr "p%d_f%d" k goalId, spr "[%s %s]" funsym args
                 end
               | "bool" ->
                   let args = String.concat " " (expandLongList ys) in
                   spr "f%d" k, spr "(%s %s)" funsym args
               | _ ->
                  let args = String.concat " " (expandLongList ys) in
                  if List.length ys = 0
                    then spr "v%d" k, spr "%s" funsym
                    else spr "v%d" k, spr "(%s %s)" funsym args
            end
        | TermKind.Var ->
(* TODO 11/6
            let i = int (z3.GetVarIndex t) in
            let x = List.nth formals i in
            spr "#%d" k, spr "%s" x
*)
            let i = int (z3.GetVarIndex t) in
            if i < List.length formals
              then spr "#%d" k, spr "%s" (List.nth formals i)
              else spr "#%d" k, spr "#%d" i
        | TermKind.Quantifier ->
            let q = z3.GetQuantifier t in
            let x = if q.IsForall then "forall" else "exists" in
            let names = Array.to_list q.Names in
            let sorts = Array.to_list q.Sorts in
            let y =
              List.map
(*
                (fun (a:Symbol,b:Sort) -> spr "%s:%s" (z3.GetSymbolString(a)) (b.GetName()))
*)
                (fun (a:Symbol,b:Sort) -> spr "(%s %s)" (z3.GetSymbolString(a)) (b.GetName()))
                (List.zip names sorts) in
            let formals' = 
              List.rev
                (List.map (fun (a:Symbol) -> z3.GetSymbolString(a)) names) in
            let z = expandLong (stringify (formals' @ formals) q.Body) in
(*
            spr "#%d" k, spr "(%s (%s) %s)" x (String.concat " " y) z
*)
            spr "#%d" k, spr "(%s %s %s)" x (String.concat " " y) z
        | _ -> raise (Never "stringify")
      in
      Hashtbl.add shortString k short;
      Hashtbl.add longString  k long;
      alreadyStringified := k :: !alreadyStringified;
      k
    end
  in
  ignore (stringify [] t); (* populate the table *)
  List.iter                (* print proof Terms  *)
    (fun i ->
      let j = Hashtbl.find formulaOfProof i in
      log oc (spr "%12s %s\n                %s\n" (expandShort i) (expandLong i) (expandLong j)))
    (List.rev !proofIndices)


(****************************************************************************)
(* CHECKING Z3 ANSWER TO QUERY                                              *)
(****************************************************************************)

let z3QueryResult = ref (false, None)

let processZ3Answer (info:info) (x:Term) =

  let model = ref (null:Model) in
  let proof = ref (null:Term) in
  let core  = ref (null:Term array) in

  let res,goalvar =
    Profiling.profile
      (fun () -> 
         (*          let nt = termToTerm (Not t) in *)
         (*          let p = z3.MkFreshConst("p", z3.MkBoolSort()) in *)
         (*          let _ = z3.AssertCnstr(z3.MkIff(p, nt)) in *)
         (*            processZ3Answer p *)
         let p = z3.MkFreshConst("__p", z3.MkBoolSort()) in
           z3.AssertCnstr(z3.MkIff(p, x));
           z3.CheckAssumptions (model, [|x|], proof, core), p) "Z3" in
    
  match res with
    | LBool.True -> begin
        (!model).Dispose ();
        log streamQueries "; unsat? false\n"; streamQueries.Flush ();
        z3QueryResult := (false, None);
      end
    | LBool.False -> begin
        assignIdsToTerm !proof;

        let stream = new System.IO.StreamWriter (proofFile ()) in
        profile "ProofZZZ.printCompactProof" (fun () -> printCompactProof stream !proof);
        stream.Flush ();
        stream.Close ();

        let stream = new System.IO.StreamWriter (fineproofFile ()) in
        let proofOpt =
          if !Util.extract_proofs
            then Some (ProofExtract.generateProofTerm info stream goalvar x !proof).v
            else None in
        stream.Flush ();
        stream.Close ();
        log streamQueries "; unsat? true\n"; streamQueries.Flush ();

        z3QueryResult := (true, proofOpt);
      end
    | LBool.Undef -> begin
        (!model).Dispose ();
        log streamQueries "; unsat? unknown\n"; streamQueries.Flush ();
        z3QueryResult := (false, None);
      end
    | _ -> raise (Never "processZ3Answer")



(****************************************************************************)
(* Z3 TERM BUILDING                                                         *)
(****************************************************************************)

(* Z3 creates sorts without requiring them to be explicitly defined first.
   This wrapper ensures that Def(Sort s) explicitly precedes all uses of
   the sort s. *)
let (defSort : string -> unit),
    (getSort : string -> Sort) =
  let x = ref (Zset.empty compare) in
  (fun s -> x := Zset.add s !x),
  (fun (s:string) ->
     if Zset.mem s !x then z3.MkSort s
     else raise (Never (spr "undefined sort: [%s]" s)))

let sortToSort (tiio:sort) : Sort = getSort (strSort tiio)

let (defFunSym : string -> FuncDecl -> unit),
    (getFunSym : string -> FuncDecl) =
  let x = ref [] in (* assoc list of FuncDecls *)
  (fun s fd -> x := (s,fd) :: !x),
  (fun s ->
     try find ((=) s) !x
     with e -> raise (Never (spr "getFunSym: [%s]" s)))

let boolSort          = z3.MkBoolSort()
let trueTerm          = z3.MkTrue()
let falseTerm         = z3.MkFalse()





let rec termToSmt (info:info) = function
  | True          -> "(true)"
  | False         -> "(false)"
  | BoundV(i,_)   -> List.nth info.binders i
  | FreeV(x,_)    -> x
  | App(f,es)     -> let a = Array.map (termToSmt info) es in
                     let s = String.concat " " (Array.to_list a) in
                     spr "(%s %s)" f s
  | Not(x)        -> spr "(not %s)" (termToSmt info x)
  | And(x,y)      -> spr "(and %s %s)" (termToSmt info x) (termToSmt info y)
  | Or(x,y)       -> spr "(or %s %s)" (termToSmt info x) (termToSmt info y)
  | Imp(x,y)      -> spr "(implies %s %s)"(termToSmt info x)(termToSmt info y)
  | Iff(x,y)      -> spr "(iff %s %s)" (termToSmt info x) (termToSmt info y)
  | Eq(x,y)       -> spr "(= %s %s)" (termToSmt info x) (termToSmt info y)
  | Forall(x,y,z) ->
      let s = String.concat " "
                (Array.map (fun (a,b) -> spr "(%s %s)" a (strSort b))
                           (Array.zip y x)) in
      let binders' = (List.rev (Array.to_list y)) @ info.binders in
      let info' = { info with binders=binders' } in
      spr "(forall %s %s)" s (termToSmt info' z)
  | Exists(x,y,z) ->
      let info' = { info with binders=y::info.binders } in
      spr "(exists (%s %s) %s)" y (strSort x) (termToSmt info' z)


let rec termToTerm (term:term) : Term =
(*
  try begin
*)
    let tm = termToTerm_ term in
    tm
(*
  end with :? Microsoft.Z3.Z3Error -> begin
    stdout.Flush ();
    streamQueries.Flush ();
    failwith (spr "\n\n  Z3 died on term:\n\n  %s\n\n"
      (termToSmt emptyInfo term))
  end
*)

and termToTerm_ = function
  | True          -> trueTerm
  | False         -> falseTerm
  | BoundV(i,s)   -> let tm = z3.MkBound (uint32 i, sortToSort s) in
                     let _ = Ht.add varTermToSort tm s in
                     tm
  | FreeV(x,tiio) -> z3.MkConst (x, sortToSort tiio) 
  | App(f,es)     -> z3.MkApp (getFunSym f, Array.map termToTerm es) 
  | Not(x)        -> z3.MkNot (termToTerm x)
  | And(x,y)      -> z3.MkAnd [|termToTerm x; termToTerm y|]
  | Or(x,y)       -> z3.MkOr [|termToTerm x; termToTerm y|]
  | Imp(x,y)      -> z3.MkImplies (termToTerm x, termToTerm y)
  | Iff(x,y)      -> z3.MkIff (termToTerm x, termToTerm y)
  | Eq(x,y)       -> z3.MkEq (termToTerm x, termToTerm y)
  | Forall(x,y,z) -> 
      let rec flatten_forall out = function
        | Forall (x',y', z) -> flatten_forall ((x',y')::out) z
        | z -> 
            let sorts, names = List.unzip (List.rev out) in
              (Array.concat sorts, Array.concat names, z) in
      let sorts, names, term = flatten_forall [(x,y)] z in
        z3.MkForall (0u, [||], Array.map sortToSort sorts, names, termToTerm term)
  | Exists(x,y,z) -> z3.MkExists (0u, [||], [|sortToSort x|], [|y|], termToTerm z)
      

let opToSmt info = function
  | DefSort(tiio) -> spr "\n:extrasorts (%s)" (strSort tiio)
  | Query(t)      -> spr "\n:formula    (not %s)" (termToSmt info t)
  | Comment(c)    -> spr "\n; %s" c
  | Warning(c)    -> spr "\n; WARNING: %s" c
  | DefPred(f,tiios,_) ->
      let l = Array.to_list (Array.map strSort tiios) in
      spr "\n:extrapreds ((%s %s))" f (String.concat " " l)
  | DefFun(f,ts,t,_) ->
      if Array.length ts = 0 then spr "\n:extrafuns  ((%s %s))" f (strSort t)
      else
        let l = Array.to_list (Array.map strSort ts) in
        spr "\n:extrafuns  ((%s %s %s))" f (String.concat " " l) (strSort t)
  | Assume(t,co,aid) ->
      let s = strOfAssumptionId aid in
      let c = match co with Some c -> spr ";;;;;;;;;;; %s (aid %s)\n" c s
                          | None -> spr "; (aid %s)\n" s in
      spr "\n%s:assumption %s" c (termToSmt info t)
(*   | JustMake(_,_) -> "; JustMake" *)


let opToZ3_ info bg = function
  | DefSort(s)       -> defSort (strSort s)
  | DefPred(f,ts,_)  -> defFunSym f (z3.MkFuncDecl (f, Array.map sortToSort ts, boolSort))
(*
*)
  | DefFun(f,ts,t,_) -> defFunSym f (z3.MkFuncDecl (f, Array.map sortToSort ts, sortToSort t))
(* TODO does using MkInjectiveFunction break compatability with SMT? *)
(*
  | DefFun(f,ts,t,_) -> defFunSym f (z3.MkInjectiveFunction (f, Array.map sortToSort ts, sortToSort t))
*)
  | Assume(t,_,aid)  -> let tm = termToTerm t in
(*
                        let _ = Ht.add assumeTerm tm aid in
*)
                        let _ = Ht.add (if bg then bgAssumeTerm else fgAssumeTerm) tm aid in
                        z3.AssertCnstr tm
(* TODO 11/14 try exists instead of not forall *)
  | Query(t)         -> 
      processZ3Answer info (termToTerm (Not t))
        
(*   | JustMake(t,i)    -> let tm = termToTerm t in *)
(*                         let _ = Ht.add skTermToInt tm i in *)
(*                         () *)
  | Comment _        -> ()
  | Warning _        -> ()

let opToZ3 (info:info) (bg:bool) (op:op) : unit =
  (* this profiling call includes a bit of overhead in recursive calls
     and other small work before making z3 calls *)
  try Profiling.profile (fun () -> opToZ3_ info bg op) "Z3"
  with :? Microsoft.Z3.Z3Error -> begin
    stdout.Flush ();
    streamQueries.Flush ();
    failwith (spr "\n\n  Z3 died on op:\n%s\n\n" (opToSmt info op))
    (* when this error gets thrown, it's useful to run make smt *)
  end


let backgroundOps = ref []

let writeSmtFile info oc title ops =
  if !Util.silent then () else
    begin
      log oc "\n;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;\n";
      log oc (spr "\n(benchmark %s\n" title);
      log oc (spr "%s" (String.concat "\n" (List.map (opToSmt info) ops)));
      log oc "\n)\n";
      oc.Flush ()
    end

let processOps (info:info) (l:taggedop list) title =

  (* write all operations to all.smt *)
  writeSmtFile info streamQueries title (List.map snd l);
  
  (* partition operations into foreground and background *)
  let fgOps, bgOps =
    List.fold_left
      (fun (a,b) -> function (FG,op) -> op::a, b | (BG,op) -> a, op::b) 
      ([],[]) l in
  let fgOps = List.rev fgOps in
  let bgOps = List.rev bgOps in

  (* process background ops *)
  backgroundOps := !backgroundOps @ bgOps;
  List.iter (opToZ3 info true) bgOps;

  (* if any foreground ops, then process them in new a context. also
     write all background and foreground ops to a new query-xxx.smt file *)
  if List.length fgOps > 0 then begin

    let stream = new System.IO.StreamWriter (queryFile ()) in
    writeSmtFile info stream "background" !backgroundOps;
    writeSmtFile info stream (spr "query_%04d" !queryCount) fgOps;
    stream.Flush ();
    stream.Close ();

    Profiling.profile (fun () -> z3.Push ()) "Z3";
    IdTable.clear idTerm;
    Ht.clear fgAssumeTerm;
    List.iter (opToZ3 info false) fgOps;
    Profiling.profile (fun () -> z3.Pop ()) "Z3";

  end;

  ()


