#light "off"
// (c) Microsoft Corporation. All rights reserved
module Microsoft.Fine.Target

open Util
open Const

type tIdent = string          (* identifier *)
type tLident = tIdent list    (* long identifier *)
type tBvDef = tIdent * tIdent (* pretty-printing name and real name *)

type tVar<'t> = tIdent * 't   (* variables, including type var and value var *)
type tName<'t> = tLident * 't (* names [for external types, etc] *)
type tClassName = tLident     (* full name of a class *)
type tFieldName = string
type tMethodName = string

type tPrimop = | AND | OR | ADD | SUB | MUL | DIV | NOT

type tFieldRef = tFieldName

type tMethodRef = tMethodName * tType list

and tKind =
  | TKind_star                     (* normal kind *)
  | TKind_affine                   (* affine kind *)
  | TKind_arrow of tType * tKind   (* t -> k where t cannot be an affine type *)
  | TKind_karrow of tKind * tKind  (* k1 -> k2 where k1 and k2 are * or A *)

and tType =
  | TType_index of uint16
  | TType_var of tVar<tKind>                    (* type variable *)
  | TType_name of tName<tKind>                  (* type names [for external types] *)
  | TType_fun of option<tBvDef> * tType * tType (* dependent type function *)
  | TType_tfun of tBvDef * tKind * tType        (* \a::k.t *)
  | TType_dep of tType * tValue                 (* dependent type function application *)
  | TType_tapp of tType * tType                 (* type application *)
  | TType_concrete of tTypeConcrete             (* data class *)
  | TType_affine of tType                       (* affine type *)

and tTypeConcrete = tClassName * tType list * tValue list (* abstract/data class *)

and tValue = 
  | TVal_var of tVar<tType>                 (* variable *)
  | TVal_obj of tTypeConcrete * list<tType> * tValue list (* object *)
  | TVal_constant of Sugar.sconst           (* constant *)
  | TVal_ldfld of tValue * tFieldRef

type tExp =
  | TExp_val of tValue                                  (* value *)
  | TExp_name of tName<tType>                           (* external name *) 
  | TExp_ldfld of tValue * tFieldRef                    (* field access *)
  | TExp_call of tValue * tValue list * tMethodRef      (* function call *)
  | TExp_extern_call of Sugar.externref * tMethodRef * tType list * tType * tValue list            (* call to an external function *)
  | TExp_let of tVar<tType> * tExp * tExp               (* let binding *)
  | TExp_isinst of tValue * (tClassName * tType list * tVar<tType> list * tExp) list * tExp * tType (* type test *)
  | TExp_cond of tExp * tExp * tExp                     (* conditional exp *)
  | TExp_primop of tPrimop * list<tValue>               (* n-ary primitive op *)
  | TExp_bot                                            (* error exp, for Exp_bot in Fine *)
                                                        (* translates to exception in MSIL *)
(* visibility qualifier *)
type tVisibility = 
  | TVis_public
  | TVis_internal

(* instance field declaration *)
type tFieldDecl = tFieldName * tType

(* static field declaration, with optional initialization value *)
type tStaticFieldDecl = tFieldName * tType * tExp option 

(* still keep the list representation to minimize the change *)
(* will change later to options instead of lists for type/value parameters *)
type tMethodDecl = tType * tMethodName * tVar<tKind> list * tVar<tType> list * tExp option
    (* t m<a:k>(x:t) {e} *)

type tEvidences = (tVar<tType> * tValue) list (* same as the target variable environment *)

type tClassDecl = {
  (* psi C<a1:k1, ..., am:km, x1:t1, ..., xn:tn> : B {methods} *)
  visibility: tVisibility;
  name: tClassName;
  namestring: string;
  ptvars: tVar<tKind> list;
  tvars: tVar<tKind> list;
  pvvars: tVar<tType> list;
  vvars: tVar<tType> list;
  evidences: tEvidences;
  extends: tType option;
  fields: tFieldDecl list;
  staticFields: tStaticFieldDecl list;
  methods: tMethodDecl list;
  kind: tKind option;
  extqual: option<Sugar.externqual>
}

type tCdecls = (string, tClassDecl) Hashtbl.t

let createTCdecls cdecls : tCdecls =
  let env = Hashtbl.create 100 in
  List.map (fun cdecl -> Hashtbl.add env (cdecl.namestring) cdecl) cdecls;
  env

type tModule = {
  name: tLident;          (* module name *)
  extends: tLident option; (* extends module name *)
  modCdecl: tClassDecl;   (* the class for module itself *)
  decls: tCdecls;         (* class declarations *)
  externs: tCdecls; (* references to extern F# classes *)
  entry : option<tExp>;   (* entry point *)
}

(* ================= utilities ======================== *)

let objType = TType_name(["System"; "object"], TKind_star) 

let newSym = (* create a new symbol *)
  fun() -> Microsoft.Fine.Sugar.nng.Apply "tGenSym" Absyn.dummyRange

(* create a new type variable with kind knd *)
let newTvar (k: tKind): tVar<tKind> = (newSym(), k)

(* create a new value variable with type typ *)
let newVvar (t: tType): tVar<tType> = (newSym(), t)
  
(* check if an expression is a value *)
let isval (e: tExp) : bool =
  match e with
    | TExp_val _ -> true
    | _ -> false

let coerceEToVals (e: tExp) : tValue list =
  match e with
    | TExp_val v -> [v]
    | _ -> []

(* get free type variables of a type *)
let freeTvars (t: tType): tVar<tKind> list = 
  let rec freetvs t result: tVar<tKind> list =
    match t with
      | TType_var tvar ->
        if (List.exists (fun id -> ((fst id) = (fst tvar))) result) then
          result           (* tvar already exists in result *)
       else result@[tvar] (* tvar is a new free type variable *)
      | TType_name _ -> result
      | TType_fun (bvd_opt, t1, t2) ->
        let result = freetvs t1 result in
        freetvs t2 result
      | TType_concrete tconcrete ->
        let (_, ts, _) = tconcrete in
        List.fold_left (fun fvs ty -> freetvs ty fvs) result ts
      | TType_dep (tfun, v) ->
        freetvs tfun result
      | TType_affine ty -> freetvs ty result in
  freetvs t []

(* get free value variables of a type *)
let freeVvars (t: tType): tIdent list =
  (* freevs_v: get the free value variables in a vlue *)
  let rec freevs_v vs v : tIdent list = 
    match v with
      | TVal_var var ->
        if (List.exists (fun id -> (id = (fst var))) vs) then []
        else [fst var]
      | TVal_constant _ -> []
      | TVal_obj ((_, targs, vargs), _, vals) ->
        let fvs1 = List.concat (List.map (freevs_t vs) targs) in
        let fvs2 = List.concat (List.map (freevs_v vs) vargs) in
        let fvs3 = List.concat (List.map (freevs_v vs) vals) in
        fvs1 @ fvs2 @ fvs3
      | TVal_ldfld (v, _) -> freevs_v vs  v
  and freevs_t vs t : tIdent list =
    match t with
      | TType_var _ 
      | TType_index _
      | TType_name _ -> []
      | TType_fun (bvd_opt, t1, t2) ->
        let fvs_1 = freevs_t vs t1 in
        let fvs_2 = (match bvd_opt with
                       | None -> freevs_t vs t2
                       | Some bvd -> freevs_t ((snd bvd)::vs) t2) in
        fvs_1 @ fvs_2
      | TType_concrete tconcrete ->
        let (_, ts, vals) = tconcrete in
        let vs_ts = List.concat (List.map (freevs_t vs) ts) in
        let vs_vals = List.concat (List.map (freevs_v vs) vals) in
        vs_ts @ vs_vals
      | TType_dep (tfun, v) ->
        (freevs_t vs tfun) @ (freevs_v vs v)
      | TType_affine ty -> freevs_t vs ty in
  freevs_t [] t

(* ================= built-in class declarations =========== *)

let is_abstract = function
    ["DepArrowSS"]
  | ["DepArrowSA"]
  | ["DepArrowAS"]
  | ["DepArrowAA"] -> true
  | _ -> false

(* abstract class DepArrow**<a1:*, a2:a1->*> { *)
(*     public a2<x> App(a1 x);                 *)
(* }                                           *)

let tClassDeparrowSS: tClassDecl = 
  let tvar_ss1: tVar<tKind> = newTvar TKind_star in
  let type_tvar_ss1 = TType_var tvar_ss1 in
  let tvar_ss2: tVar<tKind> =
    let karrow = TKind_arrow(type_tvar_ss1, TKind_star) in
      newTvar karrow in
  let type_tvar_ss2 = TType_var tvar_ss2 in
  let par_ss: tVar<tType> = newVvar type_tvar_ss1 in
  let method_app_ss: tMethodDecl = (
    TType_dep(type_tvar_ss2, TVal_var(par_ss)), (* return type *)
    "Invoke",                                      (* method name *)
    [],                                         (* type variable list *)
    [par_ss],                                   (* parameter list *)
    None                                        (* method body *)
  ) in
  {extqual=None;
   visibility = TVis_public;
  name = ["DepArrowSS"];
  namestring = "DepArrowSS";
  ptvars = [];
  tvars = [tvar_ss1; tvar_ss2];
  pvvars = [];
  vvars = [];
  evidences = [];
  extends = Some (TType_concrete (["Microsoft";"FSharp";"Core";"FastFunc`2"], [TType_var tvar_ss1; TType_var tvar_ss2], []));
  fields = [];
  staticFields = [];
  methods = [method_app_ss];
  kind = Some TKind_star;
  }

(* abstract class DepArrow*A<a1:*, a2:a1->A> { *)
(*     public a2<x> App(a1 x);                 *)
(* }                                           *)

let tClassDeparrowSA = 
  let tvar_sa1 = newTvar TKind_star in
  let type_tvar_sa1 = TType_var tvar_sa1 in
  let tvar_sa2 =
    let karrow = TKind_arrow(type_tvar_sa1, TKind_affine) in
      newTvar karrow in
  let type_tvar_sa2 = TType_var tvar_sa2 in
  let par_sa: tVar<tType> = newVvar type_tvar_sa1 in
  let method_app_sa: tMethodDecl = (
    TType_dep(type_tvar_sa2, TVal_var(par_sa)), (* return type *)
    "Invoke",                                      (* method name *)
    [],                                         (* type variable list *)
    [par_sa],                                   (* parameter list *)
    None                                        (* method body *)
  ) in
  {extqual=None;
visibility = TVis_public;
  name = ["DepArrowSA"];
  namestring = "DepArrowSA";
  ptvars = [];
  tvars = [tvar_sa1; tvar_sa2];
  pvvars = [];
  vvars = [];
  evidences = [];
  extends = None;
  fields = [];
  staticFields = [];
  methods = [method_app_sa];
  kind = Some TKind_star;
}


(* abstract class DepArrowAA<a1:A, a2:A> { *)
(*     public a2 App(a1 x);                *)
(* }                                       *)

let tClassDeparrowAA = 
  let tvar_aa1 = newTvar TKind_affine in
  let type_tvar_aa1 = TType_var tvar_aa1 in
  let tvar_aa2 = newTvar TKind_affine in
  let type_tvar_aa2 = TType_var tvar_aa2 in
  let par_aa: tVar<tType> = newVvar type_tvar_aa1 in
  let method_app_aa: tMethodDecl = (
    type_tvar_aa2, (* return type *)
    "Invoke",                           (* method name *)
    [],                              (* type variable list *)
    [par_aa],                        (* parameter list *)
    None                             (* method body *)
  ) in
  {extqual=None;
visibility = TVis_public;
  name = ["DepArrowAA"];
  namestring = "DepArrowAA";
  ptvars = [];
  tvars = [tvar_aa1; tvar_aa2];
  pvvars = [];
  vvars = [];
  evidences = [];
  extends = None;
  fields = [];
  staticFields = [];
  methods = [method_app_aa];
  kind = Some TKind_star;
}

(* abstract class DepArrowA*<a1:A, a2:*> { *)
(*     public a2 App(a1 x);                *)
(* }                                       *)

let tClassDeparrowAS = 
  let tvar_aa1 = newTvar TKind_affine in
  let type_tvar_aa1 = TType_var tvar_aa1 in
  let tvar_aa2 = newTvar TKind_star in
  let type_tvar_aa2 = TType_var tvar_aa2 in
  let par_aa: tVar<tType> = newVvar type_tvar_aa1 in
  let method_app_aa: tMethodDecl = (
    type_tvar_aa2, (* return type *)
    "Invoke",                           (* method name *)
    [],                              (* type variable list *)
    [par_aa],                        (* parameter list *)
    None                             (* method body *)
  ) in
  {extqual=None;
visibility = TVis_public;
  name = ["DepArrowAS"];
  namestring = "DepArrowAS";
  ptvars = [];
  tvars = [tvar_aa1; tvar_aa2];
  pvvars = [];
  vvars = [];
  evidences = [];
  extends = None;
  fields = [];
  staticFields = [];
  methods = [method_app_aa];
  kind = Some TKind_star;
  }

(* dependent tuple classes. There are 4 cases: *)
(*   DepTuple2SS for \x:a1 X a2 where both a1 and a2 have kind * *)
(*   DepTuple2SA for \x:a1 X a2 where a1:* and a2:A, *)
(*   Tuple2AA for a1 X a2 where a1:A and a2: A *)
(*   Tuple2AS for a1 X a2 where a1:A and a2: * *)
(* The data constructors corresponding to Tuple_UU, Tuple_UA, Tuple_AU, and Tuple_AA are *)
(* DconDepTuple2SS, DconDepTuple2SA, DconTuple2AA, DconTuple2AS. *)

(* class DepTuple2SS<a1: *, a2: a1 ->*> {} *)
let tClassDepTuple2SS = 
  let tvar_t1 : tVar<tKind> = newTvar TKind_star in
  let type1 = TType_var tvar_t1 in
  let tvar_t2 : tVar<tKind> = newTvar (TKind_arrow(type1, TKind_star)) in
  {extqual=None;
  visibility = TVis_public;
  name = ["DepTuple2SS"];
  namestring = "DepTuple2SS";
  ptvars = [];
  tvars = [tvar_t1; tvar_t2];
  pvvars = [];
  vvars = [];
  evidences = [];
  extends = None;
  fields = [];
  staticFields = [];
  methods = [];
  kind = Some TKind_star;
}

(* class DconDepTuple2SS<a1: *, a2: a1 ->*, fst:a1, snd: a2 fst> : DepTuple2SS<a1, a2> {} *)
let tClassDconDepTuple2SS =
  let tvar_t1 : tVar<tKind> = newTvar TKind_star in
  let type1 = TType_var tvar_t1 in
  let tvar_t2 : tVar<tKind> = newTvar (TKind_arrow(type1, TKind_star)) in
  let fst = newVvar type1 in
  let type2 = TType_dep(TType_var tvar_t2, TVal_var fst) in
  let snd = newVvar type2 in
  let super = TType_concrete(tClassDepTuple2SS.name, 
                             [TType_var tvar_t1; TType_var tvar_t2], 
                             []) in
  {extqual=None;
  visibility = TVis_public;
  name = Const.dcil_tuple_UU_name;
  namestring = Const.dcil_tuple_UU_namestring;
  ptvars = [];
  tvars = [tvar_t1; tvar_t2];
  pvvars = [];
  vvars = [fst; snd];
  evidences = [];
  extends = Some super;
  fields = [];
  staticFields = [];
  methods = [];
  kind = Some TKind_star;
}

(* dependent tuple class *)
(* class DepTuple2SA<a1: *, a2: a1 ->A> {} *)
let tClassDepTuple2SA = 
  let tvar_t1 : tVar<tKind> = newTvar TKind_star in
  let type1 = TType_var tvar_t1 in
  let tvar_t2 : tVar<tKind> = newTvar (TKind_arrow(type1, TKind_affine)) in
  {extqual=None;
  visibility = TVis_public;
  name = ["DepTuple2SA"];
  namestring = "DepTuple2SA";
  ptvars = [];
  tvars = [tvar_t1; tvar_t2];
  pvvars = [];
  vvars = [];
  evidences = [];
  extends = None;
  fields = [];
  staticFields = [];
  methods = [];
  kind = Some TKind_affine;
}

(* class DconDepTuple2SA<a1: *, a2: a1 ->A, fst: a1, snd: a2 fst>: DepTuple2SA<a1, a2> {} *)
let tClassDconDepTuple2SA = 
  let tvar_t1 : tVar<tKind> = newTvar TKind_star in
  let type1 = TType_var tvar_t1 in
  let tvar_t2 : tVar<tKind> = newTvar (TKind_arrow(type1, TKind_affine)) in
  let fst = newVvar type1 in
  let type2 = TType_dep(TType_var tvar_t2, TVal_var fst) in
  let snd = newVvar type2 in
  let super = TType_concrete(tClassDepTuple2SA.name, 
                             [TType_var tvar_t1; TType_var tvar_t2],
                             []) in
  {extqual=None;
  visibility = TVis_public;
  name = Const.dcil_tuple_UA_name;
  namestring = Const.dcil_tuple_UA_namestring;
  ptvars = [];
  tvars = [tvar_t1; tvar_t2];
  pvvars = [];
  vvars = [fst; snd];
  evidences = [];
  extends = Some super;
  fields = [];
  staticFields = [];
  methods = [];
  kind = Some TKind_affine;
}

(* dependent tuple class *)
(* class Tuple2AA<a1: A, a2: A> {} *)
let tClassTuple2AA = 
  let tvar_t1 : tVar<tKind> = newTvar TKind_affine in
  let tvar_t2 : tVar<tKind> = newTvar TKind_affine in
  {extqual=None;
  visibility = TVis_public;
  name = ["Tuple2AA"];
  namestring ="Tuple2AA";
  ptvars = [];
  tvars = [tvar_t1; tvar_t2];
  pvvars = [];
  vvars = [];
  evidences = [];
  extends = None;
  fields = [];
  staticFields = [];
  methods = [];
  kind = Some TKind_affine;
}

(* class DconTuple2AA<a1: A, a2: A, fst: a1, snd: a2>: Tuple2AA<a1, a2> {} *)
let tClassDconTuple2AA = 
  let tvar_t1 : tVar<tKind> = newTvar TKind_affine in
  let tvar_t2 : tVar<tKind> = newTvar TKind_affine in
  let fst = newVvar (TType_var tvar_t1) in
  let snd = newVvar (TType_var tvar_t2) in
  {extqual=None;
  visibility = TVis_public;
  name = Const.dcil_tuple_AA_name;
  namestring = Const.dcil_tuple_AA_namestring;
  ptvars = [];
  tvars = [tvar_t1; tvar_t2];
  pvvars = [];
  vvars = [fst; snd];
  evidences = [];
  extends = Some (TType_concrete(tClassTuple2AA.name, 
                                 [TType_var tvar_t1; TType_var tvar_t2], 
                                 []));
  fields = [];
  staticFields = [];
  methods = [];
  kind = Some TKind_affine;
}

(* dependent tuple class *)
(* class Tuple2AS<a1: A, a2: *> {} *)
let tClassTuple2AS = 
  let tvar_t1 : tVar<tKind> = newTvar TKind_affine in
  let tvar_t2 : tVar<tKind> = newTvar TKind_star in
  {extqual=None;
  visibility = TVis_public;
  name = ["Tuple2AS"];
  namestring = "Tuple2AS";
  ptvars = [];
  tvars = [tvar_t1; tvar_t2];
  pvvars = [];
  vvars = [];
  evidences = [];
  extends = None;
  fields = [];
  staticFields = [];
  methods = [];
  kind = Some TKind_affine;
}

(* class DconTuple2AS<a1: A, a2: *, fst: a1, snd: a2>: Tuple2AS<a1, a2> {} *)
let tClassDconTuple2AS = 
  let tvar_t1 : tVar<tKind> = newTvar TKind_affine in
  let tvar_t2 : tVar<tKind> = newTvar TKind_star in
  let fst = newVvar (TType_var tvar_t1) in
  let snd = newVvar (TType_var tvar_t2) in
  {extqual=None;
  visibility = TVis_public;
  name = Const.dcil_tuple_AU_name;
  namestring = Const.dcil_tuple_AU_namestring;
  ptvars = [];
  tvars = [tvar_t1; tvar_t2];
  pvvars = [];
  vvars = [fst; snd];
  evidences = [];
  extends = Some (TType_concrete(tClassTuple2AA.name, 
                                 [TType_var tvar_t1; TType_var tvar_t2], 
                                 []));
  fields = [];
  staticFields = [];
  methods = [];
  kind = Some TKind_affine;
}

(* the class for polymorphic types forall a::*.t *)
(* All <b:*->*> {(b a) TyApp<a::*>; } *)
let tClassAllSS: tClassDecl = 
  let tvar_a: tVar<tKind> = newTvar TKind_star in
  let tvar_b: tVar<tKind> = newTvar (TKind_karrow(TKind_star, TKind_star)) in
  let tyapp : tMethodDecl = (
    TType_tapp (TType_var tvar_b, TType_var tvar_a), (* return type *)
    "TyApp",                     (* name *)
    [tvar_a],                    (* type variables *)
    [],                          (* value parameters *)
    None                         (* method body *)
  ) in
  {extqual = None;
  visibility = TVis_public;
  name = ["All**"];
  namestring = "All**";
  ptvars = [];
  tvars = [tvar_b];
  pvvars = [];
  vvars = [];
  evidences = [];
  extends = None;
  fields = [];
  staticFields = [];
  methods = [tyapp];
  kind = Some TKind_star;
}

(* the class for polymorphic types forall a::*.t *)
(* All <b:*->A> {(b a) TyApp<a::*>; } *)
let tClassAllSA: tClassDecl = 
  let tvar_a: tVar<tKind> = newTvar TKind_star in
  let tvar_b: tVar<tKind> = newTvar (TKind_karrow(TKind_star, TKind_affine)) in
  let tyapp : tMethodDecl = (
    TType_tapp (TType_var tvar_b, TType_var tvar_a), (* return type *)
    "TyApp",                     (* name *)
    [tvar_a],                    (* type variables *)
    [],                          (* value parameters *)
    None                         (* method body *)
  ) in
  {extqual = None;
  visibility = TVis_public;
  name = ["All*A"];
  namestring = "All*A";
  ptvars = [];
  tvars = [tvar_b];
  pvvars = [];
  vvars = [];
  evidences = [];
  extends = None;
  fields = [];
  staticFields = [];
  methods = [tyapp];
  kind = Some TKind_star;
}

(* the class for polymorphic types forall a::A.t *)
(* All <b:A->*> {(b a) TyApp<a::A>; } *)
let tClassAllAS: tClassDecl = 
  let tvar_a: tVar<tKind> = newTvar TKind_affine in
  let tvar_b: tVar<tKind> = newTvar (TKind_karrow(TKind_affine, TKind_star)) in
  let tyapp : tMethodDecl = (
    TType_tapp (TType_var tvar_b, TType_var tvar_a), (* return type *)
    "TyApp",                     (* name *)
    [tvar_a],                    (* type variables *)
    [],                          (* value parameters *)
    None                         (* method body *)
  ) in
  {extqual = None;
  visibility = TVis_public;
  name = ["AllA*"];
  namestring = "AllA*";
  ptvars = [];
  tvars = [tvar_b];
  pvvars = [];
  vvars = [];
  evidences = [];
  extends = None;
  fields = [];
  staticFields = [];
  methods = [tyapp];
  kind = Some TKind_star;
}

(* the class for polymorphic types forall a::A.t *)
(* All <b:A->A> {(b a) TyApp<a::A>; } *)
let tClassAllAA: tClassDecl = 
  let tvar_a: tVar<tKind> = newTvar TKind_affine in
  let tvar_b: tVar<tKind> = newTvar (TKind_karrow(TKind_affine, TKind_affine)) in
  let tyapp : tMethodDecl = (
    TType_tapp (TType_var tvar_b, TType_var tvar_a), (* return type *)
    "TyApp",                     (* name *)
    [tvar_a],                    (* type variables *)
    [],                          (* value parameters *)
    None                         (* method body *)
  ) in
  {extqual = None;
  visibility = TVis_public;
  name = ["AllAA"];
  namestring = "AllAA";
  ptvars = [];
  tvars = [tvar_b];
  pvvars = [];
  vvars = [];
  evidences = [];
  extends = None;
  fields = [];
  staticFields = [];
  methods = [tyapp];
  kind = Some TKind_star;
}

let predefined_classes = 
  [tClassDeparrowSS; tClassDeparrowSA; tClassDeparrowAA; tClassDeparrowAS; 
   (* classes for tuple types *)
   tClassDepTuple2SS; tClassDepTuple2SA; tClassTuple2AA; tClassTuple2AS; 
   tClassDconDepTuple2SS; tClassDconDepTuple2SA; tClassDconTuple2AA; tClassDconTuple2AS;
   tClassAllSS; tClassAllSA; tClassAllAS; tClassAllAA]

let predefined = (* predefined class declarations *)
  createTCdecls predefined_classes

(*================== base types  ===========*)
(* Should be the same as the translation of the prim absyn types *)
let intType = TType_name(["Prims"; "int"], TKind_star)
let boolType = TType_name(["Prims"; "bool"], TKind_star)
let stringType = TType_name(["Prims"; "string"], TKind_star)
let unitType = TType_name(["Prims"; "unit"], TKind_star)
let botType = TType_name(["Prims"; "bot"], TKind_star)

let trueVal = TVal_constant(Sugar.Const_bool true)
let falseVal = TVal_constant(Sugar.Const_bool false)
