#light "off"
// (c) Microsoft Corporation. All rights reserved
module Microsoft.Fine.TargetChecker

(* type checker for the target language DMSIL *)

open Target
open Util
open PrettyTarget
open Profiling 

exception TargetCheckerErr of string
let err msg = raise (TargetCheckerErr msg)

let lookupCdecl_ctr = new_counter "lookup cdecl"

let pr = Printf.printf
let spr = Printf.sprintf

(* ======================== equalities =============== *)

(* type variable equivalence *)
let eqTvar (tv1: tVar<tKind>) (tv2: tVar<tKind>) = (tv1 = tv2)

(* value variable equivalence *)
let eqVvar (vv1: tVar<tType>) (vv2: tVar<tType>) = (fst vv1 = fst vv2)

(* class declaration equivalence *)
let eqCdecl (cdecl1: tClassDecl) (cdecl2: tClassDecl) =
  cdecl1.name = cdecl2.name

let eqKind (kind1: tKind) (kind2: tKind) = kind1 = kind2

(*================= type substitution =============*)

type substMap =
  | TvarMap of tVar<tKind> * tType  (* mapping a type variable to a type *)
  | VvarMap of tVar<tType> * tValue (* mapping a value variable to a value *)
  | TindexMap of tType list         (* mapping from indices to types: !n -> nth element in the list *)

type tyMap = substMap list

let emptyTyMap: tyMap = []

let addTvarMap tmap tvar ty = (TvarMap(tvar, ty)) :: tmap

let addVvarMap tmap vvar v = (VvarMap(vvar, v)) :: tmap

let rec substTvar tmap tvar: tType =
  match tmap with
    | [] -> TType_var tvar
    | TvarMap(tv, ty) :: rest ->
      if (eqTvar tv tvar) then ty else substTvar rest tvar
    | _ :: rest -> substTvar rest tvar

let rec substTindex tmap n: tType =
  match tmap with
    | [] -> TType_index n
    | (TindexMap ts) :: _ ->
       let nint = (int n) in 
       if (nint < List.length ts) then List.nth ts nint
       else err (spr "substTindex: cannot find %d" nint)
    | _ :: rest -> substTindex rest n

let rec substVvar tmap vvar: tValue =
  match tmap with
    | [] -> TVal_var vvar
    | VvarMap(var, v) :: rest ->
      if (eqVvar var vvar) then v else substVvar rest vvar
    | _ :: rest -> substVvar rest vvar

and substTConcrete tmap (cname, ts, vs) =
  (cname, List.map (substType tmap) ts, List.map (substVal tmap) vs)

and substVal tmap v: tValue =
  match v with
    | TVal_var var -> substVvar tmap var
    | TVal_obj(tconcrete, ts, vs) ->
      TVal_obj(substTConcrete tmap tconcrete, ts, List.map (substVal tmap) vs)
    | TVal_ldfld (vobj, fld) -> TVal_ldfld(substVal tmap vobj, fld)
    | _ -> v

(* kind substitution *)
and substKind tmap k = 
  match k with
    | TKind_arrow(t, k') -> TKind_arrow (substType tmap t, 
                                         substKind tmap k')
    | _ -> k

(* type substitution *)
and substType tmap t = 
  match t with
    | TType_index n -> substTindex tmap n
    | TType_var tv -> substTvar tmap tv
    | TType_name name -> t
    | TType_fun (bvdopt, t1, t2) -> 
      TType_fun (bvdopt, substType tmap t1, substType tmap t2)
    | TType_concrete(cname, ts, vs) ->
      TType_concrete (substTConcrete tmap (cname, ts, vs))
    | TType_dep (tfun, v) ->
      TType_dep (substType tmap tfun, substVal tmap v)
    | TType_affine t -> TType_affine (substType tmap t)
    | TType_tfun(bvd, kind, body) -> TType_tfun(bvd, kind, substType tmap body)
    | TType_tapp(tfun, targ) -> TType_tapp(substType tmap tfun, substType tmap targ)
    | _ -> err (spr "substType: unexpected type %s" (strType t))

let buildTyMap (cdecl:tClassDecl) ts vs =
  DebugLog.Log(Printf.sprintf "Building tymap for instantiaion of class %s\n %d type formals %d actuals\n %d value formals %d actuals"
     (Sugar.text_of_path cdecl.name) (List.length (cdecl.ptvars @ cdecl.tvars))  
     (List.length ts) 
     (List.length (cdecl.pvvars @ cdecl.vvars)) (List.length vs));
  let tmap = List.map TvarMap (List.zip (cdecl.ptvars @ cdecl.tvars) ts) in
  let vmap = List.map VvarMap (List.zip (cdecl.pvvars @ cdecl.vvars) vs) in
  tmap @ vmap

let instantiateType (cdecl: tClassDecl) ts vs t : tType = 
  substType (buildTyMap cdecl ts vs) t

(*=================Environments==============*)

(* class env : list of class declarations *)
type tCenv = tCdecls list

type hashcenv = tCdecls list

(* lookup a class decl by name *)
let lookupCdeclByName (env: hashcenv) (name: tClassName) : tClassDecl =
  (*try *)
    let cname = String.concat "." name in
    (*let debug =  (cname = "CtGenSym_0_207") in
    let _ = if debug then (pr "\n looking in %d" (List.length env);
                           List.map (fun cds -> match Hashtbl.tryfind cds cname with
                                                | Some decl -> pr " \n ## found one %s" (strClassDecl decl)
                                                | None -> pr " \n ## not found here") env; ()) in *)
    let rec find ev =
      match ev with
        | [] -> err(spr "lookupCdeclByName: %s not found" cname)
        | cdecls::rest -> (match Hashtbl.tryfind cdecls cname with
                             | Some decl -> decl
                             | None -> find rest) in
    find env
  (*with
    | :? System.Collections.Generic.KeyNotFoundException -> 
      (Printf.printf "Name %s not found in the environment" (strClassName name); raise Not_found)*)
    
let tryLookupCdeclByName (env: hashcenv) (name: tClassName) : tClassDecl option =
  (*try *)
    let cname = String.concat "." name in
    let rec find ev =
      match ev with
        | [] -> None
        | cdecls::rest -> (match Hashtbl.tryfind cdecls cname with
                             | Some decl -> Some decl
                             | None -> find rest) in
    find env


(* get the super class of a type concrete. Returns itself if no super class *)
let getSuper env (sub: tTypeConcrete) : tType =
  let (subName, ts, vs) = sub in
  let subdecl = lookupCdeclByName env subName in
  match subdecl.extends with
    | Some supert -> 
      instantiateType subdecl ts vs supert 
    | None -> TType_concrete sub
      (*let errmsg = spr "class %s has no super class" (strClassName subName) in
              raise (TargetCheckerErr errmsg) *)

(* find out if a class declaration declares an affine class *)
(* check if Affine is a superclass of cdecl *)
let isAffine cdecl =
  match (cdecl.kind) with
    | Some TKind_affine -> true
    | _ -> false

(* kind env: list of in-scope type variables *) 
type tKenv = tVar<tKind> list

let emptyKenv : tKenv = []

(* utility functions for accessing kind env *)
(* add a new type variable to kind env *)
(* report error if tvar is already in env *)
let addTvar (env: tKenv) (tvar: tVar<tKind>) : tKenv =
  if (List.exists (fun (tv: tVar<tKind>) -> (eqTvar tv tvar)) env) then
    let errmsg = spr "addTvar: %s already exists" (fst tvar) in
    err errmsg
  else (tvar :: env)

(* see if a new type variables in the kenv *)
let containsTvar (env: tKenv) (tvar: tVar<tKind>): bool =
  List.exists (fun (tv: tVar<tKind>) -> (eqTvar tv tvar)) env

(* type env: list of in-scope value variables *)
type tTenv = (tVar<tType> * tValue) list

let emptyTenv: tTenv = []

(* add a value variable to type env *)
(* There might be more than one entry for the same variable because of *)
(* type cast. The first entry is the most accurate one *)
let addVvar (env: tTenv) (var: tVar<tType>) : tTenv = (var, TVal_var var) :: env

(* look up value variables *)
let rec lookupVvarType (env: tTenv) (var: tVar<tType>) =
  match env with
    | [] -> err (spr "lookupVVarType: var %s not found in tenv" (strVvar var))
    | (vvar, _) :: rest ->
      if (eqVvar var vvar) then snd vvar else lookupVvarType rest var

(* return the representative of var *)
let rec lookupVvarRep env var =
  match env with
    | [] -> var
    | (vvar, TVal_var rep) :: rest ->
      if (eqVvar var vvar) then rep else lookupVvarRep rest var
    | _ :: rest -> lookupVvarRep rest var

let addEq env var vl = (var, vl) :: env

(* add var1 -> rep of var2 in tenv. *)
let addEqVvar tenv var1 var2 =
  let rep = lookupVvarRep tenv var2 in
  (var1, TVal_var rep) :: tenv

(* var equivalence taking into account the equivalence in env *)
let rec isEqVvar env var1 var2 =
  if (eqVvar var1 var2) then true
  else let rep1, rep2 = (lookupVvarRep env var1, lookupVvarRep env var2) in
  eqVvar rep1 rep2

(* affine env: list of in-scope affine value variables *)
type tAenv = tVar<tType> list

let emptyAenv: tAenv = []

(* add an affine value variable *)
(* report error if var is already in env *)
let rec addAvar (env: tAenv) (var: tVar<tType>): tAenv =
  (* check if var exists in env *)
  if (List.exists (fun (avar: tVar<tType>) -> eqVvar avar var) env) then
    err (spr "addAvar: %s already exists" (fst var))
  else (var :: env) 

let rec removeAvar env var : tAenv = 
  match env with
    | [] -> []
    | (v::rest) -> if (eqVvar v var) then rest
                   else v::(removeAvar rest var)

(* union two affine envs, duplication disallowed *)
let rec unionAenv aenv1 aenv2 = 
  match aenv2 with
    | [] -> aenv1 
    | var::rest -> unionAenv (addAvar aenv1 var) rest

let unionAenvs aenvs  =
  List.fold_left unionAenv emptyAenv aenvs


let rec unionBranchAenv aenv1 aenv2 =
  match aenv2 with
    | [] -> aenv1
    | var::rest -> if (List.exists (fun v -> eqVvar v var) aenv1) then
                     unionBranchAenv aenv1 rest
                   else var :: (unionBranchAenv aenv1 rest)

let unionBranchAenvs aenvs =
  List.fold_left unionBranchAenv emptyAenv aenvs

(* test if aenv1 is a superset of aenv2 *)
let contains aenv1 aenv2 =
  let rec test aev2 =
    match aev2 with
      | [] -> true
      | (var::rest) -> (List.exists (fun avar -> eqVvar avar var) aenv1 &&
                        test rest)
  in test aenv2

let addVar tenv aenv kind var =
  let newTenv = addVvar tenv var in
  let newAenv = 
    match kind with
      | TKind_affine -> addAvar aenv var
      | _ -> aenv in
  (newTenv, newAenv)

(*==================== Utilities ===================*)

(* look up a field from a type concrete *)
(* return a field declaration if found, and None otherwise *)
let lookupField cenv tconcrete fieldRef =
  let (cname, ts, vs) = tconcrete in
  let cdecl = lookupCdeclByName cenv cname in
  let field = List.find (fun f -> (fst f = fieldRef)) (cdecl.fields) in
  (fst field, instantiateType cdecl ts vs (snd field))

(* lookup a method declaration from a type concrete *)
(* returns a method decl if found, raise Not_found otherwise *)
let lookupMethod cenv tconcrete methodRef = 
  let (cname, ts, vs) = tconcrete in
  let cdecl = lookupCdeclByName cenv cname in
    List.find (fun (_, name, _, _, _) -> (name = fst methodRef)) cdecl.methods

(* normalization of a type *)
(* beta reduction *)
let rec normalize ty =
  match ty with
    | TType_dep(TType_fun(bvd, targ, tbody), v) ->
      (match bvd with
         | Some var ->
           let tmap = [VvarMap((snd var, targ), v)] in
           normalize(substType tmap tbody)
         | _ -> normalize tbody)
    | TType_tapp(TType_tfun(bvd, kind, tbody), targ) ->
      (* (\a::k.t)t' ==> t[a->t']*)
      let tmap = [TvarMap((snd bvd, kind), targ)] in
      normalize(substType tmap tbody)
    | _ -> ty

(* check if two types are equivalent *)
let rec isEqType tenv t1 t2 = 
  if (t1 = t2) then true
  else  
    let type1 = normalize t1 in
    let type2 = normalize t2 in
    match (type1, type2) with
    | (TType_fun(bvd1, targ1, tbody1), TType_fun(bvd2, targ2, tbody2)) -> 
      let isEqTargs = isEqType tenv targ1 targ2 in
      (match (bvd1, bvd2) with
        | (Some var1, Some var2) ->
          let tmap = [VvarMap((snd var1, targ1), TVal_var(snd var2, targ2))] in
          isEqTargs && (isEqType tenv (substType tmap tbody1) tbody2)
        | _ -> isEqTargs && (isEqType tenv tbody1 tbody2))
    | (TType_tfun(bvd1, kind1, tbody1), TType_tfun(bvd2, kind2, tbody2)) ->
      (* alpha equivalence of type functions *)
      if (eqKind kind1 kind2) then
        let tvar1, tvar2 = (snd bvd1, kind1), (snd bvd2, kind2) in
        let tmap = [TvarMap(tvar1, TType_var tvar2)] in
        isEqType tenv (substType tmap tbody1) tbody2
      else false
    | (TType_concrete(cname1, ts1, vs1), TType_concrete(cname2, ts2, vs2)) ->
      isEqTypeConcrete tenv (cname1, ts1, vs1) (cname2, ts2, vs2)
    | (TType_dep(tfun1, v1), TType_dep(tfun2, v2)) ->
      isEqType tenv tfun1 tfun2 && isEqVal tenv v1 v2
    | (TType_affine t1, TType_affine t2) ->
      isEqType tenv t1 t2
    | (TType_name cname1, TType_concrete(cname2, [], [])) -> (fst cname1) = cname2
    | (TType_concrete(cname1, [], []), TType_name cname2) -> cname1 = (fst cname2)
    | _ -> type1 = type2

(* check if two values are equivalent *)
(* v1: I == v2: I if I has no fields *)
and isEqVal tenv v1 v2 =
  if (v1 = v2) then true
  else
    match (v1, v2) with
      | (TVal_var var, _) -> isEq tenv var v2
      | (_, TVal_var var) -> isEq tenv var v1
      | (TVal_obj(t1, _, vs1), TVal_obj(t2, _, vs2)) ->
        if (isEqTypeConcrete tenv t1 t2) then
          let rec matchvs vls1 vls2 =
            match (vls1, vls2) with
              | ([], []) -> true
              | (v1::rest1, v2::rest2) ->
                isEqVal tenv v1 v2 && matchvs rest1 rest2
              | _ -> false in
          matchvs vs1 vs2
        else false
      | (TVal_constant cons1, TVal_constant cons2) -> isEqCons tenv cons1 cons2
      | _ -> false 

and isEq tenv var vl =
  let rec comp env =
    match env with
      | [] -> false
      | (x, v) :: rest ->
  let eqObj val1 val2 = 
    match (val1, val2) with
      | (TVal_obj _, TVal_obj _) -> isEqVal tenv val1 val2
      | _ -> val1 = val2 in          
  if (isEqVvar tenv var x && eqObj vl v) then true else comp rest
  in (*pr "\n isEq: %s, %s" (strVvar var) (strVal vl);
     List.map (fun (v, binding) -> pr "\n    %s -> %s" (strVvar v) (strVal binding)) tenv;*)
  comp tenv

and isEqCons tenv cons1 cons2 =
  match (cons1, cons2) with
    | (Sugar.Const_decimal d1, Sugar.Const_decimal d2) -> d1.Equals(d2)
    | (Sugar.Const_bigint i1, Sugar.Const_bigint i2) -> isEqByteArray i1 i2
    | (Sugar.Const_bignum n1, Sugar.Const_bignum n2) -> isEqByteArray n1 n2
    | (Sugar.Const_string(a1, _), Sugar.Const_string(a2, _)) -> isEqByteArray a1 a2
    | (Sugar.Const_bytearray(a1, _), Sugar.Const_bytearray(a2, _)) -> isEqByteArray a1 a2
    | (Sugar.Const_uint16array a1, Sugar.Const_uint16array a2) -> isEqUint16Array a1 a2
    | _ -> false

and isEqByteArray a1 a2 = 
 (Bytes.compare a1 a2 = 0)

and isEqUint16Array a1 a2 = a1.Equals(a2)

(* check if two type concrete are equivalent *)
and isEqTypeConcrete tenv (cname1, ts1, vs1) (cname2, ts2, vs2) =

  (*let _ = if (!fdebug) then (pr "\n !!!!!isEqTypeConcrete: ts1 = ";
               print_any ts1;
               pr "\n    vs1 ="; print_any vs1;
               pr "\n    ts2 ="; print_any ts2;
               pr "\n    vs2 ="; print_any vs2;) in *)

  let rec isEqTypes types1 types2 =
     match (types1, types2) with
       | ([], []) -> true
       | (t1::restts1, t2::restts2) ->
         (isEqType tenv t1 t2) && (isEqTypes restts1 restts2)
       | _ -> false in
  let rec isEqVs vals1 vals2 =
    match (vals1, vals2) with
      | ([], []) -> true
      | (v1::restvs1, v2::restvs2) ->
        (isEqVal tenv v1 v2) && (isEqVs restvs1 restvs2)
      | _ -> false in
  (cname1 = cname2) && (isEqTypes ts1 ts2) && (isEqVs vs1 vs2)

(* return if type1 is a subtype of type2 TODO *)
and isSubtype cenv tenv (type1:tType) (type2:tType) = 
  if (isEqType tenv type1 type2) then true
  else if (type1 = botType) then true
  else
    let type1 = normalize type1 in
    let type2 = normalize type2 in
    match (type1, type2) with
    | (TType_fun(bvd1, targ1, tbody1), TType_fun(bvd2, targ2, tbody2)) ->
      (match (bvd1, bvd2) with
        | (None, None) ->
          (isSubtype cenv tenv targ2 targ1) &&
          (isSubtype cenv tenv tbody1 tbody2)
        | (Some var1, Some var2) ->
          if (isSubtype cenv tenv targ2 targ1) then
            let tmap = [VvarMap((snd var1, targ1), TVal_var(snd var2, targ2))] in
            isSubtype cenv tenv (substType tmap tbody1) tbody2
          else false
        | _ -> false)
    | (TType_affine t1, TType_affine t2) ->
      isSubtype cenv tenv t1 t2
    | _ -> false

(*============================== kinding ==========================*)
(* compute the kind of a type and add to aenv the the free affine variables 
   used in the type *)
and kinding cenv kenv tenv (t: tType) : tKind = 
  match t with
    | TType_var tv -> 
      (* type variable: it must exist in the kenv *)
      if (containsTvar kenv tv) then snd tv
      else let errmsg = spr "kinding: free tvar %s" (strType t) in
        err errmsg
    | TType_name name -> (snd name)
    | TType_fun (bvdopt, t1, t2) -> 
      (* [\x:]t1 -> t2 *)
      (match (kinding cenv kenv tenv t1) with
        | TKind_star -> 
          let newtenv = match bvdopt with
                       | None -> tenv
                       | Some bvdef -> addVvar tenv (snd bvdef, t1) in
          let k2 = kinding cenv kenv newtenv t2 in
          TKind_arrow(t1, k2)
        | knd ->
          let errmsg = spr "kinding: TType_fun: %s has kind %s"
                       (strType t1) (strKind knd) in
          err errmsg
      )
    | TType_concrete tconcrete ->
      (*let _ = Printf.printf "KINDING: name = " in*)
      let (tname, tvs, vs) = tconcrete in
      (*let _ = print_any tname in
      let _ = Printf.printf "\n, TS = " in
      let _ = print_any tvs in
      let _ = Printf.printf "\n, VS = " in
      let _ = print_any vs in*)
      kindingTConcrete cenv kenv tenv tconcrete
    | TType_dep (tfun, v) -> 
      (match (kinding cenv kenv tenv tfun) with
        | TKind_arrow(t, knd) ->
          let (tyOfv, aenv_v) = vTyping cenv kenv tenv emptyAenv v in
          if (aenv_v = emptyAenv &&
              isSubtype cenv tenv tyOfv t) then knd
          else err (spr "kinding: type mismatch: require %s given %s" (strType t) (strType tyOfv))
        | _ -> err (spr "kinding: invalid tfun in %s" (strType tfun)))
    | TType_affine ty ->
      (match (kinding cenv kenv tenv ty) with
        | TKind_star -> TKind_affine
        | _ -> err (spr "kinding: invalid affine %s" (strType t)))
    | TType_tfun (bvd, kind, body) -> 
      let tvar = (snd bvd, kind) in
      let kenv' = if (containsTvar kenv tvar) then kenv  (* TODO: why are there duplicated tvars *)
                  else addTvar kenv tvar in
      TKind_karrow(kind, kinding cenv kenv' tenv body)
    | TType_tapp (tfun, targ) -> 
        (match (kinding cenv kenv tenv tfun) with
         | TKind_karrow (k1, k2) ->
           let karg = kinding cenv kenv tenv targ in
           if eqKind k1 karg then k2
           else err (spr "kinding: kind mismatch: need %s but given %s" (strKind k1) (strKind karg))
         | _ -> err (spr "kinding: unexpected kind for %s in %s" (strType tfun) (strType t)))
    | _ -> err (spr "kinding: unexpected type %s" (strType t))

(* kinding of a concrete type  *)
and kindingTConcrete cenv kenv tenv (tc: tTypeConcrete) : tKind = 
  let (classname, tys, vls) = tc in
  (*let _ = Printf.printf "\nkindingTConcrete: %s" (PrettyTarget.strTypeConcrete tc) in *)
  let cdecl = lookupCdeclByName cenv classname in
  (* first check if the type arguments have the appropriate kinds *)
  (*let _ = Printf.printf "\n %s" (PrettyTarget.strClassDecl cdecl) in*)
  let rec checkTypes tmap tvars ts =
    match (tvars, ts) with
      | ([], []) -> tmap
      | (tvar::restTvars, ty::restTys) ->
        let kind_ty = kinding cenv kenv tenv ty in
        let substKind_tvar = substKind tmap (snd tvar) in
        if (eqKind substKind_tvar kind_ty) then
          let newtmap = addTvarMap tmap tvar ty in
          checkTypes newtmap restTvars restTys
        else
          let errmsg = spr "kindingTConcrete: %s has kind %s but expects %s"
                       (strType ty) (strKind kind_ty) (strKind substKind_tvar) in
          err errmsg
       | _ ->
         let errmsg = spr "kindingTConcrete: %s: mismatched tvars and types"
                      (strTypeConcrete tc) in
         err errmsg in 
   let tmap = checkTypes emptyTyMap (cdecl.ptvars @ cdecl.tvars) tys in
   (* check if the value arguments have the appropriate types *)
   let rec checkVals (tmap: tyMap) (vvars: tVar<tType> list)
     (vs: tValue list) =
     match (vvars, vs) with
       | ([], []) -> ()
       | (vvar::restVvars, v::restVs) ->
         let (ty_v, aenv_v) = vTyping cenv kenv tenv emptyAenv v in
         let substTy_vvar = substType tmap (snd vvar) in
         if (aenv_v = emptyAenv &&
             isSubtype cenv tenv ty_v substTy_vvar) then
           let newtmap = addVvarMap tmap vvar v in
           checkVals newtmap restVvars restVs
         else let errmsg = spr "kindTConcrete: %s has type %s but expects %s" (strVal v) (strType ty_v) (strType substTy_vvar) in
           err errmsg
       | _ -> let errmsg = spr "kindingTConcrete: %s: mismatched vvars and values" (strTypeConcrete tc) in
         err errmsg in
   let _ = checkVals tmap (cdecl.pvvars @ cdecl.vvars) vls in
   let kind_tc = if (isAffine cdecl) then TKind_affine else TKind_star in
   kind_tc 

(* compute the type of a value *)
and vTyping cenv (kenv:tKenv) (tenv:tTenv) (aenv:tAenv) (v: tValue) : (tType * tAenv) = 
  DebugLog.Log(spr "\n vtyping: %s" (strVal v));
  let (t, newaenv) = match v with
    | TVal_var var -> 
      let vtype = lookupVvarType tenv var in
      DebugLog.Log(spr "\n got var type %s" (strType vtype));
      (match (kinding cenv kenv tenv vtype) with
         | TKind_affine -> (vtype, addAvar aenv var)
         | _ -> (vtype, aenv))
    | TVal_obj (tconcrete, _, vs) ->
        objectTyping cenv kenv tenv aenv tconcrete vs 
    | TVal_constant cons -> (constTyping cons, aenv)
    | TVal_ldfld(vobj, fieldRef) ->
      checkLdfld cenv kenv tenv aenv vobj fieldRef in
 (t, newaenv)

and checkLdfld cenv kenv tenv aenv vobj fieldRef =
  let (tobject, newaenv) = vTyping cenv kenv tenv aenv vobj in
  let check tconcrete =
    let (cname, ts, vs) = tconcrete in
    if ((cname = tClassDepTuple2SS.name) &&
       ((fieldRef = Const.dcil_proj_one_name) ||
        (fieldRef = Const.dcil_proj_two_name))) then
      (* project the first or second of a non-dependent tuple *)
      match ts with
        | [t1; TType_fun(_, _, t2)] -> 
          if (fieldRef = Const.dcil_proj_one_name) then (t1, newaenv)
          else (t2, newaenv)
        | _ -> err (spr "checkLdfld: unexpected targs for DepTuple2SS %s" 
                        (strType tobject))
    else let (fieldname, fieldtype) = lookupField cenv tconcrete fieldRef in
      (fieldtype, newaenv) in
  match tobject with
    | TType_concrete tconcrete -> check tconcrete
    (* allow cname as tconcrete (cname, [], []), for external records *)
    | TType_name cname -> check (fst cname, [], [])      
    | _ -> err (spr "checkLdfld: unexpected obj type %s" (strType tobject))

(* check if vs (may be values or exps) match fields fs *)
and checkObjFields (getType: tAenv -> 'a -> (tType * tAenv)) isSubTy aev fs (vs: 'a list) =
 match (fs, vs) with
   | ([], []) -> aev
   | (f::restfs, v::restvs) ->
     let fty = snd f in
     let (vty, newaenv) = getType aev v in
     if (isSubTy vty fty) then
       checkObjFields getType isSubTy newaenv restfs restvs
     else let errmsg = spr "checkObjFields: has type %s but expects %s"
                       (strType vty) (strType fty) in
          err errmsg
   | _ -> err "checkObjFields: mismatched types and values"

(* check if vs (may be values or exps) match fields fs *)
(*and checkObjvFields (getType: tAenv -> tValue -> (tType * tAenv)) isSubTy aev fs (vs: tValue list) =
 match (fs, vs) with
   | ([], []) -> aev
   | (f::restfs, v::restvs) ->
     let fty = snd f in
     let (vty, newaenv) = getType aev v in
     if (isSubTy vty fty) then
       checkObjvFields getType isSubTy newaenv restfs restvs
     else let errmsg = spr "checkObjvFields: has type %s but expects %s"
                       (strType vty) (strType fty) in
          err errmsg
   | _ -> err "checkObjvFields: mismatched types and values"*)

(* check if an object has the specified type *)
and objectTyping cenv kenv tenv aenv tconcrete vls =
 (*pr "\nobjectTyping %s" (PrettyTarget.strTypeConcrete tconcrete);*)

 let (cname, ts, vs) = tconcrete in
 let cdecl = lookupCdeclByName cenv cname in
 let tyMap = buildTyMap cdecl ts vs in
 (* check constrains *)
 let rec checkEvidences evs =
   match evs with
     | [] -> ()
     | (var, v)::rest ->
       let newv1 = substVvar tyMap var in
       let newv2 = substVal tyMap v in
       if (isEqVal tenv newv1 newv2)
       then checkEvidences rest
       else err (spr "objectTyping: check evidence %s = %s but got %s <> %s"
                     (strVvar var) (strVal v) (strVal newv1) (strVal newv2)) in
 let _ = checkEvidences cdecl.evidences in
 (* check fields *)
 let fields = List.map (fun f -> (fst f, substType tyMap (snd f))) cdecl.fields in
 let rec checkFields aev fs vs =
   match (fs, vs) with
     | ([], []) -> aev
     | (f::restfs, v::restvs) ->
       let fty = snd f in
       let (vty, aev) = vTyping cenv kenv tenv aev v in
       if (isSubtype cenv tenv vty fty) 
       then checkFields aev restfs restvs
       else let errmsg = spr "objectTyping: has type %s but expects %s"
                             (strType vty) (strType fty) in
            err errmsg
     | _ -> err "objectTyping: mismatched types and values" in
 let aenv = checkFields aenv fields vls in
 (* get super *)
 match cdecl.extends with
   | Some super -> (substType tyMap super, aenv)
   | _ -> (TType_concrete tconcrete, aenv)

(* compute the type of a constant *)
(* currently support unit, bool, int32, and string only *)
and constTyping cons = 
  match cons with
    | Sugar.Const_unit -> unitType
    | Sugar.Const_bool _ -> boolType
    | Sugar.Const_int32 _ -> intType
    | Sugar.Const_string _ -> stringType
    | _ ->
      let errmsg = spr "constTyping: unsupported const %s"
                   (cons.ToString()) in
      err errmsg
    (*| Sugar.Const_int8 _ ->
    | Sugar.Const_uint8 _ ->
    | Sugar.Const_int16 _->
    | Sugar.Const_uint16 _ ->
    | Sugar.Const_uint32 _ ->
    | Sugar.Const_int64 _ ->
    | Sugar.Const_uint64 _ ->
    | Sugar.Const_nativeint _ ->
    | Sugar.Const_unativeint _ ->
    | Sugar.Const_float32 _ ->
    | Sugar.Const_float _ ->
    | Sugar.Const_char _ ->
    | Sugar.Const_decimal _ ->
    | Sugar.Const_bigint _ ->
    | Sugar.Const_bignum _ ->
    | Sugar.Const_bytearray _ ->
    | Sugar.Const_uint16array _ -> *)

(* check if a kind is well-formed *)
and checkKind cenv kenv tenv (k: tKind) : unit = 
  match k with
    | TKind_star 
    | TKind_affine -> ()
    | TKind_arrow (t, knd) ->
      (match (kinding cenv kenv tenv t) with
        | TKind_star -> checkKind cenv kenv tenv knd
        | _ -> let errmsg = spr "checkKind: invalid %s" (strKind k) in
          err errmsg)
    | TKind_karrow(k1, k2) ->
      (match (k1, k2) with
        | (TKind_star, TKind_star)
        | (TKind_star, TKind_affine)
        | (TKind_affine, TKind_star)
        | (TKind_affine, TKind_affine) -> ()
        | _ -> err (spr "checkKind: unexpected karrow %s" (strKind k))) 
    | _ -> err (spr "checkKind: unexpected kind %s" (strKind k))

(*============================= typing ===========================*)
(* compute the type of an expression and update environments *)
(* add the free affine values used in e to aenv. *)
(* report error if there are duplicated affine values *)
let rec eTyping cenv (kenv: tKenv) (tenv: tTenv) (aenv: tAenv) (e: tExp)
    : (tType * tAenv) =
  let erre msg = 
    err (spr "eTyping: %s has error: %s" (strExp e) msg) in
  DebugLog.Log (spr "\n eTyping %s" (strExp e));
  match e with
    | TExp_val v -> vTyping cenv kenv tenv aenv v
    | TExp_name name -> (snd name, aenv)
    | TExp_ldfld (vobj, fieldRef) -> 
      checkLdfld cenv kenv tenv aenv vobj fieldRef
    | TExp_call (vobj, ags, methodRef) -> 
      let (tobj, aenv_obj) = vTyping cenv kenv tenv aenv vobj in
      (match tobj with
        | TType_affine (TType_concrete(cname, ts, vs))
        | TType_concrete(cname, ts, vs) ->
          DebugLog.Log(spr "\n Call: "; List.map (fun (v, binding) -> pr "\n    %s -> %s" 
                       (strVvar v) (strVal binding)) tenv); 
          let cdecl = lookupCdeclByName cenv cname in
          let (retTy, mname, tvars, ps, _) =
            List.find (fun (_, name, _, _, _) -> (name = fst methodRef)) cdecl.methods in
          let tmap = (buildTyMap cdecl ts vs) @ 
                     (List.map TvarMap (List.zip tvars (snd methodRef))) in
          let rec checkArgs aev tymap params args =
            (match (params, args) with
              | ([], []) -> 
                 (DebugLog.Log(spr "\n: return type = %s, \n subst = %s"
                              (strType retTy) (strType (substType tymap retTy)));
                  (substType tymap retTy, aev))
              | (param::restps, arg::restas) ->
                let substParamTy = substType tymap (snd param) in
                let (argTy, newaenv) = vTyping cenv kenv tenv aev arg in
                if (isSubtype cenv tenv argTy substParamTy) then 
                  (DebugLog.Log(spr "\n adding %s -> %s" (strVvar param) (strVal arg));
                  checkArgs newaenv (VvarMap(param, arg)::tymap) restps restas)
                else begin
                  (*print_any tenv;*)
                  erre (spr "%s has type %s but expects %s" (fst param)
                          (strType argTy) (strType substParamTy))
                end
              | _ -> erre (spr "mismatched params and args for %s" mname)) in
          checkArgs aenv tmap ps ags
        | _ -> erre (spr "%s has type %s, not a TType_concrete"
                     (strVal vobj) (strType tobj)))
    | TExp_extern_call (extref, mref, ptypes, rtype, vals) ->
      (* check external calls: *)
      (* ptypes are the parameter types and rtype is the return type *)
      let tymap = [TindexMap(snd mref)] in
      let rec checkArgs aev ptys args =
        (match (ptys, args) with
          | ([], []) -> (substType tymap rtype, aev)
          | (pty::restptys, arg::restargs) ->
            let (argTy, newaenv) = vTyping cenv kenv tenv aev arg in
            if (isSubtype cenv tenv argTy (substType tymap pty)) then
              checkArgs newaenv restptys restargs
            else erre (spr "%s has type %s but expects %s" 
                           (strVal arg) (strType argTy) (strType (substType tymap pty)))
          | _ -> erre (spr "mismatched param types and args")) in
      checkArgs aenv ptypes vals
    | TExp_let (var, e1, e2) ->
      let (type1, aenv1) = eTyping cenv kenv tenv aenv e1 in
      DebugLog.Log(spr "\n let %s: %s <- %s" (strVvar var) (strType (snd var)) (strType type1));
      (*let var = (fst var, normalize type1) in*)
      let newtenv = 
        match e1 with
          | TExp_val (TVal_var var1) -> addEqVvar tenv var var1
          | TExp_val vl -> addEq tenv var vl
          | _ -> addVvar tenv var in
      let (type2, aenv2) = eTyping cenv kenv newtenv emptyAenv e2 in
      (type2, unionAenv aenv1 (removeAvar aenv2 var))
    | TExp_isinst (v, tconcrete_es, edefault, finalTy) ->
      let (vty, aenv_v) = vTyping cenv kenv tenv aenv v in
      let checkBranch (cname, ts, vars, e) =
        (* check one branch (cname, ts, vars) -> e*)
        let tenv_v = 
          match v with
            | TVal_var var ->
              (* add new value equivalence facts to tenv *)
              DebugLog.Log(spr "\n isinst: adding %s == %s" 
                           (strVvar var) 
                           (strVal (TVal_obj((cname, ts, List.map TVal_var vars), [], []))));
              addEq tenv var (TVal_obj((cname, ts, List.map TVal_var vars), [], [])) in
        let super = getSuper cenv (cname, ts, List.map TVal_var vars) in
        let rec matchvs tenvv vs1 vs2 =
          match (vs1, vs2) with
            | ((TVal_var var1)::restvs1, v2::restvs2) ->
              matchvs (addEq tenvv var1 v2) restvs1 restvs2
            | (_::restvs1, _::restvs2) -> matchvs tenvv restvs1 restvs2
            | _ -> tenvv in
        let tenv_v = 
          match (super, vty) with
            (TType_concrete(cname1, _, vs1), TType_concrete(cname2, _, vs2)) ->
            if (cname1 = cname2) then matchvs tenv_v vs1 vs2
            else tenv_v
            | _ -> tenv_v in

        (* add newly introduced variables to the environments *)
        let rec addVars tev vars =
          match vars with
            | [] -> tev
            | var::rest -> addVars (addVvar tev var) rest in
        let newTenv = addVars tenv_v vars in
        (*let _ = (Printf.printf "\n checking branch: tenv = "; print_any newTenv) in*)
        let (ety, aenv_e) = eTyping cenv kenv newTenv emptyAenv e in
        let rec removeAvars aev vars =
            match vars with
              | [] -> aev
              | var::rest -> removeAvars (removeAvar aev var) rest in
        if (isSubtype cenv newTenv ety finalTy) then (removeAvars aenv_e vars) 
        else erre (spr "branch %s has type %s but expects %s"
                   (strExp e) (strType ety) (strType finalTy))  in
      let aenvs = List.map checkBranch tconcrete_es in 
      let (ty_default, aenv_default) =
        eTyping cenv kenv tenv emptyAenv edefault in
      let aenvs = unionBranchAenvs (aenv_default :: aenvs) in
      if (isSubtype cenv tenv ty_default finalTy) then 
        let newaenv = unionAenv aenv aenvs in
        (finalTy, newaenv)
      else erre (spr "default branch has type %s but expects %s"
                 (strType ty_default) (strType finalTy))
    | TExp_cond (eb, et, ef) ->
      let (tyb, aenv) = eTyping cenv kenv tenv aenv eb in
      if (tyb = boolType) then
        match eb with
          | TExp_val(TVal_var var) ->
            let (type1, aenv1) = eTyping cenv kenv (addEq tenv var trueVal) emptyAenv et in
            let (type2, aenv2) = eTyping cenv kenv (addEq tenv var falseVal) emptyAenv ef in
            let newaenv = unionAenv aenv (unionBranchAenv aenv1 aenv2) in
            (if (isSubtype cenv tenv type1 type2) then (type2,newaenv) 
             else if (isSubtype cenv tenv type2 type1) then (type1,newaenv)
            else erre (spr "true branch has type %s and false has %s" 
                      (strType type1) (strType type2)))
          | _ -> erre (spr "%s not a var" (strExp eb))
      else erre (spr "%s not a bool" (strExp eb))
    | TExp_primop (bop, el) -> 
        let aenv, et_l_rev = 
          List.fold_left (fun (aenv, et_l) e -> 
                            let t_e, aenv = vTyping cenv kenv tenv aenv e in
                              (aenv,  (e,t_e)::et_l)) (aenv,[]) el in
        let et_l = List.rev et_l_rev in
          (match et_l with 
               [(e1,t1); (e2,t2)] -> 
                 (match bop with
                    | AND
                    | OR -> 
                        if (t1 = boolType && t2 = boolType) then (boolType, aenv)
                        else erre (spr "%s or %s not a bool" (strVal e1) (strVal e2))
                    | ADD 
                    | SUB
                    | MUL
                    | DIV -> if (t1 = intType && t2 = intType) then (intType, aenv)
                      else erre (spr "%s or %s not an int" (strVal e1) (strVal e2))
                    | _ -> erre ( "unexpected binary operator"))
             | [(e1,t1)] -> 
                 (match bop with 
                    | NOT -> 
                        if (t1 = boolType) then (boolType, aenv)
                        else erre (spr "%s not a bool" (strVal e1))
                    | _ -> erre ( "unexpected unary operator"))
             | _ -> erre (spr "unexpected operator arity (%d)" (List.length et_l)))
    | TExp_bot -> (botType, aenv)
    | _ -> erre "unexpected exp"

(*==================== class declarations ==================*)
(* check type variables in a class declaration or in a method declarations*)
(* Create a kind env that includes all the type variables *)
let checkTvars cenv initKenv tenv tvars : tKenv = 
  let rec checktvars kenv tvs =
    match tvs with
      | [] -> kenv 
      | tv :: rest -> 
        let (_, kind) = tv in
        let _ = checkKind cenv kenv tenv kind in
        checktvars (addTvar kenv tv) rest
  in checktvars initKenv tvars

(* check value variables in a class declaration *)
(* Types of value variables can use type variables in kenv,
   and values introduced earlier *)
let checkCdeclVvars cenv kenv genv vvars : tTenv * tAenv =
  DebugLog.Log("\n checking value variables"); 
  let rec checkVvars tenv aenv vars =
    (match vars with
      | [] -> (tenv, aenv)
      | var::rest ->
        let (_, varTy) = var in
        let kind = kinding cenv kenv tenv varTy in
        let (newTenv, newAenv) = addVar tenv aenv kind var in
        checkVvars newTenv newAenv rest) in
  checkVvars genv emptyAenv vvars

let checkEvidences cenv kenv tenv evidences =
  let rec checkEvs (newtenv:tTenv) (evs: tEvidences) =
    match evs with
      | [] -> newtenv
      | (var, v) :: rest ->
        (* check that var is in scope and v is valid in scope *)
        let _ = lookupVvarType tenv var in
        let _ = vTyping cenv kenv tenv emptyAenv v in
        checkEvs ((var, v)::newtenv) rest in
  checkEvs tenv evidences

let checkStaticField cenv kenv tenv sf =
  let (name, ty, eopt) = sf in
  DebugLog.Log(spr "\n checking static field %s" name);
  match (kinding cenv kenv tenv ty) with
    | TKind_star
    | TKind_affine -> ()
    | k -> let errmsg = spr "checkStaticField: %s with unexpected kind %s"
                        (strStaticFieldDecl sf) (strKind k) in
           err errmsg;
  match eopt with
    | None -> ()
    | Some e ->
      (eTyping cenv kenv tenv emptyAenv e; ())
       
(* check static field declarations in a class declaration *)
let checkStaticFields cenv kenv tenv sfields = 
  List.map (fun sf -> checkStaticField cenv kenv tenv sf) sfields

(* check a field declaration in a class declaration *)
let checkField cenv kenv tenv field =
  let (name, ty) = field in
  DebugLog.Log(spr "\n checking field %s" name);
  match (kinding cenv kenv tenv ty) with
    | TKind_star
    | TKind_affine -> ()
    | k ->
       let errmsg = spr "checkField: %s with unexpected kind %s" 
                    (strFieldDecl field) (strKind k) in
       err errmsg

(* check field declarations in a class declaration *)
let checkFields cenv kenv tenv fields : unit =
  List.map (fun f -> checkField cenv kenv tenv f) fields; ()

(* check parameters of a method *)
(* Currently allow later parameters to use previous ones in their types
   although we don't need that at the moment *)
let checkParams cenv kenv tenv_0 aenv_0 params: (tTenv * tAenv) = 
  let rec check tenv aenv pars =
    match pars with
      | [] -> (tenv, aenv)
      | par::rest ->
        let knd = kinding cenv kenv tenv (snd par) in
        let (newtenv, newaenv) = addVar tenv aenv knd par in
        check newtenv newaenv rest in
  check tenv_0 aenv_0 params

(* check a method declaration in a class declaration *)
(* The method can have its own type variables.
   The return type may depend on the parameters *)
let checkMethod cenv kenv tenv_0 aenv_0 (m: tMethodDecl): unit =
  let (retType, name, tvars, params, body) = m in
  DebugLog.Log(spr "\n checking method %s " name); 
  let newKenv = checkTvars cenv kenv tenv_0 tvars in
  DebugLog.Log(spr "\n checking params %s " name); 
  let (tenv, aenv) = checkParams cenv newKenv tenv_0 aenv_0 params in
  DebugLog.Log(spr "\n checking return type : %s " (strType retType)); 
  (*let k_ret = kinding cenv newKenv tenv retType in*)
  DebugLog.Log(spr "\n checking body %s " name); 
  match body with
    | None -> ()
    | Some e ->
      let (ty_e, aenv_e) = eTyping cenv newKenv tenv emptyAenv e in
      if (contains aenv aenv_e) then
        (if (isSubtype cenv tenv ty_e retType) then ()
         else let errmsg = spr "checkMethod: body has type %s but expects %s"
                           (strType (normalize ty_e)) (strType (normalize retType)) in
              err errmsg)
      else err "checkMethod: free affine variables"

(* check a class declaration *)
(* Type variables introduced eariler can be used in types later in the class
   declaration. Value variables introduced earlier can be used in types of
   other value variables introduced later in the declaration. 
   Types of fields and methods can use the type variables introduced by the
   class, but not the value variables.
*)
let checkClassDecl cenv (genv:tTenv) (c: tClassDecl) : unit = 
  let name = String.concat "." c.name in
  let _ = DebugLog.Log (spr "\n >>>>>> checking class %s" (PrettyTarget.strClassDecl c)) in 
  let kenv = checkTvars cenv emptyKenv emptyTenv (c.ptvars @ c.tvars) in   (* check the type variables *)
  let (tenv, aenv) = checkCdeclVvars cenv kenv genv (c.pvvars @ c.vvars) in   (* check the value variables *)
  let tenv = checkEvidences cenv kenv tenv c.evidences in
  checkStaticFields cenv kenv tenv c.staticFields;
  checkFields cenv kenv tenv c.fields;
  List.map (fun m -> checkMethod cenv kenv tenv aenv m) c.methods;
  DebugLog.Log(spr "\n <<<<<<< finished checking %s" name);
  ()

(*========================= module ========================*)
(* check a module *)
(* first check the class declarations in the module *)
(* then check the main expression *)
let checkModule  imported (m: tModule) : tCdecls list =
  (*let _ = pr "\n dcil checking module %s" (String.concat "." m.name) in
  let _ = PrettyTarget.printModule m in*)
  let cdecls = m.decls in
  let moduleClass = m.modCdecl in
  let rec genGlobalEnv gvals =
    match gvals with
      | [] -> emptyTenv
      | (name, ty, _)::rest -> addVvar (genGlobalEnv rest) (name, ty) in
  let genv = genGlobalEnv (moduleClass.staticFields) in
  let cenv = [cdecls; m.externs]@imported in
  Hashtbl.iter (fun name cdecl -> checkClassDecl cenv genv cdecl) cdecls;
  DebugLog.Log("\n checking module body");
  cenv
