#light "off"

// (c) Microsoft Corporation. All rights reserved

module Microsoft.Fine.Tc

open Absyn
open AbsynUtils
open Util
open Const
open Tcenv
open Profiling
open Tcutil

exception ErrorProofTerm of string * exp
let tc_ctr = new_counter "TC"
let tc_pf_ctr = new_counter "TC proofs"
let kinding_ctr = new_counter "Kinding" 
let kinding_type_index_ctr = new_counter "Kinding type index" 
let inst_ctr = new_counter "Instantiation"
let try_inst_ftyp_ctr = new_counter "Try Inst Final Type"

let unify uv t = Unionfind.change uv (Fixed t)
        
let unify_subst_vars subst = 
  let unify_eq_class (uvl, topt) = match uvl with 
    | [] -> raise (NYI "Unexpected empty equivalence class")
    | (uv,_)::tl  -> 
        List.iter (fun (uv',_) -> Unionfind.union uv uv') tl;
        match topt with
          | None -> ()
          | Some t -> unify uv t in
    List.iter unify_eq_class subst

let contains_bvar l b = List.exists (bvar_eq b) l

let generalize env t e : (typ * exp) = 
  if not (is_value e) then t,e 
  else
    let find uv sub = List.tryFind (fun (uv', _) -> Unionfind.equivalent uv uv') sub in
    let uvars_in_env = Tcenv.uvars_in_env env in
    let is_uvar_in_env uv = match List.tryFind (fun uv' -> Unionfind.equivalent uv uv') uvars_in_env with 
      | None -> false
      | Some _ -> true in
    let subst_and_collect_uvars_typ sub () t = match t with 
      | Typ_uvar (uv, k) -> 
          (match find uv sub with
             | Some (_,tv) -> sub, tv, None
             | _ -> 
                 if is_uvar_in_env uv then sub, t, None
                 else
                   let fn = genident None in
                   let btv = (Typ_btvar (withinfo (fn, fn) k dummyRange))  in
                     (uv, btv)::sub, btv, None)
      | _ -> sub, t, None in
    let exp_folder_noop env () e = (env,e) in
    let ext_benv_noop () x = (),x in
    let sub, tsub = typ_fold_map subst_and_collect_uvars_typ exp_folder_noop ext_benv_noop [] () t in
      (*     let sub', e = exp_fold_map subst_uvars exp_folder_noop ext_benv_noop sub () e in *)
      (*     let _ = if (sub <> sub') then raise (Err "Unexpected uvars in a term") in *)
    let univ_typ, Lambda_e = List.fold_right 
      (fun (uv, tv) (out_typ, Lambda_e) -> match tv with 
           Typ_btvar btv -> 
             let bvname = bvar_real_name btv in
             let out_typ = withsort (Typ_univ ((bvname, bvname), btv.sort, out_typ)) Kind_star  in
             let Lambda_e = withinfo (Exp_tabs ((bvname, bvname), btv.sort, Lambda_e)) out_typ Lambda_e.p in
             let _ = unify uv (withsort tv btv.sort) in
               (out_typ, Lambda_e)
         | _ -> raise Impos) sub (tsub, e) in
      univ_typ, Lambda_e
        
let new_uvar env k r =
  let wf typ k =
    let kind_ok = TypeRelations.equivalent_kinds env k typ.sort in
    let fvs_ok = match freevars_not_in_env env typ with 
        [],[] -> true
      | _ -> false in
    let result = (fvs_ok && kind_ok) in
(*     let _ = pr "For unification: tested well-formedness of type (%s) at kind (%s)... Result: (%s)\n" (Pretty.strTyp typ) (Pretty.strKind k) (if result then "success" else "failed") in       *)
      result in
(*   let _ = pr "Creating new unification variable of kind (%s)\n" (Pretty.strKind k) in *)
  let uvb = Unionfind.fresh (Uvar wf) in
    withsort (Typ_uvar (uvb, k)) k
      (*     | _ ->  *)
      (*         let msg = spr "Unable to infer instantiation of higher-kinded type parameters: %s; Use an explicit type application." (Pretty.strKind k) in *)
      (*           raise (Error (msg, r)) *)

let is_equality_op e = match e.v with
  | Exp_tapp({v=Exp_fvar fv; sort=_;p=_}, _) -> Sugar.lid_equals fv.v Const.op_Eq 
  | _ -> false

let try_instantiate_final_typ env tps to_inst with_t = 
  let env', tparams = 
    List.fold_left (fun (env, abductibles) -> function
                      | Tparam_term(bvd, t) -> 
                          let env' = push_local_binding env (Binding_var (real_name bvd, t)) in
                          let abductibles' = (bvd_to_bvar bvd)::abductibles in
                            env', abductibles'
                      | _ -> raise Impos) (env, []) tps in
  let rec inst_final_typ env abductibles t = match t.v with (* don't try to instantiate when the function is dependent *)
    | Typ_fun (None, t1, t2) -> inst_final_typ env abductibles t2
    | Typ_fun (Some bvd, t1, t2) -> 
        let env' = push_local_binding env (Binding_var (real_name bvd, t1)) in
        let abductibles' = (bvd_to_bvar bvd)::abductibles in
          inst_final_typ env' abductibles'  t2
    | _ -> 
        let env = set_solver_flag env false in
        let do_instantiate env abductibles t with_t =
          (match TypeRelations.instantiable env abductibles t with_t with
                 None -> 
                   false, (to_inst, [])
               | Some (s, matches) -> 
                   let _ = unify_subst_vars s in 
                   let out, substs = List.fold_left 
                     (fun (out, substs) -> function
                          Binding_match({v=Exp_bvar bv1;sort=_;p=_}, e) -> 
                            if contains_bvar tparams bv1 then 
                              let bvd1 = bvar_to_bvd bv1 in
                              let out = substitute_exp out bvd1 e in
                                (out, (bvd1, e)::substs)
                            else out, substs
                        | _ -> raise Impos) (to_inst, []) matches in
                     true, (out, substs)) in
            match do_instantiate env abductibles t with_t with
                true, (t, substs) -> t, substs
              | false, (out, _) -> (* try once more, by looking inside affine type *)
                  match t.v, with_t.v with
                      Typ_affine _, Typ_affine _ -> out, []  (* cannot do anything further *)
                    | _, Typ_affine with_t' -> 
                        (match current_value env with
                           | Some v when is_non_variable_value env v -> 
                               (* try to instantiate underneath a !, relying on value promotion to fix things up later *)
                               let _, (out, substs) = do_instantiate env abductibles t with_t' in 
                                 out, substs
                           | _ -> out, [])
                    | _ -> out, []  (* cannot do anything further *) in
    profile (fun () -> inst_final_typ env' tparams to_inst) try_inst_ftyp_ctr

let retry f = 
  printfn "%A" "Retrying!";
  Util.verbosity := 1;
  f ()

let kind_star_or_A = function Kind_star | Kind_affine -> true | _ -> false 
            
let rec kinding_dep_list env (l:list<option<bvdef>*typ*(kind -> bool)>) : list<option<bvdef> * typ * kind> = 
  let (_, result) = List.fold_left 
    (fun (env, out) (bvd_opt, t, check_k) -> 
       let t, k = kinding env t in
         if check_k k then 
           let env' = match bvd_opt with
             | Some bvd -> push_local_binding env (Binding_var(snd bvd,t))
             | None -> env in
             (env', (bvd_opt, t, k)::out)
         else 
           let msg = Printf.sprintf "Unexpected kind %s" (Pretty.strKind k) in
             raise (Err msg)) (env, []) l in
    List.rev result

and well_formed_kind env k = 
  let rec wfk env index allow_tcon k = match index, k with 
    | 0, Kind_star -> k
    | _, Kind_affine -> k
    | i, Kind_tcon(Kind_star, k') when allow_tcon -> Kind_tcon(Kind_star, wfk env i allow_tcon k')
    | i, Kind_tcon(Kind_affine, k') when allow_tcon -> Kind_tcon(Kind_affine, wfk env 1 allow_tcon k')
    | i, Kind_tcon(k, k') when allow_tcon -> 
        let k = wfk env i false k in 
        let k' = wfk env i allow_tcon k' in
          Kind_tcon(k, k')
    | i, Kind_dcon(t, k') -> 
        let t', tk = kinding env t in
          (match tk with
             | Kind_star -> Kind_dcon(t', wfk env i allow_tcon k')
             | _ -> raise (Error ("Dependent indices must be elements of star types", t'.p)))
    | i, Kind_qkind(qs, k) when allow_tcon -> 
        let env = List.fold_left 
          (fun env (bvd:bvdef,k) -> match k with 
             | Kind_star -> push_local_binding env (Binding_typ (snd bvd, k)) 
             | _ -> 
                 let pp = fst bvd in
                   raise (Error ((spr "Kind-level quantification only permitted on *-kinded types; got %s\n" (Pretty.strKind k)), pp.idRange)))
          env qs in
        let k' = wfk env i allow_tcon k in
          Kind_qkind (qs, k')
    | _ -> raise (Error ("Unexpected kind in source term", dummyRange)) in
    wfk env 0 true k

and kinding env top : (typ * kind) = 
  let rec do_kinding env top =
    let env, _ = clear_expected_typ (set_kind_level env) in
    let typ', kind = match top.v with
      | Typ_btvar bv -> (* we include this case because we're not doing any opening of exps/typs unlike in Coq *)
          (match lookup_btvar env bv with 
             | Kind_unknown -> 
                 raise (Error ("Type variable with unknown kind", bv.p))
             | k -> 
                 let bv' = withinfo (bv.v) k bv.p in
                   Typ_btvar bv', k)

      | Typ_const fv -> 
          (match lookup_typ_const env fv with 
             | Kind_unknown -> 
                 let msg = spr "Type constant with unknown kind: %s" (Pretty.strTyp top) in
                   raise (Error (msg, fv.p))
             | k -> 
                 let fv' = withinfo (fv.v) k fv.p in
                   Typ_const fv', k)
            
      | Typ_univ (bvd, k, t) when Tcutil.kind_abstractable k -> 
          let k = well_formed_kind env k in
          let binding = Binding_typ (snd bvd, k) in 
          let env' = push_local_binding env binding in
          let t', k' = kinding env' t in
            (match k' with 
                 Kind_star 
               | Kind_affine -> 
                   Typ_univ(bvd, k, t'), Kind_star
               | _ -> 
                   let msg = 
                     Printf.sprintf 
                       ("Expected a *- or A-kinded type.\n\
                      Got type %s of kind %s") (Pretty.strTyp t) (Pretty.strKind k') in
                     raise (Err msg))

      | Typ_fun (bvd_opt, t1, t2) ->
          let dlist = (bvd_opt,t1,kind_star_or_A)::[(None,t2,kind_star_or_A)] in
            (match kinding_dep_list env dlist with 
                 [(bvd_opt, t1', k1); (_, t2', k2)] -> Typ_fun (bvd_opt, t1', t2'), Kind_star
               | _ -> raise Impos)

      | Typ_refine (bvd, t1, t2) -> 
          let dlist = (Some bvd, t1, kind_star_or_A)::[(None,t2, fun k -> TypeRelations.equivalent_kinds env k Kind_star)] in
            (match kinding_dep_list env dlist with 
                 [(Some bvd, t1', k1); (_, t2', k2)] -> Typ_refine (bvd, t1', t2'), kind_lub k1 k2
               | _ -> raise Impos)

      | Typ_dtuple dlist -> 
          let dlistk = kinding_dep_list env (List.map (fun (x,t) -> (x,t,kind_star_or_A)) dlist) in
          let k, dlistrev = List.fold_left (fun (k, out) (n, t, k') -> (kind_lub k k', (n,t)::out)) (Kind_star,[]) dlistk in
            Typ_dtuple (List.rev dlistrev), k
              
      | Typ_app(t1, ({v=Typ_unknown; sort=_; p=_} as t2)) -> 
          let t1', k1 = kinding env t1 in
            (match k1 with             
               | Kind_qkind((bvd, k)::tl, k') -> 
                   let t2' = new_uvar env k t2.p in
                   let k' = substitute_kind k' bvd t2' in
                   let k' = match tl with 
                     | [] -> k'
                     | _ -> Kind_qkind(tl, k') in
                     Typ_app(t1', t2'), k'
               | _ -> raise (Err "Missing type argument"))
              
      | Typ_app (t1, t2) -> 
          let t1', k1 = kinding env t1 in
          let t2', k2 = kinding env t2 in
            (match k1 with 
               | Kind_tcon (k, k') when TypeRelations.equivalent_kinds env k k2 -> 
                   Typ_app(t1', t2'), k'                

               | Kind_qkind((bvd, k)::tl, k') when TypeRelations.equivalent_kinds env k k2 -> 
                   let k' = substitute_kind k' bvd t2' in
                   let k' = match tl with 
                       [] -> k'
                     | _ -> Kind_qkind(tl, k') in
                     Typ_app(t1', t2'), k'
                       
               | Kind_tcon (k, _) -> 
                   let msg = 
                     Printf.sprintf 
                       "Type constructor %s expected argument of kind %s but got %s"
                       (Pretty.strTyp t1) (Pretty.strKind k) (Pretty.strKind k2) in
                     raise (Err msg)
               | _ -> 
                   let msg = 
                     Printf.sprintf 
                       "Expected a type constructor of kind %s => k, but got a type of kind %s"
                       (Pretty.strKind k2) (Pretty.strKind k1) in
                     raise (Err msg))

      | Typ_dep (t, e) -> 
          (if not (is_value e) then
             let msg = spr "Type-level terms mush be syntactic values; got %s" (Pretty.strExp e) in
               raise (Error (msg, top.p)));
          let t, k = kinding env t in
            (match k with 
               | Kind_dcon (t', k') -> 
                   let e, t_e, used_affines = profile (fun () -> typing (set_expected_typ env t') e) kinding_type_index_ctr in
                     (* TODO: could assert here that used_affines = []? *)
                     Typ_dep(t, e), k'
                       (*                    (match t.v with  *)
                       (*                         Typ_const v when Sugar.lid_equals v.v Const.prop_Eq ->  *)
                       (*                           (\* need to slot in the real equality predicate here *\) *)
                       (*                           let env, _ = clear_expected_typ env in *)
                       (*                           let e, t_e, used_affines = profile (fun () -> typing env e) kinding_type_index_ctr in *)
                       (*                             (\* TODO: could assert here that used_affines = []? *\) *)
                       (*                           let eqT_opt = Proof.eq_typ_of_typ (Tcenv.expand_typ env t_e) in *)
                       (*                             (match eqT_opt with  *)
                       (*                                | None ->  *)
                       (*                                    let msg = spr "Equality on type %s is unsupported" (Pretty.strTyp t_e) in *)
                       (*                                      raise (Error (msg, e.p)) *)
                       (*                                | Some eqT ->  *)
                       (*                                    let eqT, k = kinding env eqT in *)
                       (*                                      match k with *)
                       (*                                        | Kind_dcon(_, kout) -> Typ_dep (eqT, e), kout *)
                       (*                                        | _ -> raise Impos) *)
               | _ -> 
                   let e, t_e, _ = typing env e in
                   let msg = 
                     Printf.sprintf 
                       "Expected a dependent-type constructor of kind %s -> k, but got a type of kind %s"
                       (Pretty.strTyp t_e) (Pretty.strKind k) in
                     raise (Err msg))

      | Typ_affine t -> 
          let t, k = kinding env t in
            (match k with
                 Kind_star -> Typ_affine t, Kind_affine
               | _ -> 
                   let msg = 
                     Printf.sprintf 
                       "The affine qualifier can only be applied to *-kinded types\n\
                     The type %s has %s-kind"
                       (Pretty.strTyp t) (Pretty.strKind k) in
                     raise (Err msg))

      | Typ_record fn_t_l -> 
          let fn_t_l_rev, out_k = List.fold_left 
            (fun (out_l, out_k) (fn, t) -> 
               (*
                 if in_typ_decl env && Tcenv.field_name_exists env fn then
                 let msg = spr "Record field names must be unique; %s is already used" (Sugar.text_of_lid fn) in
                 raise (NYI msg)
               *)
               if false then raise Impos
               else
                 (* TODO: check that this type is an instance of the declared record type *)
                 let t', k' = kinding env t in 
                   match k' with 
                       Kind_star
                     | Kind_affine -> ((fn, t')::out_l, kind_lub k' out_k)
                     | _ -> 
                         let msg = spr "Record fields must have ground types; %s has kind %s" 
                           (Sugar.text_of_lid fn) (Pretty.strKind k') in
                           raise (Err msg)) ([], Kind_star) fn_t_l in
            Typ_record (List.rev fn_t_l_rev), out_k

      | Typ_uvar (uv,k) as t -> 
          (match Unionfind.find uv with
               Uvar _ -> t, k
             | Fixed t -> 
                 let t', k' = kinding env t in
                   assert (TypeRelations.equivalent_kinds env k k');
                   t'.v, k')

      | Typ_lam (x, t1, t2) -> 
          let dlist = (Some x, t1, function Kind_star -> true | _ -> false)::[(None,t2,fun k -> true)] in
            (match kinding_dep_list env dlist with 
                 [(Some x, t1', k1); (_, t2', k2)] -> Typ_lam(x, t1', t2'), Kind_dcon(t1', k2)
               | _ -> raise Impos)
              
      | Typ_unknown -> raise (Err "The unknown type is not well-formed") in
    let typ = withinfo typ' kind (top.p) in
      match kind with
          Kind_unknown -> 
            let msg = spr "Got a type with an unknown kind %s\n" (Pretty.strTyp typ) in
              raise (Err msg)
        | _ -> whnf typ, kind in
    profile (fun () -> do_kinding env top) kinding_ctr 
      
and instantiate_typ env typ t_actuals : (typ * list<typ>) = 
  let do_instantiate_typ env typ t_actuals = 
    let rec n_typ_params out tt = match tt.v with
        Typ_univ (bvd, k, t) -> 
          n_typ_params ((bvd, k, tt.p)::out) t 
      | _ -> List.rev out, tt in
(*     let fold_inst typ withargs =  *)
(*       List.fold_left (fun typ actual -> match typ.v with *)
(*                         | Typ_univ (bvd, k, tt) ->  *)
(*                             substitute tt bvd actual *)
(*                         | _ -> raise Impos) typ withargs  in *)
    let t_formals, mono_typ = n_typ_params [] typ in
    let inst_typ, args = 
      match t_formals, t_actuals with 
          [], [] -> typ, []  (* no instantiation necessary *)
        | _, [] ->       (* need to infer instantiation *)
            let bvd_uvars = 
              List.fold_left (fun bvd_uvars (bvd, k, r) -> 
                                let k' = substitute_kind_l k bvd_uvars in
                                let uv = new_uvar env k' r in
                                  (bvd,uv)::bvd_uvars) [] t_formals in
              List.fold_left (fun (ityp, uvars) (bvd, uv) -> 
                                let ityp = substitute ityp bvd uv in
                                  ityp, uv::uvars) (mono_typ,[]) bvd_uvars 
        | _, _ when (List.length t_formals) = (List.length t_actuals) ->        (* user provided instantiation *)
            let check_kind actual formal_kind = 
              let (actual', kind) = kinding env actual in
                if (TypeRelations.equivalent_kinds env kind formal_kind) then actual'
                else
                  let msg = Printf.sprintf 
                    "Expected a type argument of kind %s but got an argument %s of kind %s"
                    (Pretty.strKind formal_kind) (Pretty.strTyp actual') (Pretty.strKind kind) in
                    raise (Err msg) in
            let bvd_actuals = 
              List.fold_left2 (fun bvd_actuals actual (bvd, formal_kind, _) -> 
                                 let formal_kind = substitute_kind_l formal_kind bvd_actuals in
                                 let actual = check_kind actual formal_kind in
                                   (bvd,actual)::bvd_actuals) [] t_actuals t_formals in
            let res = List.fold_left (fun ityp (bvd, actual) -> substitute ityp bvd actual) mono_typ bvd_actuals in
              res, List.rev (snd (List.unzip bvd_actuals))
        | _ -> raise (Err (spr "Unexpected number of type arguments:\n%s" (Pretty.strTyp typ))) in
      inst_typ, args in
    profile (fun () -> do_instantiate_typ env typ t_actuals) inst_ctr 

and typing env top_exp : (exp*typ*list<bvar<typ>>) =
  let rec do_typing env = function 
    | Exp_bvar bv -> 
        let t = lookup_bvar env bv in  (* TODO: freshen typ vars and introduce a type application here *)
        let bv = setsort bv t in
        let bvar = Exp_bvar bv in
        let used_affines = match t.sort with 
          | Kind_affine -> [bv]
          | _ -> [] in
          if expect_tyapps env then bvar, t, used_affines
          else (* introduce tyapps *)
            let t_inst, targs = instantiate_typ env t [] in
              (match targs with
                   [] -> bvar, t, used_affines
                 | _ -> 
                     let bvar = withinfo bvar t top_exp.p in
                     let tapp = List.fold_left 
                       (fun lhs arg -> 
                          let e = Exp_tapp (lhs, arg) in
                          let t_e = match lhs.sort.v with
                            | Typ_univ(bvd,_,tret) -> 
                                substitute tret bvd arg
                            | _ -> raise Impos in
                            withinfo e t_e top_exp.p) bvar targs in
                       tapp.v, t_inst, used_affines)

    | Exp_fvar fv -> (* TODO: add affine variable tracking *)
        if Tcenv.is_datacon env fv then
          let e, t, x = typing env (withpos (Exp_constr_app(fv, [], [], []))  top_exp.p) in
            e.v, t, x
        else
          let t = match lookup_fvar env fv with
              None -> raise (Err ("Variable not found: "^(Sugar.text_of_lid fv.v)))
            | Some t -> t in (* TODO: freshen typ vars and introduce type applications here *)
          let fv' = setsort fv t in
          let fvar = Exp_fvar fv' in
          let used_affines = match kinding env t with 
              _, Kind_affine -> 
                let msg = 
                  Printf.sprintf "Free variable %s cannot have an affine type"  (Pretty.strExp (withpos fvar dummyRange)) in
                  raise (Err msg)
            | _ -> [] in
            if expect_tyapps env then fvar, t, used_affines
            else
              let t_inst, targs = instantiate_typ env t [] in
                (match targs with
                     [] -> fvar, t, used_affines
                   | _ -> 
                       let fvar' = withinfo fvar t top_exp.p in
                       let tapp = List.fold_left 
                         (fun lhs arg -> 
                            let e = Exp_tapp (lhs, arg) in
                            let t_e = match lhs.sort.v with
                              | Typ_univ(bvd,_,tret) -> 
                                  substitute tret bvd arg
                              | _ -> raise Impos in
                              withinfo e t_e top_exp.p) fvar' targs in
                         tapp.v, t_inst, used_affines)

    | Exp_constant c -> 
        let t = typing_const env c in 
          Exp_constant c, t, []

    | Exp_constr_app (cvar, [], [], [e1;e2]) when Sugar.lid_equals cvar.v Const.tuple_UU_lid -> 
        (* special case: need to figure out the suitable variant of the tuple constructor
           based on the types of the arguments. Desugar just inserts tuple_UU_lid uniformly *)
        (* This case is also responsible for packing a dependent tuple *)
        let env, topt = clear_expected_typ env in 
        let packed_t, (e1, e2), (t1, t2), used_affines = match tensor_opt topt (expand_typ_fix env) with 
          | Some ({v=Typ_dtuple([(x1,t1);(_, t2)]); sort=_; p=_} as packed_t)-> 
              let env = set_expected_typ env t1 in
              let e1, t1, used_affines_1 = typing env e1 in
              let env' = match x1 with 
                | None -> set_expected_typ env t2
                | Some x -> 
                    if (is_value e1) then  (* pack dep tuple *)
                      let t2 = substitute_exp t2 x e1 in
                        set_expected_typ env t2
                    else env in
              let e2, t2, used_affines_2 = typing env' e2 in
                Some packed_t, (e1, e2), (t1, t2), (used_affines_1@used_affines_2)
          | _ -> 
              let e1, t1, used_affines_1 = typing env e1 in
              let e2, t2, used_affines_2 = typing env e2 in
                None, (e1, e2), (t1, t2), (used_affines_1@used_affines_2) in
        let _ = disjoint_affine_usages used_affines in
        let cname = Const.tuple_data_lid t1.sort t2.sort in 
        let ctyp = lookup_lid env cname in
        let cvar = withsort cname ctyp in
        let inst_ctyp, _ = instantiate_typ env ctyp [t1;t2] in
        let constr_app_typ = List.fold_left 
          (fun out arg -> match out.v with 
             | Typ_fun (None, arg_typ, ret_typ) -> ret_typ
             | _ -> raise Impos) inst_ctyp [e1;e2] in
        let _ = kinding env constr_app_typ in
        let capp = Exp_constr_app(cvar, [t1;t2], [], [e1;e2]) in
          (match packed_t with 
               None -> capp, constr_app_typ, used_affines
             | Some t -> 
                 let capp = (withinfo capp constr_app_typ top_exp.p) in
                   Exp_ascribed(capp,t,[]), t, used_affines)
            
    | Exp_constr_app (cname, targs, _, vargs) ->
        (* TODO: Add module name check on cname *)
        let ctyp, tps = match lookup_fvar_with_params env cname with
            None -> raise (Err ("Variable not found: "^(Sugar.text_of_lid cname.v)))
          | Some t -> 
              t in (* TODO: constructors may have extraneous type and value params *)
        let inst_ctyp, targs' = instantiate_typ env ctyp targs in  (* reconstructs targs if targs = [] *)
        let env, topt = clear_expected_typ env in
        let inst_ctyp, subst = match topt, tps with
          | None, [] -> inst_ctyp, []
          | None, _ -> let msg = spr "Cannot infer term indices for constructed type %s; a type annotation may help" (Pretty.strTyp ctyp) in
              raise (Error (msg, top_exp.p))
          | Some et, _ -> 
              let inst_ctyp, subst = try_instantiate_final_typ (set_current_value env top_exp) tps inst_ctyp et in 
                inst_ctyp, subst in
        let vargs_rev, result_typ, used_affines = List.fold_left 
          (fun (arglist, ctyp, used_affines) arg -> match ctyp.v with 
               Typ_fun (bvd_opt, arg_typ, ret_typ) -> 
                 let env' = set_expected_typ env arg_typ in 
                 let arg', t', used_affines' = typing env' arg in 
                 let ret_typ' = match bvd_opt with 
                     None -> ret_typ
                   | Some bvd -> substitute_exp ret_typ bvd arg' in
                 let ret_typ', _ = kinding env ret_typ' in (* ensure non-value expressions don't escape into types *)
                   (arg'::arglist, ret_typ', used_affines'@used_affines)
             | _ -> 
                 let msg = spr "Too many arguments to constructor %s, (e.g., %s is extra)" (Pretty.str_of_lident cname.v) (Pretty.strExp arg) in
                   raise (Err msg)) ([], inst_ctyp, []) vargs in
        let _ = match result_typ.v with 
            Typ_fun _ 
          | Typ_univ _ -> 
              raise (Error (spr "All data constructors must be fully-applied\n%s" (Pretty.strExp top_exp), top_exp.p))
          | _ -> () in
        let _ = disjoint_affine_usages used_affines in
        let ctyp = List.fold_left (fun ctyp (bvd, e) -> 
                                     substitute_exp ctyp bvd e) ctyp subst in
        let cname' = setsort cname ctyp in
        let phantomVargs = List.map (function 
                                       | Tparam_term(bvd, t) -> List.assoc bvd subst
                                       |  _ -> raise Impos) tps in
        let vargs' = List.rev vargs_rev in
          Exp_constr_app(cname', targs', phantomVargs, vargs'), result_typ, used_affines

    | Exp_primop(op, args) -> 
        (match lookup_operator env op with
          | None -> raise (Err ("Unknown operator: "^op.idText))
          | Some t -> 
              let args_rev, result_typ, used_affines = List.fold_left 
                (fun (args, t, affines) arg -> match t.v with 
                     Typ_fun(xopt, t_arg, t_ret) -> 
                       let env = set_expected_typ env t_arg in
                       let arg, t', used_affines = typing env arg in
                       let t_ret = match xopt with
                         | None -> t_ret
                         | Some x -> substitute_exp t_ret x arg in
                         arg::args, t_ret, used_affines@affines
                   | _ -> 
                       let msg = spr "Too many arguments to operator %s, (e.g., %s is extra)" op.idText (Pretty.strExp arg) in
                         raise (Err msg)) ([], t, []) args in
              let _ = disjoint_affine_usages used_affines in
              let _ = match result_typ.v with 
                | Typ_fun _ -> 
                    raise (Error (spr "All operators must be fully-applied\n%s" (Pretty.strExp top_exp), top_exp.p))
                | _ -> () in
                Exp_primop(op, List.rev args_rev), result_typ, used_affines)
          
    | Exp_abs (bvd, t, body) -> 
        let (t_formal, k_t) = match t.v with 
          | Typ_unknown -> 
              let t' = new_uvar env Kind_star t.p in
                t', Kind_star 
          | _ -> 
              let (t', k) = kinding env t in
                match k with 
                  | Kind_star 
                  | Kind_affine ->  (t', k)
                  | _ -> 
                      let msg = 
                        Printf.sprintf 
                          ("Bound term variables must always have *-kinded or A-kinded types.\n \
                          The formal parameter of type %s has %s kind")
                          (Pretty.strTyp t') (Pretty.strKind k) in
                        raise (Error (msg, top_exp.p)) in
        let binding = Binding_var (snd bvd, t_formal) in
        let env' = push_local_binding env binding in
        let env', topt = clear_expected_typ env' in
        let env' = match topt with
            None -> env'
          | Some ({v=Typ_fun(bvdopt, t', bodyt); sort=_; p=_}) -> env' (*TODO: propagate type annotation under a lambda?*)
          | _ -> env' in
        let body', t_body, used_affines = typing env' body in
        let mk_affine t = match t.v with 
            Typ_affine _ -> t
          | _ -> withsort (Typ_affine t) Kind_affine in
        let t_body' = match used_affines with 
          | [] -> t_body (* no constraints on body if it didn't use any affine assumptions *)
          | _ -> (* body captured affine assumptions, so it should be given an affine type *)
              let (t_body', k') = kinding env' t_body in
                match k' with 
                  | Kind_star when is_closure_typ env' t_body' -> mk_affine t_body' 
                  | _ -> t_body' in
        let typ = withsort (Typ_fun (Some bvd, t_formal, t_body')) Kind_star in
        let exp' = Exp_abs (bvd, t_formal, body') in
          let minus = bvd_to_bvar bvd in
          let used_affines = subtract used_affines [minus] in
            exp', typ, used_affines
              
    | Exp_tabs (bvd, k, body) ->
        let binding = 
          if Tcutil.kind_abstractable k then 
            let k = well_formed_kind env k in
              Binding_typ (snd bvd, k)
          else raise (Error ((spr "Abstraction over types of kind %s is not permitted" (Pretty.strKind k)), top_exp.p)) in
        let env' = push_local_binding env binding in
        let env', topt = clear_expected_typ env' in
        let env' = match topt with
          | None -> env'
          | Some ({v=Typ_univ(bvd', k', tbody);sort=_;p=_}) -> env' (*TODO: propagate type annotation to body  *)
          | _ -> env' in
        let body', t_body, used_affines = typing env' body in
        let t_body = match used_affines with
            [] -> t_body 
          | _ -> 
              let (t_body', k') = kinding env' t_body in
                match k' with 
                  | Kind_star when is_closure_typ env' t_body' -> withsort (Typ_affine t_body') Kind_affine
                  | _ -> t_body' in
        let res_exp = Exp_tabs (bvd, k, body') in
        let res_typ = withsort (Typ_univ(bvd, k, t_body)) Kind_star in
          res_exp, res_typ, used_affines
        
    | Exp_app (e1, e2) -> (* TODO: signal for e1 to include type insantiations *)
(*         let try_specialize_equality env e1 e2 = *)
(*           match e1.v with  *)
(*               Exp_fvar fv when Sugar.lid_equals fv.v Const.op_Eq ->  *)
(*                 let env, _ = clear_expected_typ env in *)
(*                 let e2', t2, used_affines_2 = typing env e2 in *)
(*                   bind_opt (Proof.op_eq_of_typ t2) *)
(*                     (fun lid ->  *)
(*                        let e1 = W (Exp_fvar (W lid)) in *)
(*                        let e1', t1, used_affines_1 = typing env e1 in *)
(*                        let used_affines = used_affines_1@used_affines_2 in *)
(*                        let _ = disjoint_affine_usages used_affines in *)
(*                        let t1 = expand_typ_fix env t1 in *)
(*                        let tret = open_typ_with_exp t1 e2' in *)
(*                        let typ_res, k = kinding env tret in (\* Non-value expressions do not escape into types *\) *)
(*                        let exp_res = Exp_app(e1', e2') in *)
(*                          Some (exp_res, typ_res, used_affines))  *)
(*             | _ -> None in *)
        let standard_app_typing env e1 e2 = 
          let env, expected_ret_typ = clear_expected_typ env in
          let e1', t1, used_affines_1 = typing env e1 in
          let t1 = match expected_ret_typ with
            | None -> t1
            | Some et -> 
                let t1', s = try_instantiate_final_typ (clear_current_value env) [] t1 et in
                  match s with 
                    | [] -> t1'
                    | _ -> raise Impos in
          let t1 = expand_typ_fix env t1 in
          let tformal = match t1.v with 
            | Typ_fun (bvd_opt, tformal, tret) 
            | Typ_affine ({v=Typ_fun(bvd_opt, tformal, tret); sort=_; p=_}) -> tformal
            | _ -> 
                let msg = Printf.sprintf "Expected a function typed term; got a term %s of type %s" (Pretty.strExp e1') (Pretty.strTyp t1) in
                  raise (Error (msg, e1'.p)) in
          let env' = set_expected_typ env tformal in
          let e2' = Absyn.withinfo (e2.v) (e2.sort) (top_exp.p) in
          let e2', t2', used_affines_2 = typing env' e2' in
          let used_affines = used_affines_1@used_affines_2 in
          let _ = disjoint_affine_usages used_affines in
          let tret = open_typ_with_exp t1 e2' in
          let typ_res, k = kinding env tret in (* Non-value expressions do not escape into types *)
          let exp_res = Exp_app(e1', e2') in
            exp_res, typ_res, used_affines in
          
          standard_app_typing env e1 e2
(*           (match try_specialize_equality env e1 e2 with *)
(*                Some result -> result *)
(*              | None -> standard_app_typing env e1 e2) *)
    | Exp_tapp (e, targ) -> 
        let env, expected_ret_typ = clear_expected_typ env in (* TODO: propagate expected type? *)
        let e', t_e, used_affines = typing (set_expect_tyapps env) e in
        let targ', res_typ = match t_e.v with 
          | Typ_univ (bvd, k, tt) -> 
              let targ', k' = kinding env targ in
                if not (TypeRelations.equivalent_kinds env k k') then
                  let msg = spr
                    "Type application expected a type of kind %s but got a %s-kinded type %s"
                    (Pretty.strKind k) (Pretty.strKind k') (Pretty.strTyp targ') in
                    raise (Error (msg, top_exp.p))                 
                else targ', substitute tt bvd targ' 
          | _ -> 
              let msg = 
                Printf.sprintf 
                  "Type application expected a polymorphic function on the LHS. Got a  term of type %s"
                  (Pretty.strTyp t_e) in
                raise (Error (msg, top_exp.p)) in
        let res_exp = Exp_tapp (e', targ') in
          res_exp, res_typ, used_affines

    | Exp_ascribed (e, t, ev) -> 
        let t', k = kinding env t in
        let env' = set_expected_typ env t' in
        let e', res_t, used_affines = typing env' e in
        let res_e = match e'.v with (* collapse multiple ascriptions *)
          | Exp_ascribed _ -> e'.v
          | _ -> Exp_ascribed (e', t', []) in
          res_e, res_t, used_affines
            
    | Exp_match (e, eqns, def) -> 
        let env', _ = clear_expected_typ env in
        let e', t', used_affines_e = typing env' e in
        let expanded_t' = expand_typ_fix env t' in
        let def', t_def, used_affines_def = typing env def in
        let env = match expected_typ env with 
          | None -> set_expected_typ env t_def 
          | _ -> env in
        let typed_eqns, used_affines_l = List.unzip 
          (List.map
             (fun (pi, branchi) -> 
                let pi', e_pi, bindings = typing_pat (set_expected_typ env expanded_t') pi in
(*                 List.iter (fun (Binding_var(b,t)) ->  *)
(*                     pr "Pattern binding for %s at type %s\n" b.idText (Pretty.strTyp t);  *)
(*                     printfn "%A" t) bindings; *)
                let env_true = List.fold_left (fun env binding -> push_local_binding env binding) env bindings in
                let env_true = push_local_binding env_true (Binding_match(e', e_pi)) in
                let branchi', tbranchi, used_affines_i = typing env_true branchi in
                let used_affines_i = subtract used_affines_i (match pi' with 
                                                                | Pat_variant (_, _, _, bvars) -> bvars
                                                                | _ -> []) in
                  (pi', branchi'), used_affines_i) eqns) in
        let used_affines_branches = List.fold_left union used_affines_def used_affines_l in
        let _ = disjoint_affine_usages (used_affines_e@used_affines_branches) in
        let res_exp = Exp_match(e', typed_eqns, def') in
          res_exp, t_def, union used_affines_e used_affines_branches
            
    | Exp_cond (e, e1, e2) -> 
        let e', t_e, used_affines_e = typing (set_expected_typ env bool_typ) e in
        let env_true, env_false = 
          if (is_value e') then 
            let b1 = Binding_match (e', Const.exp_true_bool) in
            let b2 = Binding_match (e', Const.exp_false_bool) in
              push_local_binding env b1, push_local_binding env b2 
          else env, env in
        let e1', t_branch, used_affines_1 = typing env_true e1 in
        let e2', _, used_affines_2 = typing (set_expected_typ env_false t_branch) e2 in
        let used_affines = used_affines_e @ (union used_affines_1 used_affines_2) in
        let _ = disjoint_affine_usages used_affines in        
        let res_exp = Exp_cond(e', e1', e2') in
        let res_typ = t_branch in
          res_exp, res_typ, used_affines
            
    | Exp_let (isRec, bindings, body) when isRec=false -> 
        (match bindings with 
           | [(bvd, {v=Typ_unknown;sort=_;p=_}, e)] -> (* TODO: Warning, ignoring type argument in let binding *)
               let env_orig = env in
                 (*                let env' = match body.v with *)
                 (*                  | Exp_bvar bv when (bvar_real_name bv)=(real_name bvd) -> env *)
                 (*                  | _ -> let env', _ = clear_expected_typ env_orig in env' in *)
               let env', _ = clear_expected_typ env_orig in
               let e', t', used_affines_e = typing env' e in
               let forall_t', Lambda_e' = generalize env' t' e' in
               let binding = Binding_var (snd bvd, forall_t') in
               let env_orig' = push_local_binding env_orig binding in 
               let bvv = withsort (pp_name bvd, real_name bvd) forall_t' in
               let bvexp = withsort (Exp_bvar bvv) forall_t' in
               let env_orig' = if is_value e' then push_local_binding env_orig' (Binding_match(bvexp, e')) else env_orig' in
               let body', t_body, used_affines_body = typing env_orig' body in
               let used_affines = used_affines_e@(subtract used_affines_body [bvv]) in
               let _ = disjoint_affine_usages used_affines in
                 (*                let _ = kinding env_orig t_body in (\* ensure let-bound var does not escape scope *\) *)
               let exp' = Exp_let(isRec, [(bvd, forall_t', Lambda_e')], body') in
                 exp', t_body, used_affines
           | _ -> raise (NYI "Explicit type annotations on let-bound variables"))

    | Exp_let (isRec, bindings, body) when isRec -> 
        let env',bindings_rev = List.fold_left 
          (fun (env, out) (bvd,t,e) -> 
             match t.v with 
               | Typ_unknown -> raise (NYI "Type inference for recursive let-bound variables")
               | _ -> 
                   let t, k = kinding env t in
                     match k with
                         Kind_affine ->
                           raise (Err ("Recursive functions cannot have affine types"))
                       | _ -> 
                           let binding = Binding_var (snd bvd, t) in
                           let env' = push_local_binding env binding in
                             env', (bvd,t,e)::out) (env,[]) bindings in
        let bindings' =  List.fold_left 
          (fun out (bvd,t,exp) -> 
             let env' = set_expected_typ env' t in
             let exp', t_exp, used_affines = typing env' exp in
               match used_affines with 
                   [] -> (bvd, t, exp')::out
                 | _ -> raise (Error ("Recursive functions cannot use affine assumptions", exp'.p))) [] bindings_rev in
        let body', t_body, used_affines = typing env' body in
        let res_exp = Exp_let(isRec, bindings', body') in
        let res_typ = t_body in
          res_exp, res_typ, used_affines        

    | Exp_recd (_, _, _, (((fn,e)::tl) as fn_e_l)) -> 
        let recd_lident, tabbrv, tps, record = 
          (try 
             lookup_record_typ env fn 
           with 
               Not_found_binding _ -> 
                 let msg = spr "%s is not a valid field name" (Sugar.text_of_lid fn) in
                   raise (Error(msg, top_exp.p))) in             
        let (record, t_x_vars_rev, abductibles) = List.fold_left 
          (fun (record, t_x_vars, abductibles) -> function
             | Tparam_typ (bvd, k) -> 
                 let uv = new_uvar env k top_exp.p in
                 let record' = substitute record bvd uv in
                   record', Inl uv::t_x_vars, abductibles
             | Tparam_term (bvd, t) -> 
                 let bv = bvd_to_bvar bvd in
                   record, Inr bv::t_x_vars, bv::abductibles)  (record, [], []) tps in
        let env, _ = clear_expected_typ env in
        let _, used_affines, fn_e_l_rev, abductions = List.fold_left
          (fun (env, affines, fn_e_l, abducts) (fn,e) -> 
             let t = Tcutil.lookup_field_in_record_typ env fn record in
             let e', t', used_affines = typing env e in
             let sopt = TypeRelations.instantiable env abductibles t' t in
               match sopt with 
                 | None -> 
                     let sopt = TypeRelations.convertible (set_current_value env e') t' t in
                       (match sopt with 
                          | None -> terr env t t' top_exp.p
                          | Some subst -> 
                              let _ = unify_subst_vars subst in
                                env, affines@used_affines, (fn,e')::fn_e_l, abducts)
                 | Some (subst, abducts') -> 
                     let _ = unify_subst_vars subst in 
                     let env' = List.fold_left push_local_binding env abducts in
                       env', affines@used_affines, (fn,e')::fn_e_l, abducts'@abducts) (env, [], [], []) fn_e_l in
        let find_abductive_assumption e1 = 
          let finder = function 
            | Binding_match (e1', e2) -> equalsExp e1 e1' 
            | _ -> false in 
            try
              let (Binding_match(_, e2)) = List.find finder abductions in
                Some e2
            with 
                _ -> None in
        let result_typ, targs, phantoms = List.fold_right
          (fun tp (out_typ, targs, phantoms) -> match tp with
               Inl tv -> (match out_typ.sort with
                            | Kind_tcon (_, k') -> 
                                let result_typ = withsort (Typ_app(out_typ, tv)) k' in
                                let targs = tv::targs in
                                  result_typ, targs, phantoms
                            | _ -> raise Impos)
             | Inr bv -> 
                 let bvexp = withinfo (Exp_bvar bv) bv.sort bv.p in
                   (match out_typ.sort with
                      | Kind_dcon (t, k') -> 
                          (match find_abductive_assumption bvexp with
                               None -> let msg = 
                                 spr "Unable to infer witness for a dependent type index %s in type %s. An explicit annotation may help."
                                   (Pretty.strExp bvexp) (Pretty.strTyp tabbrv) in
                                 raise (Error (msg, top_exp.p))
                                   
                             | Some e' -> 
                                 let result_typ = withsort (Typ_dep (out_typ, e')) k' in
                                 let phantoms = e'::phantoms in
                                   result_typ, targs, phantoms)
                      | _ -> raise Impos)) t_x_vars_rev (tabbrv,[],[]) in
        let _ = disjoint_affine_usages used_affines in 
          Exp_recd(Some recd_lident, targs, phantoms, (List.rev fn_e_l_rev)), result_typ, used_affines 
            
    | Exp_proj (e, fn) when Sugar.lid_equals fn Const.tuple_proj_one -> 
        let env, topt = clear_expected_typ env in
        let e', t, used_affines = typing env e in (* TODO: instead of synthesizing a type for e, can we check it? *)
        let t' = expand_typ_fix env t in 
        let err () = 
          let msg = spr "Expected a tuple expression; got %s" 
            (Pretty.strTyp t') in
            raise (Error(msg, top_exp.p)) in
        let fieldTyp =  match t'.v with 
          | Typ_dtuple [(_,t1);_] -> t1
          | _ -> err() in
          Exp_proj(e', fn), fieldTyp, used_affines                    
        
    | Exp_proj (e, fn) when Sugar.lid_equals fn Const.tuple_proj_two -> 
        let env, topt = clear_expected_typ env in
        let e', t, used_affines = typing env e in (* TODO: instead of synthesizing a type for e, can we check it? *)
        let t' = expand_typ_fix env t in 
        let err () = 
          let msg = spr "Expected a non-dependent tuple expression; got %s" 
            (Pretty.strTyp t') in
            raise (Error(msg, top_exp.p)) in
        let fieldTyp =  match t'.v with 
          | Typ_dtuple [(_,t1);(_,t2)] -> 
              let _ = 
                try 
                  kinding env t2                 (* check to make sure that t2 is not dependent on t1 *)
                with 
                    _ -> err () in
                t2
          | _ -> err() in
          Exp_proj(e', fn), fieldTyp, used_affines
            
    | Exp_proj (e, fn) -> 
        let env, topt = clear_expected_typ env in
        let e', t, used_affines = typing env e in (* TODO: instead of synthesizing a type for e, can we check it? *)
        let fieldTyp = Tcutil.lookup_field_in_record_typ env fn t in
        let fieldTyp = if is_value e' && not (is_kind_level env) then 
          let eqT = Tcutil.eqT_of_typ env fieldTyp in 
          let id = genident None in
          let bvd = id, id in
          let eqf = Wt (Typ_dep(Wt (Typ_dep(eqT, W (Exp_bvar (bvd_to_bvar bvd)))), W (Exp_proj(e',fn)))) in
          let refined_t, _ = kinding env (Wt (Typ_refine(bvd, fieldTyp, eqf))) in
            refined_t
        else fieldTyp in
          Exp_proj(e', fn), fieldTyp, used_affines
            
    | Exp_bot -> 
        let t = match expected_typ env with 
          | Some t -> t 
          | _ -> new_uvar env Kind_star top_exp.p in
          Exp_ascribed(withsort Exp_bot t, t, []), t, []
  in 
  let do_typing () =     
    try 
      let exp', t, A = do_typing env top_exp.v in
      let t = compress t in
      let env', t_expected = clear_expected_typ env in
        match t_expected with
            None -> withinfo exp' t top_exp.p, t, A
          | Some t' -> 
              let t' = compress t' in
              let cv = withinfo exp' t top_exp.p in
(*               if !Util.mainmodule then pr "For sub-term (%s), testing convertability of computed typ (%s) with expected typ (%s)\n"  *)
(*                 (Pretty.strExp (W exp')) (Pretty.strTyp t) (Pretty.strTyp t'); *)
              let sopt = TypeRelations.convertible (set_current_value env' cv) t t' in (* TODO: stash the proof with e for later phase? *)
                match sopt with 
                  | None -> terr env t' t top_exp.p
                  | Some subst -> 
                      let _ = unify_subst_vars subst in  (* this has a side effect; changes uvar pointers in Unionfind *) 
                      let result = withinfo exp' t' top_exp.p in
                      let result = ascribe result t' [] in
                        result, t', A
    with
        Err msg -> 
          raise (Error(msg, top_exp.p)) in
    profile do_typing tc_ctr
      
and typing_pat env  pat : (pat * exp * list<binding>) = match pat, expected_typ env with

  | Pat_variant(lid, targs, _, pat_bvars), Some ({v=(Typ_dtuple [(nopt, t1); (_,t2)]);sort=_;p=_} as et)-> 
      (* Special case: pattern matching for unpacking a dependent tuple *)
      let lid' = Const.tuple_data_lid t1.sort t2.sort in
      let lid_t = lookup_lid env lid' in
      let datacon = withinfo lid' lid_t (Sugar.range_of_lid lid) in
      let (fst, tfst), (snd, tsnd), pat_bindings = match targs, pat_bvars with 
        | [], [fst; snd] -> 
            let fst = setsort fst t1 in
            let snd, t2 = match nopt with 
              | None -> setsort snd t2, t2
              | Some x -> 
                  let t2' = substitute_exp t2 x (withsort (Exp_bvar fst) t1) in
                    setsort snd t2', t2' in
              (fst, t1), (snd, t2), [Binding_var(bvar_real_name fst, t1); Binding_var(bvar_real_name snd, t2)]
        | _ -> raise (Err "Unexpected number of pattern variables when unpacking a dependent tuple") in
      let pat_exp = withsort (Exp_constr_app(datacon, [tfst;tsnd], [], [withsort (Exp_bvar fst) tfst;
                                                                        withsort (Exp_bvar snd) tsnd])) et in 
      let pat' = Pat_variant(lid', [tfst;tsnd], [], [fst;snd]) in
        pat', pat_exp, pat_bindings
          
  | Pat_variant(lid, targs, [](* no explicit phantoms in source*), pat_bvars), _ -> 
      let t_d, phantom_params = match lookup_fvar_with_params env (W lid) with
        | None -> raise (Err (spr "Cosntructor %s not found" (Pretty.str_of_lident lid)))
        | Some (t,tps) -> t, tps in
      let datacon = withinfo lid t_d (Sugar.range_of_lid lid) in
      let inst_t_d, targs' = instantiate_typ env t_d targs in
      let rec mk_vargs (outb,outv) (vargs:list<bvar<typ>>) inst_t = match inst_t.v, vargs with 
        | Typ_fun (formal_opt, targ, tres), bv::vrest -> 
            let bv' = setsort bv targ in
            let exp = withinfo (Exp_bvar bv') targ bv'.p in
            let trest = open_typ_with_exp inst_t exp in  (* open_typ ((x:t) -> t') e  subs e for x in t' *)
            let binding = Binding_var (bvar_real_name bv', targ) in
              mk_vargs (binding::outb, bv'::outv) vrest trest
        | Typ_fun _, [] -> raise (Err "Insufficient pattern variables")
        | _, hd::tl -> raise (Err "Too many pattern variables")
        | _, [] -> (outb,outv), inst_t in
      let (bindings_rev, pat_bvars_rev), final_typ = mk_vargs ([],[]) pat_bvars inst_t_d in
      let final_typ, bindings_rev, pat_bvars_rev, phantoms = match expected_typ env, phantom_params with 
        | _, [] -> final_typ, bindings_rev, pat_bvars_rev, []
        | Some expected_t, tps -> 
            let pat_exp = Exp_constr_app(datacon, 
                                         targs', [], 
                                         List.rev (List.map (fun bv -> withsort (Exp_bvar bv) bv.sort)
                                                     pat_bvars_rev)) in
            let env = set_current_value env (W pat_exp) in
            let env = List.fold_right (fun binding env -> push_local_binding env binding) bindings_rev env in
            let inst_final_typ, subst = try_instantiate_final_typ env tps final_typ expected_t in
            let bindings_rev = List.map (fun (Binding_var (bv, t)) -> Binding_var(bv, substitute_exp_l t subst)) bindings_rev in
            let pat_bvars_rev = List.map (fun bv -> setsort bv (substitute_exp_l bv.sort subst)) pat_bvars_rev in
            let phantoms = List.map (function 
                                       | Tparam_term(bvd, t) -> List.assoc bvd subst
                                       |  _ -> raise Impos) tps in
              inst_final_typ, bindings_rev, pat_bvars_rev, phantoms in
      let pat' = Pat_variant(lid, targs', phantoms, List.rev pat_bvars_rev) in
      let pat_exp_args =  List.fold_left 
        (fun out bv -> 
           let ebv = withinfo (Exp_bvar bv) bv.sort ((bvar_ppname bv).idRange) in
             ebv::out) [] pat_bvars_rev in 
      let pat_exp = Exp_constr_app(datacon, targs', phantoms, pat_exp_args) in
      let abducts = match expected_typ env with 
          None -> []
        | Some et -> 
            let env' = List.fold_left push_local_binding env (List.rev bindings_rev) in
            let sopt = TypeRelations.instantiable env' pat_bvars_rev et final_typ in
              match sopt with 
                  None -> 
                    let sopt = TypeRelations.convertible (set_current_value env' (withsort pat_exp final_typ)) et final_typ in
                      (match sopt with 
                         | None -> 
                             let msg = spr "Expected a pattern of type %s but got pattern of type %s" 
                               (Pretty.strTyp et) (Pretty.strTyp final_typ) in
                               raise (Err msg)
                         | Some subst -> 
                             let _ = unify_subst_vars subst in [])
                | Some (subst, abducts) -> let _ = unify_subst_vars subst in abducts in
        pat', withsort pat_exp final_typ, (List.rev bindings_rev)@abducts
          
let rec annotate exp t = match t.v with 
  | Typ_univ (bvd, k, t') -> 
      let exp' = annotate exp t' in
        withinfo (Exp_tabs (bvd, k, exp')) t exp.p

  | Typ_fun (bvd_opt, targ, tret) -> 
      let exp' = match exp.v with 
        | Exp_abs (bvd, {v=Typ_unknown;sort=_;p=_}, body) -> 
            let tret = match bvd_opt with
              | None -> tret 
              | Some bvd' -> 
                  let pp, rn = bvd in
                  let bv =  (withinfo (pp, rn) targ pp.idRange)  in
                  let e = withinfo (Exp_bvar bv) targ pp.idRange in
                    substitute_exp tret bvd' e in
            let body' = annotate body tret in
              Exp_abs (bvd, targ, body')

        | Exp_abs (bvd, t, body) -> 
            raise (Error ("Type ascriptions both in argument patterns and value declarations are not allowed", exp.p))
              
        | _ -> raise (Error ("Value declaration assigns a function-type to a non-lambda abstraction", exp.p)) in
        
        withinfo exp' t exp.p

  | Typ_affine t' -> 
      let e = annotate exp t' in
        withpos (Exp_ascribed (e, t, [])) e.p
          
  | _ -> withinfo (Exp_ascribed (exp, t, [])) t exp.p
      
let check_signature (env:env) (s:signature) : (env * signature) = 
  let check_tparams lid env tps = 
    let tps' = List.fold_left  (* NB: no functional dependences between tparams *)
      (fun out tp -> 
         let tp' = match tp with
           | Tparam_typ (i,k) -> 
               let k = well_formed_kind env k in
                 Tparam_typ (i,k)
           | Tparam_term (i,t) -> 
               let t,k = kinding env t in 
                 match k with 
                   | Kind_star -> Tparam_term(i,t)
                   | _ -> 
                       let msg = 
                         Printf.sprintf 
                           "Types can only be indexed by *-kinded terms; %s has a %s-kinded argument"
                           (Sugar.text_of_lid lid) (Pretty.strKind k) in
                         raise (Error(msg, (Sugar.range_of_lid lid))) in
           tp'::out) [] tps in
      List.rev tps' in
  let rec check_sigelt env = function (* TODO: handle recursive type declarations *)
    | Sig_tycon_kind (lid, tps, k, isProp) -> 
        (* parameterized types can now appear in kinds; 
           e.g. prop Eq<'a>::'a => 'a => * IS allowed *)
        let tps =check_tparams lid env tps in
        let env = push_tparams env tps in
        let k = well_formed_kind env k in 
        let result = Sig_tycon_kind(lid, tps, k, isProp) in
          result
            
    | Sig_datacon_typ (lid, tps, t, isAssumption, aq, tcn) -> 
        let tps = check_tparams lid env tps in
        let env' = push_tparams env tps in 
        let t, k = kinding env' t in
          (match k with 
               Kind_star
             | Kind_affine -> Sig_datacon_typ (lid, tps, t, isAssumption, aq, tcn)
             | _ -> 
                 let msg = 
                   Printf.sprintf 
                     "Data constructors must construct data of *- or A-kinded type; %s is given a %s-kinded type"
                     (Sugar.text_of_lid lid) (Pretty.strKind k) in
                   raise (Error(msg, (Sugar.range_of_lid lid))))
            
    | Sig_value_decl (lid, t) ->
        let t, k = kinding env t in
          (match k with 
             | Kind_star 
             | Kind_affine -> Sig_value_decl (lid, t)
             | _ -> 
                 let msg = 
                   Printf.sprintf 
                     "Value declarations must be given *- or A-kinded types; %s is given a %s-kinded type"
                     (Sugar.text_of_lid lid) (Pretty.strKind k) in
                   raise (Error(msg, (Sugar.range_of_lid lid))))

    | Sig_record_typ (lid, tps, kk, t) -> 
        let tps = check_tparams lid env tps in
        let env' = push_tparams env tps in
        let t, k = kinding env' t in
          begin
            match k with 
              | Kind_star
              | Kind_affine -> 
                  Sig_record_typ (lid, tps, kk, t) 
              |  _ -> 
                   let msg = 
                     spr "Type abbreviations are only supported for ground types; %s abbreviates a %s-kinded type"
                       (Sugar.text_of_lid lid) (Pretty.strKind k) in
                     raise (Error(msg, (Sugar.range_of_lid lid)))
          end

    | Sig_typ_abbrev (lid, tps, kk, t) -> 
        let tps = check_tparams lid env tps in
        let env' = push_tparams env tps in
        let t, k = kinding env' t in
          begin
            match k with 
              | Kind_star
              | Kind_affine -> 
                  if (TypeRelations.equivalent_kinds env kk Kind_unknown || 
                        TypeRelations.equivalent_kinds env k kk) then
                    Sig_typ_abbrev (lid, tps, kk, t) 
                  else
                    let msg = 
                      spr "Expected a type of kind %s; got %s"
                        (Pretty.strKind k) (Pretty.strKind kk) in
                      raise (Error(msg, (Sugar.range_of_lid lid)))
              |  _ -> 
                   let msg = 
                     spr "Type abbreviations are only supported for ground types; %s abbreviates a %s-kinded type"
                       (Sugar.text_of_lid lid) (Pretty.strKind k) in
                     raise (Error(msg, (Sugar.range_of_lid lid)))
          end

    | Sig_extern_value (eref, lid, typ) -> 
        let typ,k = kinding env typ in
          Sig_extern_value(eref, lid, typ)
            
    | Sig_extern_typ (eref, s) as ss-> 
        let env' = set_extern_typ env in
        let se = check_sigelt env s in
          Sig_extern_typ(eref, se)
            
    | _ -> raise Impos
        
  in
  let env', outsig = List.fold_left 
    (fun (env, outsig) sigelt -> 
       try 
         let sigelt' = check_sigelt env sigelt in
         let genfrom = match sigelt' with 
             Sig_extern_typ(eref, s) -> s
           | _ -> sigelt' in
           let gens = Proof.process_sigelt env genfrom in 
           if !Util.interleave_proof_principles then
             let env' = push_sigelt env sigelt' in
             let env', gens' = List.fold_left (fun (env', ses) se -> 
                                                 let se' = check_sigelt env' se in
                                                   (push_sigelt env' se', se'::ses)) (env',[]) gens in
               (env', (gens')@[sigelt']@outsig)
           else (push_sigelt env sigelt', sigelt'::outsig)
             (*          else push_sigelt env sigelt', sigelt'::outsig *)
       with 
           Err msg -> raise (Error(msg, range_of_sigelt sigelt)))
    (env,[]) s in
    env', List.rev outsig
      
let typing_modul env m =       
  let typing_modul env m = 
    let check_let_binding env = function
      | ([(bvd, _, exp)], isRec) when isRec=false -> 
          let qname = (current_module env) @ [(fst bvd)] in (* the only case where the pp name is actually used *)
          let env = set_current_unit env qname in
          let exp, val_decl_t = 
            try
              let t = lookup_val_decl env qname in
                annotate exp t, t
            with 
                Not_found_binding _ -> 
                  let msg = spr "No val-declaration for %s. Top-level let-bindings must be ascribed a type" ((pp_name bvd).idText) in
                    raise (Error (msg, (pp_name bvd).idRange)) in 
          let exp', t_exp, used_affines = typing env exp in
          let t_forall, Lambda_exp = generalize env t_exp exp' in
            if not (TypeRelations.equivalent env t_forall val_decl_t) then 
              terr env val_decl_t t_forall exp'.p
            else
              let _ = match used_affines with 
                  hd::tl -> raise (Error ("Top-level expression cannot capture affine assumptions", exp'.p))
                | _ -> () in
              let binding = ([bvd, t_forall, Lambda_exp], isRec) in
              let rn = snd bvd in 
              let env' = push_local_binding env (Binding_var (rn, t_forall)) in
                env', binding
                  
      | (bindings, isRec) when isRec -> 
          let rec_env, bindings' = List.fold_left 
            (fun (env, bindings') (bvd, t, exp) -> 
               let (pp, rn):bvdef = bvd in
                 try
                   let qname = (current_module env) @ [pp] in (* again use ppname *)
                   let env = set_current_unit env qname in
                   let t = lookup_val_decl env qname in
                   let exp' = annotate exp t in
                   let binding = Binding_var (rn, t) in
                   let env' = push_local_binding env binding in
                     (env', (bvd, t, exp')::bindings')
                 with 
                     Not_found_binding _ -> 
                       raise (Error ("Top-level let recs must be ascribed a type", pp.idRange))) (env, []) bindings in
          let bindings' = List.rev bindings' in
          let typed_bindings = List.map 
            (fun (bv, t, exp) -> 
               let exp', t_exp, used_affines = typing (set_expected_typ rec_env t) exp in
               let _ = match used_affines with 
                   hd::tl -> raise (Error ("Top-level expressions cannot capture affine assumptions", exp'.p))
                 | _ -> () in
                 (bv, t, exp')) bindings' in
            rec_env, (typed_bindings, isRec)
              
      | _ -> raise Impos in
      
(*     let _ = (if not (Sugar.lid_equals m.name Const.prims_lid) then (Printf.printf "Desugared module:\n"; printfn "%A" m)) in *)
    let env, signature = check_signature (set_in_typ_decl env) m.signature in 
    let env = clear_in_typ_decl env in
    let env, bindings' = List.fold_left 
      (fun (env, out) binding -> 
         let env', binding' = check_let_binding env binding in
           env', binding'::out) (env,[]) m.letbindings in
    let main' = match m.main with 
      | None -> None
      | Some main -> 
          let main', t_main, used_affines = typing env main in
            Some main' in
      {name=m.name; extends=m.extends; pos=m.pos; signature=signature;
       letbindings=List.rev bindings'; main=main'; exports=signature} 
        (* TODO: currently everything in a signature is exported; hide abstract symbols *)
  in
    try 
      let result = Some(typing_modul env m) in
(*         (Printf.printf "NCanaries = %d\n" !canary); *)
        result
    with 
      | e when handleable e -> handle_err false None e
          
  
