#light "off"
// (c) Microsoft Corporation. All rights reserved

module Microsoft.Fine.Tcenv

open Absyn 
open AbsynUtils
open Util
open Const
open Profiling 

exception Err of string

let lid_ctr = new_counter "lookup_lid"
let bvar_ctr = new_counter "lookup_bvar"
let btvar_ctr = new_counter "lookup_btvar"
let vd_ctr = new_counter "lookup_val_decl"
let fvar_ctr = new_counter "lookup_fvar"
let tc_ctr = new_counter "lookup_typ_const"
let ta_ctr = new_counter "lookup_typ_abbrev"
   
type binding =
  | Binding_var of ident * typ
  | Binding_typ of ident * kind
  | Binding_match of exp * exp

type sigtable = HashMultiMap<string, sigelt>

let rec strlid_of_sigelt se = match se with 
  | Sig_tycon_kind(lid, _, _, _)
  | Sig_typ_abbrev(lid, _, _, _) 
  | Sig_record_typ(lid, _, _, _) 
  | Sig_datacon_typ(lid, _, _, _, _, _) 
  | Sig_extern_value(_, lid, _) 
  | Sig_value_decl(lid, _) -> Sugar.text_of_lid lid
  | Sig_extern_typ(eref, s) -> strlid_of_sigelt s
  | _ -> raise Impos
      
let signature_to_sigtable s = 
  let ht = Hashtbl.create 100 in
  let _ = List.iter (fun se -> Hashtbl.add ht (strlid_of_sigelt se) se) s in
    ht

let modules_to_sigtable mods = 
  signature_to_sigtable (List.concat (List.map (fun (_,m) -> m.signature) mods))

type phase = | TC | DEREFINE 
type env = {
  phase:phase;
  curmodule: lident;             (* Name of this module---the "principal" subscript *)
  gamma:list<binding>;           (* Main typing environment *)
  externs: signature;            (* Extern declarations accumulated from other modules *)
  e_signature: signature;        (* Type and data constructors for this module *)
  curunit: lident;               (* the name of the current top-level function unit *)
  in_typ_decl: bool;             (* are we currently checking a type declaration*)
  modules:list<lident * modul>;  (* already fully type checked modules *)
  expected_typ:option<typ>;
  recall_expected_typ:option<typ>;
  expected_kind:option<kind>;
  expect_tyapps:bool;
  current_value:option<exp>;
  use_solver:bool;
  kind_level:bool;
  sigtab:sigtable;
  extern_typ:bool;
  expand_externs:bool;
  uvars:list<uvar>
}


let in_tc_phase env = env.phase=TC
let set_tc_phase env = {env with phase=TC}
let in_derefinement_phase env = env.phase = DEREFINE
let set_derefinement_phase env = {env with phase=DEREFINE}


let empty_env =  
  { phase=TC;
    curmodule= [];
    curunit=[];
    in_typ_decl=false;
    gamma= [];
    externs=[];
    e_signature= [];
    modules= [];
    expected_typ=None;
    recall_expected_typ=None;
    expected_kind=None;
    expect_tyapps= false;
    current_value=None;
    use_solver=true;
    kind_level=false;
    sigtab=Hashtbl.create 100;
    extern_typ=false;
    expand_externs=true;
    uvars=[]
  }

let set_kind_level env = {env with kind_level=true}
let is_kind_level env = env.kind_level

let set_solver_flag env f = {env with use_solver = f}
let use_solver env = env.use_solver

let modules env = env.modules
let current_module env = env.curmodule
let set_current_module env lid = {env with curmodule=lid} 

let current_unit env = env.curunit
let set_current_unit env lid = {env with curunit=lid} 

let set_in_typ_decl env = {env with in_typ_decl=true}
let in_typ_decl env = env.in_typ_decl
let clear_in_typ_decl env = {env with in_typ_decl=false}

(* let W t = withsort t Kind_unknown *)

let initial_env module_lid = 
  let ds_env = DesugarEnv.initial_env [] module_lid in
  let open_mods = DesugarEnv.open_modules ds_env in
  let sigtab = modules_to_sigtable open_mods in
    {phase=TC;
     curmodule = DesugarEnv.current_module ds_env; 
     curunit=[];
     in_typ_decl=false;
     gamma=[];
     externs=[];
     e_signature=[];
     modules=open_mods;
     expected_typ=None;
     recall_expected_typ=None;
     expect_tyapps = false;
     expected_kind=None;
     current_value=None;
     use_solver=true;
     kind_level=false;
     sigtab=sigtab;
     extern_typ=false;
     expand_externs=true;
     uvars=[]
    }
    
exception Not_found_binding of env * Disj<typ',exp'>
exception Assertion_failure
  
let assert_eq v1 v2 = 
  if v1=v2 then () 
  else 
    (printfn "%A" v1;
     printfn "%A" v2;
     raise Assertion_failure)

let lookup_bvar_ident env (rn:ident) : option<typ> = 
  try
    let b = List.find 
      (function
         | Binding_var (id, t) when id.idText = rn.idText -> true
         | _ -> false) env.gamma in
      match b with
          Binding_var (_,t) -> Some t
        | _ -> raise Impos
  with _ -> None
      
let lookup_bvar env bv = 
  let lb () =
    let rn:ident = (bvar_real_name bv) in
      match lookup_bvar_ident env rn with
        | None -> raise (Not_found_binding(env, Inr (Exp_bvar bv))) 
        | Some t -> t in
    profile lb bvar_ctr
    
let lookup_val_decl env lid = 
  let lvd () = 
    let find_vd path = function
      | Sig_extern_value(_, lid, _) 
      | Sig_value_decl (lid, _) when (Sugar.path_of_lid lid)=path -> true
      | _ -> false in
    try 
      match List.find (find_vd (Sugar.path_of_lid lid)) (env.externs@env.e_signature) with 
        | Sig_value_decl (_, t) 
        | Sig_extern_value(_, _, t) -> t
        | _ -> raise Impos
    with
      | :? System.Collections.Generic.KeyNotFoundException ->
        raise (Not_found_binding (env, Inr (Exp_fvar (withpos lid (Sugar.range_of_lid lid))))) in
    profile lvd vd_ctr
      
let err_t s ftv = Inl (Typ_const (withsort ftv s))
let err_e s fv = Inr (Exp_fvar (withsort fv s))

let lookup_qname env lid pred err = 
  let err () = raise (Not_found_binding (env, err lid)) in
  let rec finder lid = function
      [] -> None
    | hd::tl -> match pred lid hd with
        | None -> finder lid tl 
        | Some x -> Some x in
  let module_name, id = Util.pfx lid in
  let resultOpt = 
    if Sugar.lid_equals module_name env.curmodule ||
       Sugar.is_extern_ref lid then
      finder lid (env.externs@env.e_signature)
    else
      let (_, modul) = 
        (try 
           List.find (fun (lid, modul) -> Sugar.lid_equals lid module_name) env.modules 
         with
             _ -> err ()) in
        finder lid modul.signature  in
    match resultOpt with 
        None -> err ()
      | Some x -> x

let lookup_lid env lid = 
  let ll () = 
    let rec finder lid' = function
      | Sig_datacon_typ (lid, tps, t, _, aq, _) when Sugar.lid_equals lid lid' -> 
          let t = List.fold_right 
            (fun tp t -> match tp with 
                 Tparam_typ (bvd,k) -> withsort (Typ_univ (bvd, k, t)) Kind_star
               | _ -> raise (NYI "Extraneous term indices")) tps t in
            Some t
      | Sig_extern_value(_, lid, t) 
      | Sig_value_decl (lid, t) when Sugar.lid_equals lid lid' -> Some t
      | Sig_extern_typ (eref, s) as ss -> finder lid' s
      | _ -> None
    in
      lookup_qname env lid finder (err_e (Wt Typ_unknown)) in
    profile ll lid_ctr
      
let lookup_fvar env (fv:var<typ>) = 
  let lf () =
    let key = Sugar.text_of_lid fv.v in
    let rec extract = function
        | Sig_datacon_typ (lid, tps, t, _, aq,_) -> 
            let t = List.fold_right 
              (fun tp t -> match tp with 
                   Tparam_typ (bvd,k) -> withsort (Typ_univ (bvd, k, t)) Kind_star
                 | _ -> raise (NYI "Extraneous term indices")) tps t in
              Some t
        | Sig_extern_value (_, lid, t)
        | Sig_value_decl (lid, t) -> Some t
        | Sig_extern_typ(_, s) -> extract s
        | _ -> None in
      match Hashtbl.tryfind env.sigtab key with 
        | Some s -> extract s
        | None -> None 
  in
    profile lf fvar_ctr
      
let lookup_fvar_with_params env (fv:var<typ>) = 
  let lfp () = 
    let key = Sugar.text_of_lid fv.v in
    let rec extract = function
        | Sig_datacon_typ (lid, tps, t, _, aq, _) -> 
            let out = List.fold_right 
              (fun tp (t, params) -> match tp with 
                 | Tparam_typ (bvd,k) -> withsort (Typ_univ (bvd, k, t)) Kind_star, params
                 | Tparam_term (bvd, bv_t) -> 
                     let nbvd = (pp_name bvd, genident None) in
                     let nbv = withsort (Exp_bvar (bvd_to_bvar nbvd)) bv_t in
                     let t = substitute_exp t bvd nbv in
                       t, (Tparam_term(nbvd, bv_t)::params)) tps (t,[]) in
              Some out
        | Sig_extern_value(_, lid, t)
        | Sig_value_decl (lid, t) -> Some (t, [])
        | Sig_extern_typ (_, s) -> extract s
        | _ -> None in
      match Hashtbl.tryfind env.sigtab key with
        | Some s -> extract s
        | None -> None
  in
    profile lfp fvar_ctr

let is_datacon env (fv:var<typ>) = 
  let key = Sugar.text_of_lid fv.v in
  let rec is_dc = function
    | Sig_datacon_typ _ -> true
    | Sig_extern_typ(_, s) -> is_dc s
    | _ -> false in
    match Hashtbl.tryfind env.sigtab key with
      | Some s -> is_dc s
      | None -> false

let is_prop env t = 
  match get_tycon t with 
    | Some (Typ_const tc) -> 
        let lid = tc.v in
          (match Hashtbl.tryfind env.sigtab (Sugar.text_of_lid lid) with
             | Some (Sig_tycon_kind(_, _, _, prop)) -> prop
             | _ -> false)
    | Some (Typ_btvar _) -> true
    | _ -> false
        
let lookup_field_name env fn = 
  let rec finder full_lid = function
    | Sig_record_typ (_, tps, _, {v=Typ_record fn_t_l; sort=_; p=_}) -> 
        (try 
           let (fn,t) = List.find 
             (fun (fn,t) -> Sugar.lid_equals full_lid fn) fn_t_l in
             Some (tps, t)
         with
             _ -> None) 
    | _ -> None
  in
    lookup_qname env fn finder (err_e (Wt Typ_unknown))

let field_name_exists env fn =
  try
    let _ = lookup_field_name env fn in
      true
  with
      Not_found_binding _ -> false

let tycon_kind_from_tparams tps k = 
  let tyK = List.fold_right 
    (fun tp k -> match tp with
         Tparam_typ (_,k') -> Kind_tcon(k', k)
       | Tparam_term (_, t) -> Kind_dcon(t, k)) tps k in
    tyK
      
let lookup_record_typ env fn = 
  let rec finder fn = function
    | Sig_record_typ (lident, tps, _, ({v=Typ_record fn_t_l; sort=_; p=_} as t)) -> 
        (try 
           let _ = List.find 
             (fun (fn',_) -> Sugar.lid_equals fn' fn) fn_t_l in
           let tK = tycon_kind_from_tparams tps t.sort in
           let var_abbrv = withsort lident tK in
           let tabbrv = withsort (Typ_const var_abbrv) tK in
             Some (lident, tabbrv, tps, t)
         with
             _ -> None) 
    | Sig_extern_typ(_, s) -> finder fn s
    | _ -> None
  in
    lookup_qname env fn finder (err_e (Wt Typ_unknown))

let genid_ctr = new_counter "LTA.genident"
let hash_ctr = new_counter "LTA.hashtbl.tryfind"
let lookup_typ_abbrev env (ftv:var<kind>) =
  let lta () =
    let genid () = profile (fun () -> genident None) genid_ctr in
    let key = Sugar.text_of_lid ftv.v in 
    let rec extract = function
      | Sig_typ_abbrev (lid, tps, _, t) -> 
          let benv, tps' = 
            List.fold_right (fun tp (benv, tps) -> match tp with 
                               | Tparam_typ (bvd, k) ->
                                   let newname = genid () in
                                   let bvd' = newname,newname in
                                     (bvd, bvd')::benv, Tparam_typ(bvd', k)::tps
                               | Tparam_term (bvd, t) ->
                                   let newname = genid () in
                                   let bvd' = newname,newname in
                                     (bvd, bvd')::benv, Tparam_term(bvd', t)::tps) tps ([],[]) in
          let t' = freshen_typ t benv in
            Some(tps', t')
      | Sig_extern_typ(eref, s) -> 
          if env.expand_externs then extract s else None 
      | _ -> None in
    let sigopt = profile (fun () -> Hashtbl.tryfind env.sigtab key) hash_ctr in
      match sigopt with
          None -> None
        | Some s -> extract s
  in
    profile lta ta_ctr

let rec lookup_record_typ_by_name env recname = 
  let rec finder recname = function
    | Sig_record_typ (lident, tps, _, ({v=Typ_record fn_t_l; sort=_; p=_} as t)) when Sugar.lid_equals lident recname -> 
        let tK = tycon_kind_from_tparams tps t.sort in
        let var_abbrv = withsort lident tK in
        let tabbrv = withsort (Typ_const var_abbrv) tK in
          Some (lident, tabbrv, tps, t)
    | Sig_extern_typ(_, s) -> finder recname s
    | Sig_typ_abbrev(lid, tps, k, t) when Sugar.lid_equals lid recname -> 
        let var_abbrv = withsort recname k in
        let tabbrv = withsort (Typ_const var_abbrv) k in
(*           pr "Looking up type abbreviation for %s\n" (Pretty.str_of_lident lid); *)
        (match lookup_typ_abbrev env var_abbrv with 
           | Some(tps, t) ->
(*                pr "Found abbrev %s\n" (Pretty.strTyp t); *)
               let rec elim_refinement t = match t.v with
                 | Typ_record _ -> Some (lid, tabbrv, tps, t)
                 | Typ_refine(_, t, _) -> elim_refinement t
                 | Typ_const v when List.length tps = 0 -> Some (lookup_record_typ_by_name env v.v)
                 | _ -> None in
                 elim_refinement t 
           | _ -> None)
    | _ -> None
  in
    lookup_qname env recname finder (err_e (Wt Typ_unknown))

let get_record_fields env t = 
  let rec aux (tparams, eparams) t = match t.v with
    | Typ_record fn_t_l -> if not (List.isEmpty tparams) || not (List.isEmpty eparams) then 
        raise (Err "Unexpected parameters in a record type") else fn_t_l
    | Typ_const v -> 
        let (_, _, tformals, rectype) = lookup_record_typ_by_name env v.v in
        let tparams, eparams = List.rev tparams, List.rev eparams in
        let inst_rectype = instantiate_tparams rectype tformals tparams eparams in
          (match inst_rectype.v with 
             | Typ_record fn_t_l -> fn_t_l
             | _ -> raise (Err (spr "Expected a record type; got %s" (Pretty.strTyp inst_rectype))))
    | Typ_refine(_, t, _) -> aux (tparams, eparams) t 
    | Typ_app(t, t') -> aux  (t'::tparams, eparams) t
    | Typ_dep(t, e) -> aux (tparams,  e::eparams) t
    | _ -> raise (Err (spr "Expected a record type; got %s" (Pretty.strTyp t))) in
    aux ([],[]) t
        
let lookup_btvar_ident env (rn:ident): option<kind> = 
  try
    let b = List.find 
      (function
         | Binding_typ (id, k) when id.idText = rn.idText -> true
         | _ -> false) env.gamma in
      match b with
          Binding_typ (_,k) -> Some k
        | _ -> raise Impos
  with _ -> None
    
let lookup_btvar env btv = 
  let lbt () = 
  let rn:ident = bvar_real_name btv in
    match lookup_btvar_ident env rn with
        None -> raise (Not_found_binding(env, Inl (Typ_btvar btv)))
      | Some k -> k in
    profile lbt btvar_ctr

let lookup_typ_const env (ftv:var<kind>) =
  let ltc () =
    let rec tycon_kind_from_sig = function
      | Sig_tycon_kind (lid, tps, k, _) -> 
          tycon_kind_from_tparams tps k

      | Sig_record_typ (lid, tps, _, t)
      | Sig_typ_abbrev (lid, tps, _, t) -> 
          tycon_kind_from_tparams tps t.sort
            
      | Sig_extern_typ(lang, s) -> tycon_kind_from_sig s
          
      | _ -> Kind_unknown in
      match Hashtbl.tryfind env.sigtab (Sugar.text_of_lid ftv.v) with
        | Some s -> tycon_kind_from_sig s 
        | None -> Kind_unknown 
  in
    profile ltc tc_ctr

let lookup_operator env (opname:ident) = 
  let primName = W (Sugar.lid_of_path ["Prims"; ("_dummy_" ^ opname.idText)] dummyRange) in
    lookup_fvar env primName

      
let push_sigelt : env -> sigelt -> env = 
  fun env s -> 
    let lid = strlid_of_sigelt s in
    let _ = Hashtbl.add env.sigtab lid s  in
      {env with e_signature=s::env.e_signature}
    
let push_local_binding env b = 
  let check_typ_unknown e = 
    exp_fold_map 
      (fun () () -> function
           Typ_unknown -> raise (Bad "Found unknown typ")
         | t -> (), t, None)
      (fun () () e -> match e with 
           Exp_fvar v -> 
             (match v.sort.v with 
                | Typ_unknown -> raise (Bad "Found unknown typ")
                | _ -> (), e)
         | _ -> (), e)
      (fun () x -> (),x)
      () () e in
  let uvars, b = match b with 
    | Binding_var(x,t) -> uvars_in_typ t, b
    | Binding_match(e1, e2) -> [], Binding_match(unascribe e1, unascribe e2)
    | _ -> [],b in
  let env = {env with gamma=b::env.gamma; uvars=uvars@env.uvars} in
    env
      
let uvars_in_env env = List.concat (List.map uvars_in_uvar env.uvars)

let push_module : env -> modul -> env = 
  fun env m ->
    let _ = List.iter (fun sigelt -> Hashtbl.add env.sigtab (strlid_of_sigelt sigelt) sigelt) m.signature in
    let externs = List.filter (function 
                                 | Sig_extern_value _ 
                                 | Sig_extern_typ _ -> true
                                 | _ -> false) m.signature in
      {env with modules=(m.name,m)::env.modules; externs=externs@env.externs}

let clear_current_state : env -> env = 
  fun env -> 
    let sigtab = modules_to_sigtable env.modules in
      {empty_env with modules=env.modules; sigtab=sigtab}
    
let set_expected_typ : env -> typ -> env = 
  fun env t -> {env with expected_typ = Some t; recall_expected_typ=Some t}

let expected_typ : env -> option<typ> = 
  fun env -> env.expected_typ
    
let clear_expected_typ : env -> (env * option<typ>) = 
  fun env -> {env with expected_typ=None; recall_expected_typ=env.expected_typ}, env.expected_typ

let recall_expected_typ : env -> option<typ> = 
  fun env -> env.recall_expected_typ

let set_expected_kind : env -> kind -> env = 
  fun env t -> {env with expected_kind = Some t}

let expected_kind : env -> option<kind> = 
  fun env -> env.expected_kind
    
let clear_expected_kind : env -> (env * option<kind>) = 
  fun env -> {env with expected_kind=None}, env.expected_kind

let fold_env : env -> ('a -> binding -> 'a) -> 'a -> 'a = 
  fun env f a -> List.fold_right (fun e a -> f a e) env.gamma a

let find env f =
  try 
    let b = List.find f env.gamma in
      Some b
  with 
    | :? System.Collections.Generic.KeyNotFoundException -> None
      

let set_expect_tyapps env = {env with expect_tyapps=true}  
let clear_expect_tyapps env = {env with expect_tyapps=false}  
let expect_tyapps env = env.expect_tyapps
let set_current_value env v = {env with current_value=Some v}
let current_value env = env.current_value
let clear_current_value env = {env with current_value=None}

let dump (env:env) = printfn "%A" env

let freevars_not_in_env env t = 
  let ftvs, fxvs = freevarsTyp t in
  let tvs_not_in_env = (List.filter
       (fun tv -> match lookup_btvar_ident env (bvar_real_name tv) with 
            None -> true
          | Some _ -> false) ftvs) in
  let xvs_not_in_env = 
   (List.filter 
      (fun xv -> match lookup_bvar_ident env (bvar_real_name xv) with 
           None -> true
         | Some _ -> false) fxvs) in
    tvs_not_in_env, xvs_not_in_env

let rec expand_typ env torig = 
  let expand_typ_once env t = 
    let rec typ_const_and_params out t = match t.v with
        Typ_const fv -> Some (fv, out)
      | Typ_app (t1, t2) -> typ_const_and_params (Inl t2::out) t1
      | Typ_dep (t, e) -> typ_const_and_params (Inr e::out) t
      | _ -> None in
    let t = compress torig in
    let c_parms_opt = typ_const_and_params [] t in
      match c_parms_opt with 
        | None -> t
        | Some (fv, actuals) -> 
            match lookup_typ_abbrev env fv with
                None -> t
              | Some (formals, e_typ) -> 
                  if List.length formals <> List.length actuals then 
                    let msg = spr "Application of type constructor %s with an unexpected number of parameters" (Sugar.text_of_lid fv.v) in
                      raise (Err msg)
                  else
                    let fa = List.zip formals actuals in
                    let out_t = List.fold_left 
                      (fun out_typ -> function
                         | Tparam_typ(bvd, k), Inl t -> 
                             substitute out_typ bvd t
                         | Tparam_term(bvd, t), Inr e -> 
                             substitute_exp out_typ bvd e
                         | _ -> raise (Err "Ill-formed typ abbrev application should be ruled out by kinding")) e_typ fa in
                      out_t in
  let t' = expand_typ_once env torig in if LanguagePrimitives.PhysicalEquality torig t' then torig else expand_typ env t'                

let expand_typ_fix env t = expand_typ env t
      
let expand_typ_rec env t = 
(*   Printf.printf "Expand_typ_rec %s" (Pretty.strTyp t); *)
  let et () t = (), (expand_typ env (t_withpos t dummyRange)).v in
  let ee () e = (), e in
    snd (descend_typ_map et ee () t)

let expand_exp_rec env e = 
  let et () t = (), (expand_typ env (t_withpos t dummyRange)).v in
  let ee () e = (), e in
    snd (descend_exp_map et ee () e)

let get_sigtab_keys env =
  Hashtbl.fold (fun key value out -> key::out) env.sigtab [] 
    
let set_extern_typ env = {env with extern_typ=true}
let extern_typ env = env.extern_typ
let clear_extern_typ env = {env with extern_typ=false}
let clear_expand_externs env = {env with expand_externs=false}
let set_expand_externs env = {env with expand_externs=true}
