#light "off"
// (c) Microsoft Corporation. All rights reserved

module Microsoft.Fine.Tcutil

open Absyn
open Util
open Tcenv

exception Err of string
exception Error of string * Range.range

let print_not_found = function
  | Not_found_binding (env, Inl t) -> 
      Printf.printf "(Compiler bug?) Type name not found: %s\n" (Pretty.strTyp (t_withpos t dummyRange))
  | Not_found_binding (env, Inr e) -> 
      Printf.printf "(Compiler bug?) Variable not found: %s\n" (Pretty.strExp (withpos e dummyRange))
        
let handleable e = match e with
  | Error _ -> true
  | Not_found_binding _ -> print_not_found e; false  (* all bindings should have been resolved already! *)
  | _ -> false 

let handle_err warning ret e = 
  if warning then Printf.printf "\nWARNING: " else Printf.printf "\nERROR: ";
  match e with
    | Not_found_binding (env, Inl t) -> 
        Printf.printf "(Compiler bug?) Type name not found: %s\n" (Pretty.strTyp (t_withpos t dummyRange));
        Printf.printf "query %d\n" !Proof.queryCount;
        ret
    | Not_found_binding (env, Inr e) -> 
        Printf.printf "(Compiler bug?) Variable not found: %s\n" (Pretty.strExp (withpos e dummyRange));
        Printf.printf "query %d\n" !Proof.queryCount;
        ret
    | Error(msg, r) ->
        Printf.printf "(%s): %s\n" (Range.string_of_range r) msg;
        ret
    | NYI s when warning -> 
        Printf.printf "Not yet implemented feature: %s" s; ret
    | e when warning && not(!Util.proof_term_errs) -> (printfn "%A" e; ret)

let terr env t1 t2 p = 
  let msg = 
    Printf.sprintf "Expected an expression of type:\n\n%s\n\nbut got:\n\n%s"
      (Pretty.strTyp (expand_typ_fix env t1)) 
      (Pretty.strTyp (expand_typ_fix env t2)) in
    raise (Error (msg, p))

type affines = list<bvar<typ>>

let rec is_tuple_typ t = match t.v with 
  | Typ_dtuple _ -> true
  | Typ_affine t 
  | Typ_refine (_, t, _) -> is_tuple_typ t
  | _ -> false

let rec is_value e = match e.v with 
  | Exp_bot
  | Exp_constant _
  | Exp_bvar _
  | Exp_fvar _ 
  | Exp_abs _ 
  | Exp_tabs _ -> true
  | Exp_constr_app (_, _, _, el) -> List.for_all is_value el
  | Exp_ascribed(e, _, _) 
  | Exp_proj (e, _) -> is_value e
  | Exp_recd (_, _, _, fn_e_l) -> List.for_all (fun (_,e) -> is_value e) fn_e_l
  | _ -> false

let rec is_non_variable_value env e = match e.v with 
  | Exp_fvar fv -> Tcenv.is_datacon env fv
  | Exp_constant _
  | Exp_abs _ 
  | Exp_tabs _ -> true
  | Exp_ascribed(e,_,_) -> is_non_variable_value env e
  | Exp_constr_app (_, _, _, el) -> List.for_all (is_non_variable_value env) el
  | _ -> false

(* let kind_equiv k k' = (unbox k)=(unbox k') *)

let kind_dominated k k' = match k, k' with 
  | Kind_star, Kind_star 
  | Kind_affine, Kind_affine
  | Kind_star, Kind_affine -> true
  | _ -> false

let kind_lub k1 k2 = match unbox k1, unbox k2 with
  | Kind_star, Kind_star -> Kind_star
  | Kind_star, Kind_affine 
  | Kind_affine, Kind_star
  | Kind_affine, Kind_affine -> Kind_affine
  | _ -> raise (Err "Kind_lub is undefined for the specified kinds")

let rec kind_abstractable = function
  | Kind_star
  | Kind_affine -> true
  | Kind_dcon(_, k) -> kind_abstractable k
  | _ -> false
      
let disjoint_affine_usages l = 
  let rec nodups = function
    | [] 
    | [_] -> true
    | hd::tl -> 
        if not (List.exists (bvar_eq hd) tl) then
          nodups tl
        else 
          let msg = spr "Affine variable %s used more than once" (Pretty.strBvar hd) in
            raise (Err msg) in
    nodups l

let subtract l1 l2 = 
  let rec sub out l a = match l with 
    | [] -> out
    | hd::tl when bvar_eq hd a -> sub out tl a
    | hd::tl -> sub (hd::out) tl a in
    List.fold_left (sub []) l1 l2

let union l1 l2 = 
  let add l a = if List.exists (bvar_eq a) l then l else a::l in
    List.fold_left add l1 l2

let union_ev (ev1:evidence) ev2 = ev1@ev2

let lookup_field_in_record_typ env fn rt = 
  let fn_t_l = Tcenv.get_record_fields env rt in
    (match List.tryFind (fun (fn', t') -> Sugar.lid_equals fn' fn) fn_t_l with
         Some (fn', t') -> t'
       | _ -> let msg = spr "Field name %s is not a part of the type %s" 
           (Sugar.text_of_lid fn) (Pretty.strTyp rt) in
           raise (Error(msg, Sugar.range_of_lid fn)))

let is_proof_typ_constr (t:typ) = match t.v with 
  | Typ_const v ->Sugar.lid_equals v.v Const.pf_lid 
  | _ -> false
let is_proof_typ (t:typ) = match t.v with 
  | Typ_app (pf_typ, _) -> is_proof_typ_constr pf_typ
  | _ -> false
let maybe_eq_proof_typ (t:typ) = match t.v with 
  | Typ_app (pf_typ, ({v=Typ_dep({v=Typ_dep(t1, _); sort=_; p=_}, _); sort=_; p=_})) 
      when is_proof_typ_constr pf_typ -> true
  | _ -> false
let is_impl_typ_constr t = match t.v with
  | Typ_const v -> Sugar.lid_equals v.v Const.implies_lid
  | _ -> false
let impl_lhs t = match t.v with 
  | Typ_app(imp, lhs) when is_impl_typ_constr imp -> Some lhs
  | _ -> None
let destruct_implication (t:typ) = match t.v with 
  | Typ_app(pf_typ, {v=Typ_app(impl_l, rhs); sort=_; p=_}) when is_proof_typ_constr pf_typ -> 
      (match impl_lhs impl_l with 
           Some lhs -> Some (lhs, rhs)
         | _ -> None)
  | _ -> None

let is_lid t l = match t.v with 
  | Typ_app({v=Typ_const v; sort=_;p=_}, lhs) when Sugar.lid_equals v.v l ->  Some lhs
  | _ -> None
let destruct_pf_binop lid (t:typ) = match t.v with 
  | Typ_app(pf_typ, {v=Typ_app(constr, rhs); sort=_; p=_}) when is_proof_typ_constr pf_typ -> 
      (match is_lid constr lid with 
           Some lhs -> Some (lhs, rhs)
         | _ -> None)
  | _ -> None
let is_pf_binop lid (t:typ) = match destruct_pf_binop lid t with 
  | None -> false
  | _ -> true

let unwrap_pf (t:typ) = match t.v with 
    Typ_app(pf_typ, t) when is_proof_typ_constr pf_typ -> t
  | _ -> raise Impos

let is_implication t = is_pf_binop Const.implies_lid t

let typing_const (env:Tcenv.env) = function
  | Sugar.Const_unit -> Const.unit_typ
  | Sugar.Const_bool _ -> Const.bool_typ
  | Sugar.Const_int32 _ -> Const.int_typ
  | Sugar.Const_string _ -> Const.string_typ
  | _ -> raise (NYI "Unsupported constant")

open Tcenv      

let push_tparams env tps = 
  List.fold_left (fun env tp -> 
                    let binding = match tp with
                      | Tparam_typ ((pp,rn),k) -> Binding_typ (rn,k) 
                      | Tparam_term ((pp,rn), t) -> Binding_var (rn,t) in
                      push_local_binding env binding) env tps 

let get_evidence e = 
  let rec aux out e = match e.v with 
    | Exp_ascribed (e, t, ev) -> aux (ev@out) e
    | _ -> out
  in aux [] e
       
let get_evidencel el = List.flatten (List.map get_evidence el)

let subtract_ev_exact (ev:evidence) ev' = 
  let sub_one ev ev' =
    let f (e1,e2) (e1',e2') = (LanguagePrimitives.PhysicalEquality e1' e1 &&
                                 LanguagePrimitives.PhysicalEquality e2' e2) in
      List.filter (fun e -> not (f ev' e)) ev in
    List.fold_left sub_one ev ev'
      
let subtract_ev (ev:evidence) el = 
  let warn e = Printf.printf "Warning! Got a match binding with an unexpected LHS: %s" (Pretty.strExp e) in
  let aux ev e = match (unascribe e).v with 
      Exp_bvar bv -> 
        List.filter (fun (bv', _) -> match (unascribe bv').v with
                       | Exp_bvar bv' -> not(bvar_eq bv bv')
                       | _ -> warn bv'; true) ev
    | _ -> warn e; ev in
    List.fold_left aux ev el

let collect_evidence e = 
(*   Printf.printf "Collecting evidence on expr %s\n" (Pretty.strExp e); *)
  let pat_vars_as_exp_l = function
      Pat_variant (_, _, _, bvs) -> List.map (fun bv -> (withsort (Exp_bvar bv) bv.sort)) bvs in
  let add_ev e = function
    | [] -> e
    | ev -> ascribe e e.sort ev in
  let is_match_assumption (e1, e2) (e, cn) = 
(*     Printf.printf "Checking if match assumption (%s,%s) = (%s,%s)?\n" (Pretty.strExp e1) (Pretty.strExp e2) (Pretty.strExp e) (Pretty.str_of_lident cn.v); *)
    match (unascribe e2).v with 
      | Exp_constr_app(cn', _, _, _) when Sugar.lid_equals cn.v cn'.v -> equalsExp (unascribe e) (unascribe e1)
      | _ -> false in
  let is_boolean_assumption (e1, e2) e = 
    match (unascribe e2).v with 
      | Exp_constant (Sugar.Const_bool _) -> equalsExp (unascribe e) (unascribe e1)
      | _ -> false in    
  let ev =  match e.v with
    | Exp_ascribed _ 
    | Exp_gvar _      
    | Exp_bvar _
    | Exp_fvar _ 
    | Exp_bot                                 
    | Exp_constant _ -> []
    | Exp_primop(_, el)
    | Exp_extern_call(_, _, _, _, el) 
    | Exp_constr_app (_, _, _, el) -> get_evidencel el
    | Exp_abs(bvd, t, e) -> subtract_ev (get_evidence e) [AbsynUtils.bvd_to_exp bvd t]
    | Exp_tapp(e, _)   
    | Exp_proj(e, _) 
    | Exp_tabs(_, _, e) -> get_evidence e 
    | Exp_app(e1, e2) -> union_ev (get_evidence e1) (get_evidence e2) 
    | Exp_match (e, eqns, def) -> 
        let result = union_ev 
          (union_ev 
             (List.flatten (List.map (fun (Pat_variant(cname, _, _, _) as pat, br) -> 
                                        let evbranch = (get_evidence br) in
(*                                           Printf.printf "Got evidence from branch %s\n" (Pretty.strEv evbranch); *)
                                          let ev0 = subtract_ev evbranch (pat_vars_as_exp_l pat) in
(*                                             Printf.printf "After subtracting pat vars %s\n" (Pretty.strEv ev0); *)
                                            let result = List.filter (fun ev -> not (is_match_assumption ev (e, AbsynUtils.W cname))) ev0 in
(*                                               Printf.printf "After subtracting match assumption vars %s\n" (Pretty.strEv result); *)
                                              result) eqns))
             (get_evidence def)) 
          (get_evidence e) in
(*           Printf.printf "Collected evidence from match %s\n" (Pretty.strEv result); *)
          result
    | Exp_cond(e1, e2, e3) -> 
        union_ev (List.filter (fun ev -> not(is_boolean_assumption ev e1)) (union_ev (get_evidence e1) (get_evidence e2))) (get_evidence e1)
    | Exp_recd (_, _, phantoms, fnl) -> List.flatten (List.map (fun (_, e) -> get_evidence e) fnl)
    | Exp_let(false, [bvd, t, e1], e2) -> 
        union_ev 
          (subtract_ev (get_evidence e2) [AbsynUtils.bvd_to_exp bvd t])
          (get_evidence e1)
    | Exp_let(true, bvd_e_l, e2) -> 
        union_ev (List.flatten (List.map (fun (bdv, t, e) -> get_evidence e) bvd_e_l)) (get_evidence e2)
  in
    add_ev e ev
  

let bot t = withsort (Exp_ascribed (withsort Exp_bot t, t, [])) t 

let is_closure_typ env t = match (expand_typ_fix env t).v with 
    Typ_fun _ 
  | Typ_univ _ -> true
  | _ -> false

let is_boxed = function
    Kind_boxed _ -> true
  | _ -> false

let is_xvar_free (x:bvdef) t = 
  let _, xvs = freevarsTyp t in
(*     Printf.printf "Free variables of type (%s)\n [%s]\n" (Pretty.strTyp t) (String.concat "; " (List.map Pretty.strBvar xvs)); *)
    match List.tryFind (fun (bv:bvar<typ>) -> (bvar_real_name bv).idText = (snd x).idText) xvs with 
      | None -> false
      | _ -> true

let reflexivity_var env =
  let axT = Tcenv.lookup_lid env Const.reflexivity_lid in
  let axVar = withsort Const.reflexivity_lid axT in
    axVar

let eqT_of_typ env t = 
  let eqK = Tcenv.lookup_typ_const env (AbsynUtils.Wt Const.eq_lid) in
    (withsort (Typ_app(withsort (Typ_const (withsort Const.eq_lid eqK)) eqK, t))
       (Kind_dcon(t, Kind_dcon(t, Kind_star))))



