#light "off"
// (c) Microsoft Corporation. All rights reserved
module Microsoft.Fine.Translation

(*
  NS: I put in several hacks to get the translation to go through to
  IL generation.
  
  1. In transTDApp: when translating a (T<e>) if the index expression
  is not a value, then I just drop it and translate the type as the
  translation of T. What we really should do is to establish an
  invariant that ensures that every source language value is
  translated to a target language value. Currently, this invariant
  does not hold true for record projections. (x.f in the source is
  translated to let y = x in y.f in the target.)

  2. Consider the following source program:

  module A
  type rec1 = {fld1:int; fld2:int}
  end
  module B 
  open A
  val foo: rec1 -> int
  let foo x = 0
  end

  The translation of A produces a class in

  A.dll
  class A_fld1-A_fld2 { ... }
  
  module B after type expansion includes
  val foo : {A.fld1:int; A.fld2:int} -> int

  When translating this type, we have to translate the record type
  {A.fld1:int; A.fld2:int}. But, we have to notice that this type is
  really an external record type and translate it as a 

  TType_name("A.A_fld1-A_fld2") 

  rather than producing a fresh class decl for it in B.dll.

  SO, I added an ugly hack to transRecordTyp that uses a bit of
  mutable state (!) to keep track of which module is currently being
  translated and then uses a nasty hack based on the structure of the
  field names to figure out when to generate class decls for a record
  type and when to generate an external type name. 

  See curmodname:option<lident> ref and transRecordTyp

  3. Another related problem with the translation of record types is
  that after type expansion, every syntactic occurence of a record
  type is translated into a fresh class declaration. Ilgen has a hack
  to detect duplicates and suppress class generation appropriately.
  
  The real solution should address 2 and 3 together and generate
  record declarations or references to external record declarations as
  needed.

  Another problem with our current strategy for record translation is
  that it does not work for polymorphic record types---we will end up
  with name collisions and unnecessary code duplication.

  3. In transLetBinding, we have to produce static field declarations
  for all top-level bindings. 

  module A
  let rec foo x = foo x
  end
  module B 
  let bar x = foo x
  end

  Now in module A, the name foo is a top-level let binding, but in the
  body of foo the occurence of foo is a bound variable (Exp_bvar). In
  module B, the occurrence of foo is a free variable
  (Exp_fvar). Unfortunately, the source type checker binding rules
  cause the (Exp_bvar) in module A to be indexed using the
  bvar_real_name. Whereas, the Exp_fvar in module B uses the ppname.

  So, I included a hack in transLetBinding to generate static field
  declarations for both the realname and the pp_name. 

  The real solution is to sort out the source type checker and make
  sure that a consistent naming scheme is used for top-level let
  bindings.
*)

open Absyn
open Target
open Util
open Profiling
open Tcenv
open Tcutil

exception TransErr of string


let pr = Printf.printf
let spr = Printf.sprintf

let transType_ctr = new_counter "Trans Type"
let lifting_ctr = new_counter "lifting"
let alifting_ctr = new_counter "alifting"
let encoding1_ctr = new_counter "encoding1"
let encoding2_ctr = new_counter "encoding2"
let encoding3_ctr = new_counter "encoding3"
let encoding4_ctr = new_counter "encoding4"
let transExp_ctr = new_counter "Trans exp"
let transSig_ctr = new_counter "Trans sig"

let text_of_id = Microsoft.Fine.Sugar.text_of_id
let text_of_lid = Microsoft.Fine.Sugar.text_of_lid
let path_of_lid = Microsoft.Fine.Sugar.path_of_lid

let csp = new System.Security.Cryptography.MD5CryptoServiceProvider() 

let mk_affine t = match t with 
    TType_affine _ -> t
  | _ -> TType_affine t 
      
let typeAsEncodedString t = 
  let encodeString (s:string) : string =
    profile (fun () -> let bs = System.Text.Encoding.UTF8.GetBytes(s) in
             let bs = csp.ComputeHash(bs) in
               System.Convert.ToBase64String bs) encoding1_ctr in
  let rec encodeType t =
    match t with
      | TType_index u -> "#" ^ (u.ToString())
      | TType_var tvar -> fst tvar
      | TType_name name -> 
          let (lid, _) = name in
            String.concat "." lid
      | TType_fun (bvdopt, t1, t2) ->
          let s1 = encodeType t1 in
          let s2 = encodeType t2 in
            (match bvdopt with
               | None -> encodeString(s1 ^ s2)
               | Some bvd -> encodeString((snd bvd) ^ s1 ^ s2))
      | TType_concrete tconcrete -> encodeString(encodeTypeConcrete tconcrete)
      | TType_dep (tfun, v) -> 
          let sfun = encodeType tfun in
          let sv = encodeVal v in
            encodeString(sfun ^ sv)
      | TType_affine ty -> encodeString("!" ^ (encodeType ty)) 
  and encodeVal v =
    match v with
      | TVal_var vvar -> fst vvar
      | TVal_obj (tconcrete, _, vs) ->
          let strVs = String.concat "," (List.map encodeVal vs) in
            encodeString((encodeTypeConcrete tconcrete)^strVs)
      | TVal_constant cons -> cons.ToString()
      | TVal_ldfld (v, fname) -> encodeString((encodeVal v) ^ fname)
  and encodeTypeConcrete(tconcrete: tTypeConcrete) = 
    profile (fun () -> let (classname, targs, vargs) = tconcrete in
             let name = String.concat "." classname in
               match targs, vargs with
                 | [], [] -> name 
                 | [], _ -> let strVargs = String.concat "," (List.map encodeVal vargs) in
                     name ^ strVargs
                 | _ , [] -> let strTargs = String.concat "," (List.map encodeType targs) in
                     name ^ strTargs
                 | _, _ ->
                     let strTargs = String.concat "," (List.map encodeType targs) in
                     let strVargs = String.concat "," (List.map encodeVal vargs) in
                       (String.concat "." classname)^strTargs^strVargs) encoding2_ctr in
    encodeType t

let strExp = Microsoft.Fine.PrettyTarget.strExp
  
(* translate a bound variable definition to a target identifier *)
let transBvDef (bvd: bvdef) : tBvDef =
  (text_of_id (fst bvd), text_of_id (snd bvd))
let transBvDefId (bvd: bvdef) : tIdent = snd (transBvDef bvd)

let curmodname : (option<lident> ref) = ref None
let evidenceStack: (tEvidences list) ref = ref []

let curtcenv : (option<Tcenv.env> ref) = ref None

let newTvar = Target.newTvar
let newVvar = Target.newVvar

(* remove duplicated elements from a list *)
let deDup (ls : bvar<kind> list) : bvar<kind> list =
  let rec d lls (res: bvar<kind> list) =
    match lls with
      | [] -> res
      | el::rest ->
          let id1 = bvar_real_name el in
          let isEqBvar (x:bvar<kind>) =
            let id2 = bvar_real_name x in
              id1.idText = id2.idText in
            if (List.exists isEqBvar res) then d rest res
            else d rest (res@[el])
  in d ls []

let dedupVvars l = 
  let test (vv1,_) (vv2, _)= vv1=vv2 in
  let rec aux out = function
    | [] -> List.rev out
    | hd::tl -> 
        match List.tryFind (fun e -> test hd e) out with
            Some _ -> aux out tl
          | None -> aux (hd::out) tl in
    aux [] l 
      

(* environment for translation, including accumulated class declarations*)
type transEnv = (string, tClassDecl) Hashtbl.t

let createTransEnv(): transEnv = Hashtbl.create 100 

(* look up a class declaration in transenv, return None if not found and Some cdecl otherwise *)
let rec lookupTransEnv env cname : tClassDecl = 
  Hashtbl.find env cname

(* add a class declaration to env *)
let addToTransEnv env cdecl : transEnv =
  if (Hashtbl.mem env (cdecl.namestring)) then env
  else (Hashtbl.add env cdecl.namestring cdecl; env)

(* filter evidences to find those for free variables *)
let filterEvidences evidences fvs =
  DebugLog.Log(Printf.sprintf "\n: before filtering: %s, \n  fvs = %s " 
                 (String.concat ", " (List.map (fun (var, v) -> Printf.sprintf "\n %s = %s" 
                                                  (PrettyTarget.strVvar var) 
                                                  (PrettyTarget.strVal v)) 
                                        evidences))
                 (String.concat ", " (List.map PrettyTarget.strVvar fvs)));
  let rec filter evs =
    match evs with
      | [] -> []
      | (var, v)::rest -> 
          if (List.exists (fun vvar -> fst vvar = fst var) fvs)
          then (var,v) :: (filter rest)
          else filter rest in
  let result = filter evidences in
    DebugLog.Log(Printf.sprintf "\n: after filtering: %s " 
                   (String.concat ", " (List.map (fun (var, v) -> Printf.sprintf "\n %s = %s" 
                                                    (PrettyTarget.strVvar var) 
                                                    (PrettyTarget.strVal v))
                                          evidences)));
    result


let unbox = function
    Kind_boxed k -> k
  | k -> k

let mangled_op_to_pop (op:ident) = match op.idText with 
    "op_AmpAmp" -> AND
  | "op_BarBar" -> OR
  | "op_Addition" -> ADD
  | "op_Multiply" -> MUL
  | "op_Subtraction" -> SUB
  | "op_Divide" -> DIV
  | "op_Negation" -> NOT

(* translation of kinds *)
let rec transKind (transenv: transEnv) (k: Absyn.kind) : tKind * transEnv =
  match k with
    | Kind_boxed knd -> transKind transenv knd
    | Kind_star -> (TKind_star, transenv)
    | Kind_affine -> (TKind_affine, transenv)
    | Kind_dcon (ty, knd) ->
        let (newTy, transenv) = transType transenv ty in
        let (newKind, transenv) = transKind transenv knd in
          (TKind_arrow(newTy, newKind), transenv)
    | _ -> 
        let errmsg = Printf.sprintf "transKind: unexpected kind %s" 
          (Pretty.strKind k) in
          raise (TransErr errmsg)

(* translation of bound type variable *)
and transBTvar transenv (tvar: bvar<kind>) : tVar<tKind> * transEnv = 
  let id = bvar_real_name tvar in 
  let (newKind, transenv) = transKind transenv (tvar.sort) in
    ((text_of_id id, newKind), transenv)

(* translation of external names *)
and transTvar transenv (tv: var<kind>) : tName<tKind> * transEnv = 
  let (newKind, transenv) = transKind transenv (tv.sort) in
    ((path_of_lid tv.v, newKind), transenv)

(* translation of types, return a list of new class declarations *) 
and transTypeWithEnv (env:list<ident>) (transenv: transEnv) (t: typ): tType * transEnv =  
  //let _ = Printf.printf "\n === %s" (Pretty.strTyp t) in
  let rec transt trenv t : tType * transEnv =
    (match (compress t).v with
       | Typ_btvar bv -> 
           let id = (bvar_real_name bv).idText in
             (match Util.try_find_position (fun (a:ident) -> a.idText=id) env with 
                  None ->
                    let (newvar, trenv) = transBTvar trenv bv in
                      (TType_var newvar, trenv)
                | Some i -> TType_index i, trenv)
       | Typ_const var ->
           let (newvar, transenv) = transTvar trenv var in
             (TType_name newvar, trenv)
       | Typ_record fts -> transRecordType env trenv fts t
       | Typ_fun (bvd_opt, t1, t2) ->
           let (newt1, trenv) = transt trenv t1 in
           let (newt2, trenv) = transt trenv t2 in
           let tfun: tType =
             match bvd_opt with
               | None -> TType_fun (None, newt1, newt2)
               | Some bvd -> 
                   TType_fun(Some (transBvDef bvd), newt1, newt2) in
           let tconcrete: tTypeConcrete = 
             match (unbox t1.sort, unbox t2.sort) with
               | (Kind_star, Kind_star) ->
                   (tClassDeparrowSS.name, [newt1; tfun], [])
               | (Kind_star, Kind_affine) ->
                   (tClassDeparrowSA.name, [newt1; tfun], [])
               | (Kind_affine, Kind_affine) ->
                   (tClassDeparrowAA.name, [newt1; newt2], [])
               | (Kind_affine, Kind_star) ->
                   (tClassDeparrowAS.name, [newt1; newt2], [])
               | _ -> let errmsg = Printf.sprintf "transType: unexpected kind  %s -> %s"
                   (Pretty.strKind (t1.sort))
                   (Pretty.strKind (t2.sort)) in
                   raise (TransErr errmsg) in
           let dcilTy = TType_concrete tconcrete in
             (dcilTy, trenv)
       | Typ_univ (bvd, k, body) -> transTPoly env trenv bvd k body t
       | Typ_dtuple (bvd_opt_ts) -> transDTuple env trenv bvd_opt_ts t
       | Typ_app _ 
       | Typ_dep _ -> transTDApp env trenv t
       | Typ_affine (t') -> 
           let (tt, trenv) = transt trenv t' in
           let dcilTy = TType_affine tt in
             (dcilTy, trenv)  
       | Typ_refine (bvd, t, f) -> (* Currently ignore the formula *)
           transt trenv t
       | Typ_lam (bvd, t1, t2) ->
           let (newt1, trenv) = transt trenv t1 in
           let (newt2, trenv) = transt trenv t2 in
             (TType_fun(Some (transBvDef bvd), newt1, newt2), trenv)
       | _ -> 
           let errmsg = Printf.sprintf "transType: unexpected type %s " 
             (Pretty.strTyp t) in
             raise (TransErr errmsg))
  in transt transenv t 

and transType = transTypeWithEnv []

and transFieldTypes env trenv types =
  let rec transFs trenv srctypes ftypes =
    match srctypes with
      | [] -> (trenv, ftypes)
      | fty :: rest ->
          let (fty', trenv) = transTypeWithEnv env trenv fty in
            transFs trenv rest (ftypes@[fty']) in
    transFs trenv types [] 

(* translation of record type *)
and transRecordType env transenv fts origt: tType * transEnv =
  let is_external_record_type nm = 
    let modname = List.hd nm in
      match !curmodname with
        | Some x when not (Sugar.lid_equals x [modname]) -> Some (modname.idText)
        | _ -> None in
  let (fieldnames, types) = List.split fts in
  let concatenatedFieldNames =
    List.map (fun fname -> String.concat "_" (path_of_lid fname)) fieldnames in
  let newClassName = (String.concat "-" concatenatedFieldNames) in
    match is_external_record_type (List.hd fieldnames) with
        Some extMod -> 
          let dcilTy = TType_name(([extMod; newClassName], TKind_star)) in
            dcilTy, transenv
      | None -> 
          let (transenv, newtypes) = transFieldTypes env transenv types in
            (*             // create a new class that corresponds to the record type. *)
            (*             // The class name is the concatenation of field names. *)
            (*             // The fields of the class are the fields of the records *)
          let newFields = List.combine concatenatedFieldNames newtypes in
          let recordClassDecl : tClassDecl =
            { extqual=None;
              visibility = TVis_public;
              name = [newClassName];
              namestring = newClassName;
              ptvars = [];
              tvars = [];
              pvvars = [];
              vvars = [];
              evidences = [];
              extends = None;
              fields = newFields;
              staticFields = [];
              methods = [];
              kind = Some TKind_star;
            } in
          let newRecordType = TType_concrete([newClassName], [], []) in
            (newRecordType, addToTransEnv transenv recordClassDecl) 

(* ||forall a::k.t|| -> All<\a::||k||.||t||> *)
and transTPoly env transenv (bvd: bvdef) (knd: kind) (tbody: typ) origt =
  let (newbody,transenv) = transTypeWithEnv env transenv tbody in
  let tClassAll = 
    match (unbox knd, unbox (tbody.sort)) with
      | (Kind_star, Kind_star) -> tClassAllSS
      | (Kind_star, Kind_affine) -> tClassAllSA
      | (Kind_affine, Kind_star) -> tClassAllAS
      | (Kind_affine, Kind_affine) -> tClassAllAA in
  let (newKind, transenv) = transKind transenv knd in
  let tfun = TType_tfun(transBvDef bvd, newKind, newbody) in
    (TType_concrete(tClassAll.name, [tfun], []), transenv) 

(* translation of dependent tuple types *)
(* deal with (x:t1, t2) only at the moment *)
and transDTuple env transenv (bvdts) origt: tType * transEnv = 
  match bvdts with
    | (bvdef_opt, t1) :: [(None, t2)] ->
        let (newt1, transenv) = transTypeWithEnv env transenv t1 in
        let (newt2, transenv) = transTypeWithEnv env transenv t2 in
        let tfun = 
          match bvdef_opt with
            | Some bvdef -> TType_fun(Some (transBvDef bvdef), newt1, newt2)
            | None -> TType_fun(None, newt1, newt2) in
        let (newKind1, transenv) = transKind transenv t1.sort in
        let (newKind2, transenv) = transKind transenv t2.sort in
        let tconcrete =
          match (newKind1, newKind2) with
            | (TKind_star, TKind_star) ->
                (tClassDepTuple2SS.name, [newt1; tfun], [])
            | (TKind_star, TKind_affine) ->
                (tClassDepTuple2SA.name, [newt1; tfun], [])
            | (TKind_affine, TKind_affine) ->
                (tClassTuple2AA.name, [newt1; newt2], [])
            | (TKind_affine, TKind_star) ->
                (tClassTuple2AS.name, [newt1; newt2], [])
            | _ -> let errmsg = Printf.sprintf "transDTuple: unexpected kinds %s, %s"
                (Pretty.strKind t1.sort) (Pretty.strKind t2.sort) in
                raise (TransErr errmsg) in
        let dcilTy = TType_concrete tconcrete in
          (dcilTy, transenv)
    | _ -> let errmsg = spr "transDTuple: unexpected tuple type %s" (bvdts.ToString()) in
        raise (TransErr errmsg)

(* translation of type and dependent application *)
(* Find the type constructor and all the type and value arguments *)
and transTDApp env transenv (t: typ) : tType * transEnv = 
  let rec transApp trenv ty targs vargs =
    match (compress ty).v with
      | Typ_const var -> 
          let tyconName = path_of_lid (var.v) in
          let tconcrete : tTypeConcrete = (tyconName, targs, vargs) in 
          let dcilTy = TType_concrete tconcrete in
            (dcilTy, trenv)
      | Typ_app (tfun, targ) ->
          let (new_targ, trenv) = transTypeWithEnv env trenv targ in
            transApp trenv tfun (new_targ :: targs) vargs
      | Typ_dep (tdfun, earg) ->
          let (new_earg, trenv) = transExp trenv earg in
            (match new_earg with
               | TExp_val newv ->
                   transApp trenv tdfun targs (newv::vargs)
               | _ -> (* skip non-value arguments! *)
                   transApp trenv tdfun targs vargs
               | _ -> 
                   let errmsg = spr 
                     "transTDApp: %s not translated to a value \n (%s)"
                     (Pretty.strExp earg) (PrettyTarget.strExp new_earg) in
                     raise (TransErr errmsg))
      | Typ_btvar bvar ->
          (* 'a => 'a targs vargs *)
          let (newvar, trenv) = transBTvar trenv bvar in
          let applyTargs ty targs =
            (match targs with
               | [] -> ty
               | _ ->
                   let errmsg = spr "transTDApp: unexpected higher-orderness in %s" 
                     (Pretty.strTyp t) in 
                     raise (TransErr errmsg)) in
          let rec applyArgs ty args =
            (match args with
               | [] -> ty
               | arg::rest -> applyArgs (TType_dep(ty, arg)) rest) in
          let dcilTy = applyArgs (applyTargs (TType_var newvar) targs) vargs in
            (dcilTy, trenv)
      | Typ_lam (bvd, t1, t2) ->
          let (newt1, trenv) = transType trenv t1 in
          let (newt2, trenv) = transType trenv t2 in
            (TType_fun(Some (transBvDef bvd), newt1, newt2), trenv)
      | _ -> 
          let errmsg = spr "transTDApp: unexpected type %s in %s" 
            (Pretty.strTyp ty) (Pretty.strTyp t)
          in raise (TransErr errmsg) in
    transApp transenv t [] []

(* translation of bound value variable *)
(* Does not translate type *)
and transBVvar transenv (bvar: bvar<typ>) : tVar<tType> * transEnv =
  let id = bvar_real_name bvar in
  let (newt, transenv) = transType transenv bvar.sort in
  let newvar = (text_of_id id, newt) in
    (newvar, transenv)

(* translation of a free value variable *)
and transVvar transenv (var: var<typ>): tName<tType> * transEnv =
  let (newt, transenv) = transType transenv var.sort in
  let newvar: tName<tType> = (path_of_lid (var.v), newt) in
    (newvar, transenv)

(* add and translate evidences (e1 = e2) to class declarations *)
and addEvidences env evidences =
  (* translate (exp * exp) to (tVar<tType> * TExp) *)
  let rec transEvidences ees ves =
    match ees with
      | [] -> ves
      | (e1, e2) :: rest -> 
          DebugLog.Log(Printf.sprintf "\n adding evidence %s = %s" 
                         (Pretty.strExp e1) (Pretty.strExp e2));
          (match e1.v with
             | Exp_bvar var
             | Exp_ascribed ({v=Exp_bvar var; sort = _; p = _}, _, _) -> 
                 (* translate var and e2, ignoring the new env and class *)
                 (* declarations because var and e2 should be translated already *)
                 let (newvar, _) = transBVvar env var in
                 let (newExp, _) = transExp env e2 in
                   (match newExp with
                      | TExp_val vl -> transEvidences rest ves @ [(newvar, vl)]
                      | _ -> transEvidences rest ves)
             | _ -> transEvidences rest ves) in
  let newevs = transEvidences evidences [] in
    evidenceStack := newevs :: (!evidenceStack)

and removeEvidences() =
  match !evidenceStack with
    | _::rest -> evidenceStack := rest;
    | _ -> raise (TransErr " no evidences to remove")

and currentEvidences () = 
  match !evidenceStack with
    | cur :: rest -> cur
    | _ -> []

(* translation of expressions returns the target expression and *)
(* a list of new class declarations *)
and transExp transenv (e: exp): tExp * transEnv = 
  transE [] transenv e

and transE freevs transenv (e: exp) : tExp * transEnv =
  (*let _ = pr "\n translating exp %s" (Pretty.strExp e) in*)
  match e.v with
    | Exp_bvar bv -> 
        let (newvar, transenv) = transBVvar transenv bv in
          (TExp_val (TVal_var newvar), transenv)
    | Exp_fvar var ->
        let (newvar, transenv) = transVvar transenv var in
          (TExp_name newvar, transenv)
    | Exp_constant const ->
        (TExp_val (TVal_constant const), transenv)
    | Exp_constr_app (var, ts, phantom_es, es) ->
        if (Sugar.lid_equals var.v Const.tuple_UU_lid or
              Sugar.lid_equals var.v Const.tuple_UA_lid or
              Sugar.lid_equals var.v Const.tuple_AU_lid or
              Sugar.lid_equals var.v Const.tuple_AA_lid) 
        then 
          (match (ts, es) with
             | ([t1; t2], [e1; e2]) ->
                 transTuple2Constr freevs transenv var None t1 t2 e1 e2
             | _ -> let msg = Printf.sprintf "unexpected tuple constr %s" (Pretty.strExp e) in
                 raise (TransErr msg))
        else transConstrApp freevs transenv var ts phantom_es es 
    | Exp_abs _ -> transAbs freevs transenv e
    | Exp_tabs (bvd, k, body) ->
        transTAbs freevs transenv e 
    | Exp_primop(op, args) -> 
        let targs_rev, transenv = List.fold_left 
          (fun (targs, env) e ->
             let texp, env = transE freevs env e in
             let texp = match texp with 
                 TExp_val v -> v
               | _ -> 
                   let msg = spr "Unexpected non-value in operator application: %s" (Pretty.strExp e) in
                     raise (TransErr msg) in
               texp::targs, env) ([], transenv) args in
        let pop = mangled_op_to_pop op in
          TExp_primop(pop, List.rev targs_rev), transenv
    | Exp_app (efun, earg) -> transApp freevs transenv efun earg
    | Exp_tapp (efun, targ) -> transTApp freevs transenv efun targ
    | Exp_match (em, pes, edefault) ->
        transMatch freevs transenv em pes edefault e.sort
    | Exp_cond (eb, et, ef) ->
        let (neweb, transenv) = transE freevs transenv eb in
        let (newet, transenv) = transE freevs transenv et in
        let (newef, transenv) = transE freevs transenv ef in
        let newe = TExp_cond(neweb, newet, newef) in
          (newe, transenv)
    | Exp_recd(lidopt, targs, phantoms, fes) -> transRecordExp freevs transenv fes e.sort
    | Exp_proj (erecord, fieldname) ->
        let (newerecord, transenv) = transE freevs transenv erecord in
        let newFieldName = 
          (* Juan: change dcil_proj_one and dcil_proj_two in const.fs, *)
          (* if the convention below changes!! *)
          (match Sugar.lid_to_eref fieldname with
               Some (dll, rest) -> let cn, fn = Util.pfx rest in fn.idText
             | None -> String.concat "_" (path_of_lid fieldname)) in
          (match newerecord with
             | TExp_val v -> (TExp_val (TVal_ldfld(v, newFieldName)), transenv)
             | _ -> let (recordTy, transenv) = transType transenv erecord.sort in
               let newvar = newVvar recordTy in
                 (TExp_let(newvar, newerecord, 
                           TExp_ldfld(TVal_var newvar, newFieldName)), transenv))
    | Exp_ascribed (e', t, evidences) -> 
        (match (e'.v, t.v) with
           | (Exp_constr_app(dconvar, _, _, es), Typ_dtuple ttups) ->
               let _ = addEvidences transenv evidences in
               let (newe, transenv) = transDtupleConstr freevs transenv dconvar es ttups e' in
                 removeEvidences(); (newe, transenv)
           | _ -> 
               let _ = addEvidences transenv evidences in
               let (newe, transenv) = transE freevs transenv e' in
                 removeEvidences(); (newe, transenv))
    | Exp_let (is_rec, bvdtes, e2) -> transLet freevs transenv bvdtes e2
    | Exp_extern_call (eref, mn, t, tl, el) -> 
        let fail s = 
          let msg = spr "Error translating extern call %s.%s.%s: %s" eref.dll (Sugar.text_of_lid eref.classname) mn.idText s in
            raise (TransErr msg) in
        let rec transArgs trenv args es =
          match args with
            | [] -> (es, trenv)
            | arg::rest -> 
                let (earg, trenv) = transE freevs trenv arg in
                  transArgs trenv rest (es@[earg]) in
        let (exps, transenv) = transArgs transenv el [] in
        let vals = List.map (function 
                               | TExp_val v -> v
                               | _ -> fail "Unexpected non-value in arguments") exps in
        let rec transTref env argTypes t = match t.v with
          | Typ_univ(bvd, k, t) -> 
              transTref (env@[real_name bvd]) argTypes t 
          | Typ_fun(_, targ, tret) -> 
              let tt, transenv = transTypeWithEnv env transenv targ in

                transTref env (tt::argTypes) tret
          | _ -> 
              let tret, transenv = transTypeWithEnv env transenv t in
                List.rev argTypes, tret in
        let typInsts = List.map (fun t -> let (newt, _) = transType transenv t in newt) tl in
        let argTypes, retType = transTref [] [] t in
          TExp_extern_call(eref, (text_of_id mn,typInsts), argTypes, retType, vals), transenv
            
    | Exp_bot -> TExp_bot, transenv
    | _ -> 
        let errmsg = Printf.sprintf "transExp: unexpected exp %s" 
          (Pretty.strExp e) in
          raise (TransErr errmsg)

(* translate data constructor application *)
(*     S; G; A |- ei : ti ==> ei' forall i = 1..n
       -----------------------------------------------------------
       S; G; A |- D[t1, ..., tm]e1 ... en: t  ==>
       x1 = e1' in ... in xn = en' in
       x1 ... xn obj D<[|t1|], ..., [|tm|], x1, ..., xn>
*)
and transConstrApp freevs transenv var ts phantom_es es : (tExp * transEnv) = 
  let rec collect_fun_args t = match t.v with
    | Typ_fun(_, t1, t2) -> t1::collect_fun_args t2 
    | _ -> [] in
  let rec ctor_arg_types ixs t = match t.v with
    | Typ_univ(bvd, _, t) -> ctor_arg_types ((real_name bvd)::ixs) t 
    | _ -> List.map (fun t -> let (newt, _) = transTypeWithEnv (List.rev ixs) transenv t in newt) (collect_fun_args t) in
  let tgtArgTypes = ctor_arg_types [] var.sort in
    (*       Printf.printf "Translated source ctor type %s\n to target type %s\n" (Pretty.strTyp var.sort) *)
    (*         (String.concat ", " (List.map PrettyTarget.strType tgtArgTypes)); *)
  let (dconvar, transenv) = transVvar transenv var in
  let rec transts trenv ts newts =
    match ts with 
      | [] -> (newts, trenv) 
      | (t::rest) ->
          let (newt, trenv) = transType trenv t in
            transts trenv rest (newts@[newt]) in
  let (newts, transenv) = transts transenv ts [] in
  let rec getValues trenv es ves =
    match es with
      | [] -> (ves, trenv) 
      | e::rest ->
          let (newe, trenv) = transE freevs trenv e in
            match newe with
              | TExp_val v -> getValues trenv rest (ves@[(v, None)])
              | _ -> let (newt, trenv) = transType trenv e.sort in
                let newvar = newVvar newt in
                  getValues trenv rest (ves@[(TVal_var newvar, Some newe)]) in
  let (newves, transenv) = getValues transenv es [] in
  let (newpves, transenv) = getValues transenv phantom_es []  in
  let vals = List.map fst (newpves@newves) in
  let tconcrete: tTypeConcrete = (fst dconvar, newts, vals) in
  let laste = TExp_val (TVal_obj(tconcrete,tgtArgTypes,[])) in
  let rec buildExp ves laste =
    match ves with
      | [] -> laste
      | ((TVal_var var, Some e)::rest) ->
          TExp_let(var, e, buildExp rest laste)
      | (_::rest) -> buildExp rest laste in
    (buildExp (newpves@newves) laste, transenv)

and filterAndTransFvs transenv fvs_env fvs = 
  let rec aux out env_vs = 
    let isEqId (vid:ident) bvar =
      let id = bvar_real_name bvar in
        (vid.idText = id.idText) in            
      match env_vs with
        | [] -> List.rev out
        | (env_v::env_vs) ->
            match List.tryFind (fun fv -> isEqId env_v fv) fvs with 
              | None -> aux out env_vs
              | Some bvar -> 
                  let (newvar, _) = transBVvar transenv bvar in
                    aux (newvar::out) env_vs in
    aux [] fvs_env

(* translation of dependent functions
   (G; A) \\// x: t = (G'; A')
   S; G'; A' |- e: t' ==> e'
   class C<[|G|]_a, [|G|]_x >: [|\x:t -> t'|] {
   public override [|t'|] app([|t|] x) {e'}
   }
   -----------------------------------------------------------------
   S; G; A |- \x:t. e : \x:t -> t' ==>  
   obj C<dom([|G|]_a), dom([|G|]_x)>
*)
(* TODO: fixed the order of type variables as in transAbs *)
and transAbs freevs transenv e : (tExp * transEnv) = 
  match e.v with
    | Exp_abs (bvd, t, body) ->
        let (super, transenv) = transType transenv (e.sort) in
          (* get free type and value variables from the exp *)
        let (freetvs, freevs_e) = Absyn.freevarsExp e in
        let isAffine = (* to test if affine *)
          List.exists (fun (v:bvar<typ>) -> match (v.sort).sort with 
                         | Kind_affine -> true
                         | _ -> false ) freevs_e in
        let freetvs = deDup freetvs in
        let _ = DebugLog.Log(spr "\n transAbs: freetvs in %s = " (Pretty.strExp e)) in
          (*let _ = pr " %s" (String.concat "\n" (List.map (fun var -> Printf.sprintf " %s" (Pretty.strBtvar var)) freetvs)) in
            let _ = pr "\n transAbs: freevs in = %s"
            (String.concat "\n" (List.map (fun var -> Printf.sprintf "%s" (Pretty.strBvar var)) freevs_e)) in*)
          (* check which vs appears in freevs_e and add in that order *)
(*         let rec getFreevs fvs vs = *)
(*           let isEqId (vid:ident) bvar = *)
(*             let id = bvar_real_name bvar in *)
(*               (vid.idText = id.idText) in             *)
(*             match vs with *)
(*               | [] -> List.rev fvs *)
(*               | (vid::restvs) -> *)
(*                   if (List.exists (fun va -> isEqId vid va) freevs_e) then *)
(*                     let bvar = List.find (fun va -> isEqId vid va) freevs_e in *)
(*                     let (newvar, _) = transBVvar transenv bvar in *)
(*                       getFreevs (newvar::fvs) restvs  *)
(*                   else getFreevs fvs restvs in *)
        let (ftvs, transenv) = transFreeTvs transenv freetvs [] in 
        let fvs = filterAndTransFvs transenv freevs freevs_e in
        let _ = DebugLog.Log(Printf.sprintf "\n after filtering: freevs in = %s";
                             (String.concat "\n" (List.map (fun var -> Printf.sprintf "%s" (PrettyTarget.strVvar var)) fvs))) in
          (* x : t *)
        let (_, id) = bvd in
        let (tParam, transenv) = transType transenv t in
        let param = (text_of_id id, tParam) in
          (* body *)
        let (newbody, transenv) = transE (freevs@[id]) transenv body in
          (* new class for the function *)
        let classname = "C" ^ (newSym()) in
          (*let _ = pr "\n Translating: %s  -------------> \n %s, \n currentEvidences = %s" 
            (Pretty.strExp e)
            classname (PrettyTarget.strEvidence (currentEvidences())) in*)
        let retTyp = body.sort in
        let (tRetType, transenv) = transType transenv retTyp in
        let appMethod = (
          tRetType,    (* return type *)
          "Invoke",       (* method name *)
          [],          (* tvars *)
          [param],     (* params *)
          Some newbody (* body *) ) in
        let fvs = dedupVvars fvs in
        let newclass = {
          extqual=None;
          visibility = TVis_internal;
          name = [classname];
          namestring = classname;
          ptvars = [];
          tvars = ftvs;
          pvvars = [];
          vvars = fvs;
          evidences = filterEvidences (currentEvidences()) fvs;
          extends = Some (if isAffine then mk_affine super else super);
          fields = [];
          staticFields = [];
          methods = [appMethod];
          kind = Some (if isAffine then TKind_affine else TKind_star);
        } in
        let _ = DebugLog.Log(Printf.sprintf "\n %s " (PrettyTarget.strClassDecl newclass)) in
        let tconcrete = ([classname], List.map (fun ftv -> TType_var ftv) ftvs, 
                         List.map (fun fv -> TVal_var fv) fvs) in 
          (TExp_val (TVal_obj(tconcrete, [], [])), addToTransEnv transenv newclass)
    | _ ->
        let errmsg = Printf.sprintf "transAbs: unexpected exp %s" (Pretty.strExp e) in
          raise (TransErr errmsg)


and transFreeTvs trenv tvs newtvs =
  match tvs with
    | [] -> (newtvs, trenv)
    | (bv::rest) -> 
        let (newbv, trenv) = transBTvar trenv bv in
          transFreeTvs trenv rest (newtvs@[newbv])
            
(* translation of polymorphic functions 
   S, a:k; G; A |- e:t ==> e'
   class C<[|G|]_a, [|G|]_x>: [|forall a.t|] {
   public override [|t|] TyApp<a>() { e' }
   }
   -----------------------------------------------------------------
   S; G; A |- /\a::k.e : forall a.t ==> obj C<dom([|G|]_a), dom([|G|]_x)>
*)
(* TODO: fixed the order of type/value variables as in transAbs *)
and transTAbs outer_freevs transenv e : (tExp * transEnv) = 
  match e.v with
    | Exp_tabs(bvd, k, body) ->
        let (super, transenv) = transType transenv e.sort in
          (* free type and value variables in e *)
        let (freetvs, freevs) = Absyn.freevarsExp e in 
        let isAffine = (* to test if affine *)
          List.exists (fun (v:bvar<typ>) -> match (v.sort).sort with 
                         | Kind_affine -> true
                         | _ -> false) freevs in
        let freetvs = deDup freetvs in
        let (ftvs, transenv) = transFreeTvs transenv freetvs [] in 
(*         let rec transfvs trenv bvars tgtvars = *)
(*           match bvars with *)
(*             | [] -> (tgtvars, trenv)  *)
(*             | bvar::rest -> *)
(*                 let (tgtvar, trenv) = transBVvar trenv bvar in *)
(*                   transfvs trenv rest (tgtvars@[tgtvar]) in *)
        let fvs = filterAndTransFvs transenv outer_freevs freevs in 
          (* a: k *)
        let (_, id) = bvd in
          (* body *)
        let (newbody, transenv) = transE outer_freevs transenv body in
          (* new class for the polymorphic function *)
        let classname = "C" ^ (newSym()) in
          (*let _ = pr "\n Translating: %s \n -------------> \n %s " 
            (Pretty.strExp e) classname in *)
        let (tret, transenv) = transType transenv body.sort in
        let (newKind, transenv) = transKind transenv k in
        let tyappMethod = (
          tret,                        (* return type *)
          "TyApp",                     (* method name *) 
          [(text_of_id id, newKind)], (* type vars *)
          [],                          (* params *)
          Some newbody                 (* body *)
        ) in
        let fvs = dedupVvars fvs in
        let newclass = {
          extqual=None;
          visibility = TVis_internal;
          name = [classname];
          namestring = classname;
          ptvars = [];
          tvars = ftvs;
          pvvars = [];
          vvars = fvs;
          evidences = filterEvidences (currentEvidences()) fvs;
          extends = Some (if isAffine then mk_affine super else super);
          fields = [];
          staticFields = [];
          methods = [tyappMethod];
          kind = Some (if isAffine then TKind_affine else TKind_star);
        } in
        let tconcrete = ([classname], 
                         List.map (fun ftv -> TType_var ftv) ftvs, 
                         List.map (fun fv -> TVal_var fv) fvs) in
        let newexp = TExp_val (TVal_obj(tconcrete,[],[])) in
          (newexp, addToTransEnv transenv newclass) 
    | _ ->
        let errmsg = Printf.sprintf "transTabs: unexpected exp %s"
          (Pretty.strExp e) in
          raise (TransErr errmsg)

(* translation of function application 
   S; G; A |- e1: \x:t -> t' ==> e1'
   S; G; A |- e2: t ==> e2'
   --------------------------------------
   S; G; A |- e1 e2 : t' ==> 
   x1 = e1' in 
   x2 = e2' in
   x1 x2 call [|t'[x2/x]|] C::App([|t1|])
*)
and transApp freevs transenv efun earg : (tExp * transEnv) = 
  (*let _ = Printf.printf "\ntransApp: efun = %s \n with type %s \n and earg = %s \n with type %s" (Pretty.strExp efun) (Pretty.strTyp (efun.sort)) (Pretty.strExp earg) (Pretty.strTyp (earg.sort)) in *)
  let (newfun, transenv) = transE freevs transenv efun in
  let (newarg, transenv) = transE freevs transenv earg in
  let appRef = ("Invoke", []) in
  let (argType, transenv) = transType transenv (earg.sort) in
  let newvar = newVvar argType in
  let (newtfun, transenv) = transType transenv (efun.sort) in
  let newvfun = newVvar newtfun in
    match newfun with
      | TExp_val vfun ->
          (match newarg with
             | TExp_val varg -> TExp_call(vfun, [varg], appRef), transenv
             | _ -> 
                 let call = TExp_call(vfun, [TVal_var newvar], appRef) in
                   (TExp_let(newvar, newarg, call), transenv))
      | _ -> 
          (match newarg with
             | TExp_val varg ->
                 let call = TExp_call(TVal_var newvfun, [varg], appRef) in
                   (TExp_let(newvfun, newfun, call), transenv)
             | _ ->
                 let call = TExp_call(TVal_var newvfun, [TVal_var newvar], appRef) in
                   (TExp_let(newvar, newarg, TExp_let (newvfun, newfun, call)), transenv))
            
(* translation of type application 
   S; G; A |- e: forall a.t' ==> e'
   ----------------------------------------------------------------
   S; G; A |- e[t] : t' ==> e' call [|t'[t/a]|] C::TyApp<[|t|]> 

*)
and transTApp freevs transenv efun targ : (tExp * transEnv) = 
  let (newfun, transenv) = transE freevs transenv efun in
  let (newtfun, transenv) = transType transenv (efun.sort) in
  let newvfun = newVvar newtfun in  
  let (newtarg, transenv) = transType transenv targ in
  let tyappRef = ("TyApp", [newtarg]) in
    match newfun with
      | TExp_val funval -> (TExp_call(funval, [], tyappRef), transenv)
      | _ ->
          (TExp_let(newvfun, newfun,
                    TExp_call(TVal_var newvfun, [], tyappRef)), transenv)
            
(* translation of pattern match 

   -----------------------------------------------------------------
   S; G; A |- [|match e with pattern_exp_list edefault|]: t ==> 
   x = [|e|] in
   x isinst ([|pattern|] -> [|e|]) [|edefault|]
*)
and transMatch freevs transenv em pes edefault finalTy : (tExp * transEnv) = 
  let (newe, transenv) = transE freevs transenv em in
  let transPatternExp trenv (p, e) =
    match p with 
      | Pat_variant (lid, tys, phantoms, bvs) -> (* TODO: handle phantoms in a pattern *)
          let (newe, trenv) = transE (freevs@(List.map Absyn.bvar_real_name bvs)) trenv e in
          let rec transtys tenv ts newts =
            match ts with
              | [] -> (newts, tenv)
              | t::rest ->
                  let (newt, tenv) = transType tenv t in
                    transtys tenv rest (newts@[newt]) in
          let (cname, newts, trenv) =
            if (Sugar.lid_equals lid Const.tuple_UU_lid or
                  Sugar.lid_equals lid Const.tuple_UA_lid)
            then (* construct the two types for Tuple_UU/UA *)
              match bvs with
                | [var1; var2] ->
                    let (newt1, trenv) = transType trenv (var1.sort) in
                    let (newt2, trenv) = transType trenv (var2.sort) in
                    let id = (bvar_real_name var1).idText in
                    let tfun = TType_fun(Some(id, id), newt1, newt2) in
                    let tupleCname = if (Sugar.lid_equals lid Const.tuple_UU_lid)
                    then tClassDconDepTuple2SS.name
                    else tClassDconDepTuple2SA.name in
                      (tupleCname, [newt1; tfun], trenv)
                | _ -> let msg = "transMatch: unexpected Tuple pattern" in
                    raise (TransErr msg)
            else let (newts, trenv) = transtys trenv tys []in
              (path_of_lid lid, newts, trenv) in
          let rec transbvs tenv bvs newvs =
            match bvs with
              | [] -> (newvs, tenv) 
              | bv::rest ->
                  let (newv, tenv) = transBVvar tenv bv in
                    transbvs tenv rest (newvs@[newv]) in
          let (newvs, trenv) = transbvs trenv bvs [] in
          let patterne = (cname, newts, newvs, newe) in
            (patterne, trenv)
              (*       | Pat_constant cons -> *)
              (*         let errmsg = Printf.sprintf "transMatch: unexpected pattern %s" *)
              (*                      (cons.ToString()) *)
              (*         in raise (TransErr errmsg)  *)in
  let rec transPats trenv pes tces =
    match pes with
      | [] -> (tces, trenv) 
      | pe::rest ->
          let (tce, trenv) = transPatternExp trenv pe in
            transPats trenv rest (tces@[tce]) in
  let (tces, transenv) = transPats transenv pes [] in
  let (newedefault, transenv) = transE freevs transenv edefault in
  let (newFinalTy, transenv) = transType transenv finalTy in
  let finale = 
    match newe with
      | TExp_val ve -> TExp_isinst(ve, tces, newedefault, newFinalTy)
      | _ -> let (newt, transenv) = transType transenv (em.sort) in
        let newvar = newVvar newt in
          TExp_let (newvar, newe,
                    TExp_isinst(TVal_var newvar, tces, newedefault, newFinalTy)) in
    (finale, transenv)

(* translation of a record expression *)
(* t is the type of the whole record expression *)
and transRecordExp freevs transenv fes t : (tExp * transEnv) = 
  let (tRecordType, transenv) = transType transenv t in
  let tconcrete =
    (match tRecordType with 
       | TType_concrete tconcr -> tconcr
       | TType_name (n,_) -> (n,[],[])
       | _ ->
           let errmsg =
             Printf.sprintf "transRecordExp: unexpected src type %s and target type %s" 
               (Pretty.strTyp t)
               (PrettyTarget.strType tRecordType) in
             raise (TransErr errmsg)) in
  let rec transFes trenv fieldExps newes newvs = 
    match fieldExps with
      | [] -> (newes, newvs, trenv) 
      | (_, e) :: rest -> 
          let (newe, trenv) = transE freevs trenv e in
          let (newt, trenv) = transType trenv e.sort in
          let newvar = newVvar newt in
            transFes trenv rest (newes @ [newe]) (newvs@[newvar]) in
  let (newes, newvs, transenv) = transFes transenv fes [] [] in
  let rec genNewobjExp es nvs =
    match (es, nvs) with
      | ([], []) -> TExp_val(TVal_obj(tconcrete, [], List.map TVal_var newvs)) (* TODO: Records can have polymorphic args too; populate arg types *)
      | (e::restes, nv::restnvs) ->
          TExp_let(nv, e, genNewobjExp restes restnvs) in
  let newObjExp = genNewobjExp newes newvs in
    (newObjExp, transenv)

(* translation of dependent tuple constructor applications *)
(* Tuple_** <_>(e1, e2) : Typ_dtuple (x:t1, t2) => *)
(* obj(DconTuple_**<||t1||, \x:||t1||.||t2||, ||e1||, ||e2||>, [] []) *)
and transDtupleConstr freevs transenv dconvar es ttups orige =
  if (Sugar.lid_equals dconvar.v Const.tuple_UU_lid or
        Sugar.lid_equals dconvar.v Const.tuple_UA_lid) then
    match (ttups, es) with
      | ((Some bvdef, t1) :: [(None, t2)], [e1; e2]) ->
          transTuple2Constr freevs transenv dconvar (Some bvdef) t1 t2 e1 e2
      | _ -> transE freevs transenv orige
  else transE freevs transenv orige

and transTuple2Constr freevs transenv dconvar bvdefopt t1 t2 e1 e2 =
  let (newt1, transenv) = transType transenv t1 in
  let (newt2, transenv) = transType transenv t2 in
  let tfun = match bvdefopt with
    | Some bvdef -> TType_fun(Some (transBvDef bvdef), newt1, newt2) 
    | None -> TType_fun(None, newt1, newt2) in
  let (newe1, transenv) = transE freevs transenv e1 in
  let (newe2, transenv) = transE freevs transenv e2 in
  let (newvar1, newvar2) = (newVvar newt1, newVvar newt2) in
  let (newval1, eopt1) =
    (match newe1 with | TExp_val val1 -> (val1, None)
       | _ -> (TVal_var newvar1, Some newe1)) in
  let (newval2, eopt2) =
    (match newe2 with | TExp_val val2 -> (val2, None)
       | _ -> (TVal_var newvar2, Some newe2)) in
  let (newKind1, transenv) = transKind transenv t1.sort in
  let (newKind2, transenv) = transKind transenv t2.sort in
  let tconcrete =
    match (newKind1, newKind2) with
      | (TKind_star, TKind_star) ->
          (tClassDconDepTuple2SS.name, [newt1; tfun], [newval1; newval2])
      | (TKind_star, TKind_affine) ->
          (tClassDconDepTuple2SA.name, [newt1; tfun], [newval1; newval2])
      | (TKind_affine, TKind_affine) ->
          (tClassDconTuple2AA.name, [newt1; newt2], [newval1; newval2])
      | (TKind_affine, TKind_star) ->
          (tClassDconTuple2AS.name, [newt1; newt2], [newval1; newval2])
      | _ -> let errmsg = Printf.sprintf "transDTuple: unexpected kinds %s, %s"
          (Pretty.strKind t1.sort) (Pretty.strKind t2.sort) in
          raise (TransErr errmsg) in
  let constre = TExp_val(TVal_obj(tconcrete, [], [])) in
  let finale =
    match (eopt1, eopt2) with
      | (None, None) -> constre
      | (Some e, None) -> TExp_let (newvar1, e, constre)
      | (None, Some e) -> TExp_let (newvar2, e, constre)
      | (Some e1, Some e2) -> TExp_let (newvar1, e1, TExp_let (newvar2, e2, constre))
  in (finale, transenv)
       
(* translation of let binding *)
and transLet freevs transenv bvdtes e : (tExp * transEnv) = 
  let rec getFreevs fvs bvdtes =
    match bvdtes with
      | [] -> fvs
      | (bvd, _, _) :: rest ->
          let (_, id) = bvd in
            getFreevs (fvs@[id]) rest in
  let fvs = getFreevs freevs bvdtes in
  let (newbody, transenv) = transE fvs transenv e in
  let rec transBvdtes fvs trenv bvdtes laste =
    match bvdtes with
      | [] -> (laste, trenv) 
      | bvdte :: rest ->
          let ((bvd:bvdef), ty, exp) = bvdte in
          let newid = transBvDefId bvd in
          let (newt, trenv) = transType trenv ty in
          let newvar = (newid, newt) in
          let newfvs = fvs @ [snd bvd] in
          let (newe, trenv) = transE newfvs trenv exp in
            transBvdtes newfvs trenv rest (TExp_let(newvar, newe, laste)) in
    transBvdtes freevs transenv bvdtes newbody 

(* translation of let-bindings *)
(* Let-bindings are translated into static fields of the class corresponding *)
(* to the module, which has the same name as the module *)
(* type letbinding = list<bvdef * typ * exp> *)
let transLetbinding transenv (let_binding: letbinding) isrec : transEnv * tStaticFieldDecl list =
  let transBinding trenv (bvd: bvdef, t: typ, e: exp): transEnv * list<tStaticFieldDecl>  = 
    let (external_name, internal_name) = bvd in
    let (newt, trenv) = transType trenv t in
      (*      let _ = Printf.printf "Translates source type %s in let binding for %s to %s\n" 
              (Pretty.strTyp t) (external_name.idText) (PrettyTarget.strType newt) in *)
    let (newe, trenv) = transExp trenv e in
      (* Nik: Ugly hack duplicates a static field; once at the
         external name and once at the internal name *)
    let sfdecls: list<tStaticFieldDecl> = [(text_of_id external_name, newt, Some newe)] in
    let sfdecls =
      if internal_name.idText = external_name.idText then sfdecls
      else (text_of_id internal_name, newt, Some newe)::sfdecls in
      (trenv, sfdecls) in
  let rec transBindings trenv bindings out_fdecls = 
    match bindings with
      | [] -> (trenv, out_fdecls)
      | binding::rest -> 
          let (trenv, fdecls) = transBinding trenv binding in
            transBindings trenv rest (out_fdecls@fdecls)
  in transBindings transenv let_binding []

(* translation of type parameters *)
(* type tparam =
   | Tparam_typ  of bvdef * kind (* idents for pretty printing *)
   | Tparam_term of bvdef * typ
*)
let transTParams transenv (tparams: tparam list) : (transEnv * tVar<tKind> list * tVar<tType> list ) =
  let rec transTps trenv tps tvars vvars =
    match tps with
      | [] -> (trenv, tvars, vvars)
      | (Tparam_typ (bvd, knd)) :: rest -> 
          let (newKind, trenv) = transKind trenv knd in
          let newTvars = (transBvDefId bvd, newKind) in
            transTps trenv rest (tvars @ [newTvars]) vvars
      | (Tparam_term (bvd, ty)) :: rest ->
          let (newty, trenv) = transType trenv ty in
          let newVvar = (transBvDefId bvd, newty) in
            transTps trenv rest tvars (vvars @ [newVvar]) in
    transTps transenv tparams [] [] 

(* translation of kinds for type constructors *)
let transKindTc transenv (k: kind): (transEnv * tVar<tKind> list * tVar<tType> list * tKind) =
  let rec transkTc trenv k tvars vvars =
    match k with 
      | Kind_boxed _ 
      | Kind_star 
      | Kind_affine -> (trenv, tvars, vvars,  fst(transKind trenv k))
      | Kind_tcon (Kind_star, k2) ->
          let newTv = newTvar TKind_star in
            transkTc trenv k2 (tvars @ [newTv]) vvars 
      | Kind_tcon (Kind_affine, k2) ->
          let newTv = newTvar TKind_affine in
            transkTc trenv k2 (tvars @ [newTv]) vvars
      | Kind_tcon (k1, k2) ->
          let (newK1, trenv) = transKind trenv k1 in
          let newTv = newTvar newK1 in
            transkTc trenv k2 (tvars @ [newTv]) vvars
      | Kind_dcon (t, k2) ->
          let (newt, trenv) = transType trenv t in
            transkTc trenv k2 tvars (vvars @ [(newVvar newt)])
      | Kind_qkind (bvdef_ks, k) ->
          let rec transTvs transenv tvs newtvs =
            (match tvs with
               | [] -> (newtvs, transenv)
               | ((bvd, knd)::rest) ->
                   let (newKind, transenv) = transKind transenv knd in
                     transTvs transenv rest (newtvs@[(text_of_id(snd bvd), newKind)])) in
          let (newtvs, trenv) = transTvs trenv bvdef_ks [] in
            transkTc trenv k (tvars @ newtvs) vvars
      | _ ->
          let errmsg = Printf.sprintf "transKindTc: unexpected kind %s" 
            (Pretty.strKind k) in
            raise (TransErr errmsg)
  in transkTc transenv k [] [] 

(* translation of type constructor *)
(* Type parameters are translated to additional type/value parameters *)
(* in the result generic class *)
let transTycon transenv (lid: lident) (tparams: list<tparam>) (k: kind) (b: bool) 
    : transEnv = 
  let className = path_of_lid lid in
  let (transenv, tvars_tp, vvars_tp) = transTParams transenv tparams in
  let (transenv, tvars_k, vvars_k, knd) = transKindTc transenv k in
  let newclass: tClassDecl = 
    { extqual=None;
      visibility = TVis_public;
      name = className;
      namestring = String.concat "." className;
      ptvars = tvars_tp;
      tvars = tvars_k;
      pvvars = vvars_tp;
      vvars = dedupVvars vvars_k;
      evidences = [];
      extends = None;
      fields = [];
      staticFields = [];
      methods = [];
      kind = Some knd;
    } in
    addToTransEnv transenv newclass

(* translate the dependent type of a data constructor *)
(* Return a list of new value variables, a list of class declarations, *)
(* and the result type *)
let transDconDep transenv (dept: typ)
    : transEnv * tVar<tType> list * tType =
  let rec transDconDepfun count trenv dept vvars  =
    match dept.v with 
      | Typ_fun (bvd_opt, t1, t2) -> 
          let (newt1, trenv) = transType trenv t1 in
          let newvar =
            match bvd_opt with
              | None -> 
                  let name = Printf.sprintf "field_%d" count in
                    (name, newt1)
              | Some bvd -> (transBvDefId bvd, newt1) in
            transDconDepfun (count+1) trenv t2 (vvars @ [newvar])
      | t -> 
          let (newt, trenv) = transType trenv dept in
            (trenv, vvars,  newt) in
    transDconDepfun 1 transenv dept []

(* translate the type of a data constructor to a list of type variables, 
   a list of value variables, a list of class declarations, and the
   result type (type constructor instantiation)
*)
let transDconType transenv (t: typ) : (transEnv * tVar<tKind> list * tVar<tType> list  * tType) =
  let rec transDconPoly trenv (ty: typ) tvars =
    match ty.v with
      | Typ_univ (bvd, knd, tbody) ->
          let (newKind, trenv) = transKind trenv knd in
          let newtvar = (transBvDefId bvd, newKind) in
            transDconPoly trenv tbody (tvars @ [newtvar])
      | dept -> 
          let (trenv, vvars,  resultType) = transDconDep trenv ty in
            (trenv, tvars, vvars, resultType) in
    transDconPoly transenv t [] 

let genericEqualityRef = 
  let eref = {Sugar.language=Sugar.FSharp; 
              Sugar.dll="FSharp.Core"; 
              Sugar.classname=Const.p2l ["Microsoft.FSharp.Core.LanguagePrimitives"];
              Sugar.innerclass=Some "HashCompare"; Sugar.extqual=None} in
    eref, "GenericEqualityIntrinsic"

let addEqualityMethod trenv (cdecl:tClassDecl) = 
  (*   if List.length cdecl.tvars  <> 0 then cdecl else *)
  let is_value_type = function
      TType_name ((["Prims"; "bool"], _))
    | TType_name ((["Prims"; "unit"], _))
    | TType_name ((["Prims"; "int"], _))
    | TType_name ((["Prims"; "string"], _)) -> true
    | _ -> false in
  let objType = Target.objType in 
  let boolType, trenv = transType trenv Const.bool_typ in 
  let truev = TExp_val(TVal_constant (Sugar.Const_bool true)) in
  let falsev = TExp_val(TVal_constant (Sugar.Const_bool false)) in
  let branch = 
    let cn = cdecl.name in
    let targs = List.map (fun t -> TType_var t) (cdecl.ptvars@cdecl.tvars) in
    let fields = (cdecl.vvars@cdecl.fields) in
    let vargs = List.map (fun (fn, t) -> let loc = newSym () in (loc,t)) fields in
    let eref, geneq_mname = genericEqualityRef in
    let rec mk_body xs fields locals = match fields, locals with
      | field::fields, local::locals -> 
          let x = newSym(), boolType in
          let body = mk_body (x::xs) fields locals in
          let t = snd field in
          let this_eq = 
            TExp_extern_call(eref, (geneq_mname, [t]), [TType_index 0us; TType_index 0us], 
                             boolType, [TVal_var field; TVal_var local]) in
            TExp_let(x, this_eq, body)
      | [], [] -> 
          List.fold_left (fun out x -> TExp_cond(TExp_val (TVal_var x), out, falsev)) truev xs in
    let body = mk_body [] fields vargs in
      (cn, targs, vargs, body) in
  let methodName = "Equals" in
  let tparams = [] in
  let argument = (newSym (), objType) in
  let newloc = newSym (), objType in 
  let body = TExp_let(newloc, TExp_val (TVal_var argument), TExp_isinst(TVal_var newloc, [branch], falsev, boolType)) in
  let md = (boolType, methodName, tparams, [argument], Some body) in
    trenv, {cdecl with methods=md::cdecl.methods}

(* translation of data constructors *)
(* "tparams" are inherited from the enclosing type constructor *)
(* The type of data constructor is in curried format *)
let transDcon transenv (aq:aqual) (lid: lident) (tparams: list<tparam>) (t: typ) (b: bool) 
    : transEnv  =
  let className = path_of_lid lid in
  let (transenv, tvars_tp, vvars_tp) = transTParams transenv tparams in
  let (transenv, tvars_t, vvars_t, super) = transDconType transenv t in
  let vis = match aq with Public -> TVis_public | _ -> TVis_internal in
  let newclass: tClassDecl = 
    { extqual=None;
      visibility = vis;
      name = className;
      namestring = String.concat "." className;
      ptvars = tvars_tp;
      tvars = tvars_t;
      pvvars = vvars_tp;
      vvars = dedupVvars vvars_t;
      evidences = [];
      extends = Some super; 
      fields = [];
      staticFields = [];
      methods = [];
      kind = None;   (* data constructors have the same type as their tcon *)
    } in
  let transenv, newclass = 
    if not !Util.extract_proofs then addEqualityMethod transenv newclass 
    else transenv, newclass in
    addToTransEnv transenv newclass

(* translation signelt *)
(* type sigelt =
   | Sig_tycon_kind   of lident * list<tparam> * kind * bool (* bool identifies a prop *)
   | Sig_typ_abbrev   of lident * list<tparam> * typ
   | Sig_typ_abstract of lident * list<tparam>
   | Sig_datacon_typ  of lident * list<tparam> * typ * bool (* bool identifies an assumption *)
   | Sig_value_decl   of lident * typ 
   | Sig_extern_value of lang * lident * typ
   | Sig_extern_typ   of Sugar.externref * sigelt (* lident * typ * kind *)
   | Sig_record_typ   of lident * list<tparam> * typ
*)
let generateExternWrapper transenv eref extid top = 
  let thisModuleName, extid = Util.pfx extid in  
  let rec mk_fun args t = match t.v with
    | Typ_fun(bvd_opt, targ, tret) -> 
        let id = genident None in
        let bvd = (id, id) in
        let bv = withsort (Exp_bvar (bvd_to_bvar_s bvd targ)) targ in
        let body = mk_fun (bv::args) tret in
          withsort (Exp_abs(bvd, targ, body)) t
    | _ -> 
        let methodName = extid in
        let rec mk_call letvars = function
            [] -> withsort (Exp_extern_call(eref, methodName, top, [], letvars)) t 
          | arg::tl -> 
              let id = genident None in
              let bvd = id,id in
              let bv = withsort (Exp_bvar (bvd_to_bvar_s bvd arg.sort)) arg.sort in
                withsort (Exp_let(false, [(bvd, arg.sort, arg)], mk_call (bv::letvars) tl)) t in
          mk_call [] args in
  let wrapper = mk_fun [] top in
  let letbinding = [(extid,extid), top, wrapper] in
    transLetbinding transenv letbinding false

  let new_bvd_bvar t =  
    let x = new_bvd() in 
    let xx = bvd_to_bvar_s x t in 
    x, withsort (Exp_bvar xx) t

  let w t = withsort t Kind_star
  let W t s = withsort t s 

  let boolTyp = Const.bool_typ
  let pf_typ t = w (Typ_app(Const.pf_typ, t))  (* pf<t> *)
  let and_typ = Const.and_typ
  let implies_typ = Const.implies_typ
  let or_typ = Const.or_typ
  let starStarKind = Kind_tcon(Kind_star, Kind_star)

  let lookupDcon tcenv lid =
    let t = Tcenv.lookup_lid tcenv lid in
    withsort lid t
  let f_refl_var tcenv = Tcutil.reflexivity_var tcenv 
  let f_and_intro_var tcenv = lookupDcon tcenv Const.and_intro_lid
  let f_or_intro1_var tcenv = lookupDcon tcenv Const.or_intro1_lid
  let f_implies_intro_var tcenv = lookupDcon tcenv Const.implies_intro_lid 
  let f_tuple_var tcenv = lookupDcon tcenv Const.tuple_UU_lid
  let f_eq_typ tcenv t e = (* Eq<t; e, e> *)
    w (Typ_dep(W (Typ_dep(Tcutil.eqT_of_typ tcenv t, e))
                 (Kind_dcon(t, Kind_star)),
               e)) 

  let mkImpliesEqTy tcenv t1 t2 e1 e2 = (* l_implies<pf<Eq<t1; e1, e1>>, pf<Eq<t2; e2, e2>>> *)
    w (Typ_app(W (Typ_app(implies_typ, pf_typ(f_eq_typ tcenv t1 e1))) starStarKind,
               pf_typ(f_eq_typ tcenv t2 e2)))
  let mkImplies tcenv t1 t2 e1 e2 =
    (* ImpliesIntro<Eq<t1; e1, e1>, Eq<t2; e2, e2>> \x:pf<Eq<t1; e1, e1>>.(Reflexivity<t2> e2 *)
    (*                                             : pf<Eq<t2; e2, e2>>) *)
    (* : pf<l_implies<pf<Eq<t1; e1, e1>>, pf<Eq<t2; e2, e2>>>> *)
    let eqTy1, eqTy2 = f_eq_typ tcenv t1 e1, f_eq_typ tcenv t2 e2 in
    let pfEqTy1, pfEqTy2 = pf_typ eqTy1, pf_typ eqTy2 in
    let exp_fun =
      let body = W (Exp_constr_app(f_refl_var tcenv, [t2], [], [e2])) pfEqTy2 in
      let fun_typ = w (Typ_fun(None, pfEqTy1, pfEqTy2)) in
      W (Exp_abs(new_bvd(), pfEqTy1, body)) fun_typ in
    W (Exp_constr_app(f_implies_intro_var tcenv, [eqTy1; eqTy2], [], [exp_fun]))
      (pf_typ (mkImpliesEqTy tcenv t1 t2 e1 e2))

  (* types/terms for proofs of boolean type*)
  let f_eqBoolTyp tcenv = (* Eq<Bool; true, true> *)
    f_eq_typ tcenv boolTyp Const.exp_true_bool
  let f_pfEqTyp tcenv = (* pf<Eq<Bool; true, true> *)
    pf_typ (f_eqBoolTyp tcenv)
  let f_impliesPfEqTyp tcenv = (* l_imples(pf<Eq<bool>>, pf<Eq<bool>>) *)
    mkImpliesEqTy tcenv boolTyp boolTyp Const.exp_true_bool Const.exp_true_bool
  let f_pfImpliesPfEqTyp tcenv = (* pf<l_imples(pf<Eq<bool>>, pf<Eq<bool>>)> *)
    pf_typ (f_impliesPfEqTyp tcenv)

  let f_pfEq tcenv = (* reflexivity<boolTyp> Const.exp_true_bool : pf<Eq<bool; true; true>> *)
  W (Exp_constr_app(f_refl_var tcenv, [boolTyp], [], [Const.exp_true_bool])) (f_pfEqTyp tcenv)

  let f_pfImpliesPfEq tcenv = mkImplies tcenv boolTyp boolTyp Const.exp_true_bool Const.exp_true_bool

 (* add /\'a \x:'a \y:\a. call eq(x, y) as the body of Prims.op_Equality *)
 let generateEqualityOp transenv top = match top.v with  
   | Typ_univ(tvar, knd, {v=Typ_fun(_, ttvar, {v=Typ_fun(_, t', result_t); sort=_; p=_}); sort=_; p=_}) ->  
     let eref = {Sugar.language=Sugar.FSharp;  
                 Sugar.dll="FSharp.Core";  
                 Sugar.classname=Const.p2l ["Microsoft.FSharp.Core.LanguagePrimitives"]; 
                 Sugar.innerclass=Some "HashCompare"; Sugar.extqual=None} in 
     let (x, xx), (y, yy) = new_bvd_bvar ttvar, new_bvd_bvar ttvar in 
     let tv = new_bvd () in 
     let bv = w (Typ_btvar(bvd_to_bvar_s tv Kind_star)) in 
     let univ = w (Typ_univ(tv, Kind_star,  
                            w (Typ_fun(None, bv, w (Typ_fun(None, bv, Const.bool_typ)))))) in 
     let callExp = W (Exp_extern_call(eref, Sugar.text_to_id0 "GenericEqualityIntrinsic", univ, [ttvar], [xx;yy])) Const.bool_typ in
     let bodyExp = 
       if (not (!Util.extract_proofs)) then callExp
       else match !curtcenv with
         | None -> callExp
         | Some tcenv ->
         let impliesBoolTvarTy = (* l_implies<pf<Eq<bool; true, true>>, pf<Eq<tvar; x, x>>> *)
           mkImpliesEqTy tcenv Const.bool_typ ttvar Const.exp_true_bool xx in
         let impliesTvarBoolTy = (* l_implies<pf<Eq<tvar; x, x>>, pf<Eq<bool; true, true>>> *)
           mkImpliesEqTy tcenv ttvar Const.bool_typ xx Const.exp_true_bool in
         let proofTy = pf_typ(w (Typ_app(W (Typ_app(and_typ, pf_typ(impliesBoolTvarTy)))
                                        (Kind_tcon(Kind_star, Kind_star)),
                                 pf_typ(impliesTvarBoolTy)))) in
         let proofTerm = 
           (* AndIntro<l_implies<pf<Eq<bool; true, true>>, pf<Eq<tvar; x, x>>>, *)
           (*          l_implies<pf<Eq<tvar; x, x>>, pf<Eq<bool; true, true>>>> *)
           (*          (mkImplies bool tvar true x) *)
           (*          (mkImplies var bool x true)  *)
           (* : pf<l_and<pf<l_implies<pf<Eq<bool; true, true>>, pf<Eq<tvar; x, x>>>>, *)
           (*            pf<l_implies<pf<Eq<tvar; x, x>>, pf<Eq<boo; true, true>>>>>>> *)
           W 
            (Exp_constr_app(f_and_intro_var tcenv, [impliesBoolTvarTy; impliesTvarBoolTy], [], 
                            [mkImplies tcenv Const.bool_typ ttvar Const.exp_true_bool xx;
                             mkImplies tcenv ttvar Const.bool_typ xx Const.exp_true_bool])) 
            (proofTy) in
         (* tuple_UU<bool, proofTy> callExp proofTerm: Typ_dtuple(bool, proofTy) *)
         W (Exp_constr_app(f_tuple_var tcenv, [Const.bool_typ; proofTy], [],
                           [callExp; proofTerm]))
           (w(Typ_dtuple([(None, Const.bool_typ); (None, proofTy)]))) in
     let inner_fun = w (Typ_fun(None, ttvar, result_t)) in 
     let wrapper = W (Exp_abs(x, ttvar,  
                             (W (Exp_abs(y, ttvar, bodyExp))
                                inner_fun))) 
                     (w (Typ_fun(None, ttvar, inner_fun))) in 
     let polyWrapper = W (Exp_tabs(tvar, knd, wrapper)) top in
     let _, mname = Util.pfx (Const.op_Eq) in 
     let letbinding = [(mname, mname), top, polyWrapper] in 
(* let _  = Printf.printf "Adding let bindnig for equality op %s at typ %s\n: %s" mname.idText (Pretty.strTyp top) 
         (Pretty.strExp polyWrapper) in *)
     transLetbinding transenv letbinding false 
   | _ -> transenv, [] 

let generateAndOrOp transenv top lid = match top.v with
  | Typ_fun(_, _, {v=Typ_fun(_, _, result_t); sort=_; p= _}) ->
    let eref = {Sugar.language=Sugar.FSharp;  
                Sugar.dll="FSharp.Core";  
                Sugar.classname=Const.p2l ["Microsoft.FSharp.Core.LanguagePrimitives"]; 
                Sugar.innerclass=Some "IntrinsicOperators"; Sugar.extqual=None} in 
    let (x, xx), (y, yy) = new_bvd_bvar Const.bool_typ, new_bvd_bvar Const.bool_typ in
    let funcName = if (Sugar.lid_equals lid Const.op_And_lid) then "op_BooleanAnd" else "op_BooleanOr" in
    let funcTyp = w (Typ_fun(None, boolTyp,
                             w (Typ_fun(None, boolTyp, boolTyp)))) in
    let callExp = W (Exp_extern_call(eref, Sugar.text_to_id0 funcName, funcTyp, [], [xx;yy])) boolTyp in
    let bodyExp =
      if (not (!Util.extract_proofs)) then callExp 
      else match !curtcenv with
        | None -> callExp
        | Some tcenv ->
          let pfEqTyp = f_pfEqTyp tcenv in
          let eqBoolTyp = f_eqBoolTyp tcenv in
          let pfEq = f_pfEq tcenv in
          let andPfEqTyp = (* l_and pfEqTyp pfEqTyp *)
            w (Typ_app(W (Typ_app(and_typ, pfEqTyp)) starStarKind, pfEqTyp)) in
          let orPfEqTyp = (* l_or pfEqTyp pfEqTyp *)
            w (Typ_app(W (Typ_app(or_typ, pfEqTyp)) starStarKind, pfEqTyp)) in
          (* proof term for and: b=true => x=true &&  y=true *)
          let proofTyp_and = (*l_implies pfEqTyp (l_and pfEqTyp pfEqTyp) *)
           pf_typ (w (Typ_app(W (Typ_app(implies_typ, pfEqTyp)) starStarKind, pf_typ andPfEqTyp))) in
          let proofTerm_and = (*impliesIntro<EqBoolTyp, l_and pfEqTyp pfEqTyp> \x:pfEqTyp.andIntro<eqBoolTyp, eqBoolTyp> pfEq pfEq*)
            W (Exp_constr_app(f_implies_intro_var tcenv, [eqBoolTyp; andPfEqTyp], [], 
                              [W (Exp_abs(new_bvd(), pfEqTyp, 
                                          W (Exp_constr_app(f_and_intro_var tcenv, [eqBoolTyp; eqBoolTyp], [], [pfEq; pfEq]))
                                             (pf_typ andPfEqTyp)))
                                 (w (Typ_fun(None, pfEqTyp, pf_typ andPfEqTyp)))]))
              proofTyp_and in

         (* proof term for or: (z=true => x=true ||  y=true) && (z=false => x=false &&  y=false) *)
         let impliesEqOrTyp = w (Typ_app(W (Typ_app(implies_typ, pfEqTyp)) starStarKind, pf_typ orPfEqTyp)) in
         let impliesEqAndTyp = w (Typ_app(W (Typ_app(implies_typ, pfEqTyp)) starStarKind, pf_typ andPfEqTyp)) in
         let proofTyp_or = (*pf<l_and pf<impliesEqOrTyp> pf<impliesEqAndTyp>> *)
           pf_typ (w (Typ_app(W (Typ_app(and_typ, pf_typ impliesEqOrTyp)) starStarKind, pf_typ impliesEqAndTyp))) in
         let proofTerm_or = (* andIntro<impliesEqOrTyp, impliesEqAndTyp>  *)
                            (*         (impliesIntro<eqBoolTyp, orPfEqTyp> \x:pfEqTyp.orIntro<eqBoolTyp, eqBoolTyp> pfEq) *)
                            (*         (impliesIntro<eqBoolTyp, andpfEqTyp> \x:pfEqTyp.andIntro<eqBoolTyp, eqBoolTyp> pfEq pfEq) *)
           W (Exp_constr_app(f_and_intro_var tcenv, [impliesEqOrTyp; impliesEqAndTyp], [],
                             [W (Exp_constr_app(f_implies_intro_var tcenv, [eqBoolTyp; orPfEqTyp], [],
                                                [W (Exp_abs(new_bvd(), pfEqTyp, 
                                                            W (Exp_constr_app(f_or_intro1_var tcenv, 
                                                                              [eqBoolTyp; eqBoolTyp], 
                                                                              [], [pfEq]))  
                                                              (pf_typ orPfEqTyp)))
                                                    (w (Typ_fun(None, pfEqTyp, pf_typ orPfEqTyp)))]))
                                (pf_typ impliesEqOrTyp);
                              W (Exp_constr_app(f_implies_intro_var tcenv, [eqBoolTyp; andPfEqTyp], [],
                                                [W (Exp_abs(new_bvd(), pfEqTyp, 
                                                            W (Exp_constr_app(f_and_intro_var tcenv, [eqBoolTyp; eqBoolTyp], [], [pfEq; pfEq]))
                                                              (pf_typ andPfEqTyp)))
                                                   (w (Typ_fun(None, pfEqTyp, pf_typ andPfEqTyp)))]))
                                (pf_typ impliesEqAndTyp)]))
             proofTyp_or in
      let proofTy, proofTerm = 
        if (Sugar.lid_equals lid Const.op_And_lid)
        then (proofTyp_and, proofTerm_and)
        else (proofTyp_or, proofTerm_or) in
     (* tuple_UU<bool, proofTy> callExp proofTerm: Typ_dtuple(bool, proofTy) *)
      W (Exp_constr_app(f_tuple_var tcenv, [Const.bool_typ; proofTy], [],
                        [callExp; proofTerm]))
        (w(Typ_dtuple([(None, Const.bool_typ); (None, proofTy)]))) in
    let inner_fun = w (Typ_fun(None, boolTyp, result_t)) in
    let wrapper = W (Exp_abs(x, boolTyp,
                             (W (Exp_abs(y, boolTyp, bodyExp))
                                inner_fun)))
                    (w (Typ_fun(None, boolTyp, inner_fun))) in
    let _, mname = Util.pfx lid in
    let letbinding = [(mname, mname), top, wrapper] in
    transLetbinding transenv letbinding false
  | _ -> transenv, []

let generateNotOp transenv top = match top.v with
  | Typ_fun(_, _, result_t) ->
    let (x, xx) = new_bvd_bvar boolTyp in
    let mainExp = (* Exp_cond(xx, Const.exp_false_bool, Const.exp_true_bool) *)
      W (Exp_cond(xx, Const.exp_false_bool, Const.exp_true_bool)) boolTyp in
    let bodyExp =
      if (not (!Util.extract_proofs)) then mainExp
      else match !curtcenv with
         | None -> mainExp
         | Some tcenv ->
           (* a proof term with type l_and<pf<l_implies(pf<Eq<bool>>, pf<Eq<bool>>)>, *)
           (*                              pf<l_implies(pf<Eq<bool>>, pf<Eq<bool>>)>> *)
           let pfImpliesPfEqTyp = f_pfImpliesPfEqTyp tcenv in
           let pfImpliesPfEq = f_pfImpliesPfEq tcenv in
           let impliesPfEqTyp = f_impliesPfEqTyp tcenv in
           let proofTy = pf_typ(w (Typ_app(W (Typ_app(and_typ, pfImpliesPfEqTyp)) starStarKind, pfImpliesPfEqTyp))) in
           let proofTerm = (*andIntro<impliesPfEqTyp, impliesPfEqTyp>  pfImpliesPfEq pfImpliesPfEq*)
             W (Exp_constr_app(f_and_intro_var tcenv, [impliesPfEqTyp; impliesPfEqTyp], [], [pfImpliesPfEq; pfImpliesPfEq]))
                proofTy
           in
           W (Exp_constr_app(f_tuple_var tcenv, [boolTyp; proofTy], [], [mainExp; proofTerm])) 
             (w (Typ_dtuple([(None, boolTyp); (None, proofTy)])))
    in
    let funcTyp = w (Typ_fun(None, boolTyp, result_t)) in
    let wrapper = W (Exp_abs(x, boolTyp, bodyExp)) funcTyp in
    let _, mname = Util.pfx Const.op_Not_lid in
    let letbinding = [(mname, mname), top, wrapper] in
(*let _  = Printf.printf "\nAdding let bindnig for negation op %s at typ %s\n: %s" mname.idText (Pretty.strTyp top) 
         (Pretty.strExp wrapper) in*)
    transLetbinding transenv letbinding false
  | _ -> transenv, []

let is_autogen = function
  | Sig_tycon_kind (lid, _, _, _) 
  | Sig_datacon_typ (lid, _, _, _, _, _) -> 
      let _, id = Util.pfx lid in
      let id = id.idText in
        (String.length id) > 6 && (String.sub id 0 6) = "__AUTO" 
  | _ -> false
      
(* add this prefix to every extern class declaration *)
(* Should be the same as in Sugar.fs::extern_id *)
let exnDeclPrefix = "__Extern"

let is_primop_dummy (lid:lident) = match lid with 
  | [prim;opname] when prim.idText="Prims" ->
      let opstr = opname.idText in
        opstr.Length >= 10 && String.sub opstr 0 10 = "_dummy_op_"
  | _ -> false

let rec transSigelt transenv (sige: sigelt): transEnv * tStaticFieldDecl list * tClassDecl list =
  (*let _ = pr "transSigelt: %s " (Pretty.strSigelt sige) in*)
  if is_autogen sige then transenv,[], [] else
    match sige with
      | Sig_tycon_kind (lid, tparams, knd, bl) ->
          let name = text_of_lid lid in 
            if name = "Prims.int" || name = "Prims.bool" || name= "Prims.string" || name = "Prims.unit" then 
              transenv, [], []
            else
              let _ = DebugLog.Log(Printf.sprintf "\n sigelt: tcon %s" (Pretty.str_of_lident lid)) in
              let transenv = transTycon transenv lid tparams knd bl in
                transenv, [], []
      | Sig_datacon_typ (lid, tparams, ty, bl, aq, _) -> 
          let _ = DebugLog.Log(Printf.sprintf "\n sigelt: dcon %s" (Pretty.str_of_lident lid)) in
            if (Sugar.lid_equals lid Const.tuple_UU_lid or
                  Sugar.lid_equals lid Const.tuple_UA_lid or
                  Sugar.lid_equals lid Const.tuple_AU_lid or
                  Sugar.lid_equals lid Const.tuple_AA_lid) then 
              transenv, [], []
            else let transenv = transDcon transenv aq lid tparams ty bl in
              transenv, [], []
      | Sig_extern_value (lang, lid, t) -> 
          let _ = DebugLog.Log(Printf.sprintf "\n sigelt: exn value %s" (Pretty.str_of_lident lid)) in
          let transenv, sf = generateExternWrapper transenv lang lid t in
            transenv, sf, []
              (* Just skip the operator decls in prims for now *)
      | Sig_value_decl(lid, t) -> 
        if (Sugar.lid_equals lid Const.op_Eq)
        then let transenv, sf = generateEqualityOp transenv t in transenv, sf, []
        else if (Sugar.lid_equals lid Const.op_And_lid or
                 Sugar.lid_equals lid Const.op_Or_lid)
        then let transenv, sf = generateAndOrOp transenv t lid in transenv, sf, []
        else if (Sugar.lid_equals lid Const.op_Not_lid)
        then let transenv, sf = generateNotOp transenv t in transenv, sf, []
        else if (Sugar.lid_equals lid Const.op_Add_lid or
                 Sugar.lid_equals lid Const.op_Subtraction_lid or
                 Sugar.lid_equals lid Const.op_Multiply_lid or
                 Sugar.lid_equals lid Const.op_GreaterThanOrEqual_lid)
        then transenv, [], []
        else let (newt, transenv) = transType transenv t in
             let lid' = (match lid with 
                           [l] -> [l]
                         | modname::lid -> lid) in 
             let sf = [(text_of_lid lid', newt, None)] in
             transenv, sf, [] 
      | Sig_extern_typ(eref, Sig_typ_abbrev(lid, tps, _, typ)) -> 
          let _ = DebugLog.Log(Printf.sprintf "\n exn typ: tcon %s" (Pretty.str_of_lident lid)) in
            (if List.length tps <> 0 then 
               let errmsg = spr "transSigelt: unexpected parameterized type abbrev: %s" (Pretty.strSigelt sige) in
                 raise (TransErr errmsg));
            (match Sugar.lid_to_eref lid, typ.v with 
               | None, _ -> 
                   let errmsg = spr "transSigelt: expected extern type name; got %s" (Pretty.strSigelt sige) in
                     raise (TransErr errmsg)
               | Some(dll, classname), Typ_record fn_t_l -> 
                   let fieldDecls = List.map 
                     (fun (fn, t) -> 
                        match Sugar.lid_to_eref fn with 
                          | Some(dll, rest) -> 
                              let classname, fn = Util.pfx rest in
                              let tt, transenv = transType transenv t in
                                (fn.idText, tt)
                          | None -> raise Impos) fn_t_l in
                   let cname = exnDeclPrefix::dll::Sugar.path_of_lid classname in
                   let record_decl = 
                     {extqual=None;
                      visibility=TVis_public;
                      name=cname;
                      namestring = String.concat "." cname;
                      ptvars=[];
                      tvars=[];
                      pvvars=[];
                      vvars=[];
                      evidences = [];
                      extends=None;
                      fields=fieldDecls;
                      staticFields=[];
                      methods=[];
                      kind=None} in
                     transenv, [], [record_decl]
               | _ -> raise (NYI "[Translation] External references to non-record types"))

      | Sig_extern_typ(eref, Sig_tycon_kind(lid, tparams, knd, bl)) when List.length tparams = 0 ->
          let _ = DebugLog.Log(Printf.sprintf "\n sigelt: exn tcon %s" (Pretty.str_of_lident lid)) in
            (match Sugar.lid_to_eref lid with 
                 Some(dll, classname) -> 
                   let cname = exnDeclPrefix::dll::Sugar.path_of_lid classname in
                   let tcDecl = {extqual=None;
                                 visibility=TVis_public;
                                 name=cname;
                                 namestring = String.concat "." cname;
                                 ptvars=[];
                                 tvars=[];
                                 pvvars=[];
                                 vvars=[];
                                 evidences = [];
                                 extends=None;
                                 fields=[];
                                 staticFields=[];
                                 methods=[];
                                 kind=None} in
                     (transenv, [], [tcDecl])
               | _ -> raise Impos)
              
      | Sig_extern_typ(eref, Sig_datacon_typ(lid, tparams, ty, bl, aq, Some tcn)) when List.length tparams=0 ->       
          let _ = DebugLog.Log(Printf.sprintf "\n sigelt: exn dcon %s" (Pretty.str_of_lident lid)) in 
            (match Sugar.lid_to_eref tcn, Sugar.lid_to_eref lid with
               | Some(dll, variantClassName), Some(dll', constructorName) when dll=dll' ->
                   let extends = TType_name(exnDeclPrefix::dll::Sugar.path_of_lid variantClassName, TKind_star) in
                   let _, constr = Util.pfx constructorName in
                   let constr = constr.idText in
                   let rec mk_vvars (ix, out) t = match t.v with
                     | Typ_univ _ -> raise (NYI "External polymorphic constructors")
                     | Typ_fun(_, targ, t) -> 
                         let tt, transenv = transType transenv targ in
                         let vvar = (spr "_%s%d" constr ix, tt) in
                           mk_vvars (ix+1, vvar::out) t
                     | _ -> List.rev out in
                   let vvars = mk_vvars (1, []) ty in
                   let cname = exnDeclPrefix::dll::Sugar.path_of_lid constructorName in
                   let constrDecl = {extqual=None;
                                     visibility=TVis_public;
                                     name=cname;
                                     namestring = String.concat "." cname;
                                     ptvars=[];
                                     tvars=[];
                                     pvvars=[];
                                     vvars= dedupVvars vvars;
                                     evidences = [];
                                     extends=Some extends;
                                     fields=[];
                                     staticFields=[];
                                     methods=[];
                                     kind=None} in
                     (transenv, [] , [constrDecl])
               | _ -> raise Impos)

      | Sig_extern_typ(_, Sig_record_typ(lid, tps, _, t)) 
      | Sig_record_typ(lid, tps, _, t) -> 
          (match t.v with
             | Typ_record fn_t_l ->
                 let transenv, tvars_tps, vvars_tps = transTParams transenv tps in
                 let (fnames, ftypes) = List.split fn_t_l in
                 let newfnames = List.map (fun fname -> String.concat "_" (path_of_lid fname)) fnames in
                 let transenv, newftypes = transFieldTypes [] transenv ftypes in
                 let newfields = List.combine newfnames newftypes in
                 let recCdecl: tClassDecl =
                   {extqual = None;
                    visibility = TVis_public;
                    name = path_of_lid lid;
                    namestring = String.concat "." (path_of_lid lid);
                    ptvars = [];
                    tvars = tvars_tps;
                    pvvars = [];
                    vvars = dedupVvars vvars_tps;
                    evidences = [];
                    extends = None;
                    fields = newfields;
                    staticFields = [];
                    methods = [];
                    kind = Some TKind_star;
                   } in
                   (addToTransEnv transenv recCdecl, [], [])
             | _ -> raise (TransErr (Printf.sprintf "unexpected record type: %s" (Pretty.strSigelt sige))))
            
      | _ -> raise (NYI "[Translation] Unhandled sig elt")
          
(* translation of signature  *)
(* type signature = list<sigelt> *)
let transSignature transenv (sg: signature) : transEnv * tStaticFieldDecl list * tClassDecl list =
  let rec transsigs trenv sigs fields externs =
    match sigs with
      | [] -> (trenv, fields, externs)
      | sigelt :: rest ->
          let (trenv, flds, exts) = transSigelt trenv sigelt in
            transsigs trenv rest (fields@flds) (externs@exts) in
    transsigs transenv sg [] []


(* translation of modules *)
(*  type modul = {
    name: lident;
    pos: Range.range;
    signature: signature;
    letbindings: list<letbinding * bool>; (* the boolean signals a let rec *)
    main:option<exp>;
    exports: signature;
    }
*) 
let transModule tcenv_opt (pfx:list<modul>)  (m: modul): tModule =
  (*if (Pretty.str_of_lident m.name <> "ProofLib") then Pretty.printModule m;*)
  curmodname := Some m.name;
  let _ = match (tcenv_opt) with
    | Some tcenv -> curtcenv := Some tcenv
    | _ -> () in
  let externs = List.concat (List.map (fun m -> List.filter (function 
                                                               | Sig_extern_typ _ -> true
                                                               | _ -> false) m.signature) pfx) in
  let newName = path_of_lid m.name in
  let scope, name = Util.pfx newName in
  let name = name ^ "_Container" in
  let containerClass = newName in(* scope@[name] in *)
  let transenv, sfields_signature, externs_signature = transSignature (createTransEnv()) (externs@m.signature) in
  let rec transLetbindings transenv letbds sfs =
    match letbds with
      | [] -> (transenv, sfs) 
      | (letbd, isrec) :: rest ->
          let (transenv, newsfs) = transLetbinding transenv letbd isrec in
            transLetbindings transenv rest (sfs @ newsfs) in
  let (transenv, sfields) = transLetbindings transenv m.letbindings [] in
  let sfields_signature = List.filter 
    (fun (name, t, _) -> not (List.exists (fun (name', _, _) -> name=name') sfields)) sfields_signature in
  let sfields = sfields_signature@sfields in 
    (*   let _ = List.iter (fun (name, tt, _) -> print_string ("Translation added static field " ^name^": "^(PrettyTarget.strType tt)^"\n")) sfields in *)
  let moduleClassDecl: tClassDecl = 
    (* the class corresponding to the module *)
    { extqual=None;
      visibility = TVis_public;
      name = containerClass;
      namestring = String.concat "." newName;
      ptvars = [];
      tvars = [];
      pvvars = [];
      vvars = [];
      evidences = [];
      extends = None;
      fields = [];
      staticFields = sfields;
      methods = [];
      kind = Some TKind_star;
    } in
    (* the target checker regards the static fields as globally visible *)
  let extendsName = match m.extends with
    | None -> None
    | Some lid -> Some (path_of_lid lid) in
  let result = match m.main with
      None -> 
        { name = newName;
          extends = extendsName;
          modCdecl = moduleClassDecl;
          decls = transenv;
          externs = createTCdecls(externs_signature);
          entry = None;
        }
    | Some e -> 
        let tMain, transenv = transExp transenv e in
          { 
            name = newName;
            extends = extendsName;
            modCdecl = moduleClassDecl;
            decls = transenv;
            externs = createTCdecls(externs_signature);
            entry = Some tMain;
          } in
    curmodname := None; 
    (*PrettyTarget.printModule result;*)
    result

