#light "off"

// (c) Microsoft Corporation. All rights reserved
module Microsoft.Fine.TypeRelations

open Absyn
open AbsynUtils
open Util
open Const
open Tcenv
open Profiling
open Tcutil

let debug = ref false
let eq_ctr = new_counter "TReln.equivalent"
let inst_ctr = new_counter "TReln.instantiable"
let convert_ctr = new_counter "TReln.convertible"
let refsub_ctr =  new_counter "TReln.refinement_subtyping"
let teq_ctr =  new_counter "TReln.tequiv"
(* if extract_proofs is set, formulas are checked with proofs during derefinment *)
let use_solver env = Tcenv.use_solver env
    
(* an equivalence class of unification variables, possibly unified to a typ *)
type subst = list<(list<uvar*kind> * option<typ>)> 
(* Utilities on substitution maps *)
let find_equivalence_class (subst:subst) uv = 
  let rec finder seen rest = match rest with 
    | [] -> None, seen
    | eq_cl::rest -> 
        let uvl, _ = eq_cl in
          if (List.exists (fun (uv',_) -> Unionfind.equivalent uv uv') uvl) then 
            Some(eq_cl), seen@rest
          else finder (eq_cl::seen) rest 
  in finder [] subst 

let in_equivalence_class subst uv1 uv2 = 
  let eq1, _ = find_equivalence_class subst uv1 in    
  let eq2, _ = find_equivalence_class subst uv2 in
    match eq1, eq2 with
        None, None -> false
      | Some c1, Some c2 -> LanguagePrimitives.PhysicalEquality c1 c2 
      | _ -> false 
          
(* These functions have strange signatures because that's what absyn's
   type_fold_map functions are more general than what's needed here. *)
let subst_typ subst () () t = match (compress' t) with
  | Typ_uvar (uv,_) -> 
      let eqc, _ = find_equivalence_class subst uv in
        (match eqc with
           | Some (uvl, Some t') -> (), t'.v, Some t'.sort
           | _ -> (), t, None)
  | _ -> (),t, None
      
let subst_exp () () e = (), e 
  
let apply_subst (subst:subst) (t:typ) : typ = 
  match subst with 
      []  -> t
    | _ -> 

        let _, t' = typ_fold_map (subst_typ subst) subst_exp (fun () x -> ((),x)) () () t in
          t' 

let apply_subst_exp (subst:subst) (e:exp) : exp = 
  match subst with 
      [] -> e
    | _ -> 
        let _, e' = exp_fold_map (subst_typ subst) subst_exp (fun () x -> ((),x)) () () e in
          e' 

let promote_values_affine env t1 t2 = 
  let t1, t2 = expand_typ_fix env t1, expand_typ_fix env t2 in
    match t1.v, t2.v with 
      | Typ_affine _, Typ_affine _ -> t1, t2
      | _, Typ_affine _ ->  (* implicit promotion to affine type for values *)
          (match current_value env with 
             | Some v when is_non_variable_value env v -> (withsort (Typ_affine t1) Kind_affine), t2
             | _ -> t1, t2)
      | _ -> t1, t2
          
(* assumed equivalences between bound variables *)
type equivs = list<bvdef*bvdef>
    
let bvars_equivalent (bvar1:bvar<'a>) (bvar2:bvar<'a>) equivs : bool = 
  let rn1 = (bvar_real_name bvar1).idText in
  let rn2 = (bvar_real_name bvar2).idText in
    rn1=rn2 ||List.exists 
      (fun (b1:bvdef, b2:bvdef) -> 
         let b1n = real_name b1 in
         let b2n = real_name b2 in
           (b1n.idText = rn1 && b2n.idText = rn2) ||
           (b2n.idText = rn1 && b1n.idText = rn2)) equivs

(* a list of bvars that can be equated with any term index *)
(* Used to compute the set of match assumptions in pattern matching*)
type equatables=list<bvar<typ>> 
type renv = {equatables:equatables;
             tcenv:Tcenv.env;
             subst:subst;
             eqbs:list<binding>;
             evidence:evidence}
let mk_renv equatables env subst = {equatables=equatables; tcenv=env; subst=subst; eqbs=[]; evidence=[]}
let strEquatables renv = 
  String.concat ", " (List.map (fun bv -> Pretty.strBvar bv) renv.equatables) 
let debug_rti renv t1 t2 = 
  if !debug then 
    let eqbs = strEquatables renv in
      pr "Refine_term_indices (%s):\n\t %s\n\t %s\n" eqbs (Pretty.strTyp t1) (Pretty.strTyp t2)
  else ()
let debug_equatable renv e1 b =     
  if !debug then 
    let eqbs = strEquatables renv in
      pr "Testing_equatable (%s):\t %s\t result:%s\n" eqbs (Pretty.strExp e1) (if b then "Yes!" else "no") 
  else ()
let debug_gen_eqb renv e1 e2 =     
  if !debug then 
    let eqbs = strEquatables renv in
      pr "Gen_eqb (%s): adding_equation (%s = %s)\n" eqbs (Pretty.strExp e1) (Pretty.strExp e2)
  else ()

      
(* the actual work of unifying a uv with a type t*)      
let rec try_unify env subst uv t : option<subst> = 
  let uv_occurs_in_typ uv t = 
    let rec occurs_typ t = match compress' t.v with 
      | Typ_uvar (uv',_) -> (Unionfind.equivalent uv uv')
      | Typ_btvar _
      | Typ_const _  -> false
      | Typ_dep (t, _) (* Q: is this right? Should we check e also? *)
      | Typ_affine t 
      | Typ_univ (_, _, t) -> occurs_typ t 
      | Typ_record fnt_l -> List.exists (fun (fn,t) -> occurs_typ t) fnt_l
      | Typ_dtuple nt_l -> List.exists (fun (n,t) -> occurs_typ t) nt_l
      | Typ_fun  (_, t, t') 
      | Typ_refine (_, t, t') 
      | Typ_app (t, t') -> occurs_typ t || occurs_typ t'
      | _ -> raise (NYI "Unexpected type in occurs check") in
      occurs_typ t in
  let is_well_formed t uvs =
    List.for_all (fun (uv,k) -> match Unionfind.find uv with 
                    | Uvar wf -> wf t k 
                    | _ -> raise Impos) uvs in
  let tuv = apply_subst subst uv in
  let tt = apply_subst subst t in
    match tuv.v, tt.v with
      | Typ_uvar(uv1, k1), Typ_uvar (uv2, k2) when Unionfind.equivalent uv1 uv2 -> Some subst
      | Typ_uvar(uv1, k1), Typ_uvar (uv2, k2) when in_equivalence_class subst uv1 uv2 -> Some subst
      | Typ_uvar(uv1, k1), Typ_uvar (uv2, k2) when equivalent_kinds env k1 k2 -> 
          let eqclass1, rest = find_equivalence_class subst uv1 in
          let eqclass2, rest = find_equivalence_class rest uv2 in
          let union12 = match eqclass1, eqclass2 with
              None, None -> Some ([uv1,k1;uv2,k2], None)
            | Some c, None -> 
                (match c with
                   | (uvs, None) -> Some ((uv2,k2)::uvs, None)
                   | (uvs, Some t) -> if is_well_formed t [uv2,k2] then Some ((uv2,k2)::uvs, Some t) else None)
            | None, Some c -> 
                (match c with
                   | (uvs, None) -> Some ((uv1,k1)::uvs, None)
                   | (uvs, Some t) -> if (is_well_formed t [uv1,k1]) then Some ((uv1,k1)::uvs, Some t) else None)
            | Some c1, Some c2 -> 
                (match c1, c2 with
                   | (uvs1, None), (uvs2, None) -> Some (uvs1@uvs2, None)
                   | (uvs1, Some t), (uvs2, None) -> if is_well_formed t [uv2,k2] then Some(uvs1@uvs2, Some t) else None                         
                   | (uvs1, None), (uvs2, Some t) -> if is_well_formed t [uv1,k1] then Some(uvs1@uvs2, Some t) else None
                   | _ -> None) (* cannot be unified ... already fixed to another type, 
                                   TODO: unless both these types are the same? *) in
            bind_opt union12 (fun eqclass12 -> Some (eqclass12::rest))
      | Typ_uvar (uv, k) as t1, _ -> 
          let eqclass, rest = find_equivalence_class subst uv in
          let eqclass = match eqclass with 
              None -> 
                if is_well_formed tt [uv,k] then 
                  Some ([uv,k], Some tt)
                else 
                  None
            | Some (uvs, None) ->  
                if is_well_formed tt uvs then 
                  Some (uvs, Some tt)
                else 
                  None
            | Some (uvs, Some tt') -> 
                None (* already unified with something else *) in
            bind_opt eqclass (fun eqc -> Some (eqc::rest))
      | _ -> 
          raise (NYI "Unexpected unification cases")
(* end try_unify *)

and unifiable env subst t1 t2 = 
  (* straight unification of types t1 and t2, without using the solver or the 
     matching assumptions in the context to refine term indices *)
  let rec same_typ env equivs subst t t' : option<subst> =
    let tt1 = whnf (apply_subst subst t) in 
    let tt2 = whnf (apply_subst subst t') in
    let etf = expand_typ_fix env in
      match (tt1.v, tt2.v) with 
        | Typ_btvar bv1, Typ_btvar bv2 -> 
            if bvars_equivalent bv1 bv2 equivs then Some subst else None
              
        | Typ_const fv1, Typ_const fv2 -> 
            if fvar_eq fv1 fv2 then Some subst else None
              
        | Typ_record fnt_l1, Typ_record fnt_l2 -> 
            if (List.length fnt_l1 =  List.length fnt_l2) then 
              let l = List.zip fnt_l1 fnt_l2 in
              let rec same_rec subst = function
                  [] -> Some subst
                | ((fn1, t1), (fn2, t2))::tl -> 
                    if (Sugar.path_of_lid fn1)=(Sugar.path_of_lid fn2) then
                      bind_opt (same_typ env equivs subst (etf t1) (etf t2))
                        (fun subst' -> same_rec subst' tl) 
                    else None in
                same_rec subst l
            else None
              
        | Typ_fun (nopt, t1, t2),  Typ_fun (nopt', t1', t2') -> (* equivalence of bound term vars handles by tequiv *)
            bind_opt (same_typ env equivs subst (etf t1) (etf t1'))
              (fun subst' -> 
                 (match nopt, nopt' with 
                    | None, None 
                    | Some _, None
                    | None, Some _ -> 
                        same_typ env equivs subst' (etf t2) (etf t2')
                    | Some bvd1, Some bvd2 -> 
                        let equivs' = (bvd1, bvd2)::equivs in
                          same_typ env equivs' subst' (etf t2) (etf t2')))

        | Typ_lam(x, t1, t2), Typ_lam(y, t1', t2') -> 
            bind_opt (same_typ env equivs subst (etf t1) (etf t1'))
              (fun subst' -> 
                 let equivs' = (x, y)::equivs in
                   same_typ env equivs' subst' (etf t2) (etf t2'))
              
        | Typ_univ(bvd1, k, t), Typ_univ(bvd2, k', t') ->                 
            if equivalent_kinds env k k' then same_typ env ((bvd1, bvd2)::equivs) subst (etf t) (etf t')
            else None
              
        | Typ_dtuple nt_l, Typ_dtuple nt_l' -> 
            if (List.length nt_l) = (List.length nt_l') then
              let rec equiv_dep_list equivs subst = function
                  [] -> Some subst
                | ((n1, t1), (n2, t2))::tl -> 
                    bind_opt (same_typ env equivs subst (etf t1) (etf t2))
                      (fun subst' -> 
                         match n1, n2 with
                           | None, None 
                           | Some _, None
                           | None, Some _ ->
                               equiv_dep_list equivs subst' tl
                           | Some bvd1, Some bvd2 -> 
                               let equivs' = (bvd1, bvd2)::equivs in
                                 equiv_dep_list equivs' subst' tl
                           | _ -> None) in
              let l = List.zip nt_l nt_l' in
                equiv_dep_list equivs subst l 
            else None
              
        | Typ_refine (bvd1, t1, f1),  Typ_refine (bvd2, t2, f2) -> 
            bind_opt (same_typ env equivs subst (etf t1) (etf t2))
              (fun subst' -> 
                 same_typ env ((bvd1,bvd2)::equivs) subst' (etf f1) (etf f2))

        | Typ_app (t1, t2), Typ_app (t1', t2') -> 
            bind_opt (same_typ env equivs subst t1 t1')
              (fun subst' -> same_typ env equivs subst' (etf t2) (etf t2'))
              
        | Typ_dep (t1, e1), Typ_dep (t2, e2) -> 
            bind_opt (same_typ env equivs subst t1 t2)
              (fun subst' -> same_exp env equivs subst' e1 e2)
              
        | Typ_unknown, Typ_unknown -> Some subst 
            
        | _, Typ_uvar(uv,k) ->
            try_unify env subst tt2 tt1
        | Typ_uvar(uv,k), _ -> 
            try_unify env subst tt1 tt2
              
        | Typ_affine t1, Typ_affine t2 -> 
            same_typ env equivs subst (etf t1) (etf t2)

        | _ -> None 
            
  and same_exp env equivs subst e1 e2 = (* only implemented for non-function values *)
(*     let _ = pr "Testing equivalence of expressions (%s) and (%s)\n" (Pretty.strExp e1) (Pretty.strExp e2) in *)
    let result = 
      match e1.v, e2.v with
        | Exp_bvar bv1, Exp_bvar bv2 -> 
            if bvars_equivalent bv1 bv2 equivs then Some subst else None
        | Exp_fvar fv1, Exp_fvar fv2 -> 
            if fvar_eq fv1 fv2 then Some subst else None
        | Exp_constant c1, Exp_constant c2 -> 
            if c1=c2 then Some subst else None
        | Exp_constr_app (v1, tl, _, el), Exp_constr_app (v2, tl2, _, el2) ->  
            if ((List.length tl) = (List.length tl2) &&
                (List.length el) = (List.length el2) &&
                (fvar_eq v1 v2)) then
              let rec same_typ_list subst = function
                  [] -> Some subst
                | (t1, t2)::tl -> 
                    bind_opt (same_typ env equivs subst (expand_typ_fix env t1) (expand_typ_fix env t2))
                      (fun subst' -> same_typ_list subst' tl) in
              let rec same_exp_list subst = function
                  [] -> Some subst
                | (e1, e2)::tl -> 
                    bind_opt (same_exp env equivs subst e1 e2)
                      (fun subst' -> same_exp_list subst' tl) in
                bind_opt (same_typ_list subst (List.zip tl tl2))
                  (fun subst' -> 
                     same_exp_list subst' (List.zip el el2))
            else None
              
        | Exp_recd (_, _, _, fne_l_1), Exp_recd (_, _, _, fne_l_2) ->
            if (List.length fne_l_1 = List.length fne_l_2) then
              let rec same_fn_lists subst = function
                  [] -> Some subst
                | ((f1, e1), (f2,e2))::tl -> 
                    let fn_eq = Sugar.lid_equals f1 f2 in 
                      if fn_eq then
                        bind_opt (same_exp env equivs subst e1 e2)
                          (fun subst' -> same_fn_lists subst' tl)
                      else None in
                same_fn_lists subst (List.zip fne_l_1 fne_l_2)
            else None

        | Exp_proj(e1, fn1), Exp_proj(e2, fn2) -> 
            bind_opt (same_exp env equivs subst e1 e2)
              (fun subst' -> if Sugar.lid_equals fn1 fn2 then Some subst' else None)

        | Exp_ascribed(_, _, _), _ 
        | _, Exp_ascribed(_, _, _) -> same_exp env equivs subst (unascribe e1) (unascribe e2)
              
        | _ -> None in
      result  in
  let t1' = expand_typ_fix env t1 in
  let t2' = expand_typ_fix env t2 in
    same_typ env [] subst t1' t2'
(* end unifiable *)

and tequiv renv t1 t2 : option<renv> = (* *list<binding>> =  *)
  (* With environment renv, and substitutions of unification variables
     subst; equate t1 and t2 using equality to refine term indices. If
     successful, return a substitution and a list of inferred equalities
     for the equatables in renv. *)
  (* S;G |- t1 ~=~ t2 *)
  let thunk () = 
    let is_equatable renv e = 
      let result = 
        match (unascribe e).v with 
            Exp_bvar bv -> 
              let rn = bvar_real_name bv in
                List.exists (fun (bvar:bvar<typ>) -> (bvar_real_name bvar).idText = rn.idText) renv.equatables 
          | _ -> false in
      let _ = debug_equatable renv e result in  result in
      
    let gen_eqb renv e1 e2 = 
      let e1, e2 = unascribe e1, unascribe e2 in
      let find_match_assumptions env e : list<binding> = 
        let finder out b = match b with
          | Binding_match (e1, e2) -> 
              if equalsExp e1 e then b::out
              else if equalsExp e2 e then (Binding_match(e2, e1))::out
              else out
          | _ -> out in 
          Tcenv.fold_env env finder [] in
      let env = List.fold_left push_local_binding renv.tcenv renv.eqbs in
      let ma = find_match_assumptions env e1 in
        (match ma with 
           | [] ->  
               debug_gen_eqb renv e1 e2;
               let new_eqb = Binding_match(e1, e2) in
               let renv' = {renv with eqbs=new_eqb::renv.eqbs} in
                 Some renv'
           | [Binding_match (_, e2')] -> 
               if (equalsExp e2 e2') then Some renv
               else None
           | _ -> raise Impos) in
      
    let matches renv e1 e2 : option<renv> = 
      let e1, e2 = unascribe e1, unascribe e2 in    
      let equiv env e1 e2 = 
        (*       pr "Equiv %s = %s called on environment: " (Pretty.strExp e1)  (Pretty.strExp e2);  *)
        let in_eq_class eqc e' = 
          let result = List.exists (fun e'' -> 
                                      if LanguagePrimitives.PhysicalEquality e' e'' then true
                                      else equalsExp (unascribe e') (unascribe e'')) eqc in
            (*         let _ = Printf.printf "%s %s eq class is [%s]\n"  *)
            (*           (Pretty.strExp e') (if result then "IN" else "NOT IN") *)
            (*           (String.concat ";" (List.map Pretty.strExp eqc)) in *)
            result in
        let mk_eq_class eqc matches = 
          (*         let ctr = ref 0 in *)
          let rec mk_eq_class eqc matches =
            let eqc', matches', changed = Tcenv.fold_env env
              (fun (eqc, matches, changed) b -> match b with
                 | Binding_match(e1', e2') -> 
                     (*                    pr "Environment contains %s=%s\n" (Pretty.strExp e1') (Pretty.strExp e2'); *)
                     let t1 = (in_eq_class eqc e1') in
                     let t2 = (in_eq_class eqc e2') in
                     let eqc' = if t1 || t2 then
                       let eqc1 =(if t1 then eqc else (e1'::eqc)) in
                         if t2 then eqc1 else e2'::eqc1 
                     else eqc in
                     let changed' = ((t1 || t2) && not(t1&&t2)) in
                     let matches' = if changed' then 
                       ((* pr "\tAdding to equiv class\n"; *)
                         (e1',e2')::matches) else matches in
                     let changed = changed || changed' in
                       eqc', matches', changed
                 | _ -> eqc, matches, changed) (eqc, matches, false) in
              if not changed then eqc', matches' 
              else (mk_eq_class eqc' matches') in
            mk_eq_class eqc matches in
        let eqc, matches' = mk_eq_class [e1] [] in
          if in_eq_class eqc e2 then
            Some matches'
          else None in
      let env = renv.tcenv in
        match equiv env e1 e2 with
            Some ev -> Some ({renv with evidence=ev@renv.evidence})
          | _ -> 
              if (not (in_derefinement_phase env)) && use_solver env && Solver.query_equiv env e1 e2 then Some renv
              else None in

    let rec refine_exps renv e1 e2 =  (* Only for non functional values *)
      let e1, e2 = unascribe e1, unascribe e2 in
        (*       pr "Testing equivalence of %s=%s\n" (Pretty.strExp e1) (Pretty.strExp e2); print_any e1; print_any e2; *)
        if is_equatable renv e1 then gen_eqb renv e1 e2 
        else if is_equatable renv e2 then gen_eqb renv e2 e1
        else match matches renv e1 e2 with
            Some renv' -> Some renv'
          | _ -> 
              match e1.v, e2.v with
                | Exp_bvar bv1, Exp_bvar bv2 when (bvar_real_name bv1).idText = (bvar_real_name bv2).idText -> Some renv
                | Exp_fvar fv1, Exp_fvar fv2 when (Sugar.text_of_lid fv1.v) = (Sugar.text_of_lid fv2.v) -> Some renv
                | Exp_constant c1, Exp_constant c2 when Sugar.const_eq c1 c2  -> Some renv
                | Exp_constr_app (v1, tl, elp1, el), Exp_constr_app (v2, tl2, elp2, el2) ->  
                    if ((List.length tl) = (List.length tl2) &&
                        (List.length elp1) = (List.length elp2) &&
                        (List.length el) = (List.length el2) &&
                        (Sugar.lid_equals v1.v v2.v)) then
                      let rec refine_exp_list renv = function
                          [] -> Some renv
                        | (e1, e2)::tl -> 
                            match refine_exps renv e1 e2 with
                                None -> None
                              | Some renv' -> refine_exp_list renv' tl in
                        refine_exp_list renv (List.zip (elp1@el) (elp2@el2))
                    else None

                | Exp_recd (_, _, _, fne_l_1), Exp_recd (_, _, _, fne_l_2) -> 
                    if (List.length fne_l_1 = List.length fne_l_2) then
                      let rec refine_fnl renv = function
                        | [] -> Some renv
                        | ((f1, e1), (f2,e2))::tl -> 
                            if (Sugar.lid_equals f1 f2) then 
                              match refine_exps renv e1 e2 with
                                  None -> None
                                | Some renv' -> refine_fnl renv' tl 
                            else None in
                        refine_fnl renv (List.zip fne_l_1 fne_l_2)
                    else None
                | _ -> None in    
      
    let refine_term_indices renv t1 t2 = 
      let tt1 = compress t1 in
      let tt2 = compress t2 in
        debug_rti renv t1 t2;
        match (tt1.v, tt2.v) with
          | Typ_app(t1, t1'), Typ_app(t2, t2') ->
              bind_opt 
                (tequiv renv t1 t2) 
                (fun renv' -> tequiv renv' t1' t2')
                
          | Typ_dep(t1, e1), Typ_dep(t2, e2) -> 
              bind_opt (tequiv renv t1 t2)
                (fun renv' -> 
                   let e1 = apply_subst_exp renv'.subst e1 in
                   let e2 = apply_subst_exp renv'.subst e2 in
                     refine_exps renv' e1 e2)
          | _, _ -> None in 
      
    let t1, t2 = whnf t1, whnf t2 in
      match (unifiable renv.tcenv renv.subst t1 t2) with (* first try to show that t1 and t2 are the same *)
          None -> refine_term_indices renv t1 t2 (* if that fails, then try refining indices *)
        | Some subst' -> Some {renv with subst=subst'} in
    profile thunk teq_ctr
      (* end tequiv *)

and equivalent env t1 t2 = 
  let thunk = fun () -> 
    let renv = {equatables=[]; tcenv=env; subst=[]; eqbs=[]; evidence=[]} in
    let t1, t2 = promote_values_affine env t1 t2 in
      match tequiv renv t1 t2 with
          None -> false
        | Some renv -> 
            if List.length renv.subst <> 0 then raise Impos
            else true in
    profile thunk eq_ctr

and equivalent_kinds env k1 k2 = 
  match unbox k1, unbox k2 with 
    | Kind_star, Kind_star 
    | Kind_affine, Kind_affine -> true
    | Kind_tcon(k1_1, k1_2), Kind_tcon(k2_1, k2_2) -> 
        equivalent_kinds env k1_1 k2_1 && equivalent_kinds env k2_1 k2_2
    | Kind_dcon(t1, k1), Kind_dcon(t2, k2) -> 
        equivalent env t1 t2 && equivalent_kinds env k1 k2
    | _ -> false 

(* The type conversion relation S;G |- t1 <: t2 *)
let as_ctr = new_counter "TReln.apply_subst (refsub)"
let query_ctr = new_counter "TReln.query (refsub)"
let rec convertible env t1 t2 : option<subst> =
(*   let _ = pr "Convertible called for (%s) and (%s)\n" (Pretty.strTyp t1) (Pretty.strTyp t2) in *)
  let thunk () = 
    let refinement_subtyping env t1 t2 = 
      let tt1 = compress t1 in
      let tt2 = compress t2 in
        match (tt1.v, tt2.v) with
          | Typ_refine(bvd1, t1, phi1), Typ_refine(bvd2, t2, phi2) -> 
              let t1, t2 = whnf t1, whnf t2 in
              let phi1, phi2 = whnf phi1, whnf phi2 in
              if not (use_solver env) then 
                bind_opt (tequiv (mk_renv [] env []) tt1 tt2)
                  (fun renv -> Some renv.subst)
              else
                bind_opt (convertible env t1 t2) 
                  (fun subst' -> 
                     let name = real_name bvd2 in 
                     let tt1' = profile (fun () -> apply_subst subst' tt1) as_ctr in
                     let phi2' = profile (fun () -> apply_subst subst' phi2) as_ctr in
                     let env' = push_local_binding env (Binding_var (name, tt1))in
                     let env', mklet = match current_value env' with 
                         None -> raise Impos (* env', fun x -> x *)
                       | Some v -> 
                           let bv = new_bvar name name dummyRange in
                           let ebv = withinfo (Exp_bvar bv) tt1 dummyRange in
                           let env' = 
                             if is_value v then 
                               let b = Binding_match (ebv, v) in
                                 push_local_binding env' b 
                             else env' in
                           let mklet pf = Exp_let(false, [(name, name), t_withpos Typ_unknown dummyRange, v], withpos pf dummyRange) in 
                             env', mklet in
                       if profile (fun () -> Solver.query env' phi2') query_ctr then Some subst'
                       else None)
                  
          | Typ_refine(bvd1, t1, phi1), _ -> convertible env t1 tt2 
              
          | _, (Typ_refine _) -> 
              let fn = genident None in
              let t1' = withsort (Typ_refine((fn,fn), tt1, true_typ)) tt1.sort in
                convertible env t1' tt2 
                  
          (* This case is a structural rule that is not included in the formalism *)
          | Typ_affine t1, Typ_affine t2 -> convertible env t1 t2 
              
          | _ -> None in 
    let t1, t2 = promote_values_affine env t1 t2 in
      match tequiv (mk_renv [] env []) t1 t2 with (* try unification/equivalence first *)
        | None -> profile (fun () -> refinement_subtyping env t1 t2) refsub_ctr  (* if that fails, try refinement implication *)
        | Some renv -> Some renv.subst in
    profile thunk convert_ctr
      
let instantiable env equatables t1 t2 = 
  let thunk = fun () -> 
    let renv = {equatables=equatables; tcenv=env; subst=[]; eqbs=[]; evidence=[]} in
    let t1, t2 = promote_values_affine env t1 t2 in
      bind_opt 
        (tequiv renv t1 t2)
        (fun renv -> Some (renv.subst, renv.eqbs)) in
    profile thunk inst_ctr
      
let find_all_match_assumptions env : list<binding> = 
  let finder out b = match b with
    | Binding_match (e1, e2) -> b::out
    | _ -> out in 
    Tcenv.fold_env env finder []

              
let equivalent_show_evidence env t1 t2 = 
(*   Printf.printf "Testing equivalence of types %s and %s\n" (Pretty.strTyp t1) (Pretty.strTyp t2); *)
  let renv = {equatables=[]; tcenv=env; subst=[]; eqbs=[]; evidence=[]} in
  let t1, t2 = promote_values_affine env t1 t2 in
    match tequiv renv t1 t2 with
        None -> (* let ma = find_all_match_assumptions env in print_any ("Match assumptions:", ma);  *)false, []
      | Some renv -> 
          if List.length renv.subst <> 0 then raise Impos
          else true, renv.evidence
            (*             (if renv.evidence <> [] then (Printf.printf "Reporting evidence %s" (Pretty.strEv renv.evidence)); *)

let equivalent_with_equations env eqbs t1 t2 = 
  let renv = {equatables=eqbs; tcenv=env; subst=[]; eqbs=[]; evidence=[]} in
  let t1, t2 = promote_values_affine env t1 t2 in
    match tequiv renv t1 t2 with
        None -> (* let ma = find_all_match_assumptions env in print_any ("Match assumptions:", ma);  *)None
      | Some renv -> 
          if List.length renv.subst <> 0 then raise Impos
          else Some renv.eqbs

  
            
  
