#light "off"
// (c) Microsoft Corporation. All rights reserved
module Microsoft.Fine.Util

open System.Diagnostics

exception Impos
exception NYI of string
exception Not_found
exception Bad of string
exception Failure of string

let silent=ref false
let verbosity = ref -1
let trace_counter = ref 0
let inline_proof_term = ref false
let interleave_proof_principles = ref true
let extract_proofs = ref false
let typecheck_proofs = ref false
let skipUnit = ref false
let proof_term_errs = ref false
let skip_trans = ref false
let genIL = ref false
let verbose_names = ref false
let print_real_names = ref false
let mainmodule = ref false
let dump_src = ref false
let dump_tgt = ref false
type DebugLog() =
  class
  [<Conditional("DEBUG")>]
  static member Log s = () (*Printf.printf "%s" s *)
end


let pr = Printf.printf
let spr = Printf.sprintf
let fpr = Printf.fprintf 
let soi = string_of_int

let try_find_position matcher f = 
  let rec aux pos = function
    | [] -> None
    | a::tl -> if (matcher a) then Some pos else aux (pos+1us) tl
  in aux 0us f

let debug l s = 
  if (!verbosity > l) then 
    begin
      Printf.printf "\n<DEBUG>\n"; Printf.printf s ; Printf.printf "\n</DEBUG>\n"
    end
      
let debug_dump v s a = 
  if (!verbosity > v) then 
    begin
      debug v s; printfn "%A" a
    end

let bump_trace_counter () = trace_counter:=!trace_counter+1; !trace_counter
let trace n s a = 
  if (!verbosity = n) then 
    (let n = bump_trace_counter () in
       Printf.printf "\n%d<trace>:" n;
       print_string s; printfn "%A" a;
       Printf.printf "\n</trace>%d\n" n; flush stdout)
      
type Disj<'a,'b> =
  | Inl of 'a
  | Inr of 'b

type triVal =
  | Yes 
  | No
  | Maybe
      
let (-<-) f g x = f (g x)

let pfx l = 
  match List.rev l with 
      hd::tl -> List.rev tl, hd
    | _ -> raise Impos
          
(* lookup in associative list *)
let rec find p = function
  | []       -> raise Not_found
  | (x,y)::t -> if p x then y else find p t

(* get index of first occurrence of x in l, if any *)
let findIdx x l =
  let rec f k = function
    | [] -> None
    | y::tl -> if x = y then Some k else f (succ k) tl
  in
  f 0 l


let rec intersect l1 l2 f = 
  List.fold_left (fun out l1elt -> 
                    if List.exists (fun l2elt -> f l1elt l2elt) l2 then
                      l1elt::out
                    else out) [] l1

let rec list1N_ acc k n = if k < 1 then acc else list1N_ (k::acc) (k-1) n 
let list1N n = list1N_ [] n n

let map1N n f = List.map f (list1N n)

let list0N n = map1N (n+1) pred


let interleave l1 l2 =
  if List.length l1 <> List.length l2 then failwith "interleave" else
  let rec f acc = function
    | [], []       -> acc
    | x::xs, y::ys -> f (y::x::acc) (xs,ys)
  in
  List.rev (f [] (l1,l2))

let firstN n l =
  let rec f acc i l =
    if i = n then acc else
    match l with
      | h::tl -> f (h::acc) (i+1) tl
      | _     -> failwith "firstN"
  in
  List.rev (f [] 0 l)

let is_some = function
  | Some _ -> true
  | _ -> false

let must = function
    Some x -> x
  | None -> raise Impos
     
let bind_opt opt f = match opt with
    None -> None
  | Some x -> f x

let tensor_opt opt f = match opt with
    None -> None
  | Some x -> Some (f x)

let rec find_map l f = match l with 
    [] -> None 
  | x::tl -> match f x with 
        None -> find_map tl f
      | y -> y
          
let (=^^=>) m f = match m with 
  | None -> f ()
  | _ -> m

let existsNone l = List.exists (function None -> true | _ -> false) l

let projSomes l = List.map (function (Some x) -> x) l

module Hashtbl2 = struct
  type ('a,'b) t = ('a,'b) Hashtbl.t * string * ('a -> string)
  let create s f : ('a,'b) t = (Hashtbl.create 17, s, f)
  let add ((ht,_,_):('a,'b) t) (k:'a) (v:'b) = Hashtbl.add ht k v
  let replace ((ht,_,_):('a,'b) t) (k:'a) (v:'b) = Hashtbl.replace ht k v
  let mem ((ht,_,_):('a,'b) t) (k:'a) = Hashtbl.mem ht k
  let iter ((ht,_,_):('a,'b) t) f = Hashtbl.iter f ht
  let fold f ((ht,_,_):('a,'b) t) x = Hashtbl.fold f ht x
  let find ((ht,s,f):('a,'b) t) (k:'a) =
    if Hashtbl.mem ht k then Hashtbl.find ht k
    else failwith (spr "Hashtbl2.find: %s not in table %s" (f k) s)
  let clear ((ht,_,_):('a,'b) t) = Hashtbl.clear ht
  let remove ((ht,_,_):('a,'b) t) (x:'a) = Hashtbl.remove ht x
end

module IdTable = struct
  type 'a t = int ref * (int,'a) Hashtbl.t * ('a,int) Hashtbl.t * string * ('a -> string)
  let create s f = ref 0, Hashtbl.create 17, Hashtbl.create 17, s, f
  let process ((k,ht1,ht2,_,_):'a t) (x:'a) =
    if Hashtbl.mem ht2 x then Hashtbl.find ht2 x
    else (incr k; let i = !k in Hashtbl.add ht1 i x; Hashtbl.add ht2 x i; i)
  let getVal ((_,ht1,_,s,_):'a t) i =
    if Hashtbl.mem ht1 i then Hashtbl.find ht1 i
    else failwith (spr "IdTable.getVal: %d not in table %s" i s)
  let getId ((_,_,ht2,s,f):'a t) (x:'a) =
    if Hashtbl.mem ht2 x then Hashtbl.find ht2 x
    else failwith (spr "IdTable.getId: %s not in table %s" (f x) s)
  let iter ((_,ht1,_,_,_):'a t) f = Hashtbl.iter f ht1
  let clear ((_,ht1,ht2,_,_):'a t) = Hashtbl.clear ht1; Hashtbl.clear ht2
  let mem ((_,_,ht2,_,_):'a t) = Hashtbl.mem ht2
end

module Bijection = struct
  type ('a,'b) t = ('a,'b) Hashtbl2.t * ('b,'a) Hashtbl2.t
  let create s fa fb : ('a,'b) t = Hashtbl2.create (spr "%s_ab" s) fa, Hashtbl2.create (spr "%s_ba" s) fb
  let add ((ht1,ht2):('a,'b) t) (a:'a) (b:'b) = Hashtbl2.add ht1 a b; Hashtbl2.add ht2 b a
  let findA ((ht1,_):('a,'b) t) = Hashtbl2.find ht1
  let findB ((_,ht2):('a,'b) t) = Hashtbl2.find ht2
  let memA ((ht1,_):('a,'b) t) = Hashtbl2.mem ht1
  let memB ((_,ht2):('a,'b) t) = Hashtbl2.mem ht2
  let clear ((ht1,ht2):('a,'b) t) = Hashtbl2.clear ht2; Hashtbl2.clear ht1
  let iter ((ht1,_):('a,'b) t) f = Hashtbl2.iter ht1 f
end

